SystemOrganization addCategory: #'Connectors-Base'!
SystemOrganization addCategory: #'Connectors-EToys'!
SystemOrganization addCategory: #'Connectors-Info'!
SystemOrganization addCategory: #'Connectors-Lines and Curves'!
SystemOrganization addCategory: #'Connectors-Shapes'!
SystemOrganization addCategory: #'Connectors-Text-Base'!
SystemOrganization addCategory: #'Connectors-Tools'!

!MenuMorph methodsFor: '*connectors-accessing' stamp: 'nk 9/8/2002 09:18'!
capitalize
	self items do:[:anItem | (anItem isKindOf: StringMorph) ifTrue: [ anItem contents: (anItem contents capitalized) ] ].
	^ nil! !

!Bezier2Segment methodsFor: '*connectors-vector functions' stamp: 'nk 8/19/2003 18:02'!
intersectionsWithLineFrom: e1 to: e2
	"Answer the set of intersections, if any"
	| retval l i |
	retval := OrderedCollection new: 2.
	l := LineSegment from: e1 to: e2.
	self lineSegmentsDo: [ :p1 :p2 | | s |
		s := LineSegment from: p1 to: p2.
		i := s intersectionWith: l.
		i ifNotNil: [ retval add: i ].
		retval size = 2 ifTrue: [ ^retval asArray ].
	].
	^retval asArray! !

!Bezier2Segment methodsFor: '*connectors-vector functions' stamp: 'nk 8/18/2003 21:35'!
lineSegmentsWithParametricValueDo: aBlock
	"Evaluate aBlock with the receiver's line segments and the corresponding parametric values, ranging from just above 0 to 1.0"
	"Note: We could use forward differencing here."
	| steps last deltaStep t lastT next |
	steps := 1 max: (self length // 10). "Assume 10 pixels per step"
	last := start.
	deltaStep := 1.0 / steps asFloat.
	t := deltaStep.
	lastT := 0.0.
	1 to: steps do:[:i|
		next := self valueAt: t.
		aBlock value: last value: next value: lastT value: t.
		last := next.
		lastT := t.
		t := t + deltaStep].! !

!CircleMorph methodsFor: '*connectors' stamp: 'nk 2/28/2006 09:53'!
defaultColor
	^Color green muchLighter! !

!CircleMorph methodsFor: '*connectors' stamp: 'nk 7/18/2003 17:02'!
endShapeColor: aColor
	self fillStyle isTransparent ifFalse: [ self color: aColor ].! !

!CircleMorph methodsFor: '*connectors' stamp: 'nk 7/1/2002 16:08'!
endShapeWidth: anInteger
	| oldLineAttachment oldRotationCenter |
	oldLineAttachment := self lineAttachmentOffset.
	oldRotationCenter := self rotationCenter.
	super endShapeWidth: anInteger.
	self lineAttachmentOffset: oldLineAttachment.
	self rotationCenter: oldRotationCenter.
! !

CircleMorph subclass: #NCPinMorph
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Connectors-Shapes'!

!NCPinMorph commentStamp: '<historical>' prior: 0!
I am a circle that wants to act as a connection pin.
I offer to embed myself on drop.
When embedded, I will create a Player-valued variable in my new owner.!

!NCPinMorph methodsFor: 'naming' stamp: 'cmm 12/4/2010 20:05'!
defaultNameStemForInstances
	^ self class name
		copyFrom: 3
		to: self class name size! !

!NCPinMorph methodsFor: 'submorphs-accessing' stamp: 'nk 8/24/2004 14:38'!
droppedIntoPasteUp: aMorph
	| targets |
	targets := (aMorph rootMorphsAt: self center) reject: [ :ea | ea isPinMorph ].
	targets size < 1 ifTrue: [ ^self ].
	WorldState addDeferredUIMessage: [ self offerToEmbedIn: targets ].
! !

!NCPinMorph methodsFor: 'submorphs-accessing' stamp: 'nk 8/24/2004 15:18'!
embeddedIntoMorph: aMorph
	| player |
	player := aMorph assuredPlayer.
	Cursor normal show.
	player createSlotFor: self.! !

!NCPinMorph methodsFor: 'testing' stamp: 'nk 8/24/2004 14:35'!
isPinMorph
	^true! !

!NCPinMorph methodsFor: 'submorphs-accessing' stamp: 'nk 8/24/2004 15:44'!
noteNewOwner: aMorph

	super noteNewOwner: aMorph.

	aMorph isPlayfieldLike
		ifTrue: [ ^self droppedIntoPasteUp: aMorph ].

	aMorph isHandMorph
		ifTrue: [ ]
		ifFalse: [ self embeddedIntoMorph: aMorph ]
! !

!NCPinMorph methodsFor: 'embeddings' stamp: 'nk 8/23/2004 12:10'!
offerToEmbedIn: targets 
	| selected choices labels |
	choices := OrderedCollection withAll: targets.
	labels := choices collect: [ :ea | ea externalName ].
	labels addFirst: '<do not embed>' translated.
	choices addFirst: nil.
	selected := (SelectionMenu
		labels: labels
		lines: {1}
		selections: choices) startUpWithCaption: 'Embed pin into...' translated.
	selected ifNotNil: [ selected addMorphFront: self ].
	Cursor normal show.! !

!HandleMorph methodsFor: '*connectors-connection' stamp: 'nk 8/9/2004 16:23'!
isConnectable
	^false
! !

!HandleMorph methodsFor: '*connectors-testing' stamp: 'nk 6/12/2004 16:19'!
isHandleMorph
	^true! !

PackageInfo subclass: #ConnectorsInfo
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Connectors-Info'!

!ConnectorsInfo class methodsFor: 'installation' stamp: 'nk 5/13/2004 18:07'!
createConnectorsFlaps
	"ConnectorsInfo createConnectorsFlaps"
	| flap |
	Utilities
		informUser: 'Making new flaps (delete them if you want)...'
		during: [
			(Flaps globalFlapTab: 'Connectors')
				ifNil: [flap := Flaps newConnectorsFlap openInWorld; setToPopOutOnDragOver: false.
					Flaps addGlobalFlap: flap].
			Flaps positionNavigatorAndOtherFlapsAccordingToPreference]! !

!ConnectorsInfo class methodsFor: 'class initialization' stamp: 'nk 4/30/2004 16:54'!
initialize
	[self new register] on: MessageNotUnderstood do: []! !

!ConnectorsInfo class methodsFor: 'distribution' stamp: 'nk 7/8/2003 08:16'!
makeSAR
	"ConnectorsInfo makeSAR"
	Smalltalk at: #SARPackageDumper ifPresent: [ :sarPackageDumper |
		sarPackageDumper fileOutPackage: self default
			as: ('Connectors', self default cvsID, '.sar')
	]! !

!ConnectorsInfo class methodsFor: 'installation' stamp: 'nk 4/30/2004 18:21'!
postInstall
	self createConnectorsFlaps.
	Preferences addPreference: #useSmartLabels category:  'Connectors' default:  false balloonHelp: 'if set, Connectors will use labels that try to re-position themselves to stay out of the way.'! !

!ConnectorsInfo methodsFor: 'introspection' stamp: 'nk 7/26/2003 14:26'!
additionalChangeSets
	^#()! !

!ConnectorsInfo methodsFor: 'introspection' stamp: 'nk 5/20/2003 09:22'!
changesText
	^'
"Change Set:		Connectors-nk
Date:				6 February 2003
Author:				Ned Konz <ned@bike-nomad.com>
License:				Squeak''s or MIT

3/29/03:
Added a number of fixes that came from my work with the Tablet PC.
Added graph traversal and simple layout behavior.
Improved resizing behavior and event handling.

2/6/03:
added fix for self-connections from Hernan Pablo Tylim:
This change set fix a problem when you connect both ends of a connector in the same shape and this shape is not in the world (shape pasteUpMorph ~~ shape currentWorld).

12/12/02:
added open connectors project to World menu
fixed bug with centering default text style when Accufonts loaded
suppressed re-creation of default flaps

12/4/02:
added buttonBar labeling; changed category names.
added SAR creation to ConnectorsInfo class.
Made install clean up empty class and method categories.

11/15:
changed class category names

7/1 - 10/29:

Added new Project subclass with its own World type.

Added custom NCPartsBin class for flaps

Several bug fixes.

6/22 - 6/30:

Changed to using (new) CircleMorph for round end decorations.

Made NCSchematicConnectorMorph 

Fixed up AP adjuster quite a bit.

Made end decorations scale with line width.

Made connectors respect being locked.

Added NCNoteMorph.

Made pre-defined line-end shapes look much nicer.

Changed pre-made Connectors from using PolygonMorph arrows to attached shapes.

Made transition Connector in FSM flap and toolbar start wiring with a label.

Made Connector labels delete when Connector is deleted.

Moved notification into addedOrRemovedSubmorph:

Fixed problem with unattached Connectors re-sizing when grabbed/dropped

Fixed problem with line attachment point in Connectors dropped onto MakerButtons

Added ''connect to nearest orthogonal point''

Fixed minor bugs

Removed obsolete methods in Connector that took Blocks

6/22: added some utility methods

6/20: (1.02) (085) Fixed Attachment Point Adjuster and gridding. Added ''Connect to nearest orthogonal point'' to connector end menu.

6/19: Version 1.0 hits the streets...

6/18: Made end attachments scale with line width.

6/15: Added NCSketchMorph, attacked more coordinate problems. Moved ObjectsTool pages around; reorganized demo flaps. Fixed slide-to-trash problem.

6/14: Fixed problems with connectors in translated coordinate systems. Added connection feedback.

4/19: changed category names for 3.3 compatibility. Fixed loading into old images that had Connectors loaded already. Reduced memory consumed by Constraints.

4/18: added ortho snap, new Sketch that aligns itself on grid lines, more work on gridding.

4/14: got transforms working OK for selection
Worked on gridding (for schematic editor)
Got SmartLabels working; made them the default (though there is a Preference)

4/10: put back bounds fiddling for another look at the problem.

This is a version of my Connectors that use a MessageSend instead of a BlockContext for pluggability. This (I hope) will serialize better; I''d like to use these for collaborative drawings on Swikis.

This change set will convert existing Connectors as needed. It may miss some; I''d like to hear about it if it does."
'

! !

!ConnectorsInfo methodsFor: 'introspection' stamp: 'nk 12/4/2002 18:25'!
cvsID
	^('$Revision: 1.13 $' findTokens: ' $:') second! !

!ConnectorsInfo methodsFor: 'naming' stamp: 'nk 10/18/2002 22:33'!
externalName
	^'Connectors'! !

!ConnectorsInfo methodsFor: 'naming' stamp: 'nk 5/11/2004 13:53'!
fullName
	^self! !

!ConnectorsInfo methodsFor: 'introspection' stamp: 'cmm 11/16/2010 23:37'!
postscriptText
	^'"Postscript:
First, convert all the old constraint morphs (if any).
Then make sure that all the connectors have the new FSM.
Then set up some flaps."
| flap |
Utilities informUser: ''Converting old connectors...'' during: [
NCConstraintMorph allSubInstancesDo: [ :ea | ea convertToCurrentVersion ].
NCConnectorMorph allSubInstancesDo: [ :ea | ea reinitializeFSM ].
NCScrolledCompositeStateMorph allInstances do: [ :ea |
ea transformMorph submorphs first on: #mouseDown send: #mouseDown: to: ea.
] ].


Utilities informUser: ''Making Connectors look nicer...'' during: [
NCLineEndConstraintMorph updateOldShapes26June2002 ].

Utilities informUser: ''Recreating Objects Tool thumbnails, please wait...'' during: [
PartsBin clearThumbnailCache; cacheAllThumbnails ].

NCDrawingProject newProjectNamed: ''New Connectors Drawing''.

SystemOrganization removeEmptyCategories.
Morph organization removeEmptyCategories.
Smalltalk garbageCollect.
((Class allSubInstances) do:
	[:cls | | org badCats | org := cls organization.
		(org categories anySatisfy: [ :cat | cat beginsWith2: ''*project-connect'' ])
			ifTrue: [ org removeEmptyCategories ] ]).

ConnectorsInfo instVarNamed: ''default'' put: nil.
'
! !

!ConnectorsInfo methodsFor: 'introspection' stamp: 'cmm 11/16/2010 23:37'!
preambleText
	"Compiler evaluate: ((PackageInfo named: 'Connectors') preambleText) logged: false"

	^ '"$Revision: 1.13 $"
"Convert old instances if necessary."
(Smalltalk hasClassNamed: ''NCConstraintMorph'') ifTrue: [
	| i1 i2 |
	i1 := NCConstraintMorph allInstVarNames indexOf: ''targetBlock''.
	i1 > 0 ifTrue: [
	NCConstraintMorph addInstVarName: ''constraint''.
	i2 := NCConstraintMorph allInstVarNames indexOf: ''constraint''.
	NCConstraintMorph allSubInstances do: [ :ea | ea instVarAt: i2 put: (ea instVarAt: i1) ].
	]
].

Smalltalk organization renameCategory: ''People-nk-Broom'' toBe: ''Connectors-Broom''.
'! !

!ConnectorsInfo methodsFor: 'introspection' stamp: 'nk 7/29/2003 17:13'!
prerequisiteChangeSets
	^#('ConnectorsPreload-nk')! !

GenericPropertiesMorph subclass: #ConnectorPropertiesMorph
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Connectors-Tools'!

!ConnectorPropertiesMorph methodsFor: 'as yet unclassified' stamp: 'nk 1/18/2004 17:08'!
initialize
	"initialize the state of the receiver"
	super initialize.

	thingsToRevert
		at: #lineWidth: put: myTarget lineWidth;
		at: #lineColor: put: myTarget lineColor;
		at: #borderWidth: put: myTarget borderWidth;
		at: #borderColor: put: myTarget borderColor;
		at: #beSmoothCurve: put: myTarget isCurve;
		yourself.

	self rebuild! !

!ConnectorPropertiesMorph methodsFor: 'as yet unclassified' stamp: 'nk 7/17/2004 09:16'!
rebuild

	| buttonColor menu |
	buttonColor := color lighter.
	
	self removeAllMorphs.
	self addAColumn: {
		self lockedString: ('Connector Properties for {1}' translated format:{myTarget name}).
	}.
	
	menu := MenuMorph new defaultTarget: myTarget.
	myTarget addYellowButtonMenuItemsTo: menu event: ActiveHand lastEvent.

	self addARow: { self inAColumn: (menu  submorphs collect: [ :item | item duplicateMorph: ActiveHand lastEvent ]) }.

	self addARow: {
		self inAColumn: {
			self addARow: {
				self 
					buttonNamed: 'Accept' translated action: #doAccept color: buttonColor
					help: 'keep changes made and close panel' translated.
				self 
					buttonNamed: 'Cancel' translated action: #doCancel color: buttonColor
					help: 'cancel changes made and close panel' translated.
			}.
		}.
	}.
! !

!GenericPropertiesMorph methodsFor: '*connectors-visual properties' stamp: 'nk 1/18/2004 17:03'!
doConnectorProperties
	myTarget openAConnectorPropertySheet.
	self delete.! !

!PhraseTileMorph methodsFor: '*connectors-accessing' stamp: 'nk 9/9/2004 13:16'!
playerRepresented
	"Answer the player represented by the receiver. 
	Morphs that serve as references to other morphs reimplement this; 
	by default a morph represents its own player."

	^ resultType == #Player
		ifTrue: [self try]
		ifFalse: [super playerRepresented]! !

!PhraseTileMorph methodsFor: '*connectors-embeddings' stamp: 'nk 9/9/2004 14:09'!
revealPlayerIn: aPasteUpMorph 
	(self playerRepresented ifNil: [ ^self ]) revealPlayerIn: aPasteUpMorph ! !

!StringMorph methodsFor: '*connectors-testing' stamp: 'nk 9/4/2004 17:42'!
isStringMorph
	^true! !

!PasteUpMorph methodsFor: '*connectors-connections' stamp: 'nk 5/11/2003 09:44'!
allConnectedMorphs
	"Answer the Set of my submorphs that have connections"
	| allConnected |
	allConnected := IdentitySet new: submorphs size.
	submorphs do: [ :ea | ea connections isEmpty ifFalse: [ allConnected add: ea ]].
	^allConnected.! !

!PasteUpMorph methodsFor: '*connectors-connections' stamp: 'nk 3/12/2002 09:34'!
allConnections
	"Answer a collection of my submorphs that are connectors"
	^submorphs select: [ :ea | ea isConnector ]! !

!PasteUpMorph methodsFor: '*connectors-embeddings' stamp: 'nk 8/9/2004 16:19'!
allowsConnectionToSubmorph: aMorph
	^true
! !

!PasteUpMorph methodsFor: '*connectors-embeddings' stamp: 'nk 9/10/2004 11:50'!
allowsConnectionToSubmorphs
	"Answer true if I will allow a connector to connect to any of my submorphs.
	If not, the search for attachment points in embedded morphs will not be made."

	^self valueOfProperty: #allowsConnectionToSubmorphs ifAbsent: [ self isPartsBin not ]! !

!PasteUpMorph methodsFor: '*connectors-misc' stamp: 'nk 8/25/2003 13:53'!
defaultConnectorClass
	^NCAAConnectorMorph! !

!PasteUpMorph methodsFor: '*connectors-world menu' stamp: 'nk 6/12/2004 15:16'!
grabNCDrawingFromScreen: evt
	"Allow the user to specify a rectangular area of the Display, capture the pixels from that area (except for the PasteUpMorph itself), and use them to create a new drawing morph. Attach the result to the hand."
	"World grabNCDrawingFromScreen: ActiveHand lastEvent"
	| rect form m |
	rect := Rectangle fromUser.
	form := self currentWorld patchAt: rect without: self currentWorld andNothingAbove: false.
	m := World drawingClass new form: form.
	evt hand position: Sensor cursorPoint.  "update hand pos after Sensor loop in fromUser"
	evt hand attachMorph: m.! !

!PasteUpMorph methodsFor: '*connectors-misc' stamp: 'nk 6/13/2002 09:34'!
isConnectable
	"Answer whether I will accept connections from ConnectorMorphs"
	^false! !

!PasteUpMorph methodsFor: '*connectors' stamp: 'nk 1/26/2004 12:51'!
newMorphMenu
	(TheWorldMenu new
		world: self
		project: (self project ifNil: [Project current])       "mvc??"
		hand: ActiveHand) newMorph! !

!PasteUpMorph methodsFor: '*connectors-commands' stamp: 'sw 1/7/2004 02:50'!
tellAllContents: aMessageSelector
	"Send the given message selector to all the objects within the receiver"

	self submorphs do:
		[:m |
			m playerRepresented ifNotNilDo:
				[:p | p performScriptIfCan: aMessageSelector]]! !

!PasteUpMorph methodsFor: '*connectors-world menu-override' stamp: 'nk 7/18/2004 20:12'!
yellowButtonClickOnDesktopWithEvent: evt 
	"Put up either the personalized menu or the world menu when 
	the user clicks on the morphic desktop with the yellow button. 
	The preference 'personalizedWorldMenu' governs which one 
	is used"
	| aMenu |
	Preferences personalizedWorldMenu
		ifTrue: [aMenu := MenuMorph new defaultTarget: self.
			Preferences personalizeUserMenu: aMenu.
			aMenu addLine.
			aMenu
				add: 'personalize...' translated
				target: Preferences
				action: #letUserPersonalizeMenu]
		ifFalse: [aMenu := self buildWorldMenu: evt.
			aMenu addTitle: 'World' translated].
	Preferences cmdGesturesEnabled ifTrue: [
		self addGraphModelYellowButtonItemsTo: aMenu event: evt ].
	aMenu popUpEvent: evt in: self! !

!SequenceableCollection methodsFor: '*connectors-enumerating' stamp: 'nk 12/30/2003 15:39'!
groupsOf: n atATimeCollect: aBlock 
	"Evaluate aBlock with my elements taken n at a time. Ignore any 
	leftovers at the end. 
	Allows use of a flattened  
	array for things that naturally group into groups of n. 
	If aBlock has a single argument, pass it an array of n items, 
	otherwise, pass the items as separate arguments. 
	See also pairsDo:"
	| passArray args  |
	passArray := aBlock numArgs = 1.
	^(n
		to: self size
		by: n)
		collect: [:index | 
			args := (self copyFrom: index - n + 1 to: index) asArray.
			passArray
				ifTrue: [aBlock value: args]
				ifFalse: [aBlock valueWithArguments: args]]! !

!SequenceableCollection methodsFor: '*connectors-enumerating' stamp: 'nk 12/30/2003 15:37'!
groupsOf: n atATimeDo: aBlock 
	"Evaluate aBlock with my elements taken n at a time. Ignore any leftovers at the end.
	Allows use of a flattened 
	array for things that naturally group into groups of n.
	If aBlock has a single argument, pass it an array of n items,
	otherwise, pass the items as separate arguments.
	See also pairsDo:"
	| passArray args |
	passArray := (aBlock numArgs = 1).
	n
		to: self size
		by: n
		do: [:index | 
			args := (self copyFrom: index - n + 1 to: index) asArray.
			passArray ifTrue: [ aBlock value: args ]
				ifFalse: [ aBlock valueWithArguments: args ]].! !

!SequenceableCollection methodsFor: '*connectors-enumerating' stamp: 'nk 8/23/2003 10:42'!
nextToLast
	^self at: self size - 1! !

!SequenceableCollection methodsFor: '*connectors-enumerating' stamp: 'nk 7/17/2003 17:55'!
overlappingPairsCollect: aBlock 
	"Answer the result of evaluating aBlock with all of the overlapping pairs of my elements."
	| retval |
	retval := self species new: self size - 1.
	1 to: self size - 1
		do: [:i | retval at: i put: (aBlock value: (self at: i) value: (self at: i + 1)) ].
	^retval! !

!SequenceableCollection methodsFor: '*connectors-enumerating' stamp: 'nk 7/17/2003 17:55'!
overlappingPairsDo: aBlock 
	"Emit overlapping pairs of my elements into aBlock"

	1 to: self size - 1
		do: [:i | aBlock value: (self at: i) value: (self at: i + 1)]! !

!SequenceableCollection methodsFor: '*connectors-enumerating' stamp: 'nk 7/17/2003 17:54'!
overlappingPairsWithIndexDo: aBlock 
	"Emit overlapping pairs of my elements into aBlock, along with an index."

	1 to: self size - 1
		do: [:i | aBlock value: (self at: i) value: (self at: i + 1) value: i ]! !

!SequenceableCollection methodsFor: '*connectors-enumerating' stamp: 'nk 3/30/2002 12:45'!
paddedWith: otherCollection do: twoArgBlock 
	"Evaluate twoArgBlock with corresponding elements from this collection and otherCollection.
	Missing elements from either will be passed as nil."
	1 to: (self size max: otherCollection size) do:
		[:index | twoArgBlock value: (self at: index ifAbsent: [])
				value: (otherCollection at: index ifAbsent: [])]! !

!Morph class methodsFor: '*connectors-scripting' stamp: 'nk 9/10/2004 11:26'!
additionsToViewerCategoryEmbeddings
	"Answer viewer additions for the 'embeddings' category"
	"Vocabulary initialize"

	^#(
		embeddings 
		(
			(slot containingObject 'Me or my outermost container' Player readOnly Player getRootOwner unused unused)
			(slot pinCount 'A count of my embedded objects that want to act as pins for connections' Number readOnly Player getPinCount unused unused)
			(slot wantsConnectionWhenEmbedded 'Whether I should be act as a pin for connections when embedded in another object' Boolean readWrite Player getWantsConnectionWhenEmbedded Player setWantsConnectionWhenEmbedded:)
			(slot allowsConnectionToEmbeddedPins 'Whether objects embedded in me may be used as connection pins' Boolean readWrite Player getAllowsConnectionToEmbeddedPins Player setAllowsConnectionToEmbeddedPins:)
			(command tellAllPins: 'Send a message to all of my embedded objects that want to act as pins for connections' ScriptName)
			(command tellAllEmbeddedObjects: 'Send a message to all of my embedded objects' ScriptName)
		)
	)

! !

!Morph methodsFor: '*connectors-graphs' stamp: 'nk 1/7/2004 14:56'!
addAllPredecessorsTo: aSet
	| preds |
	preds := self predecessors.
	preds do: [ :pred | (aSet includes: pred) ifFalse: [ aSet add: pred. pred addAllPredecessorsTo: aSet ]]! !

!Morph methodsFor: '*connectors-graphs' stamp: 'nk 1/7/2004 14:59'!
addAllSuccessorsTo: aSet
	| successors |
	successors := self successors.
	successors do: [ :successor | (aSet includes: successor) ifFalse: [ aSet add: successor. successor addAllSuccessorsTo: aSet ]]! !

!Morph methodsFor: '*connectors-connection' stamp: 'nk 6/11/2002 15:23'!
addConnectedMorphsTo: aCollection
	"Add myself and all my connected Morphs (recursively) to aCollection"
	aCollection add: self.
	self connections do: [ :ea |
		ea ifNotNil: [
			(aCollection includes: ea)
				ifFalse: [ ea addConnectedMorphsTo: aCollection. ]
		]
	].
	self connectedMorphs do: [ :ea |
		ea ifNotNil: [
			(aCollection includes: ea)
				ifFalse: [ ea addConnectedMorphsTo: aCollection. ]
		]
	].
	^aCollection! !

!Morph methodsFor: '*connectors-attachment points' stamp: 'nk 2/3/2004 17:04'!
addHorizontalAttachmentPoints: n 
	| incr oldSpecs newSpec |
	n <= 0
		ifTrue: [^ self].
	incr := 1.0 / n.
	oldSpecs := self attachmentPointSpecs.
	self
		attachmentPointSpecs: (Array
				streamContents: [:s | 
					s nextPutAll: oldSpecs.
					(0 to: 1.0 by: incr)
						do: [:i | 
							newSpec := {#pointAtOffset:. 0.0 @ i}.
							(oldSpecs includes: newSpec)
								ifFalse: [s nextPut: newSpec].
							newSpec := {#pointAtOffset:. 1.0 @ i}.
							(oldSpecs includes: newSpec)
								ifFalse: [s nextPut: newSpec]]])! !

!Morph methodsFor: '*connectors-attachment points' stamp: 'nk 2/3/2004 17:04'!
addVerticalAttachmentPoints: n 
	| incr oldSpecs newSpec |
	n <= 0
		ifTrue: [^ self].
	incr := 1.0 / n.
	oldSpecs := self attachmentPointSpecs.
	self
		attachmentPointSpecs: (Array
				streamContents: [:s | 
					s nextPutAll: oldSpecs.
					(0 to: 1.0 by: incr)
						do: [:i | 
							newSpec := {#pointAtOffset:. i @ 0.0 }.
							(oldSpecs includes: newSpec)
								ifFalse: [s nextPut: newSpec].
							newSpec := {#pointAtOffset:. i @ 1.0 }.
							(oldSpecs includes: newSpec)
								ifFalse: [s nextPut: newSpec]]])! !

!Morph methodsFor: '*connectors-dropping/grabbing' stamp: 'nk 8/18/2004 12:20'!
alignAttachmentPointsWithGridNear: aPoint
	"Try to align my AP nearest the grid with the prevailing grid."
	| tweak minDist tweaks best aps |
	aps := self attachmentPoints.
	aps isEmpty ifTrue: [ ^self position: (self griddedPoint: self position) ].
	minDist := 1000.
	tweak := 0@0.
	tweaks := Bag new.
	best := 0@0.
	aps do: [ :ap | | dist |
		dist := ap dist: (self griddedPoint: ap).
		tweak := (self griddedPoint: ap) - ap.
		dist < minDist ifTrue: [ best := tweak ].
		tweaks add: tweak.
	].
	tweaks sortedCounts first key > 1
		ifTrue: [ tweak := tweaks sortedCounts first value ]
		ifFalse: [ tweak := best ].
	self position: (self position + tweak) asIntegerPoint.! !

!Morph methodsFor: '*connectors-attachment points' stamp: 'nk 8/19/2004 06:38'!
allAttachmentPointSpecs
	| specs |
	specs := OrderedCollection new.
	self
		allConnectableMorphsDo: [:m |
			(m ~~ self and: [ m wantsConnectionWhenEmbedded ])
				ifTrue: [specs addLast: {#attachmentPointForEmbeddedMorph:. m}]].

	"A policy decision here: if a Morph has embedded submorphs that want connection, then those will replace the default attachment points."

	(specs isEmpty or: [ self definesAttachmentPointSpecs ])
		ifTrue: [ specs addAll: self attachmentPointSpecs ].

	^ specs! !

!Morph methodsFor: '*connectors-attachment points' stamp: 'nk 8/18/2004 12:18'!
allAttachmentPoints
	^ self allAttachmentPointSpecs
		collect: [:ea |
			(self perform: ea first withArguments: ea allButFirst) asIntegerPoint ]! !

!Morph methodsFor: '*connectors-embeddings' stamp: 'nk 8/19/2004 15:08'!
allConnectableMorphsDo: aBlock 
	"Evaluate the given block for all morphs in this composite morph
	whose immediate owner allows connection to submorphs.
	Also evaluate the block for the receiver."

	self allowsConnectionToSubmorphs
		ifTrue: [ submorphs do: [:m | m allConnectableMorphsDo: aBlock]].

	aBlock value: self.
! !

!Morph methodsFor: '*connectors-embeddings' stamp: 'nk 8/19/2004 15:09'!
allConnectedMorphsDo: aBlock
	"Evaluate the given block for all morphs in this composite morph (including the receiver)
	that have connected constraints."

	self allConnectableMorphsDo: [ :m |
		m hasConnectedConstraints ifTrue: [ aBlock value: m ]]
! !

!Morph methodsFor: '*connectors-embeddings' stamp: 'nk 8/19/2004 15:01'!
allConnectedSubmorphsDo: aBlock
	"Evaluate the given block for all morphs in this composite morph (excluding the receiver)
	whose immediate owner allows connection to submorphs."

	self allConnectableMorphsDo: [ :m |
		(m ~~ self and: [ m hasConnectedConstraints ]) ifTrue: [
			aBlock value: m ]]
! !

!Morph methodsFor: '*connectors-graphs' stamp: 'nk 1/7/2004 14:57'!
allPredecessors
	| preds |
	preds := IdentitySet new.
	self addAllPredecessorsTo: preds.
	^preds
! !

!Morph methodsFor: '*connectors-graphs' stamp: 'nk 1/7/2004 14:58'!
allSuccessors
	| preds |
	preds := IdentitySet new.
	self addAllSuccessorsTo: preds.
	^preds
! !

!Morph methodsFor: '*connectors-embeddings' stamp: 'nk 8/9/2004 16:09'!
allowsConnectionToSubmorph: aMorph
	^aMorph wantsConnectionWhenEmbedded
! !

!Morph methodsFor: '*connectors-embeddings' stamp: 'nk 8/14/2004 12:54'!
allowsConnectionToSubmorphs
	"Answer true if I will allow a connector to connect to any of my submorphs.
	If not, the search for attachment points in embedded morphs will not be made."

	self wantsConnectionWhenEmbedded ifTrue: [ ^false ].

	^self valueOfProperty: #allowsConnectionToSubmorphs ifAbsent: [ true ]! !

!Morph methodsFor: '*connectors-connection' stamp: 'nk 7/12/2003 20:55'!
allowsDetachmentFromEnd: endIndex ofConnector: aConnector
	"Answer true if I would like the given end of aConnector detaching from me."
	^true! !

!Morph methodsFor: '*connectors-embeddings' stamp: 'nk 9/9/2004 12:31'!
allowsEmbeddingOf: aRectangleMorph 
	^true! !

!Morph methodsFor: '*connectors-dropping/grabbing' stamp: 'nk 6/14/2004 16:45'!
asButtonPrototype
	"I have been dropped on a NCMakerButton as a prototype. Return a duplicate of me ready to be used as a prototype"
	| newMe |
	newMe := self usableSiblingInstance2
		fullBounds;
		isPartsDonor: true;
		stopSteppingSelfAndSubmorphs;
		suspendEventHandler;
		formerOwner: nil;
		formerPosition: nil;
		removeProperty: #undoGrabCommand;
		disconnectAllConstraints;
		yourself.
	newMe graphModel ifNotNilDo: [ :gm | gm context: nil ].
	^newMe! !

!Morph methodsFor: '*connectors-connection' stamp: 'nk 8/14/2004 12:27'!
attachFrom: aConstraintMorph at: aPoint
	"If my property #movableAttachments is not set or is set to true, then make the constraintMorph attach to the nearest attachment point at any given instant.
	If the property #movableAttachments is set to false, then make the constraint attach
	permanently to the nearest attachment point."

	| spec |
	spec := self attachmentPointSpecNearest: aPoint.
	spec first == #attachmentPointForEmbeddedMorph:
		ifTrue: [aConstraintMorph connectToMorph: spec second usingSpec: {#attachmentPointWhenEmbedded}]
		ifFalse: [
	(self valueOfProperty: #movableAttachments ifAbsent: [ true ])
		ifTrue: [ aConstraintMorph connectToNearestAttachmentPoint ]
		ifFalse: [ self attachFrom: aConstraintMorph atNearestSpecTo: aPoint ]			]

! !

!Morph methodsFor: '*connectors-connection' stamp: 'nk 8/14/2004 12:25'!
attachFrom: aConstraintMorph atNearestSpecTo: aPoint 
	| spec |
	spec := self attachmentPointSpecNearest: aPoint.
	spec first == #attachmentPointForEmbeddedMorph:
		ifTrue: [aConstraintMorph connectToMorph: spec second usingSpec: {#attachmentPointWhenEmbedded}]
		ifFalse: [aConstraintMorph connectToMorph: self usingSpec: spec]! !

!Morph methodsFor: '*connectors-attachment points' stamp: 'nk 8/18/2004 12:20'!
attachmentPointAndSpecNearest: aPoint 
	"aPoint is in my coordinate system.
	Answer an Array with a Point, and a Spec.
	Note that in the case of multiple attachment points at the same position, this will favor the first one."
	| minDist minPoint minSpec dist point |
	minDist := 1e6.
	self allAttachmentPointSpecs
		do: [:spec |
			point := self perform: spec first withEnoughArguments: (spec allButFirst copyWith: aPoint).
			dist := aPoint dist: point.
			dist < minDist ifTrue: [
				minDist := dist.
				minPoint := point.
				minSpec := spec.
			]
		].
	^Array with: minPoint asIntegerPoint with: minSpec.! !

!Morph methodsFor: '*connectors-embeddings' stamp: 'nk 8/14/2004 13:50'!
attachmentPointForEmbeddedMorph: aMorph
	^(aMorph transformFrom: self) localPointToGlobal: aMorph attachmentPointWhenEmbedded
! !

!Morph methodsFor: '*connectors-attachment points' stamp: 'nk 9/29/2000 16:53'!
attachmentPointNearest: aPoint 
	^ (self attachmentPointAndSpecNearest: aPoint) first! !

!Morph methodsFor: '*connectors-attachment points' stamp: 'nk 9/29/2000 16:53'!
attachmentPointSpecNearest: aPoint 
		^ (self attachmentPointAndSpecNearest: aPoint) second! !

!Morph methodsFor: '*connectors-attachment points' stamp: 'nk 8/18/2004 12:23'!
attachmentPointSpecs
	^ self valueOfProperty: #attachmentPointSpecs
		ifAbsent: [ self defaultAttachmentPointSpecs ]! !

!Morph methodsFor: '*connectors-attachment points' stamp: 'nk 2/25/2001 17:00'!
attachmentPointSpecs: aCollectionOrNil
	^self setProperty: #attachmentPointSpecs toValue: aCollectionOrNil
! !

!Morph methodsFor: '*connectors-embeddings' stamp: 'nk 8/14/2004 13:47'!
attachmentPointWhenEmbedded
	^self  referencePosition! !

!Morph methodsFor: '*connectors-attachment points' stamp: 'nk 6/13/2002 10:10'!
attachmentPoints
	^ self attachmentPointSpecs
		collect: [:ea |
			(self perform: ea first withArguments: ea allButFirst) asIntegerPoint ]! !

!Morph methodsFor: '*connectors-attachment points' stamp: 'MAL 1/22/2005 14:28'!
attachmentSpecNearest: aPoint 
	^ (self attachmentPointAndSpecNearest: aPoint) second! !

!Morph methodsFor: '*connectors-geometry' stamp: 'nk 4/6/2001 07:31'!
bottomLeftCorner
	"Take rounded corners into account"
	^self wantsRoundedCorners
		ifTrue: [ bounds bottomLeft + (2@-2) ]
		ifFalse: [ bounds bottomLeft ]! !

!Morph methodsFor: '*connectors-geometry' stamp: 'nk 4/6/2001 07:30'!
bottomRightCorner
	"Take rounded corners into account"
	^self wantsRoundedCorners
		ifTrue: [ bounds bottomRight + (-2@-2) ]
		ifFalse: [ bounds bottomRight ]! !

!Morph methodsFor: '*connectors-labels' stamp: 'nk 8/14/2004 13:56'!
boundsSignatureHash
	"Answer a hash value that can be used to see if I've moved or been changed significantly"
	^self boundsInWorld hash
! !

!Morph methodsFor: '*connectors-geometry' stamp: 'nk 4/25/2001 21:45'!
centerX

	^ bounds center x! !

!Morph methodsFor: '*connectors-geometry' stamp: 'nk 4/25/2001 21:45'!
centerY

	^ bounds center y! !

!Morph methodsFor: '*connectors-geometry' stamp: 'nk 8/18/2004 12:55'!
closestOrthogonalPointTo: aPoint
	| xMax yMax |
	xMax := aPoint x.
	yMax := aPoint y.
	(aPoint y between: bounds top and: bounds bottom) ifTrue: [
		xMax := (aPoint x < bounds left)
			ifTrue: [ self pasteUpMorphOrWorld right ]
			ifFalse: [ self pasteUpMorphOrWorld left ].
	].
	(aPoint x between: bounds left and: bounds right) ifTrue: [
		yMax := (aPoint y < bounds top)
			ifTrue: [ self pasteUpMorphOrWorld bottom ]
			ifFalse: [ self pasteUpMorphOrWorld top ].
	].
	^(self firstIntersectionWithLineFrom: aPoint to: xMax@yMax)
		ifNil: [ self intersectionWithLineSegmentFromCenterTo: aPoint ].! !

!Morph methodsFor: '*connectors-geometry' stamp: 'nk 5/19/2003 20:37'!
closestPointTo: aPoint
	^self fullBounds pointNearestTo: aPoint! !

!Morph methodsFor: '*connectors-private' stamp: 'nk 3/11/2002 14:22'!
connectConstraint: aConstraint
	| oldConstraints |
	oldConstraints := self valueOfProperty: #connectedConstraints ifAbsentPut: [ WeakSet new ].
	oldConstraints add: aConstraint.! !

!Morph methodsFor: '*connectors-queries' stamp: 'nk 5/11/2003 06:40'!
connectedConstraints
	"Answer the constraints that have me as an input"
	| retval ccs |
	ccs := self valueOfProperty: #connectedConstraints ifAbsent: [ ^EmptyArray ].
	retval := IdentitySet new: ccs size.
	ccs do: [ :ea | ea notNil ifTrue: [ retval add: ea ]].
	retval isEmpty ifTrue: [ self removeProperty: #connectedConstraints ].
	^retval! !

!Morph methodsFor: '*connectors-queries' stamp: 'nk 6/15/2002 18:13'!
connectedMorphs
	"Answer all the morphs I am connected to (empty unless I'm a connector)"
	^EmptyArray! !

!Morph methodsFor: '*connectors-embeddings' stamp: 'nk 8/19/2004 15:03'!
connectedSubmorphs
	"Answer a collection of all of my submorphs that have connected constraints."

	^Array streamContents: [ :s | self allConnectedSubmorphsDo: [ :m | s nextPut: m ]]
! !

!Morph methodsFor: '*connectors-notifications' stamp: 'nk 1/20/2004 18:45'!
connectedToBy: aConnector
	"notification upon connection by a NCConnectorMorph"
! !

!Morph methodsFor: '*connectors-notifications' stamp: 'nk 1/20/2004 19:09'!
connectedToEnd: oneOrTwo ofConnector: aConnector 
	"notification upon connection by a NCConnectorMorph.
	If either I or aConnector have a model, ensure that we both have a model and then notify my model."
	| gm |
	gm := self graphModel.
	gm
		ifNil: [aConnector graphModel
				ifNotNilDo: [:cm | self assuredGraphModel connectedToEnd: oneOrTwo ofEdge: cm]]
		ifNotNil: [gm connectedToEnd: oneOrTwo ofEdge: aConnector assuredGraphModel].
		

	self connectedToBy: aConnector! !

!Morph methodsFor: '*connectors-connection' stamp: 'nk 6/14/2002 16:18'!
connectionTarget
	^self
! !

!Morph methodsFor: '*connectors-attachment points' stamp: 'nk 8/4/2004 17:02'!
connectionTargetsWithin: delta ofPoint: globalPoint forEnd: endIndex ofConnector: aConnector do: aBlock
	"For me and each of my submorphs in a square with width of delta*2 of globalPoint
	that might like to have the given end (endIndex is 1 or 2) of aConnector attached,
	evaluate aBlock."

	| extent localPoint vicinity |
	extent := (delta * 2) asPoint.
	self allMorphsDo: [ :m |
		localPoint := m globalPointToLocal: globalPoint.
		vicinity := Rectangle center: localPoint extent: extent.
		(m isLocked not
			and: [ m visible 
			and: [ (m fullBounds intersects: vicinity) "or (m isShadowVisibleIn: vicinity)"
			and: [ (m wantsAttachmentFromEnd: endIndex ofConnector: aConnector)
			and: [ (aConnector wantsToAttachEnd: endIndex toMorph: m) ]]]])
			ifTrue: [ aBlock value: m ] ].
! !

!Morph methodsFor: '*connectors-queries' stamp: 'nk 8/19/2004 15:19'!
connections
	"Answer the NCConnectorMorphs that are tied to me or my submorphs."
	| retval |
	retval := IdentitySet new.
	self connectionsDo: [ :c | retval add: c ].
	^retval! !

!Morph methodsFor: '*connectors-queries' stamp: 'nk 12/26/2003 16:30'!
connectionsAtAttachmentPoint: aPoint
	"Answer a collection of all the connections to the given point (if any)"
	| retval |
	retval := Array streamContents: [ :s |
		self connections do: [ :connection |
			connection constraints do: [ :constraint |
				(constraint notNil and: [
					self flag: #TODO.	"coordinate transform"
					constraint lastTargetPoint asIntegerPoint = aPoint asIntegerPoint ])
					ifTrue: [ s nextPut: connection ]
			]
		]].
	^retval
! !

!Morph methodsFor: '*connectors-queries' stamp: 'nk 8/18/2004 12:21'!
connectionsAtAttachmentPointNumbered: anInteger
	"Answer a collection of all the connections to the given point (if any)"
	^self connectionsAtAttachmentPoint: (self allAttachmentPoints at: anInteger ifAbsent: [ ^#() ])! !

!Morph methodsFor: '*connectors-queries' stamp: 'nk 8/19/2004 15:15'!
connectionsDo: aBlock
	"Evaluate aBlock with all the NCConnectorMorphs that are connected to me directly or to any of my submorphs."
	self allConnectableMorphsDo: [ :m | m myConnectionsDo: aBlock ]! !

!Morph methodsFor: '*connectors-queries' stamp: 'nk 3/11/2002 14:13'!
constraints
	"Answer a Set of all of my submorphs that are constraints"

	| retval |
	retval := Set new.
	self allMorphsDo: [ :ea | ea isConstraint ifTrue: [ retval add: ea ] ].
	^retval! !

!Morph methodsFor: '*connectors-attachment points' stamp: 'nk 1/29/2004 09:31'!
createHorizontalAttachmentPoints: n
	| incr |
	incr := 1.0 / n.
	self attachmentPointSpecs: (Array streamContents: [ :s | (0 to: 1.0 by: incr) do: [ :i | 
		s nextPut: {#pointAtOffset: . 0 @ i }.
				s nextPut: {#pointAtOffset: . 1 @ i } ]])! !

!Morph methodsFor: '*connectors-attachment points' stamp: 'nk 1/29/2004 09:31'!
createVerticalAttachmentPoints: n
	| incr |
	incr := 1.0 / n.
	self attachmentPointSpecs: (Array streamContents: [ :s | (0 to: 1.0 by: incr) do: [ :i | 
		s nextPut: {#pointAtOffset: . i @ 0 }.
				s nextPut: {#pointAtOffset: . i @ 1 } ]])! !

!Morph methodsFor: '*connectors-attachment points' stamp: 'nk 8/18/2004 20:44'!
defaultAttachmentPointSpecs
	^{ 
		{ #topLeftCorner } .
		{ #topCenter } .
		{ #topRightCorner } .
		{ #leftCenter } .
"		{ #center } ."
		{ #rightCenter } .
		{ #bottomLeftCorner } .
		{ #bottomCenter } .
		{ #bottomRightCorner }
	}
! !

!Morph methodsFor: '*connectors-attachment points' stamp: 'nk 8/19/2004 06:36'!
definesAttachmentPointSpecs
	^ self hasProperty: #attachmentPointSpecs
! !

!Morph methodsFor: '*connectors-labels' stamp: 'nk 6/22/2002 17:52'!
deleteAllLabels
	"delete all labels from me and my submorphs"
	self allMorphsDo: [ :ea | ea deleteMorphs: ea labels  ].! !

!Morph methodsFor: '*connectors-submorphs-add/remove' stamp: 'nk 2/26/2001 23:50'!
deleteMorphs: aCollection
	aCollection do: [ :ea | ea delete ]! !

!Morph methodsFor: '*connectors-graphs' stamp: 'nk 5/10/2003 09:23'!
depthFirstSearchVisitNodeBlock: nodeBlock edgeBlock: edgeBlock states: nodeStates postOrder: postOrder
	"nodeBlock is called with the node (Morph).
	edgeBlock is called with the edge (NCConnectorMorph) and one of:
		#unexplored -- dest morph has not yet been seen
		#processed -- dest morph has been seen
		#frontier -- this edge is a back edge (i.e. creates a cycle)
	If postOrder is true, the node block will be called after visiting a node.
	"

	| state |
	state := nodeStates at: self ifAbsentPut: [ #unexplored ].
	postOrder ifFalse: [ nodeBlock value: self ].
	nodeStates at: self put: #frontier.
	self outgoingConnections do: [:conn | 
		| neighbor | 
		neighbor := conn destinationMorph.
		neighbor ifNotNil: [
			state := (nodeStates at: neighbor ifAbsentPut: [ #unexplored ]).
			edgeBlock value: conn value: state.
			state == #unexplored
				ifTrue: [neighbor
							depthFirstSearchVisitNodeBlock: nodeBlock
							edgeBlock: edgeBlock
							states: nodeStates
							postOrder: postOrder].
		]
	].
	nodeStates at: self put: #processed.
	postOrder ifTrue: [ nodeBlock value: self ].
! !

!Morph methodsFor: '*connectors-graphs' stamp: 'nk 5/15/2003 09:56'!
depthFirstSearchVisitPre: nodeBlock post: postNodeBlock edgeBlock: edgeBlock states: nodeStates 
	"nodeBlock is called on first encountering the node (Morph). 
	edgeBlock is called with the edge (NCConnectorMorph) and one of: 
	#unexplored -- dest morph has not yet been seen 
	#processed -- dest morph has been seen 
	#frontier -- this edge is a back edge (i.e. creates a cycle) 
	If postNodeBlock is non-nil, it will be called after visiting a node. 
	"
	| state |
	state := nodeStates
				at: self
				ifAbsentPut: [#unexplored].
	nodeBlock
		ifNotNil: [nodeBlock value: self].
	nodeStates at: self put: #frontier.
	self outgoingConnections
		do: [:conn | 
			| neighbor | 
			neighbor := conn destinationMorph.
			neighbor
				ifNotNil: [state := nodeStates
								at: neighbor
								ifAbsentPut: [#unexplored].
					edgeBlock
						ifNotNil: [edgeBlock value: conn value: state].
					state == #unexplored
						ifTrue: [neighbor
								depthFirstSearchVisitPre: nodeBlock
								post: postNodeBlock
								edgeBlock: edgeBlock
								states: nodeStates]]].
	nodeStates at: self put: #processed.
	postNodeBlock
		ifNotNil: [postNodeBlock value: self]! !

!Morph methodsFor: '*connectors-connection' stamp: 'nk 8/19/2004 15:09'!
disconnectAllConstraints
	self allConnectedMorphsDo: [ :m | m disconnectAllMyConstraints ].! !

!Morph methodsFor: '*connectors-connection' stamp: 'nk 8/19/2004 15:04'!
disconnectAllMyConstraints
	self connectedConstraints do: [ :ea | ea inputs: (ea inputs copyWithout: self) ]! !

!Morph methodsFor: '*connectors-private' stamp: 'nk 3/11/2002 14:48'!
disconnectConstraint: aConstraint
	| oldConstraints |
	oldConstraints := self valueOfProperty: #connectedConstraints ifAbsent: [ ^self ].
	oldConstraints := oldConstraints reject: [ :ea | ea == aConstraint or: [ ea isNil ]].
	oldConstraints isEmpty ifTrue: [ self removeProperty: #connectedConstraints. ^self ].
	self setProperty: #connectedConstraints toValue: oldConstraints.! !

!Morph methodsFor: '*connectors-notifications' stamp: 'nk 1/20/2004 18:43'!
disconnectedFromBy: aConnector
	"notification upon disconnection by a NCConnectorMorph"! !

!Morph methodsFor: '*connectors-notifications' stamp: 'nk 1/20/2004 19:09'!
disconnectedFromEnd: oneOrTwo ofConnector: aConnector
	"notification upon disconnection by a NCConnectorMorph.
	If either I or aConnector have a model, ensure that we both have a model and then notify my model."

	| gm |
	gm := self graphModel.
	gm
		ifNil: [aConnector graphModel
				ifNotNilDo: [:cm | self assuredGraphModel disconnectedFromEnd: oneOrTwo ofEdge: cm]]
		ifNotNil: [ gm disconnectedFromEnd: oneOrTwo ofEdge: aConnector assuredGraphModel].

	self disconnectedFromBy: aConnector.! !

!Morph methodsFor: '*connectors-copying' stamp: 'nk 6/14/2004 16:31'!
duplicateForEmbedding
	"This is sent to (for instance) the prototype in a NCMakerButton.
	I'm not calling duplicate here (though asButtonPrototype calls duplicate) to save time;
	the resulting object is a sibling."

	^self usableSiblingInstance2 restoreSuspendedEventHandler; isPartsDonor: false; yourself
! !

!Morph methodsFor: '*connectors-attachment points' stamp: 'nk 8/21/2004 08:17'!
embeddedPinsDo: aBlock
	self
		allConnectableMorphsDo: [:m |
			(m ~~ self and: [ m wantsConnectionWhenEmbedded ])
				ifTrue: [ aBlock value: m ]].
! !

!Morph methodsFor: '*connectors-attachment points' stamp: 'nk 9/10/2004 11:17'!
embeddedPlayersDo: aBlock
	self allMorphsWithPlayersDo: [ :m :p | m ~~ self ifTrue: [ aBlock value: p ]]! !

!Morph methodsFor: '*connectors-connection' stamp: 'nk 7/18/2003 17:00'!
endShapeBorderColor: aColor
	self borderColor: aColor! !

!Morph methodsFor: '*connectors-connection' stamp: 'nk 7/18/2003 17:17'!
endShapeBorderWidth: aWidth
	self borderWidth: aWidth! !

!Morph methodsFor: '*connectors-connection' stamp: 'nk 7/18/2003 17:01'!
endShapeColor: aColor
	self fillStyle: aColor! !

!Morph methodsFor: '*connectors-connection' stamp: 'nk 6/30/2002 21:05'!
endShapeWidth: aWidth
	| originalWidth originalExtent where newExtent |
	where := self referencePosition.
	originalWidth := self valueOfProperty: #originalWidth ifAbsentPut: [ 2 ].
	originalExtent := self valueOfProperty: #originalExtent ifAbsentPut: [ self extent ].
	newExtent := (originalExtent * aWidth / originalWidth).
	self extent: newExtent.
	self referencePosition: where.
! !

!Morph methodsFor: '*connectors-submorphs-add/remove' stamp: 'nk 8/18/2003 12:12'!
ensureMorph: back behind: front
	"Add a morph to the list of submorphs behind the specified morph,
	but only if it isn't already behind.
	Assume that it is more common for newMorph to already be in front"
	| backIndex i |
	backIndex := 0.
	i := submorphs size.
	"back to front search"
	submorphs reverseDo: [ :ea |
		ea == back
			ifTrue: [ backIndex := i ].
		ea == front
			ifTrue: [ backIndex = 0
				ifFalse: [ ^self "already in front" ]
				ifTrue: [ ^self privateAddMorph: back atIndex: i + 1  ]
			].
		i := i - 1.
 	].
! !

!Morph methodsFor: '*connectors-submorphs-add/remove' stamp: 'nk 3/3/2001 11:35'!
ensureMorph: front inFrontOf: back
	"Add a morph to the list of submorphs in front of the specified morph,
	but only if it isn't already in front. Assume that it is more common for newMorph to already be in front"
	| frontIndex |
	frontIndex := 0.
	"front to back search"
	submorphs withIndexDo: [ :ea :i |
		ea == front
			ifTrue: [ frontIndex := i ].
		ea == back
			ifTrue: [ frontIndex = 0
				ifFalse: [ ^self "already in front" ]
				ifTrue: [ ^self privateAddMorph: front atIndex: i  ]
			].
 	].
! !

!Morph methodsFor: '*connectors-geometry' stamp: 'nk 6/21/2003 13:29'!
firstIntersectionWithLineFrom: start to: end
	|  intersections |
	intersections := self fullBounds intersectionsWithLineFrom: start to: end.
	intersections isEmpty ifTrue: [ ^nil ].
	^intersections detectMin: [ :ea | ea dist: start ] ! !

!Morph methodsFor: '*connectors-queries' stamp: 'nk 8/19/2004 15:01'!
hasConnectedConstraints
	"Answer whether I have any connected constraints"
	| ccs |
	ccs := self valueOfProperty: #connectedConstraints ifAbsent: [ ^false ].
	^ccs anySatisfy: [ :c | c notNil ]! !

!Morph methodsFor: '*connectors-private' stamp: 'nk 9/10/2004 11:46'!
highlightForConnection: aBoolean 
	| highlights |
	(self valueOfProperty: #isHighlight ifAbsent: [false]) ifTrue: [^self].
	highlights := self valueOfProperty: #connectionHighlights ifAbsent: [].
	self flag: #TODO.	"use contrasting colors"
	aBoolean 
		ifTrue: 
			["self highlightForConnection: false."

			self setProperty: #highlightedForConnection toValue: aBoolean.
			highlights ifNil: 
					[| outerBorder aps pasteUp relativeFullBounds |
					pasteUp := self world.
					aps := self allAttachmentPoints collect: [:ea | self point: ea in: pasteUp].
					highlights := OrderedCollection new: 4 + aps size.
					relativeFullBounds := self bounds: self fullBounds in: pasteUp.
					outerBorder := relativeFullBounds outsetBy: 3.
					highlights 
						addAll: ({ 
								relativeFullBounds left @ outerBorder top 
									corner: outerBorder right @ relativeFullBounds top.
								relativeFullBounds topRight corner: outerBorder bottomRight.
								outerBorder left @ relativeFullBounds bottom 
									corner: relativeFullBounds right @ outerBorder bottom.
								outerBorder topLeft corner: relativeFullBounds bottomLeft} 
									collect: [:rect | self highlightMorphForBoundaryRect: rect]).
					highlights addAll: (aps collect: [:ap | self highlightMorphForAPRect: ap]).
					"self pasteUpMorph addAllMorphs: highlights"
					highlights do: [:ea | pasteUp addMorphFront: ea].
					self setProperty: #connectionHighlights toValue: highlights]]
		ifFalse: 
			[highlights 
				ifNotNil: [self pasteUpMorphOrWorld deleteMorphs: highlights].
			self removeProperty: #connectionHighlights.
			self removeProperty: #highlightedForConnection.
			self resetExtension]! !

!Morph methodsFor: '*connectors-private' stamp: 'nk 5/12/2003 21:22'!
highlightMorphForAPRect: ap 
	^(NCHighlightMorph new)
		bounds: (Rectangle center: ap extent: 8 @ 8);
		color: (Color blue alpha: 0.4);
		lock: true;
		yourself! !

!Morph methodsFor: '*connectors-private' stamp: 'nk 5/12/2003 21:21'!
highlightMorphForBoundaryRect: rect 
	^(NCHighlightMorph new)
		bounds: rect;
		color: (Color green alpha: 0.3);
		lock: true;
		yourself! !

!Morph methodsFor: '*connectors-private' stamp: 'nk 4/19/2002 11:11'!
highlightedForConnection
	^self valueOfProperty: #highlightedForConnection ifAbsent: [ false ]
! !

!Morph methodsFor: '*connectors-graphs' stamp: 'nk 5/9/2003 19:28'!
immediatelyFollows: aMorph
	^self incomingConnections anySatisfy: [ :conn | conn startConnection  == aMorph ]! !

!Morph methodsFor: '*connectors-graphs' stamp: 'nk 5/9/2003 19:28'!
immediatelyPrecedes: aMorph
	^self outgoingConnections anySatisfy: [ :conn | conn endConnection  == aMorph ]! !

!Morph methodsFor: '*connectors-queries' stamp: 'nk 8/21/2004 09:22'!
incomingConnectionCount
	| retval |
	retval := 0.
	self connectionsDo: [ :conn | conn destinationMorph == self ifTrue: [ retval := retval + 1 ]].
	^retval! !

!Morph methodsFor: '*connectors-queries' stamp: 'nk 5/11/2003 08:49'!
incomingConnections
	| retval |
	retval := IdentitySet new.
	self connectionsDo: [ :conn | conn destinationMorph == self ifTrue: [ retval add: conn ]].
	^retval! !

!Morph methodsFor: '*connectors-geometry' stamp: 'nk 6/21/2003 13:58'!
intersectionWithLineSegmentFromCenterTo: aPoint 
	^(self firstIntersectionWithLineFrom: aPoint to: self center)
		ifNil: [ self closestPointTo: aPoint ]! !

!Morph methodsFor: '*connectors-testing' stamp: 'nk 12/26/2003 17:33'!
isButton
	^false! !

!Morph methodsFor: '*connectors-connection' stamp: 'nk 6/13/2002 09:40'!
isConnectable
	"Answer whether I will accept connections from ConnectorMorphs"
	^true
! !

!Morph methodsFor: '*connectors-classification' stamp: 'nk 3/12/2002 09:27'!
isConnector
	^false! !

!Morph methodsFor: '*connectors-connection' stamp: 'nk 2/23/2001 14:37'!
isConstrainedBy: aMorph
	"Answer true if aMorph or any of its submorphs are an input to any of my constraints"
	| herSubmorphs |
	herSubmorphs := aMorph allMorphs.
	^self constraints anySatisfy: [ :constraint | constraint hasAnyInputIn: herSubmorphs ]
! !

!Morph methodsFor: '*connectors-queries' stamp: 'nk 2/23/2001 13:42'!
isConstraint
	^false! !

!Morph methodsFor: '*connectors-testing' stamp: 'nk 8/18/2003 09:15'!
isCurveHandle
	^false! !

!Morph methodsFor: '*connectors-geometry' stamp: 'nk 5/19/2003 20:38'!
isEntirelyWithinOutlineOf: aMorph

	| itsBounds |
	itsBounds := aMorph fullBounds.
	(itsBounds containsRect: self fullBounds) ifFalse: [ ^false ].
	^self isEntirelyWithinShadowForm: aMorph shadowForm! !

!Morph methodsFor: '*connectors-geometry' stamp: 'nk 10/17/2003 15:54'!
isEntirelyWithinShadowForm: itsShadow

	| myShadow xorForm retval box |
	myShadow := self shadowForm.
	box := (myShadow boundingBox translateBy: myShadow offset)
		quickMerge:  (itsShadow boundingBox translateBy: itsShadow offset).
	xorForm := Form extent: box extent depth: 1.
	xorForm copy: (self fullBounds translateBy: box origin negated)
		from: 0@0
		in: myShadow
		rule: Form over.
	xorForm
		copyBits: (itsShadow boundingBox translateBy: (itsShadow offset - box origin) negated) from: itsShadow at: 0@0 clippingBox: xorForm boundingBox rule: Form erase fillColor: nil.
	retval := xorForm primCountBits = 0.
	"retval ifFalse: [ xorForm asMorph position: box origin; openInWorld ]."
	^retval ! !

!Morph methodsFor: '*connectors-geometry' stamp: 'nk 5/19/2003 20:39'!
isEntirelyWithinShadowForm: itsShadow bounds: itsBounds 
	(itsBounds containsRect: self fullBounds) ifFalse: [^false].
	^self isEntirelyWithinShadowForm: itsShadow! !

!Morph methodsFor: '*connectors-testing' stamp: 'nk 6/12/2004 16:20'!
isHandleMorph
	"Answer true if I am handle-like"
	^false! !

!Morph methodsFor: '*connectors-private' stamp: 'nk 11/26/2002 11:25'!
isHighlight
	^false! !

!Morph methodsFor: '*connectors-testing' stamp: 'nk 8/26/2003 15:57'!
isObjectsTool
	^false! !

!Morph methodsFor: '*connectors-testing' stamp: 'nk 8/31/2004 12:29'!
isPinMorph
	^false! !

!Morph methodsFor: '*connectors-geometry' stamp: 'nk 6/22/2003 12:37'!
isShadowVisibleIn: aRectangle
	"aRectangle is in my coordinate system"
	| myShadow patch |
	myShadow := self shadowForm.
	patch := myShadow contentsOfArea: (aRectangle translateBy: bounds origin negated).
	^patch primCountBits ~= 0.
! !

!Morph methodsFor: '*connectors-testing' stamp: 'nk 9/4/2004 17:41'!
isStringMorph
	^false! !

!Morph methodsFor: '*connectors-labels' stamp: 'nk 3/11/2002 21:22'!
labels
	"Answer the NCLabelMorphs that are tied to me"
	^ self connectedConstraints select: [ :ea | ea isLabelConstraint ] thenCollect: [ :ea | ea drawingElement ]! !

!Morph methodsFor: '*connectors-connection' stamp: 'nk 4/8/2001 08:18'!
lineAttachmentOffset
	"Answer the relative offset from my rotation center at which a line is to attach if I am being used as a line end adornment"
	^self valueOfProperty: #lineAttachmentOffset ifAbsent: [ 0@0 ].! !

!Morph methodsFor: '*connectors-connection' stamp: 'nk 4/8/2001 08:18'!
lineAttachmentOffset: aPoint
	"Set the relative offset from my rotation center at which a line is to attach if I am being used as a line end adornment"
	^self setProperty: #lineAttachmentOffset toValue: aPoint.! !

!Morph methodsFor: '*connectors-buttons' stamp: 'nk 9/1/2004 20:01'!
makerButtonImage
	"Answer a Form for use on a button"
	^self makerButtonImageOfColor: Color transparent.! !

!Morph methodsFor: '*connectors-buttons' stamp: 'nk 9/1/2004 20:17'!
makerButtonImageOfColor: aColor
	"Answer a Form for use on a button"
	^(self imageForm: 32
		backgroundColor: aColor
		forRectangle: self fullBounds)
				replaceColor: aColor
				withColor: Color transparent! !

!Morph methodsFor: '*connectors-attachment points' stamp: 'nk 4/25/2001 21:35'!
movableAttachments
	^self valueOfProperty: #movableAttachments ifAbsent: [ true ]! !

!Morph methodsFor: '*connectors-attachment points' stamp: 'nk 3/27/2001 12:04'!
movableAttachments: aBoolean
	"If aBoolean is true, then allow a connection to choose the nearest of my attachment points.
	If aBoolean if false, then the attachment point chosen at connection time will be permanent"
	aBoolean
		ifTrue: [ self removeProperty: #movableAttachments ]
		ifFalse: [ self setProperty: #movableAttachments toValue: false ]! !

!Morph methodsFor: '*connectors-dropping/grabbing' stamp: 'nk 1/7/2004 21:29'!
movedFromFormerPosition
	"Answer the distance that I have been moved from my former position, as a Point"
	| formerPos |
	(formerPos := self formerPosition) ifNil: [ ^0@0 ].
	^self position - formerPos! !

!Morph methodsFor: '*connectors-queries' stamp: 'nk 8/21/2004 09:32'!
myConnectionsDo: aBlock
	"Evaluate aBlock with all the unique NCConnectorMorphs that are connected directly to me."
	| ccs totalConstraints allConnectors |
	ccs := self valueOfProperty: #connectedConstraints ifAbsent: [ ^self ].
	totalConstraints := 0.
	allConnectors := IdentitySet new.
	ccs do: [ :ea | ea notNil
		ifTrue: [ totalConstraints := totalConstraints + 1.
			ea isLineConstraint ifTrue: [ allConnectors add: ea drawingElement ]]].
	allConnectors do: aBlock.
	totalConstraints isZero ifTrue: [ self removeProperty: #connectedConstraints ].
! !

!Morph methodsFor: '*connectors-buttons' stamp: 'nk 9/2/2004 08:41'!
nameForButton
	^self valueOfProperty: #nameForButton ifAbsent: [ self externalName ]! !

!Morph methodsFor: '*connectors-buttons' stamp: 'nk 9/2/2004 08:41'!
nameForButton: aString
	self setProperty:  #nameForButton  toValue: aString! !

!Morph methodsFor: '*connectors-graphs' stamp: 'nk 5/10/2003 21:55'!
neighbors
	| retval |
	retval := IdentitySet new.
	self incomingConnections do: [ :ea | ea sourceMorph
		ifNotNilDo: [ :sm | retval add: sm ]].
	self outgoingConnections do: [ :conn | conn destinationMorph
		ifNotNilDo: [ :dm | retval add: dm ]].
	^retval! !

!Morph methodsFor: '*connectors-attachment points' stamp: 'nk 8/19/2004 06:45'!
nonDefaultAttachmentPointSpecs
	^ self valueOfProperty: #attachmentPointSpecs
		ifAbsent: [ #() ]! !

!Morph methodsFor: '*connectors-labels' stamp: 'nk 4/14/2002 10:26'!
nudgeForLabel: labelBounds
	"Make label not overlap me. Answer an increment, or 0@0 if OK"
	| flags intersection |
	intersection := labelBounds intersect: self bounds.
	intersection hasPositiveExtent ifFalse: [ ^0@0 ].
	intersection = labelBounds ifTrue: [ ^0@-1 ].
	flags := 0.
	intersection top == self top ifTrue: [ flags := flags + 1 ].
	intersection right == self right ifTrue: [ flags := flags + 2 ].
	intersection bottom == self bottom ifTrue: [ flags := flags + 4 ].
	intersection left == self left ifTrue: [ flags := flags + 8 ].
	^flags caseOf: {
	"no intersection"
			[ 0 ] -> [ 0@0 ].
	"2 adjacent sides only" 
			[ 9 ] -> [ 1@1 ].
			[ 3 ] -> [ -1@1 ].
			[ 12 ] -> [ 1@-1 ].
			[ 6 ] -> [ -1@-1 ].
	"2 opposite sides only" 
			[ 10 ] -> [ 0@-1 ].
			[ 5 ] -> [ 1@0 ].
	"only 1 side" 
			[ 8 ] -> [ -1@0 ].
			[ 1 ] -> [ 0@-1 ].
			[ 2 ] -> [ 1@0 ].
			[ 4 ] -> [ 0@1 ].
	"3 sides" 
			[ 11 ] -> [ 0@1 ].
			[ 13 ] -> [ 1@0 ].
			[ 14 ] -> [ 0@-1 ].
		[ 7 ] -> [ -1@0 ].
 "all sides" 
		[ 15 ] -> [ 1@-1 "move up and to the right" ].
	}.
! !

!Morph methodsFor: '*connectors-queries' stamp: 'nk 8/21/2004 09:23'!
outgoingConnectionCount
	| retval |
	retval := 0.
	self connectionsDo: [ :conn | conn sourceMorph == self ifTrue: [ retval := retval + 1 ]].
	^retval! !

!Morph methodsFor: '*connectors-queries' stamp: 'nk 5/11/2003 08:49'!
outgoingConnections
	| retval |
	retval := IdentitySet new.
	self connectionsDo: [ :conn | conn sourceMorph == self ifTrue: [ retval add: conn ]].
	^retval! !

!Morph methodsFor: '*connectors-geometry' stamp: 'nk 5/19/2003 20:39'!
overlapsOutlineOf: aMorph 
	"Answer true if aMorph's shadow and my shadow overlap at all"
	^self overlapsShadowForm: aMorph shadowForm bounds: aMorph fullBounds.! !

!Morph methodsFor: '*connectors-attachment points' stamp: 'nk 8/21/2004 08:31'!
pinCount
	| numberOfPins |
	numberOfPins := 0.
	self embeddedPinsDo: [ :m | numberOfPins := numberOfPins + 1 ].
	^numberOfPins
! !

!Morph methodsFor: '*connectors-geometry' stamp: 'nk 3/8/2001 10:12'!
pointAtOffset: aPoint
	^bounds topLeft + ((aPoint x * bounds width) @ (aPoint y * bounds height))! !

!Morph methodsFor: '*connectors-graphs' stamp: 'nk 5/12/2003 14:03'!
predecessors
	| retval |
	retval := IdentitySet new.
	self incomingConnections do: [ :ea | ea sourceMorph
		ifNotNilDo: [ :sm | sm isHandMorph ifFalse: [ retval add: sm ]]].
	^retval 
		! !

!Morph methodsFor: '*connectors-connection' stamp: 'nk 4/2/2001 19:49'!
preferredConnection
	^self valueOfProperty: #preferredConnection ifAbsent: []! !

!Morph methodsFor: '*connectors-connection' stamp: 'nk 3/27/2001 12:53'!
preferredConnection: anArrayOrSelector 
	^ self
		setProperty: #preferredConnection
		toValue: ((anArrayOrSelector isKindOf: Symbol)
				ifTrue: [{anArrayOrSelector}]
				ifFalse: [anArrayOrSelector])! !

!Morph methodsFor: '*connectors-connection' stamp: 'nk 4/5/2001 09:23'!
preferredConnectionSelector
	| pref |
	pref := self valueOfProperty: #preferredConnection ifAbsent: [ ^nil ].
	^pref first! !

!Morph methodsFor: '*connectors-gestures' stamp: 'nk 1/7/2004 12:19'!
preferredConnectorFor: ends ortho: isOrtho curved: isCurved
	"Answer a connector if I have any preference. The default is to return nil, which will cause the default world behavior to be run (typically a prototype duplication)"
	^nil! !

!Morph methodsFor: '*connectors-gestures' stamp: 'nk 10/28/2003 08:38'!
preferredConnectorName
	^'Connector'! !

!Morph methodsFor: '*connectors-gestures' stamp: 'nk 1/26/2004 12:05'!
prototypeContainer
	"Return a probable prototype container.
	- local flaps named 'Prototypes'
	- global flaps named 'Prototypes'
	- global flaps named 'Prototypes1' etc.
	- parts bins named 'Prototypes'
	- current world
	"
	| flap |
	flap := self currentWorld localFlapTabs
				detect: [:tab | tab flapID sameAs: 'Prototypes']
				ifNone: [Flaps globalFlapTabWithID: 'Prototypes'].
	^flap
				ifNil: [ (self currentWorld submorphNamed: 'Prototypes') ifNil: [ self currentWorld ] ]
				ifNotNil: [flap referent].
! !

!Morph methodsFor: '*connectors-gestures' stamp: 'nk 1/26/2004 12:07'!
prototypesSuchThat: selectionBlock ifNone: defaultBlock 
	"Return zero or more appropriate prototypes found in a Prototypes 
	flap (local, global), or in  
	the World, or make one from a class, or call defaultBlock with the 
	protoName and  
	gesture for a default"
	| container protos |
	container := self prototypeContainer.
	protos := container submorphs
				collect: [:m | m partRepresented renderedMorph]
				thenSelect: selectionBlock.
	protos isEmpty
		ifTrue: [ ^defaultBlock value ].
	^ protos! !

!Morph methodsFor: '*connectors-gestures' stamp: 'nk 1/27/2004 18:02'!
prototypesSuchThat: selectionBlock sortedBy: blockList ifNone: defaultBlock
	| allSelections sortedSelections |
	allSelections := self prototypesSuchThat: selectionBlock ifNone: [ #() ].
	sortedSelections := OrderedCollection new.
	blockList do: [ :block |
		sortedSelections addAllLast: ((allSelections select: block) copyWithoutAll: sortedSelections)
	].
	sortedSelections isEmpty ifTrue: [ ^defaultBlock value ].
	^sortedSelections! !

!Morph methodsFor: '*connectors-labels' stamp: 'nk 4/14/2002 13:40'!
relocateLabelFrom: aRectangle
	"Given a prospective label location, answer a new rectangle which doesn't overlap me"
	| newBounds nudge steps |
	newBounds := aRectangle.
	steps := aRectangle extent r truncated.
	[ (nudge := self nudgeForLabel: newBounds) isZero or: [ steps < 0 ] ]
		whileFalse: [ newBounds := newBounds translateBy: nudge.
			steps := steps - 1 ].
	steps < 0 ifTrue: [ ^aRectangle ].
	^ newBounds
! !

!Morph methodsFor: '*connectors-structure' stamp: 'nk 8/22/2004 05:48'!
rootInPasteUp
	"Answer myself or my containing morph up to the first pasteUpMorph"
	^self orOwnerSuchThat: [ :o | | oo | (oo := o owner) notNil and: [ oo isPlayfieldLike or: [ oo isHandMorph ]] ]! !

!Morph methodsFor: '*connectors-embeddings' stamp: 'nk 9/10/2004 06:03'!
shouldAddSlotForEmbeddedMorph: aMorph 
	"Answer true if my player should add a slot for the embedded morph's 
	player. Don't do this if neither I nor aMorph is scripted, unless aMorph 
	wantsConnectionWhenEmbedded. Also don't do this if my player already 
	points to aMorph's player."
	| myPlayer itsPlayer |
	itsPlayer := aMorph player.
	myPlayer := self player.
	(aMorph wantsConnectionWhenEmbedded
			or: [itsPlayer belongsToUniClass
					or: [myPlayer belongsToUniClass]])
		ifFalse: [^ false].
	^ myPlayer isNil
		or: [itsPlayer isNil
				or: [(myPlayer pointsTo: itsPlayer) not]]! !

!Morph methodsFor: '*connectors-embeddings' stamp: 'nk 9/9/2004 12:36'!
shouldPromptForSlotForEmbeddedMorph: aMorph
	^aMorph wantsConnectionWhenEmbedded not! !

!Morph methodsFor: '*connectors-graphs' stamp: 'nk 5/12/2003 14:03'!
successors
	| retval |
	retval := IdentitySet new.
	self outgoingConnections do: [ :conn | conn destinationMorph
		ifNotNilDo: [ :dm | dm isHandMorph ifFalse: [ retval add: dm ] ]].
	^retval! !

!Morph methodsFor: '*connectors-geometry' stamp: 'nk 4/6/2001 07:29'!
topLeftCorner
	"Take rounded corners into account"
	^self wantsRoundedCorners
		ifTrue: [ bounds topLeft + (2@2) ]
		ifFalse: [ bounds topLeft ]! !

!Morph methodsFor: '*connectors-geometry' stamp: 'nk 4/6/2001 07:30'!
topRightCorner
	"Take rounded corners into account"
	^self wantsRoundedCorners
		ifTrue: [ bounds topRight + (-2@2) ]
		ifFalse: [ bounds topRight ]! !

!Morph methodsFor: '*connectors-initialization' stamp: 'nk 1/8/2004 12:19'!
uniqueNameLike: stem
	"Choose an innocuous name for the receiver -- one that does not end in the word Morph, and that is unique among the names in the World"
	^self uniqueNameLike: stem in: self world ! !

!Morph methodsFor: '*connectors-initialization' stamp: 'nk 1/8/2004 12:27'!
uniqueNameLike: stem in: aWorld
	"Choose an innocuous name for the receiver -- one that does not end in the word Morph, and that is unique among the names in the World"

	| proposedName allKnownNames |
	proposedName := stem.
	(proposedName size > 5 and: [proposedName endsWith: 'Morph'])
		ifTrue: [proposedName := proposedName copyFrom: 1 to: proposedName size - 5.
			proposedName := proposedName asString translated.].
	
	allKnownNames := aWorld ifNil: [Set new] ifNotNil: [aWorld allKnownNames asSet].
	(allKnownNames includes: proposedName) ifFalse: [ ^proposedName ].

	^ Utilities keyLike: proposedName asString satisfying:
		[:name | (allKnownNames includes: name) not]! !

!Morph methodsFor: '*connectors-copying' stamp: 'nk 6/14/2004 16:36'!
usableSiblingInstance2
	"Return another similar morph whose Player is of the same class as 
	mine. Do not open it in the world."
	| aName usedNames newPlayer newMorph topRenderer |
	(topRenderer := self topRendererOrSelf) == self
		ifFalse: [^ topRenderer usableSiblingInstance2 ].
	self player ifNotNilDo: [ :p | p assureUniClass ].
	newMorph := self veryDeepCopySibling.
	newPlayer := newMorph player.
	newPlayer
		ifNotNil: [newPlayer resetCostumeList].
	(aName := self knownName) isNil
		ifTrue: [self player
				ifNotNil: [aName := newMorph innocuousName]].
	"Force a difference here"
	aName notNil
		ifTrue: [usedNames := (self world
						ifNil: [OrderedCollection new]
						ifNotNil: [self world allKnownNames])
						copyWith: aName.
			newMorph
				setNameTo: (Utilities
						keyLike: aName
						satisfying: [:f | (usedNames includes: f) not])].
	newMorph privateOwner: nil.
	newPlayer
		ifNotNil: [newPlayer assureEventHandlerRepresentsStatus.
			self presenter flushPlayerListCache].
	^ newMorph! !

!Morph methodsFor: '*connectors-connection' stamp: 'nk 8/19/2004 14:28'!
wantsAttachmentFromEnd: endIndex ofConnector: aConnector
	"Answer true if I would like the given end of aConnector attaching to me."
	"TODO: way too much policy in this one place!!"

	owner ifNil: [ ^false ].	
	self == aConnector ifTrue: [ ^false ].
	self isConnectable ifFalse: [ ^false ].
	(self firstOwnerSuchThat: [ :m | m == aConnector ]) ifNotNil: [ ^false ].
	(self isConstrainedBy: aConnector) ifTrue: [ ^false ].
	self isPartsDonor ifTrue: [ ^false ].

	"Allow connection to submorphs when the SHIFT key is pressed, but only to top-level morphs"
	(ActiveHand lastEvent shiftPressed not
		and: [ self topRendererOrSelf owner isPlayfieldLike not ])
			ifTrue: [ ^false ].

	"Avoid connecting from a different World (WiW exception)"
	aConnector pasteUpMorphOrWorld ifNotNilDo: [ :itsWorld |
		(itsWorld isPartsBin not
			and: [ itsWorld ~= World
			and: [ self pasteUpMorphOrWorld ~~ itsWorld ]])
				ifTrue: [ ^false ]].

	^true
! !

!Morph methodsFor: '*connectors-embeddings' stamp: 'nk 8/9/2004 16:11'!
wantsConnectionWhenEmbedded	
	"Return true if I will accept connections even though I'm embedded in a non-pasteup morph"
	^self valueOfProperty: #wantsConnectionWhenEmbedded ifAbsent: [ false ]
! !

!Morph methodsFor: '*connectors-embeddings' stamp: 'nk 8/9/2004 16:12'!
wantsConnectionWhenEmbedded: aBoolean
	"Return true if I will accept connections even though I'm embedded in a non-pasteup morph"
	self setProperty: #wantsConnectionWhenEmbedded toValue: aBoolean
! !

!Morph methodsFor: '*connectors-testing' stamp: 'nk 1/18/2004 16:36'!
wantsConnectorPropertiesSheet
	^self isConnector! !

!Morph methodsFor: '*connectors-scripting' stamp: 'nk 8/21/2004 08:42'!
wantsEmbeddingsVocabulary
	submorphs ifNil: [ ^true ].	"called from EToyVocabulary>>initialize after basicNew"

	^ (Preferences valueOfFlag: #alwaysShowConnectionVocabulary)
		or: [ self wantsConnectionWhenEmbedded 
		or: [ self pinCount > 0 ]]! !

!Morph methodsFor: '*connectors-graphs' stamp: 'nk 6/12/2004 09:28'!
wantsGraphModel
	"Answer true if I should be given a graph model when added to a world that has a graph context"
	^false! !

!Morph methodsFor: '*connectors-testing' stamp: 'nk 1/18/2004 16:36'!
wantsTextPropertiesSheet
	^self isTextMorph! !

!Morph methodsFor: '*connectors-graphs' stamp: 'nk 1/7/2004 15:02'!
withAllPredecessors
	^self allPredecessors add: self; yourself
! !

!Morph methodsFor: '*connectors-graphs' stamp: 'nk 1/7/2004 15:01'!
withAllSuccessors
	^self allSuccessors add: self; yourself
! !

Morph subclass: #NCConstraintMorph
	instanceVariableNames: 'constrained constraint dieWithInput inputs lastTarget stepTime lastTargetBounds'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Connectors-Base'!

!NCConstraintMorph commentStamp: '<historical>' prior: 0!
I apply the result of a pluggable MessageSend to some inputs, and then
take the value of that MessageSend and apply it to some other object.

Typically the result of the MessageSend is a Point, and I constrain the other object by moving it to that Point.

The MessageSend (if I have a single input) is assumed to return a Point in the coordinate system of my input. I translate this to my coordinate system prior to applying it.

I do this in my step method.!

!NCConstraintMorph class methodsFor: 'fileIn/Out' stamp: 'nk 4/18/2002 22:52'!
classVersion
	"This version has the Message Sends and additional instance variables"
	^2
! !

!NCConstraintMorph class methodsFor: 'instance creation' stamp: 'nk 2/26/2001 21:30'!
includeInNewMorphMenu
	^false! !

!NCConstraintMorph class methodsFor: 'line end shapes' stamp: 'nk 4/8/2001 12:29'!
preMadeShapeNames
	"Answer a collection of pre-made shape names for a menu"
	"NCConstraintMorph preMadeShapeNames"

	^#( )! !

!NCConstraintMorph methodsFor: 'connection-callbacks' stamp: 'nk 6/15/2002 13:23'!
absolutePoint: p whileConnectedTo: m
	"This returns a point in my coordinate system, not m's"
	^p! !

!NCConstraintMorph methodsFor: 'menus' stamp: 'dgd 8/26/2004 14:36'!
addCustomMenuItems: aCustomMenu hand: aHandMorph 
	super addCustomMenuItems: aCustomMenu hand: aHandMorph.
	aCustomMenu
		addLine;
		add: 'attach to...' translated action: #attachToSelectedMorph:;
		add: 'display attachments' translated action: #displayInputs;
		addLine.

	false
		ifTrue: [aCustomMenu
				add: 'add arbitrary shape...' translated
				target: self
				selector: #addSelectedMorph:.
			self class preMadeShapeNames isEmpty
				ifFalse: [aCustomMenu add: 'add pre-made shape...' translated subMenu: self selectShapeMenu].
			aCustomMenu
				add: 'delete shape...' translated
				target: self
				selector: #deleteSelectedMorph:;
				 addLine].

	"connectToAbsolutePoint:"

	#(#connectToCenter
		#connectToNearestAttachmentPoint
		#connectToNearestPoint
		#connectToNearestPointToCenter
		#connectToRelativePoint:
		#noConnectionPreference )
		do: [:sel | aCustomMenu
				addUpdating: #connectionWordingFor:
				target: self
				selector: #choosePreferredConnection:
				argumentList: {sel}].
	self dieWithInput
		ifFalse: [aCustomMenu
				addUpdating: #connectionWordingFor:
				target: self
				selector: #choosePreferredConnection:
				argumentList: #(#connectToNothing )]! !

!NCConstraintMorph methodsFor: 'menus' stamp: 'nk 4/7/2001 22:24'!
addSelectedMorph: anEvent
	"Allow the user to select a new submorph and add it to me"
	| hand handle |
	hand := anEvent hand.
	handle := NewHandleMorph new 
		followHand: hand
		forEachPointDo: [ :newPoint |  ]
		lastPointDo: [ :newPoint | | targets choice |
			targets := (self world morphsAt: newPoint) copyWithoutAll: { self . handle }.
			targets := targets reject: [ :ea | ea isPlayfieldLike ].
			targets isEmpty
				ifFalse: [ "choice := targets size = 1
					ifTrue: [ 1 ]
					ifFalse: [ (PopUpMenu labelArray: (targets collect: [ :ea | ea name ]))
							startUpWithCaption: 'Select Morph' ].
					choice > 0
						ifTrue: [ self addShape: (targets at: choice) ]"
					self addShape: targets first.
					self layoutChanged; changed.
				].
		].
	anEvent hand attachMorph: handle.
	handle startStepping.! !

!NCConstraintMorph methodsFor: 'constraints' stamp: 'nk 7/18/2003 17:12'!
addShape: aMorph
	| newMorph |
	aMorph ifNil: [ ^self ].
	newMorph := aMorph duplicate center: self center.
	newMorph endShapeColor: self endShapeColor.
	newMorph endShapeBorderColor: self endShapeBorderColor.
	self addMorphFront: newMorph.
	self forceRedraw.
	^newMorph! !

!NCConstraintMorph methodsFor: 'menus' stamp: 'dgd 8/26/2004 15:03'!
addYellowButtonMenuItemsTo: aMenu event: evt
	| event |
	event := (evt isKindOf: UserInputEvent) ifTrue: [ evt ] ifFalse: [ ActiveEvent ].
	
	aMenu 
		defaultTarget: self;
		addStayUpItem;
		add: 'resize' translated target: self selector: #resizeMorph: argument: event;
		add: 'attach to...' translated target: self selector: #attachToSelectedMorph:;
		add: 'properties...' translated target: self selector: #openAPropertySheet;
		add: 'display attachments' translated target: self selector: #displayInputs.

	false ifTrue: [
		aMenu
			addLine;
			add: 'add arbitrary shape...' translated target: self selector: #addSelectedMorph:.
		self class preMadeShapeNames isEmpty
			ifFalse: [ aMenu add: 'add pre-made shape...' translated subMenu: self selectShapeMenu ].
		aMenu
			addWithLabel: 'delete shape...' translated enablement: #hasSubmorphs action: #deleteSelectedMorph:.
	].

	aMenu addLine.

	self connectionsInMenu
	do: [ :sel |
		aMenu addUpdating: #connectionWordingFor: target: self selector: #choosePreferredConnection: argumentList: { sel }.
	].

	self dieWithInput ifFalse: [
		aMenu addUpdating: #connectionWordingFor: target: self selector: #choosePreferredConnection: argumentList: #( connectToNothing ).
	].
! !

!NCConstraintMorph methodsFor: 'connection' stamp: 'nk 6/21/2003 16:52'!
anchorPreferredConnection: sel at: aPoint
	"this is called at the end of the user selection of a relative or absolute preferred connection point"

	| target targets |

	targets := (self pasteUpMorphOrWorld rootMorphsAt: aPoint)
		reject: [ :ea | (ea isHandleMorph)
			or: [ ea == owner
				or: [ ea == self
					or: [ (self allMorphs includes: ea)
						or: [ (ea isConstrainedBy: self) ] ] ] ] ].

	target := targets isEmpty ifTrue: [ self pasteUpMorphOrWorld ] ifFalse: [ targets first ].

	^self anchorPreferredConnection: sel to: target at: aPoint! !

!NCConstraintMorph methodsFor: 'connection' stamp: 'nk 6/15/2002 12:57'!
anchorPreferredConnection: sel to: target at: aPoint
	"this is called at the end of the user selection of a relative or absolute preferred connection point.
	Answer my new preferredConnection."

	| relPoint pref |

	relPoint := aPoint - (self globalPointToLocal: (target localPointToGlobal: (target topLeft))).
	self input: target.
	self preferredConnection: (pref := { sel . relPoint }).
	self perform: sel with: relPoint.
	^pref
! !

!NCConstraintMorph methodsFor: 'constraints' stamp: 'nk 6/22/2003 09:43'!
applyConstraint: aPoint
	self center ~= aPoint ifTrue: [
		self isFlexed
			ifTrue: [ owner center: (self point: aPoint in: self pasteUpMorphOrWorld) ]
			ifFalse: [ self center: aPoint ]
	].
	lastTargetBounds := nil.! !

!NCConstraintMorph methodsFor: 'dropping/grabbing' stamp: 'nk 8/16/2003 11:11'!
asButtonPrototype
	^super asButtonPrototype
		inputs: #(); yourself! !

!NCConstraintMorph methodsFor: 'connection-callbacks' stamp: 'nk 2/21/2002 18:37'!
atSpec: s on: m
	^m perform: s first withArguments: s allButFirst asArray.! !

!NCConstraintMorph methodsFor: 'attachments' stamp: 'nk 8/21/2004 07:16'!
attachTo: aMorph at: myPoint 
	"myPoint is in my coordinate system. 
	 
	Precedence of preferences: 
	1. Attachment point on embedded pin, if any 
	2. My preference 
	3. aMorph's preference 
	4. aMorph's default 
	"
	| pref spec itsPoint |
	aMorph == self	"ignore self-connections"
		ifTrue: [^ self].

	self stopStepping.

	"check for embedded pins first (overrides my preference)"
	itsPoint := aMorph globalPointToLocal: (self localPointToGlobal: myPoint).
	spec := aMorph attachmentPointSpecNearest: itsPoint.
	spec first == #attachmentPointForEmbeddedMorph:
		ifTrue: [self connectToMorph: spec second usingSpec: {#attachmentPointWhenEmbedded}]
		ifFalse: [
			"check for my preference, if any"
			self input: aMorph.
			pref := self reAnchorPreferredConnectionTo: aMorph at: myPoint.
			(pref isNil or: [pref first == #noConnectionPreference])
				ifFalse: [self perform: pref first withArguments: pref allButFirst]
				ifTrue: [pref := aMorph preferredConnection.
					pref
						ifNil: [aMorph attachFrom: self at: itsPoint]
						ifNotNil: [self perform: pref first withArguments: pref allButFirst]]].

	self startStepping! !

!NCConstraintMorph methodsFor: 'menus' stamp: 'nk 4/7/2001 17:50'!
attachToSelectedMorph: anEvent
	"Allow the user to select a new input"
	| handle targets highlights |
	highlights := self highlightInputs.
	handle := NewHandleMorph new 
		followHand: anEvent hand
		forEachPointDo: [ :newPoint |  ]
		lastPointDo: [ :newPoint | | localPoint choice |
			localPoint := self globalPointToLocal: newPoint.
			highlights do: [ :ea | ea  delete ].
			targets := self connectionTargetsAt: localPoint.
			targets := targets copyWithout: handle.
			targets isEmpty
				ifTrue: [ self connectToAbsolutePoint: localPoint ]
				ifFalse: [ choice := targets size = 1
					ifTrue: [ 1 ]
					ifFalse: [ (PopUpMenu labelArray: (targets collect: [ :ea | ea name ]))
							startUpWithCaption: 'Select Morph' ].
					choice > 0
						ifTrue: [ self attachUnobnoxiouslyTo: (targets at: choice) at: localPoint ]
				].
		].
	anEvent hand attachMorph: handle.
	handle startStepping.! !

!NCConstraintMorph methodsFor: 'attachments' stamp: 'nk 2/27/2001 08:46'!
attachUnobnoxiouslyTo: aMorph at: aPoint
	self attachTo: aMorph at: aPoint.! !

!NCConstraintMorph methodsFor: 'accessing' stamp: 'nk 11/26/2002 11:26'!
beAdjacent
	^false! !

!NCConstraintMorph methodsFor: 'connection-callbacks' stamp: 'nk 2/15/2002 12:30'!
centerOf: m
	^m center! !

!NCConstraintMorph methodsFor: 'menus' stamp: 'nk 4/5/2001 17:19'!
choosePreferredConnection: sel
	"To be called from a menu choice (thus the user interaction).
	Sel should take 0 or 1 argument, or be nil"

	"if nil, forget my preference, and change my constraintBlock"
	((sel == #noConnectionPreference) or: [ sel isNil ])
		ifTrue: [ self preferredConnection: nil.
			self input ifNil: [ ^self ].
			^self attachTo: self input at: self center ].

	(sel numArgs = 1) ifTrue: [
		self selectPointWith: (self currentHand)
			thenDo: [ :pt | self anchorPreferredConnection: sel at: pt ].
	]
	ifFalse: [
		self preferredConnection: { sel }.
		self perform: sel
	].! !

!NCConstraintMorph methodsFor: 'connection' stamp: 'nk 2/22/2002 17:58'!
connectTo: m atVertexNumber: ix
	self constraint: (MessageSend receiver: self selector: #vertexNumbered:of: argument: ix)
! !

!NCConstraintMorph methodsFor: 'connection' stamp: 'nk 2/22/2002 17:58'!
connectToAbsolutePoint: aPoint

	self constraint:
		(MessageSend receiver: self selector: #absolutePoint:whileConnectedTo: argument: aPoint)! !

!NCConstraintMorph methodsFor: 'connection' stamp: 'nk 2/22/2002 17:59'!
connectToCenter
	self constraint: (MessageSend receiver: self selector: #centerOf:)
! !

!NCConstraintMorph methodsFor: 'connection' stamp: 'nk 2/22/2002 17:59'!
connectToCursorPoint
	self constraint: (MessageSend receiver: self selector: #cursorPointOf:)
! !

!NCConstraintMorph methodsFor: 'connection' stamp: 'MAL 1/22/2005 13:58'!
connectToMorph: m usingSpec: s
	self input: m.
	self constraint: (MessageSend receiver: self selector: #atSpec:on: arguments: {s. m})
	"self constraint: (MessageSend receiver: m selector: s first arguments: s allButFirst)"! !

!NCConstraintMorph methodsFor: 'connection' stamp: 'nk 2/22/2002 17:59'!
connectToNearestAttachmentPoint
	self constraint: (MessageSend receiver: self selector: #nearestAttachmentPointOn:)
! !

!NCConstraintMorph methodsFor: 'connection' stamp: 'nk 2/22/2002 17:59'!
connectToNearestPoint
	self constraint: (MessageSend receiver: self selector: #nearestPointOn:)! !

!NCConstraintMorph methodsFor: 'connection' stamp: 'nk 2/22/2002 17:59'!
connectToNearestPointToCenter
	self constraint: (MessageSend receiver: self selector: #nearestPointToCenterOf:)! !

!NCConstraintMorph methodsFor: 'connection' stamp: 'nk 2/22/2002 17:59'!
connectToNothing
	self input: nil.
	self constraint: nil.! !

!NCConstraintMorph methodsFor: 'connection' stamp: 'nk 2/22/2002 17:59'!
connectToRelativePoint: relPoint

	self constraint:
		(MessageSend receiver: self selector: #positionRelative:to: argument: relPoint).
! !

!NCConstraintMorph methodsFor: 'connection' stamp: 'MAL 1/22/2005 22:02'!
connectUsingSelectedSpec: p
	| s |
	s := self nearestAttachmentSpecOn: inputs first to: p.
	self constraint: (MessageSend receiver: self selector: #atSpec:on: argument: s)
! !

!NCConstraintMorph methodsFor: 'connection' stamp: 'MAL 1/22/2005 22:03'!
connectUsingSpec: s
	self constraint: (MessageSend receiver: self selector: #atSpec:on: argument: s)
! !

!NCConstraintMorph methodsFor: 'constraints' stamp: 'nk 9/22/2003 16:22'!
connectionTargetsAt: localPoint
	^(self pasteUpMorphOrWorld rootMorphsAt: (self point: localPoint in: self pasteUpMorphOrWorld))
		reject: [ :ea | ea == self or: [ ea isHighlight ] ].
! !

!NCConstraintMorph methodsFor: 'menus' stamp: 'dgd 8/26/2004 14:39'!
connectionWordingFor: sel
	"sel is the desired menu choice"
	| str selector |
	selector := self preferredConnection ifNotNilDo: [ :pc | pc first ].

	(sel == #connectToNothing) ifTrue: [
		^(selector == #connectToNothing)
			ifTrue: [ '<yes>', 'connect to nothing' translated ]
			ifFalse: [ '<no>', 'connect to nothing' translated ]
	].

	(sel == #noConnectionPreference) ifTrue: [
		^selector
			ifNil: [ '<yes>', 'no connection preference' translated ]
			ifNotNil: [ '<no>', 'no connection preference' translated ]
	].

	str := (sel = selector) ifTrue: [ '<yes>' ] ifFalse: [ '<no>' ].
	str := str, (sel splitOnCapBoundaries) translated.
	^str! !

!NCConstraintMorph methodsFor: 'menus' stamp: 'MAL 1/22/2005 15:02'!
connectionsInMenu
	"Answer a collection of the connection types that should appear in my menu"
	^#( connectToCenter connectToNearestAttachmentPoint connectToNearestPoint connectToNearestPointToCenter connectToRelativePoint: connectUsingSelectedSpec: noConnectionPreference )! !

!NCConstraintMorph methodsFor: 'constraints' stamp: 'nk 4/17/2001 15:23'!
constrained
	"Answer the morph that I constrain, or self if none"
	^constrained ifNil: [ self ]! !

!NCConstraintMorph methodsFor: 'constraints' stamp: 'nk 2/13/2001 14:04'!
constrained: anObject
	constrained := anObject! !

!NCConstraintMorph methodsFor: 'accessing' stamp: 'nk 2/23/2002 13:18'!
constraint
	^constraint! !

!NCConstraintMorph methodsFor: 'constraints' stamp: 'nk 2/23/2002 13:18'!
constraint: aBlockOrMessageSend 
	"aBlock is passed self and all of my inputs (some of which could be nil).
	It should return a Point"

	constraint := (aBlockOrMessageSend isKindOf: BlockContext) 
				ifTrue: [self error: 'BlockContexts aren''t used anymore']
				ifFalse: [aBlockOrMessageSend].
	aBlockOrMessageSend ifNotNil: [self startStepping]! !

!NCConstraintMorph methodsFor: 'object fileIn' stamp: 'nk 4/19/2002 10:07'!
convertToCurrentVersion
	self convertToMessageSend.
	dieWithInput ifNil: [
		dieWithInput := (self valueOfProperty: #dieWithInput ifAbsent: [ false ]).
		self removeProperty: #dieWithInput.
	].
! !

!NCConstraintMorph methodsFor: 'object fileIn' stamp: 'nk 4/19/2002 10:12'!
convertToCurrentVersion: varDict refStream: smartRefStrm 
	"I renamed targetBlock to constraint to catch these conversions."
	varDict at: 'targetBlock'
		ifPresent:  [:b | constraint := b ].
	self convertToCurrentVersion.
	^super convertToCurrentVersion: varDict refStream: smartRefStrm
! !

!NCConstraintMorph methodsFor: 'object fileIn' stamp: 'nk 9/4/2004 17:43'!
convertToMessageSend
	"If I have an old-style (i.e. BlockContext) constraintBlock,
	convert it if possible to a MessageSend that does the same thing"

	| pc |
	(constraint isMessageSend) ifTrue: [^self].	"already converted"
	self input == self ifTrue: [inputs := inputs allButFirst].
	(constraint isNil or: [inputs isEmpty or: [self input isNil]])
		ifTrue: 
			[self connectToNothing.
			^self].
	pc := self preferredConnection.
	pc ifNil: [self attachTo: self input at: self center]
		ifNotNil: [self perform: pc first withArguments: pc allButFirst]! !

!NCConstraintMorph methodsFor: 'connection-callbacks' stamp: 'nk 2/15/2002 12:31'!
cursorPointOf: m
	^m lastEvent position! !

!NCConstraintMorph methodsFor: 'initialization' stamp: 'nk 2/28/2006 09:39'!
defaultColor
	^Color red! !

!NCConstraintMorph methodsFor: 'naming' stamp: 'cmm 12/4/2010 20:03'!
defaultNameStemForInstances
	^ self class name
		copyFrom: 3
		to: self class name size! !

!NCConstraintMorph methodsFor: 'constraints' stamp: 'nk 4/8/2001 16:18'!
defaultTargetPoint
	^self center! !

!NCConstraintMorph methodsFor: 'submorphs-add/remove' stamp: 'nk 2/23/2002 13:18'!
delete
	constrained := nil.
	constraint := nil.
	super delete! !

!NCConstraintMorph methodsFor: 'menus' stamp: 'nk 4/8/2001 12:22'!
deleteSelectedMorph: anEvent
	"Allow the user to delete one of my submorphs, if any."
	| choice |
	submorphs isEmpty ifTrue: [ ^self ].
	submorphs size = 1 ifTrue: [ ^self deleteShape: submorphs first ].

	choice := (PopUpMenu labelArray: (submorphs collect: [ :ea | ea externalName ]))
							startUpWithCaption: 'Select Morph'.
	choice > 0 ifTrue: [ self deleteShape: (submorphs at: choice) ].! !

!NCConstraintMorph methodsFor: 'menus' stamp: 'nk 4/18/2001 11:05'!
deleteShape: aMorph
	aMorph delete.
	self forceRedraw.
	self step.! !

!NCConstraintMorph methodsFor: 'accessing' stamp: 'nk 4/18/2002 18:02'!
dieWithInput
	^dieWithInput! !

!NCConstraintMorph methodsFor: 'accessing' stamp: 'nk 4/18/2002 18:01'!
dieWithInput: aBoolean
	"If aBoolean is true, make sure that I delete myself when my
	input goes away"
	dieWithInput := aBoolean! !

!NCConstraintMorph methodsFor: 'menus' stamp: 'nk 2/26/2001 23:53'!
displayInputs
	| highlights |
	highlights := self highlightInputs.
	self addAlarm: #deleteMorphs: with: highlights after: 2000! !

!NCConstraintMorph methodsFor: 'testing' stamp: 'nk 3/11/2002 21:16'!
drawingElement
	"Answer the drawing element that I constrain, typically either a Connector or a label"
	^owner! !

!NCConstraintMorph methodsFor: 'attachments' stamp: 'nk 7/18/2003 17:13'!
endShapeBorderColor
	^self valueOfProperty: #endShapeBorderColor ifAbsent: [
		owner ifNotNil: [ owner color ]
			ifNil: [ Color black ]
	 ].! !

!NCConstraintMorph methodsFor: 'attachments' stamp: 'nk 7/18/2003 17:13'!
endShapeBorderColor: aColor
	self submorphsDo: [ :ea | ea allMorphsDo: [ :sm | sm endShapeBorderColor: aColor ] ].
	self setProperty: #endShapeBorderColor toValue: aColor.! !

!NCConstraintMorph methodsFor: 'attachments' stamp: 'nk 7/18/2003 17:16'!
endShapeBorderWidth: aWidth
	self submorphsDo: [ :ea | ea allMorphsDo: [ :sm | sm endShapeBorderWidth: aWidth ] ].
	self computeLineAttachmentPoint.
	self forceRedraw.! !

!NCConstraintMorph methodsFor: 'attachments' stamp: 'nk 4/18/2001 11:45'!
endShapeColor
	^self valueOfProperty: #endShapeColor ifAbsent: [
		owner ifNotNil: [ owner color ]
			ifNil: [ Color black ]
	 ].! !

!NCConstraintMorph methodsFor: 'attachments' stamp: 'nk 4/18/2001 11:45'!
endShapeColor: aColor
	self submorphsDo: [ :ea | ea allMorphsDo: [ :sm | sm endShapeColor: aColor ] ].
	self setProperty: #endShapeColor toValue: aColor.! !

!NCConstraintMorph methodsFor: 'attachments' stamp: 'nk 6/18/2002 10:41'!
endShapeWidth: aWidth
	self submorphsDo: [ :ea | ea allMorphsDo: [ :sm | sm endShapeWidth: aWidth ] ].
	self computeLineAttachmentPoint.
	self forceRedraw.! !

!NCConstraintMorph methodsFor: 'stepping and presenter' stamp: 'nk 1/10/2004 18:09'!
ensureInFront
	"Make sure that I am in front of my input."
	| meInWorld input inputInWorld pu |
	((input := self input) isNil) ifTrue: [ ^self ].
	inputInWorld := input root.
	meInWorld := self isFlexed ifTrue: [ owner ] ifFalse: [ self ].
	(meInWorld boundsInWorld intersects: (inputInWorld boundsInWorld)) ifFalse: [ ^self ].
	inputInWorld owner ~~ meInWorld owner ifTrue: [ ^self ].
	((pu := meInWorld pasteUpMorph) isNil or: [ pu ~= inputInWorld pasteUpMorph ])
		ifTrue: [ ^self ].
	pu ensureMorph: inputInWorld behind: meInWorld.
! !

!NCConstraintMorph methodsFor: 'constraints' stamp: 'nk 2/26/2001 22:26'!
firstInput
	| ips |
	ips := self inputs.
	^ips isEmpty ifTrue: [ ] ifFalse: [ ips first ] ! !

!NCConstraintMorph methodsFor: 'WiW support' stamp: 'nk 8/16/2003 10:42'!
fixMorphicLayer
	(super morphicLayerNumber) ~= (self morphicLayerNumber)
		ifTrue: [ | oldOwner |
			(oldOwner := owner) ifNotNil: [ 
				oldOwner removeMorph: self.
				oldOwner addMorphInLayer: self.
			]
		]. ! !

!NCConstraintMorph methodsFor: 'stepping and presenter' stamp: 'nk 1/20/2004 11:51'!
forceRedraw

	self changed "; layoutChanged".
	self lastTargetPoint: nil.
	lastTargetBounds := nil.! !

!NCConstraintMorph methodsFor: 'drawing' stamp: 'nk 4/5/2001 17:40'!
forcedInvisible
	^super visible not! !

!NCConstraintMorph methodsFor: 'event handling' stamp: 'nk 2/26/2001 19:04'!
handlesMouseDown: anEvent
	anEvent yellowButtonPressed ifTrue: [ ^true ].
	^super handlesMouseDown: anEvent! !

!NCConstraintMorph methodsFor: 'constraints' stamp: 'nk 2/23/2001 14:35'!
hasAnyInputIn: aCollection
	^inputs includesAnyOf: aCollection! !

!NCConstraintMorph methodsFor: 'constraints' stamp: 'nk 6/13/2002 09:42'!
hasAnyInputs
	^inputs notNil 
		and: [inputs anySatisfy: [:ea | ea ~~ nil ]]
! !

!NCConstraintMorph methodsFor: 'halos and balloon help' stamp: 'nk 4/5/2001 17:51'!
hasHalo
	| halo haloTarget |
	super hasHalo ifTrue: [ ^true ].
	halo := self currentHand halo.
	halo ifNil: [ ^false ].
	haloTarget := halo target.
	haloTarget == self ifTrue: [ ^true ].
	self allMorphsDo: [ :ea | haloTarget == ea ifTrue: [ ^true ] ].
	^false! !

!NCConstraintMorph methodsFor: 'constraints' stamp: 'nk 2/23/2001 13:44'!
hasInput: anObject
	^inputs includes: anObject! !

!NCConstraintMorph methodsFor: 'menus' stamp: 'nk 1/23/2004 15:31'!
hasYellowButtonMenu
	^true! !

!NCConstraintMorph methodsFor: 'drawing' stamp: 'nk 6/17/2002 16:19'!
highlightInputs
	| highlights |
	highlights := self inputs collect: [ :ea |
		(Morph new)
			bounds: ea fullBounds;
			color: ((Color red) alpha: 0.3);
			lock
	].
	highlights := OrderedCollection withAll: highlights.
	highlights add: ( (EllipseMorph new)
			extent: 12@12;
			center: (self point: self targetPoint in: self pasteUpMorph);
			color: ((Color green) alpha: 0.6);
			lock).
	highlights do: [ :ea | self pasteUpMorph addMorphFront: ea ].
	^highlights
! !

!NCConstraintMorph methodsFor: 'stepping and presenter' stamp: 'nk 3/2/2001 16:17'!
incrementStepTime
	stepTime := (stepTime + 1) min: self maximumStepTime! !

!NCConstraintMorph methodsFor: 'initialization' stamp: 'nk 2/28/2006 09:38'!
initialize
	super initialize.
	inputs := EmptyArray.
	dieWithInput := false.
	self lastTargetPoint: 0@0.
	self
		stepTime: self minimumStepTime;
		color: self defaultColor;
		extent: 10@10;
		connectToNearestAttachmentPoint.
! !

!NCConstraintMorph methodsFor: 'constraints' stamp: 'nk 2/15/2002 12:42'!
input
	^inputs at: 1 ifAbsent: []! !

!NCConstraintMorph methodsFor: 'constraints' stamp: 'nk 2/23/2001 16:20'!
input: anObject
	self inputs: { anObject }! !

!NCConstraintMorph methodsFor: 'constraints' stamp: 'nk 4/9/2003 08:14'!
inputs
	^Array withAll: (inputs reject: [:ea | ea isNil])! !

!NCConstraintMorph methodsFor: 'constraints' stamp: 'nk 8/16/2003 11:11'!
inputs: aCollection 
	"aCollection consists of objects that I derive my target from"

	inputs do: [:ea | ea ifNotNil: [ea disconnectConstraint: self]].
	inputs := WeakArray withAll: aCollection.
	inputs do: [:ea | ea ifNotNil: [ea connectConstraint: self]].
	lastTargetBounds := nil.
	aCollection isEmpty
		ifTrue: [ self stopStepping ]
		ifFalse: [ self startStepping ]! !

!NCConstraintMorph methodsFor: 'initialization' stamp: 'nk 1/20/2004 11:51'!
intoWorld: aWorld
	self forceRedraw.
	super intoWorld: aWorld.
! !

!NCConstraintMorph methodsFor: 'attachments' stamp: 'nk 2/23/2001 13:42'!
isConstraint
	^true! !

!NCConstraintMorph methodsFor: 'testing' stamp: 'nk 2/27/2001 13:02'!
isLabelConstraint
	^false! !

!NCConstraintMorph methodsFor: 'testing' stamp: 'nk 2/27/2001 13:01'!
isLineConstraint
	^false! !

!NCConstraintMorph methodsFor: 'accessing' stamp: 'nk 8/16/2003 10:59'!
lastTargetPoint
	^lastTarget ifNil: [ lastTarget:= 0@0 ]! !

!NCConstraintMorph methodsFor: 'accessing' stamp: 'nk 6/21/2003 16:32'!
lastTargetPoint: aPoint
	lastTarget := aPoint.! !

!NCConstraintMorph methodsFor: 'stepping and presenter' stamp: 'nk 3/2/2001 16:20'!
maximumStepTime
	^200! !

!NCConstraintMorph methodsFor: 'stepping and presenter' stamp: 'nk 3/2/2001 15:47'!
minimumStepTime
	^20! !

!NCConstraintMorph methodsFor: 'WiW support' stamp: 'nk 4/5/2001 17:54'!
morphicLayerNumber
	"Try to keep me above my input"
	^(self input notNil and: [ self input ~~ self and: [ self input isMorph ]])
		ifTrue: [ self input morphicLayerNumber - 1 ]
		ifFalse: [ super morphicLayerNumber ]
! !

!NCConstraintMorph methodsFor: 'event handling' stamp: 'nk 6/24/2004 13:27'!
mouseDown: evt
	evt yellowButtonPressed  "First check for option (menu) click"
		ifTrue: [^ self yellowButtonActivity: evt shiftPressed ].
	super mouseDown: evt
! !

!NCConstraintMorph methodsFor: 'connection-callbacks' stamp: 'nk 6/15/2002 13:17'!
nearestAttachmentPointOn: m
	^m attachmentPointNearest: (self point: self center relativeTo: m)! !

!NCConstraintMorph methodsFor: 'connection-callbacks' stamp: 'MAL 1/22/2005 14:22'!
nearestAttachmentPointOn: m to: relPoint
	^m attachmentPointNearest: (self point: relPoint relativeTo: m)! !

!NCConstraintMorph methodsFor: 'connection-callbacks' stamp: 'MAL 1/22/2005 14:59'!
nearestAttachmentSpecOn: m to: relPoint
	"Return the nearest attachment spec for a morph and a point relative to that morph"
	^m attachmentSpecNearest: m topLeft + relPoint! !

!NCConstraintMorph methodsFor: 'connection-callbacks' stamp: 'nk 6/15/2002 13:24'!
nearestPointOn: m
	^m closestPointTo: (self point: self center relativeTo: m)! !

!NCConstraintMorph methodsFor: 'connection-callbacks' stamp: 'nk 6/15/2002 13:24'!
nearestPointToCenterOf: m
	^m intersectionWithLineSegmentFromCenterTo: (self point: self center relativeTo: m)! !

!NCConstraintMorph methodsFor: 'constraints' stamp: 'nk 3/31/2002 10:39'!
needsToApplyConstraintGivenNewTarget: newTarget
	"newTarget has just been returned by 'self target'.
	Do I have to apply the constraint?
	If I return true, applyConstraint: will be called."
	^lastTarget ~= newTarget! !

!NCConstraintMorph methodsFor: 'connection-callbacks' stamp: 'nk 6/15/2002 13:53'!
point: aPoint relativeTo: morph
	| world mtfm ctfm |
	world := self world.
	ctfm := self transformFrom: world.
	mtfm := morph transformFrom: world.
	^mtfm globalPointToLocal: (ctfm localPointToGlobal: aPoint)! !

!NCConstraintMorph methodsFor: 'connection-callbacks' stamp: 'nk 2/15/2002 12:23'!
positionRelative: relPoint to: m
	^m topLeft + relPoint! !

!NCConstraintMorph methodsFor: 'connection' stamp: 'nk 3/3/2001 15:20'!
reAnchorPreferredConnectionAt: aPoint
	"Answer my preferredConnection, if any"

	| pref |
	(pref := self preferredConnection) ifNil: [ ^nil ].
	pref first numArgs = 0 ifTrue: [ ^pref ].
	^self anchorPreferredConnection: pref first at: aPoint.
! !

!NCConstraintMorph methodsFor: 'connection' stamp: 'nk 3/3/2001 15:19'!
reAnchorPreferredConnectionTo: aMorph at: aPoint
	"Answer my preferredConnection, if any"

	| pref |
	(pref := self preferredConnection) ifNil: [ ^nil ].
	pref first numArgs = 0 ifTrue: [ ^pref ].
	^self anchorPreferredConnection: pref first to: aMorph at: aPoint.! !

!NCConstraintMorph methodsFor: 'menus' stamp: 'nk 3/2/2001 18:33'!
selectPointWith: aHand thenDo: aBlock
	"Allow the user to select a point, then pass the point to aBlock.
	aBlock might never get called, if the handle gets deleted."
	| handle |
	handle := NewHandleMorph new 
		followHand: aHand
		forEachPointDo: [ :newPoint |  ]
		lastPointDo: aBlock.
	aHand attachMorph: handle.
	handle startStepping.! !

!NCConstraintMorph methodsFor: 'menus' stamp: 'nk 7/12/2002 18:50'!
selectShapeMenu

	| menu |
	menu := MenuMorph new layoutPolicy: TableLayout new.

	self class preMadeShapeNames do: [ :ea |
		| morph text ix mm sm |
		morph := self class perform: ea.
		ix := ea findString: 'Shape' startingAt: 1.
		ix := ix = 0 ifTrue: [ ea size ] ifFalse: [ ix - 1 ].
		text := (ea copyFrom: 1 to: ix) splitOnCapBoundaries.
		(mm := MenuItemMorph new)
			layoutPolicy: TableLayout new;
			listDirection: #leftToRight;
			layoutInset: 3@3;
			wrapCentering: #topLeft;
			hResizing: #spaceFill;
			vResizing: #shrinkWrap;
			contents: ' ';
			target: self;
			selector: #addShape: ;
			arguments: { morph }.
		"The SketchMorph here is because my PolygonMorphs were getting stretched."
		sm := SketchMorph withForm: morph imageForm.
		mm addMorphBack: sm.
		sm heading: -90.
		mm addMorphBack: AlignmentMorph newVariableTransparentSpacer.
		mm addMorphBack: ((StringMorph contents: text) font: Preferences standardMenuFont).
		menu addMorphBack: mm.
	].

	^menu! !

!NCConstraintMorph methodsFor: 'stepping and presenter' stamp: 'nk 7/11/2003 23:45'!
step
	| myInputs newTarget |
	myInputs := self inputs.		"hang on to them"
	myInputs isEmpty ifTrue: [
		constraint ifNotNil: [ constraint := nil. self changed].
			self dieWithInput ifTrue: [ self delete ].
		^self ].
	newTarget := self target.
	(self needsToApplyConstraintGivenNewTarget: newTarget)
		ifTrue: 
			[self applyConstraint: newTarget.
			lastTarget := newTarget.
			stepTime := self minimumStepTime]
		ifFalse: [self incrementStepTime]! !

!NCConstraintMorph methodsFor: 'stepping and presenter' stamp: 'nk 3/2/2001 15:47'!
stepTime
	^stepTime! !

!NCConstraintMorph methodsFor: 'stepping and presenter' stamp: 'nk 3/2/2001 15:46'!
stepTime: mSecsPerStep

	stepTime := mSecsPerStep truncated.
! !

!NCConstraintMorph methodsFor: 'constraints' stamp: 'nk 4/8/2001 11:15'!
target
	^self targetPoint! !

!NCConstraintMorph methodsFor: 'constraints' stamp: 'nk 8/14/2004 14:07'!
targetPoint
	| pt input isMorph |
	input := self input.
	(constraint isNil or: [input isNil]) 
		ifTrue: [^self defaultTargetPoint].

	(isMorph := input isMorph) ifTrue: [ | inputBounds |
		inputBounds := input boundsInWorld.
		lastTargetBounds = inputBounds ifTrue: [ ^self lastTargetPoint ].
		lastTargetBounds := inputBounds.
	].

	"call target block"
	pt := constraint valueWithEnoughArguments: { input }.

	"transform point"
	pt := (isMorph 
		ifTrue: [self globalPointToLocal: (input localPointToGlobal: pt)]
		ifFalse: [self globalPointToLocal: pt]) asIntegerPoint.

	^ pt
! !

!NCConstraintMorph methodsFor: 'halos and balloon help' stamp: 'nk 8/21/2004 10:54'!
transferHalo: event from: formerHaloOwner
	"Progressively transfer the halo to the next likely recipient"

	(formerHaloOwner == self) ifTrue:[
		self input ifNotNilDo: [ :in |  ^in transferHalo: event from: formerHaloOwner ] ].
	^super transferHalo: event from: formerHaloOwner! !

!NCConstraintMorph methodsFor: 'connection-callbacks' stamp: 'nk 2/21/2002 18:22'!
vertexNumbered: ix of: m
	^m vertexAt: ix! !

!NCConstraintMorph methodsFor: 'copying' stamp: 'nk 9/27/2003 15:43'!
veryDeepFixupWith: aDeepCopier
	super veryDeepFixupWith: aDeepCopier.
	constrained := aDeepCopier references at: constrained ifAbsent: [ ].
	inputs := inputs collect: [ :ea | | inp |
		inp := ea ifNotNil: [ aDeepCopier references at: ea ifAbsent: [ nil ] ].
		inp ifNotNil: [ inp connectConstraint: self ].
		inp ].
! !

!NCConstraintMorph methodsFor: 'copying' stamp: 'nk 6/21/2003 23:17'!
veryDeepInner: aDeepCopier 
	super veryDeepInner: aDeepCopier.
	constrained := constrained.
	inputs := inputs copy.	"this lets you stay attached"
	constraint := constraint veryDeepCopyWith: aDeepCopier.
	lastTarget := nil.
	lastTargetBounds := nil.
	stepTime := stepTime.
	dieWithInput := dieWithInput.! !

!NCConstraintMorph methodsFor: 'drawing' stamp: 'nk 2/23/2002 13:18'!
visible
	self forcedInvisible ifTrue: [^false].
	^constraint isNil or: [constrained isNil or: [self hasAnyInputs not]]! !

!NCConstraintMorph methodsFor: 'stepping and presenter' stamp: 'nk 2/23/2002 13:18'!
wantsSteps
	^constraint notNil! !

NCConstraintMorph subclass: #NCLineEndConstraintMorph
	instanceVariableNames: 'firstVertex lineAttachmentPoint'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Connectors-Base'!

!NCLineEndConstraintMorph commentStamp: '<historical>' prior: 0!
My constrained object is one end of an open PolygonMorph (a/k/a LineMorph).!

!NCLineEndConstraintMorph class methodsFor: 'fileIn/Out' stamp: 'nk 4/18/2002 22:52'!
classVersion
	"This version has the Message Sends and additional instance variables"
	^2
! !

!NCLineEndConstraintMorph class methodsFor: 'line end shapes' stamp: 'nk 7/28/2002 20:06'!
closedArrowheadShape
	"NCLineEndConstraintMorph closedArrowheadShape openInHand"

	^(self filledArrowheadShape)
			makeOpen;
			borderWidth: 2;
			name: 'closed arrowhead';
			lineAttachmentOffset: 0@-1! !

!NCLineEndConstraintMorph class methodsFor: 'line end shapes' stamp: 'nk 6/25/2002 18:14'!
crowsFootShape
	"NCLineEndConstraintMorph crowsFootShape openInHand"

	^(PolygonMorph vertices: { 0@15 . 7@0. 14@15  }
		color: Color black
		borderWidth: 2
		borderColor: Color black)
			makeOpen;
			setProperty: #referencePosition toValue: 7@16;
			lineAttachmentOffset: -0.01@0;
			vResizing: #rigid;
			hResizing: #rigid;
			name: 'crows foot'
! !

!NCLineEndConstraintMorph class methodsFor: 'line end shapes' stamp: 'nk 6/25/2002 17:41'!
doubleArrowheadShape
	"NCLineEndConstraintMorph doubleArrowheadShape openInHand"

	^(PolygonMorph vertices: { 0@0 . 7@9 . 14@0 . 7@9 . 7@15 . 0@6 . 7@15 . 14@6 }
		color: Color black
		borderWidth: 2
		borderColor: Color black)
			makeOpen;
			setProperty: #referencePosition toValue: 7@15;
			lineAttachmentOffset: (0@(-8/15)) asFloatPoint;
			name: 'double arrowhead'! !

!NCLineEndConstraintMorph class methodsFor: 'line end shapes' stamp: 'nk 6/25/2002 17:28'!
filledArrowheadShape
	"NCLineEndConstraintMorph filledArrowheadShape openInHand"

	^(PolygonMorph vertices: { 0@0 . 7@17 . 14@0 . 0@0 }
		color: Color black
		borderWidth: 0
		borderColor: Color black)
			setProperty: #referencePosition toValue: 7@17;
			vResizing: #rigid;
			hResizing: #rigid;
			name: 'filled arrowhead'! !

!NCLineEndConstraintMorph class methodsFor: 'line end shapes' stamp: 'nk 7/1/2002 16:03'!
filledCircleShape
	"NCLineEndConstraintMorph filledCircleShape openInHand"

	^ CircleMorph new
		borderWidth: 0;
		color: Color black;
		extent: 12@12;
		setProperty: #originalExtent toValue: 12@12;
		setProperty: #originalWidth toValue: 2;
		rotationCenter: 0.5@1.0;
		lineAttachmentOffset: 0@-1;
		name: 'filled circle'! !

!NCLineEndConstraintMorph class methodsFor: 'line end shapes' stamp: 'nk 6/25/2002 18:06'!
filledDiamondShape
	"NCLineEndConstraintMorph filledDiamondShape openInHand"

	^(PolygonMorph vertices: { 6@0 . 12@11 . 6@22 . 0@12 . 6@0 }
		color: Color black
		borderWidth: 0
		borderColor: Color black)
			setProperty: #referencePosition toValue: 6@22;
			lineAttachmentOffset: -0.01@-0.95;
			name: 'filled diamond'! !

!NCLineEndConstraintMorph class methodsFor: 'line end shapes' stamp: 'nk 7/1/2002 16:22'!
junctionDotShape
	"NCLineEndConstraintMorph junctionDotShape openInHand"

	^ CircleMorph new
		borderWidth: 0;
		color: Color black;
		extent: 8@8;
		setProperty: #originalExtent toValue: 8@8;
		setProperty: #originalWidth toValue: 2;
		rotationCenter: 0.5@0.5;
		lineAttachmentOffset: 0@0;
		name: 'junction dot'! !

!NCLineEndConstraintMorph class methodsFor: 'line end shapes' stamp: 'nk 6/25/2002 17:20'!
openArrowheadShape
	"NCLineEndConstraintMorph openArrowheadShape openInHand"
	"NCConnectorMorph newDirectionalAssociation openInHand"
	^(PolygonMorph vertices: { 0@0 . 6@15 . 12@0  }
		color: Color black
		borderWidth: 2
		borderColor: Color black)
			makeOpen;
			setProperty: #referencePosition toValue: 6@15;
			vResizing: #rigid;
			hResizing: #rigid;
			name: 'open arrowhead'! !

!NCLineEndConstraintMorph class methodsFor: 'line end shapes' stamp: 'nk 7/1/2002 07:19'!
openCircleShape
	"NCLineEndConstraintMorph openCircleShape openInHand"

	^self filledCircleShape borderWidth: 2; color: Color transparent; name: 'open circle'! !

!NCLineEndConstraintMorph class methodsFor: 'line end shapes' stamp: 'nk 4/8/2001 12:21'!
openDiamondShape
	"NCLineEndConstraintMorph openDiamondShape openInHand"

	^self filledDiamondShape
		makeOpen;
		borderWidth: 2;
		name: 'open diamond'! !

!NCLineEndConstraintMorph class methodsFor: 'line end shapes' stamp: 'nk 4/17/2002 13:29'!
preMadeShapeNames
	"Answer a collection of pre-made shape names for a menu"
	"NCLineEndConstraintMorph preMadeShapeNames"

	^#(
		filledCircleShape
		openCircleShape
		junctionDotShape
		openDiamondShape
		filledDiamondShape
		openArrowheadShape
		closedArrowheadShape
		filledArrowheadShape 
		doubleArrowheadShape
		crowsFootShape
	)
! !

!NCLineEndConstraintMorph class methodsFor: 'conversion' stamp: 'nk 6/12/2004 18:33'!
updateOldShapes26June2002
	"NCLineEndConstraintMorph updateOldShapes26June2002"
	"Change all pre-defined shapes to the better ones"
	self allSubInstancesDo: [ :lec |
		lec submorphsDo: [ :att | 
			(att isKindOf: PolygonMorph) ifTrue: [ | new |
				new := att knownName caseOf: {
				[ 'closed arrowhead' ] -> [ #closedArrowheadShape ].
				[ 'crows foot' ] -> [ #crowsFootShape ].
				[ 'double arrowhead' ] -> [ #doubleArrowheadShape ].
				[ 'filled arrowhead' ] -> [ #filledArrowheadShape ].
				[ 'filled circle' ] -> [ #filledCircleShape ].
				[ 'filled diamond' ] -> [ #filledDiamondShape ].
				[ 'junction dot' ] -> [ #junctionDotShape ].
				[ 'open arrowhead' ] -> [ #openArrowheadShape ].
				} otherwise: [ nil ].
				new ifNotNil: [ | newEnd |
					newEnd := self perform: new.
					lec deleteShape: att.
					lec addShape: newEnd.
				]
			]
		]
	].

	Smalltalk at: #NCMakerButton ifPresent: [ :ncmb |
		ncmb allSubInstancesDo: [ :ea |
		ea updateImage ]].
! !

!NCLineEndConstraintMorph methodsFor: 'rotate scale and flex' stamp: 'nk 7/4/2003 14:09'!
addFlexShellIfNecessary
	"When scaling or rotating from a halo, I can do this without a flex shell"

	^ self
! !

!NCLineEndConstraintMorph methodsFor: 'submorphs-add/remove' stamp: 'nk 7/1/2002 14:08'!
addJunctionDotShape
	self submorphs isEmpty ifTrue: [
		self addShape: self class junctionDotShape.
		self connectToNearestOrthogonalPoint.
		self preferredConnection: #(#connectToNearestOrthogonalPoint)
	].! !

!NCLineEndConstraintMorph methodsFor: 'constraints' stamp: 'nk 7/18/2003 17:26'!
addShape: aMorph
	| newMorph |
	aMorph ifNil: [ ^self ].
	newMorph := super addShape: aMorph.
	newMorph endShapeWidth: owner lineWidth.
	newMorph heading: self heading.
	newMorph referencePosition: self lastTargetPoint.
	self computeLineAttachmentPoint.
	self applyConstraint: lastTarget.
	self layoutChanged.! !

!NCLineEndConstraintMorph methodsFor: 'constraints' stamp: 'nk 8/21/2004 16:04'!
applyConstraint: anArray 
	"anArray consists of { targetPoint . angle. <input owner identity hash> }"
	| newTargetPoint newPoint oldPoint angle oldOwnerHash |

	oldOwnerHash := self lastTargetOwnerHash.
	newTargetPoint := anArray first.
	angle := anArray second + 90.0.
	newPoint := newTargetPoint.
	"Now compute actual line start"
	lineAttachmentPoint isZero
		ifFalse: [newPoint := newPoint
						+ (lineAttachmentPoint rotateBy: angle negated degreesToRadians about: 0 @ 0).
			newPoint := newPoint rounded].
	oldPoint := constrained vertexAt: self vertexIndex.
	newPoint ~= oldPoint
		ifTrue: [constrained verticesAt: self vertexIndex put: newPoint].
	oldOwnerHash ~= self inputOwnerHash
		ifTrue: [self drawingElement inputOwnerChanged].
	submorphs isEmpty
		ifTrue: [^ self center: newPoint].
	submorphs
		do: [:ea | 
			ea heading: angle.
			ea referencePosition: newTargetPoint].
	lastTargetBounds := nil.
	bounds := self submorphBounds.
	self layoutChanged.
! !

!NCLineEndConstraintMorph methodsFor: 'constraints' stamp: 'nk 6/25/2002 18:03'!
computeLineAttachmentPoint
	^lineAttachmentPoint := (submorphs isEmpty
		ifTrue: [ 0@0 ]
		ifFalse: [ (submorphs collect: [ :ea | ea lineAttachmentOffset * ea extent ]) min ]) rounded
! !

!NCLineEndConstraintMorph methodsFor: 'connection' stamp: 'nk 6/20/2002 13:22'!
connectToNearestOrthogonalPoint
	self constraint: (MessageSend receiver: self selector: #nearestOrthogonalPointOn:)! !

!NCLineEndConstraintMorph methodsFor: 'constraints' stamp: 'nk 4/7/2001 17:49'!
connectionTargetsAt: localPoint
	^owner connectionTargetsAt: localPoint for: self! !

!NCLineEndConstraintMorph methodsFor: 'menus' stamp: 'nk 6/20/2002 13:37'!
connectionsInMenu
	| retval |
	retval := OrderedCollection withAll: super connectionsInMenu.
	retval add: #connectToNearestOrthogonalPoint afterIndex: retval size - 1.
	^retval! !

!NCLineEndConstraintMorph methodsFor: 'constraints' stamp: 'nk 4/15/2001 15:58'!
constrainedVertex
	constrained ifNil: [ ^0@0 ].
	^constrained vertices at: self vertexIndex! !

!NCLineEndConstraintMorph methodsFor: 'object fileIn' stamp: 'nk 4/19/2002 10:06'!
convertToCurrentVersion
	lineAttachmentPoint ifNil: [
		lineAttachmentPoint := (self valueOfProperty: #lineAttachmentPoint ifAbsent: [ 0@0 ]).
		self removeProperty: #lineAttachmentPoint.
	].
	super convertToCurrentVersion.! !

!NCLineEndConstraintMorph methodsFor: 'object fileIn' stamp: 'nk 4/18/2002 21:57'!
convertToCurrentVersion: varDict refStream: smartRefStrm 
	lineAttachmentPoint ifNil: [
		lineAttachmentPoint := (self valueOfProperty: #lineAttachmentPoint ifAbsent: [ 0@0 ]).
		self removeProperty: #lineAttachmentPoint.
	].
	^super convertToCurrentVersion: varDict refStream: smartRefStrm
! !

!NCLineEndConstraintMorph methodsFor: 'constraints' stamp: 'nk 4/8/2001 16:30'!
defaultTargetPoint
	^self lastTargetPoint! !

!NCLineEndConstraintMorph methodsFor: 'submorphs-add/remove' stamp: 'nk 7/1/2002 14:10'!
deleteAllShapes
	self submorphs isEmpty ifFalse: [
		self submorphs do: [ :ea | self deleteShape: ea ].
		self preferredConnection: nil
	]
! !

!NCLineEndConstraintMorph methodsFor: 'menus' stamp: 'nk 2/22/2002 15:08'!
deleteShape: aMorph
	super deleteShape: aMorph.
	self computeLineAttachmentPoint.
	self forceRedraw.! !

!NCLineEndConstraintMorph methodsFor: 'drawing' stamp: 'nk 7/12/2003 00:21'!
drawOn: aCanvas

	self hasAnyInputs ifFalse: [
			aCanvas fillRectangle: ((Rectangle center: self lastTargetPoint extent: 10@10) intersect: self bounds)
			fillStyle: self fillStyle.
	]! !

!NCLineEndConstraintMorph methodsFor: 'accessing' stamp: 'nk 8/22/2004 05:52'!
drawingElement
	"Answer the drawing element that I constrain, typically either a Connector or a label"
	^constrained ifNotNilDo: [ :c | c owner ]! !

!NCLineEndConstraintMorph methodsFor: 'private' stamp: 'nk 1/20/2004 18:41'!
endIndex
	"Answer the index of the end of the line that I'm constraining"
	^firstVertex 
		ifTrue: [ 1 ]
		ifFalse: [ 2 ]! !

!NCLineEndConstraintMorph methodsFor: 'stepping and presenter' stamp: 'nk 8/25/2003 16:05'!
ensureInFront
	"Make sure that I my line is in front of my input."
	| input world |
	constrained ifNil: [ ^self ].
	(input := self input) ifNil: [ ^self ].
	input := input root.
	input isHandMorph ifTrue: [ ^self ].
	(owner isNil or: [ (owner owner) ~~ (world := input owner) ]) ifTrue: [ ^self ].
	(world isNil or: [ world isHandMorph ]) ifTrue: [  ^self ].
	(constrained bounds intersects: (input bounds)) ifFalse: [ ^self ].
	world ensureMorph: input behind: owner.! !

!NCLineEndConstraintMorph methodsFor: 'layout' stamp: 'nk 4/8/2001 14:29'!
fixColor
	submorphs isEmpty
		ifFalse: [ self color: (Color r: 1.0 g: 0 b: 0 alpha: 0.2 )].
	super visible
		ifTrue: [ self color: Color red ].
! !

!NCLineEndConstraintMorph methodsFor: 'stepping and presenter' stamp: 'nk 8/20/2003 17:08'!
forceRedraw
	self lastTargetAngle: (self lastTargetAngle + 1e-6).
	lastTargetBounds := nil.
	self changed.
! !

!NCLineEndConstraintMorph methodsFor: 'event handling' stamp: 'nk 2/26/2001 19:28'!
handlesMouseDown: evt
	^false! !

!NCLineEndConstraintMorph methodsFor: 'initialization' stamp: 'nk 4/18/2002 18:08'!
initialize
	super initialize.
	lineAttachmentPoint := 0@0.
! !

!NCLineEndConstraintMorph methodsFor: 'naming' stamp: 'nk 8/21/2004 10:57'!
innocuousName
	^'connector end' translated! !

!NCLineEndConstraintMorph methodsFor: 'constraints' stamp: 'nk 8/20/2004 06:28'!
inputOwnerHash
	"Answer the identityHash of my first input's root morph"
	^((((inputs at: 1 ifAbsent: [ ^0 ]) ifNil: [ ^0 ]) rootInPasteUp ifNil: [ ^0 ]) owner ifNil: [ ^0 ])
		identityHash 
! !

!NCLineEndConstraintMorph methodsFor: 'constraints' stamp: 'nk 11/1/2004 10:58'!
inputs: aCollection 
	inputs
		withIndexDo: [:ea :index | (ea notNil and: [ ea isHandMorph not]) ifTrue: [
				ea disconnectedFromEnd: self endIndex ofConnector: self owner.
				ea triggerEtoyEvent: #disconnectedFrom from: owner.
				owner triggerEtoyEvent: (firstVertex ifTrue: [ #sourceDisconnecting ] ifFalse: [ #destinationDisconnecting ])]].
	super inputs: aCollection.
	inputs
		do: [:ea | (ea notNil and: [ ea isHandMorph not])
				ifTrue: [ea connectedToEnd: self endIndex ofConnector: self owner.
						ea triggerEtoyEvent: #connectedTo from: owner.
						owner triggerEtoyEvent: (firstVertex ifTrue: [ #sourceConnected ] ifFalse: [ #destinationConnected ])]]
! !

!NCLineEndConstraintMorph methodsFor: 'testing' stamp: 'nk 2/27/2001 13:02'!
isLineConstraint
	^true! !

!NCLineEndConstraintMorph methodsFor: 'accessing' stamp: 'nk 6/22/2003 12:50'!
lastTargetAngle
	^lastTarget
		ifNil: [ 0 ]
		ifNotNil: [ lastTarget second ]! !

!NCLineEndConstraintMorph methodsFor: 'constraints' stamp: 'nk 8/19/2004 22:02'!
lastTargetAngle: angle
	| lastOwnerHash | 
	lastOwnerHash := self lastTargetOwnerHash.
	lastTarget := lastTarget
		ifNil: [ { self constrainedVertex . angle. lastOwnerHash } ]
		ifNotNil: [ { lastTarget first . angle. lastOwnerHash } ].
	^lastTarget! !

!NCLineEndConstraintMorph methodsFor: 'accessing' stamp: 'nk 8/19/2004 17:03'!
lastTargetOwnerHash
	^lastTarget
		ifNil: [ 0 ]
		ifNotNil: [ lastTarget at: 3 ifAbsent: [ 0 ] ]! !

!NCLineEndConstraintMorph methodsFor: 'accessing' stamp: 'nk 4/15/2001 14:32'!
lastTargetPoint
	^lastTarget
		ifNil: [ self constrainedVertex ]
		ifNotNil: [ lastTarget first ]! !

!NCLineEndConstraintMorph methodsFor: 'constraints' stamp: 'nk 8/19/2004 22:03'!
lastTargetPoint: aPoint
	| lastOwnerHash |
	lastOwnerHash := self lastTargetOwnerHash.
	lastTarget := lastTarget
		ifNil: [ { aPoint . 0. lastOwnerHash } ]
		ifNotNil: [ { aPoint . lastTarget second. lastOwnerHash } ].
	^lastTarget! !

!NCLineEndConstraintMorph methodsFor: 'layout' stamp: 'nk 4/8/2001 15:36'!
layoutChanged
	self fixColor.
	submorphs isEmpty
		ifTrue: [ self privateBounds: (Rectangle center: self targetPoint extent: 10@10) ].
	super layoutChanged.
! !

!NCLineEndConstraintMorph methodsFor: 'initialization' stamp: 'nk 4/18/2002 18:09'!
line: l firstVertex: b otherMorph: m
	self constrained: l;
		inputs: { m }.
	firstVertex := b.
! !

!NCLineEndConstraintMorph methodsFor: 'accessing' stamp: 'nk 4/18/2002 17:59'!
lineAttachmentPoint
	^lineAttachmentPoint! !

!NCLineEndConstraintMorph methodsFor: 'accessing' stamp: 'nk 4/18/2002 17:59'!
lineAttachmentPoint: aPointOrNil
	lineAttachmentPoint := aPointOrNil! !

!NCLineEndConstraintMorph methodsFor: 'connection-callbacks' stamp: 'nk 6/15/2002 13:16'!
nearestAttachmentPointOn: m
	^m attachmentPointNearest: (self nextVertexRelativeTo: m)
! !

!NCLineEndConstraintMorph methodsFor: 'connection-callbacks' stamp: 'nk 6/27/2002 12:35'!
nearestOrthogonalPointOn: m
	^m closestOrthogonalPointTo: (self nextVertexRelativeTo: m)
! !

!NCLineEndConstraintMorph methodsFor: 'connection-callbacks' stamp: 'nk 2/15/2002 12:37'!
nearestPointOn: m
	^m closestPointTo: (self nextVertexRelativeTo: m)! !

!NCLineEndConstraintMorph methodsFor: 'connection-callbacks' stamp: 'nk 2/15/2002 12:38'!
nearestPointToCenterOf: m
	^m intersectionWithLineSegmentFromCenterTo: (self nextVertexRelativeTo: m).! !

!NCLineEndConstraintMorph methodsFor: 'private' stamp: 'nk 2/13/2001 18:27'!
nextVertex
	"Answer the point at the next vertex toward the other end of the line"
	^firstVertex
		ifTrue: [ constrained nextToFirstPoint ]
		ifFalse: [ constrained nextToLastPoint ]! !

!NCLineEndConstraintMorph methodsFor: 'connection-callbacks' stamp: 'nk 6/12/2002 07:47'!
nextVertexRelativeTo: m
	^self point: self nextVertex relativeTo: m! !

!NCLineEndConstraintMorph methodsFor: 'halos and balloon help' stamp: 'nk 6/17/2002 22:17'!
okayToBrownDragEasily
	"Answer whether it is appropriate for a drag handle to be shown for the receiver"

	^ self hasAnyInputs not! !

!NCLineEndConstraintMorph methodsFor: 'halos and balloon help' stamp: 'nk 7/4/2003 14:06'!
okayToResizeEasily
	"Answer whether it is appropriate for a sizing handle to be shown for the receiver"

	^ false! !

!NCLineEndConstraintMorph methodsFor: 'drop shadows' stamp: 'nk 6/17/2002 22:13'!
okayToRotateEasily
	"Answer whether it is appropriate for a rotation handle to be shown for the receiver"

	^ false! !

!NCLineEndConstraintMorph methodsFor: 'connection' stamp: 'nk 1/20/2004 10:51'!
preferredConnection: pref
	super preferredConnection: pref.
	self forceRedraw.! !

!NCLineEndConstraintMorph methodsFor: 'private' stamp: 'nk 4/8/2001 16:49'!
privateMoveBy: delta
	super privateMoveBy: delta.
	self lastTargetPoint: self lastTargetPoint + delta
! !

!NCLineEndConstraintMorph methodsFor: 'rotate scale and flex' stamp: 'nk 7/4/2003 14:11'!
rotationDegrees: n
	"I don't get rotated."
	^self! !

!NCLineEndConstraintMorph methodsFor: 'rotate scale and flex' stamp: 'nk 7/4/2003 14:12'!
setExtentFromHalo: anExtent
	"The user has dragged the grow box such that the receiver's extent would be anExtent.
	Ignore the user's wishes completely."
! !

!NCLineEndConstraintMorph methodsFor: 'rotate scale and flex' stamp: 'nk 7/4/2003 14:16'!
setRotationCenterFrom: aPoint! !

!NCLineEndConstraintMorph methodsFor: 'stepping and presenter' stamp: 'nk 5/2/2004 15:08'!
step
	"Modified version of super method that assumes a single input"
	| myInput newTarget |

	myInput := inputs at: 1 ifAbsent: [].
	myInput ifNil: [
		constraint ifNotNil: [ constraint := nil. self changed].
			self dieWithInput ifTrue: [ self delete ].
		^self ].

	newTarget := self target.

	(self needsToApplyConstraintGivenNewTarget: newTarget)
		ifTrue: 
			[self applyConstraint: newTarget.
			lastTarget := newTarget.
			stepTime := self minimumStepTime]
		ifFalse: [self incrementStepTime].

	self ensureInFront.! !

!NCLineEndConstraintMorph methodsFor: 'constraints' stamp: 'nk 8/19/2004 17:08'!
target
	"Answer an Array with my target point and angle"
	| pt |
	pt := "self griddedPoint:" self targetPoint.
	^{ pt . 
		(self nextVertex - pt) degrees.
		self inputOwnerHash
	 }! !

!NCLineEndConstraintMorph methodsFor: 'private' stamp: 'nk 2/13/2001 18:26'!
vertexIndex
	"Answer the index of the end of the line that I'm constraining"
	^firstVertex 
		ifTrue: [ 1 ]
		ifFalse: [ constrained vertices size ]! !

!NCLineEndConstraintMorph methodsFor: 'drawing' stamp: 'nk 4/18/2001 10:09'!
visible
	self forcedInvisible ifTrue: [ ^false ].
	submorphs isEmpty ifFalse: [ ^true ].
	^self preferredConnectionSelector ~~ #connectToNothing! !

NCConstraintMorph subclass: #NCOffsetConstraintMorph
	instanceVariableNames: 'offset'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Connectors-Base'!

!NCOffsetConstraintMorph commentStamp: '<historical>' prior: 0!
I am a ConstraintMorph with an offset. The point I get from evaluating my MessageSend is offset by a point.!

NCOffsetConstraintMorph subclass: #NCLabelMorph
	instanceVariableNames: 'label'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Connectors-Text-Base'!

!NCLabelMorph commentStamp: 'tk 12/1/2004 15:29' prior: 0!
I am a ConstraintMorph that owns a TextMorph. I follow attempts to
move myself by remembering an offset.

I am movable by pressing the shift key, or by selecting my edge.

My menu has an "attach to" choice that lets the user pick what object
to connect to.

I prtetend to be a TextMorph, and delegate any message not understood
to my label.!

!NCLabelMorph class methodsFor: 'instance creation' stamp: 'nk 2/26/2001 21:31'!
includeInNewMorphMenu
	^true! !

!NCLabelMorph class methodsFor: 'instance creation' stamp: 'nk 6/17/2002 11:36'!
initializedInstance
	^self newStandAlone! !

!NCLabelMorph class methodsFor: 'instance creation' stamp: 'nk 9/2/2004 10:04'!
supplementaryPartsDescriptions

	Preferences eToyFriendly ifTrue: [ ^#() ].

	^ {
			DescriptionForPartsBin
					formalName: 'Label'
					categoryList: #(Connectors)
					documentation: 'A label that can be attached to another morph by dropping or menu choice and will follow the morph to which it is attached.'
					globalReceiverSymbol: #NCLabelMorph 
					nativitySelector: #newStandAlone.	
	}! !

!NCLabelMorph class methodsFor: 'defaults' stamp: 'nk 5/13/2004 14:33'!
textClass
	^Smalltalk at: #NCGrabbableDisplayTextMorph
		ifAbsent: [ Smalltalk at: #NCTextMorph ifAbsent: [ TextMorph ]]! !

!NCLabelMorph methodsFor: 'dropping/grabbing' stamp: 'nk 10/13/2003 20:47'!
aboutToBeGrabbedBy: aHand 
	| oldFocus |
	"label isLocked
		ifFalse: [oldFocus := aHand keyboardFocus.
			(oldFocus respondsTo: #passKeyboardFocusTo:)
				ifTrue: [oldFocus passKeyboardFocusTo: label]
				ifFalse: [aHand newKeyboardFocus: self label]]."
	^ super aboutToBeGrabbedBy: aHand! !

!NCLabelMorph methodsFor: 'accessing' stamp: 'nk 1/20/2004 11:10'!
acceptOnCR
	^label acceptOnCR! !

!NCLabelMorph methodsFor: 'accessing' stamp: 'nk 1/20/2004 11:10'!
acceptOnCR: aBoolean
	label acceptOnCR: aBoolean! !

!NCLabelMorph methodsFor: 'accessing' stamp: 'nk 1/20/2004 11:10'!
acceptOnCROnOff
	self acceptOnCR: self acceptOnCR not! !

!NCLabelMorph methodsFor: 'accessing' stamp: 'nk 1/20/2004 11:11'!
acceptOnCRString
	^label acceptOnCRString! !

!NCLabelMorph methodsFor: 'accessing' stamp: 'nk 1/20/2004 11:11'!
acceptOnFocusLoss
	^label acceptOnFocusLoss! !

!NCLabelMorph methodsFor: 'accessing' stamp: 'nk 1/20/2004 11:11'!
acceptOnFocusLoss: aBoolean
	^label acceptOnFocusLoss: aBoolean! !

!NCLabelMorph methodsFor: 'accessing' stamp: 'nk 1/20/2004 11:11'!
acceptOnFocusLossOnOff
	^label acceptOnFocusLossOnOff! !

!NCLabelMorph methodsFor: 'accessing' stamp: 'nk 1/20/2004 11:12'!
acceptOnFocusLossString
	^label acceptOnFocusLossString! !

!NCLabelMorph methodsFor: 'menus' stamp: 'dgd 8/26/2004 15:03'!
addYellowButtonMenuItemsTo: aMenu event: evt
	aMenu
		add: 'delete' translated target: self selector: #dismissMorph;
		add: 'move' translated target: evt hand selector: #attachMorph: argument: self.

	super addYellowButtonMenuItemsTo: aMenu event: evt.
	aMenu
		addLine;
		add: 'font...' translated action: #promptForFont;
		addUpdating: #unlockOrLockTextPhrase action: #toggleTextLocked.
	label addFitAndWrapItemsTo: aMenu.
	label addAcceptMenuItemsTo: aMenu.! !

!NCLabelMorph methodsFor: 'private' stamp: 'nk 6/17/2002 16:15'!
adjustTextAnchor: ignored! !

!NCLabelMorph methodsFor: 'accessing' stamp: 'nk 10/21/2003 12:17'!
appendKeyStroke: anEvent
	| index |
	label ifNil: [^self ].
	anEvent hand newKeyboardFocus: label.
	index := label lastCharacterIndex.
	label selectFrom: index + 1 to: index.
	label keyStroke: anEvent.! !

!NCLabelMorph methodsFor: 'accessing' stamp: 'nk 2/26/2001 14:32'!
borderColor
	"Make me work with the properties panel"
	^super color! !

!NCLabelMorph methodsFor: 'accessing' stamp: 'nk 2/26/2001 14:32'!
borderColor: aColor
	"Make me work with the properties panel"
	super color: aColor! !

!NCLabelMorph methodsFor: 'menus' stamp: 'nk 4/14/2002 14:59'!
changeMargins: evt
	label changeMargins: evt.
! !

!NCLabelMorph methodsFor: 'menus' stamp: 'nk 6/17/2002 17:34'!
chooseEmphasisOrAlignment
	| wasEmptyOrNil |
	label isLocked ifTrue: [ ^self ].
	wasEmptyOrNil :=  label selection isEmptyOrNil.
	wasEmptyOrNil ifTrue: [ label selectAll ].
	label chooseEmphasisOrAlignment.
	wasEmptyOrNil ifTrue: [ label selectFrom: 1 to: 0 ].
! !

!NCLabelMorph methodsFor: 'menus' stamp: 'nk 5/8/2003 14:22'!
chooseFont
	| wasEmptyOrNil oldWrap oldHResizing |
	label isLocked ifTrue: [ ^self ].
	wasEmptyOrNil :=  label selection isEmptyOrNil.
	wasEmptyOrNil ifTrue: [ label selectAll ].
	oldWrap := label isWrapped.
	oldHResizing := label hResizing.
	label wrapFlag: false.
	label chooseFont.
	label releaseCachedState; changed.
	self extent: label extent.
	wasEmptyOrNil ifTrue: [ label selectFrom: 1 to: 0 ].
	label wrapFlag: oldWrap.
	label hResizing: oldHResizing.
! !

!NCLabelMorph methodsFor: 'menus' stamp: 'nk 6/17/2002 17:35'!
chooseStyle
	| wasEmptyOrNil |
	label isLocked ifTrue: [ ^self ].
	wasEmptyOrNil :=  label selection isEmptyOrNil.
	wasEmptyOrNil ifTrue: [ label selectAll ].
	label chooseStyle.
	wasEmptyOrNil ifTrue: [ label selectFrom: 1 to: 0 ].
! !

!NCLabelMorph methodsFor: 'accessing' stamp: 'nk 2/26/2001 14:34'!
color
	"Make me work with the properties panel"
	^label ifNil: [ super color ]
		ifNotNil: [ label color ]! !

!NCLabelMorph methodsFor: 'accessing' stamp: 'nk 2/26/2001 14:34'!
color: aColor
	"Make me work with the properties panel"
	label ifNil: [ super color: aColor ]
		ifNotNil: [ label color: aColor ]! !

!NCLabelMorph methodsFor: 'geometry testing' stamp: 'nk 4/13/2002 10:58'!
containsPoint: aPoint
	^self visibleBounds containsPoint: aPoint
! !

!NCLabelMorph methodsFor: 'initialization' stamp: 'nk 12/28/2003 13:46'!
defaultFont
	^Preferences connectorsLabelFont! !

!NCLabelMorph methodsFor: 'error handling' stamp: 'nk 8/21/2004 11:39'!
doesNotUnderstand: aMessage
	"As much as possible, delegate all behavior to my label"
	^(label respondsTo: aMessage selector)
		ifTrue: [ 
"			Transcript nextPutAll: 'NCLabelMorph delegating '; print: aMessage; cr."
			label perform: aMessage selector withArguments: aMessage arguments ]
		ifFalse: [ super doesNotUnderstand: aMessage ]! !

!NCLabelMorph methodsFor: 'testing' stamp: 'nk 3/11/2002 21:19'!
drawingElement
	"Answer the drawing element that I constrain, typically either a Connector or a label"
	^self! !

!NCLabelMorph methodsFor: 'visual properties' stamp: 'nk 2/26/2001 14:36'!
fillStyle: aColor
	"Make me work with the properties panel"
	label ifNil: [ super fillStyle: aColor ]
		ifNotNil: [ label fillStyle: aColor ]! !

!NCLabelMorph methodsFor: 'objects from disk' stamp: 'tk 12/1/2004 15:26'!
fixUponLoad: aProject seg: anImageSegment
	"We are in an old project that is being loaded from disk.
Fix up conventions that have changed."

	"NCLabelMorph passes any message it does not understand to its text"
	label fixUponLoad: aProject seg: anImageSegment.

	^ super fixUponLoad: aProject seg: anImageSegment	"me,
not the label"
! !

!NCLabelMorph methodsFor: 'visual properties' stamp: 'nk 5/19/2003 22:42'!
forceTextLocked: aBoolean
	label lock: aBoolean! !

!NCLabelMorph methodsFor: 'event handling' stamp: 'nk 5/19/2003 22:53'!
handlerForMouseDown: anEvent
	"Make me movable with the shift pressed."
	self inPartsBin ifTrue: [ ^nil ].
	((anEvent shiftPressed) or: [ anEvent yellowButtonPressed])
		ifTrue: [ ^self ].
	^super handlerForMouseDown: anEvent
! !

!NCLabelMorph methodsFor: 'event handling' stamp: 'nk 10/21/2003 11:56'!
handlesKeyboard: evt
	label ifNil: [ ^false ].
	^label handlesKeyboard: evt.! !

!NCLabelMorph methodsFor: 'event handling' stamp: 'nk 5/19/2003 22:45'!
handlesMouseDown: anEvent
	"Make me movable with the shift pressed."
	self inPartsBin ifTrue: [ ^false ].
	^((anEvent shiftPressed) or: [ anEvent yellowButtonPressed])
! !

!NCLabelMorph methodsFor: 'event handling' stamp: 'nk 4/3/2002 12:21'!
handlesMouseOver: evt
	^true! !

!NCLabelMorph methodsFor: 'initialization' stamp: 'nk 6/17/2002 20:35'!
initialize
	super initialize.
	self
		hResizing: #shrinkWrap;
		vResizing: #shrinkWrap;
		layoutPolicy: TableLayout new;
		color: Color transparent;
		connectToNearestAttachmentPoint.

! !

!NCLabelMorph methodsFor: 'initialization' stamp: 'nk 10/13/2003 20:38'!
initializeToStandAlone
	self initialize; string: 'Label' wrap: false! !

!NCLabelMorph methodsFor: 'classification' stamp: 'nk 2/26/2001 15:53'!
isFlexMorph
	^label owner ~= self! !

!NCLabelMorph methodsFor: 'testing' stamp: 'nk 3/31/2002 11:28'!
isLabelConstraint
	^true! !

!NCLabelMorph methodsFor: 'visual properties' stamp: 'nk 5/19/2003 22:42'!
isTextForcedLocked
	^label isLocked! !

!NCLabelMorph methodsFor: 'menus' stamp: 'nk 4/14/2002 14:39'!
isTextMorph
	^true
! !

!NCLabelMorph methodsFor: 'dropping/grabbing' stamp: 'nk 4/14/2002 17:11'!
justDroppedInto: newOwner event: evt
	"Determine whether we need to be embedded or attached."
	(newOwner dropEnabled and: [ newOwner isPlayfieldLike not ])
		ifTrue: [ ^super justDroppedInto: newOwner event: evt ].
	(newOwner rootMorphsAtGlobal: evt cursorPoint)
		do: [ :ea | ea ~~ self ifTrue: [ ^self attachTo: ea at: evt cursorPoint ] ].
	super justDroppedInto: newOwner event: evt! !

!NCLabelMorph methodsFor: 'event handling' stamp: 'nk 10/21/2003 11:57'!
keyDown: evt
	label keyDown: evt
! !

!NCLabelMorph methodsFor: 'event handling' stamp: 'nk 10/21/2003 11:56'!
keyStroke: evt
	label keyStroke: evt
! !

!NCLabelMorph methodsFor: 'event handling' stamp: 'nk 10/21/2003 11:58'!
keyUp: evt
	label keyUp: evt
! !

!NCLabelMorph methodsFor: 'event handling' stamp: 'nk 10/21/2003 11:57'!
keyboardFocusChange: aBoolean
	label keyboardFocusChange: aBoolean
! !

!NCLabelMorph methodsFor: 'accessing' stamp: 'nk 2/27/2001 00:08'!
label
	^label! !

!NCLabelMorph methodsFor: 'event handling' stamp: 'nk 11/10/2003 23:25'!
matchForGesture: aCRGesture
	"Answer a number corresponding to how close my text height is to that of aCRGesture."
	^label matchForGesture: aCRGesture
! !

!NCLabelMorph methodsFor: 'event handling' stamp: 'nk 6/14/2004 15:26'!
mouseDown: evt
	^(evt shiftPressed or: [ evt yellowButtonPressed ])
		ifTrue: [ label releaseEditor.
			evt shiftPressed ifTrue: [ evt hand grabMorph: self ]
				ifFalse: [ super mouseDown: evt ]]
		ifFalse: [ label mouseDown: evt ].! !

!NCLabelMorph methodsFor: 'event handling' stamp: 'nk 6/17/2002 15:46'!
mouseEnter: evt
	label isLocked ifFalse: [
		Preferences mouseOverForKeyboardFocus ifTrue:
			[ evt hand newKeyboardFocus: label ].
	]
! !

!NCLabelMorph methodsFor: 'event handling' stamp: 'nk 4/28/2002 09:03'!
mouseLeave: evt
	Preferences mouseOverForKeyboardFocus ifTrue:
		[ evt hand releaseKeyboardFocus: label ]
! !

!NCLabelMorph methodsFor: 'private' stamp: 'nk 10/13/2003 21:11'!
noteNewOwner: aMorph
	super noteNewOwner: aMorph.

	"(aMorph notNil and: [ aMorph isHandMorph ])
		ifTrue: [ label isLocked ifFalse: [aMorph newKeyboardFocus: label] ]."! !

!NCLabelMorph methodsFor: 'accessing' stamp: 'nk 10/21/2003 13:04'!
openForEditingFrom: anEvent
	| index |
	label ifNil: [^self ].
	anEvent hand newKeyboardFocus: label.
	index := label lastCharacterIndex.
	label selectFrom: index + 1 to: index.
! !

!NCLabelMorph methodsFor: 'event handling' stamp: 'nk 6/17/2003 14:42'!
promptForFont
	TextStyle promptForFont: 'Choose label font:' andSendTo: self withSelector: #setFontTo:! !

!NCLabelMorph methodsFor: 'selection' stamp: 'nk 6/17/2002 12:08'!
selectAll
	label selectAll! !

!NCLabelMorph methodsFor: 'visual properties' stamp: 'nk 2/26/2001 18:51'!
setFontTo: aFont
	label beAllFont: aFont;
	composeToBounds! !

!NCLabelMorph methodsFor: 'stepping and presenter' stamp: 'nk 6/17/2002 17:19'!
step
	| inputPU |
	super step.
	self input ifNotNil: [
		"change me to the same pasteUp that my input is in"
		(self pasteUpMorph ~~ (inputPU := self input pasteUpMorph)) ifTrue: [
			owner isHandMorph ifFalse: [
				inputPU ifNotNil: [
					inputPU addMorph: (self isFlexed ifTrue: [ owner ] ifFalse: [ self ])
				]
		 	]
		]
	]! !

!NCLabelMorph methodsFor: 'accessing' stamp: 'nk 10/21/2003 11:18'!
string
	^label contents asString! !

!NCLabelMorph methodsFor: 'initialization' stamp: 'nk 10/13/2003 20:18'!
string: aString

	self string: aString wrap: false! !

!NCLabelMorph methodsFor: 'initialization' stamp: 'nk 1/29/2004 14:06'!
string: aString font: aFont wrap: shouldWrap minWidth: chars

	label ifNotNil: [ label delete minWidth ].
	label := shouldWrap
		ifTrue: [ self class textClass new contentsWrapped: aString ]
		ifFalse: [ self class textClass new contentsAsIs: aString ].
	label beAllFont: aFont.
	label releaseCachedState; changed.
	label margins: 4@4.
	self extent: label extent.
	label hResizing: #spaceFill.
	self addMorphBack: label.
! !

!NCLabelMorph methodsFor: 'initialization' stamp: 'nk 8/2/2002 07:40'!
string: aString fontName: aName size: aSize wrap: shouldWrap
	^self string: aString fontName: aName size: aSize wrap: shouldWrap minWidth: 10
! !

!NCLabelMorph methodsFor: 'initialization' stamp: 'nk 11/10/2003 23:24'!
string: aString fontName: aName size: aSize wrap: shouldWrap minWidth: chars

	label ifNotNil: [ label delete minWidth ].
	label := shouldWrap
		ifTrue: [ self class textClass new contentsWrapped: aString ]
		ifFalse: [ self class textClass new contentsAsIs: aString ].
	label fontName: aName size: aSize.
	label releaseCachedState; changed.
	label margins: 4@4.
	self extent: label extent.
	label hResizing: #spaceFill.
	self addMorphBack: label.
! !

!NCLabelMorph methodsFor: 'initialization' stamp: 'nk 8/2/2002 07:41'!
string: aString minWidth: chars

	self string: aString wrap: true minWidth: chars! !

!NCLabelMorph methodsFor: 'initialization' stamp: 'nk 1/29/2004 14:10'!
string: aString wrap: shouldWrap 
	self 
		string: aString
		font: self defaultFont
		wrap: shouldWrap
		minWidth: 10! !

!NCLabelMorph methodsFor: 'initialization' stamp: 'nk 1/29/2004 14:06'!
string: aString wrap: shouldWrap minWidth: chars 
	self 
		string: aString
		font: self defaultFont
		wrap: shouldWrap
		minWidth: chars! !

!NCLabelMorph methodsFor: 'accessing' stamp: 'nk 10/21/2003 11:18'!
text
	^label contents! !

!NCLabelMorph methodsFor: 'accessing' stamp: 'nk 6/18/2003 15:34'!
textAlignment
	^label textAlignment! !

!NCLabelMorph methodsFor: 'accessing' stamp: 'nk 6/18/2003 15:34'!
textAlignmentSymbol
	^label textAlignmentSymbol! !

!NCLabelMorph methodsFor: 'accessing' stamp: 'nk 2/26/2001 12:33'!
textColor
	^label color! !

!NCLabelMorph methodsFor: 'accessing' stamp: 'nk 2/26/2001 12:33'!
textColor: aColor
	label color: aColor! !

!NCLabelMorph methodsFor: 'visual properties' stamp: 'nk 3/27/2002 22:03'!
textWrap: aBoolean
	label wrapFlag: aBoolean! !

!NCLabelMorph methodsFor: 'menus' stamp: 'nk 5/19/2003 22:42'!
toggleTextLocked
	label lock: label isLocked not
! !

!NCLabelMorph methodsFor: 'menus' stamp: 'dgd 8/26/2004 15:03'!
unlockOrLockTextPhrase
	^ self isTextForcedLocked
		ifTrue: ['unlock text' translated]
		ifFalse: ['lock text' translated] ! !

!NCLabelMorph methodsFor: 'geometry testing' stamp: 'nk 4/13/2002 10:56'!
visibleBounds
	"Answer my visible bounds for the purposes of label motion and 
	containsPoint: "
	^ (label notNil
			and: [super color isTransparent
					and: [super borderWidth == 0
							and: [super borderColor isTransparent]]])
		ifTrue: [label bounds]
		ifFalse: [self bounds]
! !

!NCLabelMorph methodsFor: 'event handling' stamp: 'nk 6/17/2002 15:48'!
wantsKeyboardFocusFor: aSubmorph
	label isLocked ifTrue: [ ^false ].
	aSubmorph wouldAcceptKeyboardFocus ifTrue: [^ true].
	^ super wantsKeyboardFocusFor: aSubmorph! !

NCLabelMorph subclass: #NCSmartLabelMorph
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Connectors-Text-Base'!

!NCSmartLabelMorph commentStamp: '<historical>' prior: 0!
I am a label that attempts to keep out of the way of any connectors I'm attached to.!

!NCSmartLabelMorph class methodsFor: 'parts bin' stamp: 'nk 9/2/2004 10:04'!
supplementaryPartsDescriptions

	Preferences eToyFriendly ifTrue: [ ^#() ].

	^ {
		DescriptionForPartsBin
				formalName: 'Smart Label'
				categoryList: #(Connectors)
				documentation: 'A label that can be attached to another morph by dropping or menu choice and will follow the morph to which it is attached, trying to stay out of the way.'
				globalReceiverSymbol: #NCSmartLabelMorph 
				nativitySelector: #newStandAlone.
	}! !

!NCSmartLabelMorph methodsFor: 'menus' stamp: 'dgd 8/26/2004 15:05'!
addYellowButtonMenuItemsTo: aMenu event: evt
	super addYellowButtonMenuItemsTo: aMenu event: evt.
	aMenu
		add: 'change margins' translated target: self selector: #changeMargins:.
! !

!NCSmartLabelMorph methodsFor: 'constraints' stamp: 'nk 6/17/2002 17:33'!
applyConstraint: aPoint 
	| input newBounds newCenter |
	(self pasteUpMorph isNil or: [ (input := self input) isNil ])
		ifTrue: [^ super applyConstraint: aPoint].
	"self offset: 0 @ 0."
	newCenter := self constrainedCenter: aPoint.
	newBounds := self visibleBounds translateBy: newCenter - self center.
	self flag: #TODO. "need to deal with flexed labels"
	self isFlexed ifFalse: [
	newBounds := input relocateLabelFrom: newBounds.
	].
	self position: newBounds origin.
! !

!NCSmartLabelMorph methodsFor: 'attachments-nk' stamp: 'nk 4/13/2002 11:34'!
attachUnobnoxiouslyTo: aMorph at: aPoint
	self attachTo: aMorph at: aPoint.
! !

!NCSmartLabelMorph methodsFor: 'constraints' stamp: 'nk 4/14/2002 10:41'!
needsToApplyConstraintGivenNewTarget: newTarget 
	"newTarget has just been returned by 'self target'. 
	Do I have to apply the constraint? 
	If I return true, applyConstraint: will be called."
	| changed |
	changed := (super needsToApplyConstraintGivenNewTarget: newTarget)
		or: [ | same oldHash newHash myInput |
			oldHash := self valueOfProperty: #savedBoundsHash ifAbsent: [ 0 ].
			myInput := self input.
			newHash := myInput ifNotNil: [ myInput boundsSignatureHash ] ifNil: [ 0 ].
			same := oldHash = newHash.
			same ifFalse: [ self setProperty: #savedBoundsHash toValue: newHash ].
			same not
		].
	changed ifTrue: [ self offset: 0@0 ].
	^changed
! !

!NCSmartLabelMorph methodsFor: 'stepping' stamp: 'nk 8/25/2003 15:52'!
stepTime
	^500! !

!NCOffsetConstraintMorph methodsFor: 'constraints' stamp: 'nk 4/13/2002 10:01'!
applyConstraint: aPoint
	"Make sure that I stay within my owner's bounds"
	self center: (self constrainedCenter: aPoint)! !

!NCOffsetConstraintMorph methodsFor: 'attachments-nk' stamp: 'nk 6/17/2002 17:55'!
attachUnobnoxiouslyTo: aMorph at: aPoint
	"If the attachment would make me overlap aMorph, move me away from aMorph's center."
 	| incr limit newBounds |
	self offset: 0@0.
	self center: aPoint.
	self attachTo: aMorph at: aPoint.	"sets constraint block"
	incr := 0@1.
	limit := 2000.
	newBounds := self bounds.
	[ limit < 0 or: [ incr isZero ] ] whileFalse: [
		incr := aMorph nudgeForLabel: newBounds.
		limit := limit - 1.
		newBounds := newBounds translateBy: incr.
	].
	self position: newBounds origin.
	self setOffset.! !

!NCOffsetConstraintMorph methodsFor: 'constraints' stamp: 'nk 6/17/2002 17:30'!
constrainedCenter: aPoint
	"Make sure that I stay within my owner's bounds. Answer a new center position"

	| newCenter newBounds myPU |
	myPU := self pasteUpMorph.
	newCenter := aPoint + offset.
	self isFlexed ifFalse: [
	newBounds := bounds translateBy: newCenter - self center.
	myPU
		ifNotNil: [ newCenter := newCenter + (newBounds amountToTranslateWithin: myPU bounds)].
	].
	^newCenter
! !

!NCOffsetConstraintMorph methodsFor: 'initialization' stamp: 'nk 3/31/2002 11:31'!
initialize
	super initialize.
	offset := 0@0.
! !

!NCOffsetConstraintMorph methodsFor: 'constraints' stamp: 'nk 4/5/2001 17:56'!
input: aMorph
	super input: aMorph.
	offset := 0@0.
	self fixMorphicLayer! !

!NCOffsetConstraintMorph methodsFor: 'accessing' stamp: 'nk 6/17/2002 17:13'!
offset: aPoint
	offset := aPoint.! !

!NCOffsetConstraintMorph methodsFor: 'stepping and presenter' stamp: 'nk 4/4/2001 18:24'!
setOffset
	offset := self center - self target! !

!NCOffsetConstraintMorph methodsFor: 'stepping and presenter' stamp: 'nk 8/25/2003 14:29'!
step
	| outer |
	outer := self root.
	(outer hasHalo or: [ outer owner isHandMorph ])
		ifTrue: [ self setOffset ]
		ifFalse: [ super step. self ensureInFront. ]
! !

!NCOffsetConstraintMorph methodsFor: 'stepping and presenter' stamp: 'nk 4/5/2001 17:41'!
visible
	^self forcedInvisible not.! !

Morph subclass: #NCCurveHandleMorph
	instanceVariableNames: 'dragStartPoint targetPoint desiredPoint getPoint setPoint offset'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Connectors-Lines and Curves'!

!NCCurveHandleMorph methodsFor: 'event handling' stamp: 'nk 8/21/2003 09:21'!
adjustCurve
	| retval |
	retval := setPoint valueWithEnoughArguments: { desiredPoint }.
	targetPoint := getPoint value.
	^retval! !

!NCCurveHandleMorph methodsFor: 'naming' stamp: 'cmm 12/4/2010 20:04'!
defaultNameStemForInstances
	^ self class name
		copyFrom: 3
		to: self class name size! !

!NCCurveHandleMorph methodsFor: 'accessing' stamp: 'nk 8/21/2003 13:59'!
dragStartPoint
	^dragStartPoint! !

!NCCurveHandleMorph methodsFor: 'drawing' stamp: 'nk 7/15/2003 07:20'!
drawOn: aCanvas
	aCanvas frameAndFillRectangle: self bounds
		fillColor: self fillStyle
		borderWidth: 1
		borderColor: self fillStyle asColor muchDarker! !

!NCCurveHandleMorph methodsFor: 'event handling' stamp: 'nk 8/21/2003 09:31'!
handleMove: evt
	| lastPoint |
	lastPoint := desiredPoint.
	desiredPoint := evt position + offset.
	lastPoint = desiredPoint ifTrue: [ ^self ].
	self adjustCurve
		ifTrue: [ self center: evt position ]
		ifFalse: [ self stopHandleDrag: evt ].
! !

!NCCurveHandleMorph methodsFor: 'initialization' stamp: 'nk 8/22/2003 20:47'!
initialize
	super initialize.
	self extent: 10@10.
	offset := 0@0.
	self on: #mouseDown send: #startHandleDrag: to: self.
	self on: #mouseUp send: #stopHandleDrag: to: self.
	self on: #mouseMove send: #handleMove: to: self.! !

!NCCurveHandleMorph methodsFor: 'testing' stamp: 'nk 8/9/2004 16:24'!
isConnectable
	^false
! !

!NCCurveHandleMorph methodsFor: 'testing' stamp: 'nk 8/18/2003 09:15'!
isCurveHandle
	^true! !

!NCCurveHandleMorph methodsFor: 'testing' stamp: 'nk 6/12/2004 16:20'!
isHandleMorph
	^true! !

!NCCurveHandleMorph methodsFor: 'layers' stamp: 'nk 7/14/2003 22:32'!
morphicLayerNumber
	^20! !

!NCCurveHandleMorph methodsFor: 'event handling' stamp: 'nk 2/28/2006 10:35'!
mouseUpEvent
	"Synthesize a mouse up event for use by Players, etc."
	^MouseButtonEvent new setType: #mouseUp position: self center buttons: MouseEvent redButton hand: self world activeHand.! !

!NCCurveHandleMorph methodsFor: 'initialization' stamp: 'nk 1/14/2004 21:46'!
setPointSelector
	^setPoint selector! !

!NCCurveHandleMorph methodsFor: 'initialization' stamp: 'nk 1/14/2004 21:46'!
setPointSelector: sel
	setPoint selector: sel! !

!NCCurveHandleMorph methodsFor: 'event handling' stamp: 'nk 8/19/2003 12:18'!
startHandleDrag: evt
	evt hand newMouseFocus: self.
	self center: evt position.
	desiredPoint := getPoint value.
	offset := desiredPoint - evt position.
	dragStartPoint := evt position.
	self triggerEvent: #dragStart.! !

!NCCurveHandleMorph methodsFor: 'stepping and presenter' stamp: 'nk 8/22/2003 17:39'!
step
	| newCenter |
	dragStartPoint ifNotNil: [ ^self ].
	newCenter := getPoint value rounded.
	(self center ~= newCenter) ifTrue: [ self center: newCenter ].! !

!NCCurveHandleMorph methodsFor: 'stepping and presenter' stamp: 'nk 8/22/2003 21:44'!
stepTime
	^100! !

!NCCurveHandleMorph methodsFor: 'event handling' stamp: 'nk 8/18/2003 10:24'!
stopHandleDrag: evt
	dragStartPoint := desiredPoint := nil.
	evt hand releaseMouseFocus: self.
	self triggerEvent: #dragStop.! !

!NCCurveHandleMorph methodsFor: 'initialization' stamp: 'nk 8/18/2003 08:55'!
target
	^getPoint receiver! !

!NCCurveHandleMorph methodsFor: 'initialization' stamp: 'nk 8/22/2003 17:32'!
target: aMorph arguments: args getPointSelector: getPointSel setPointSelector: setPointSel
	getPoint := WeakMessageSend receiver: aMorph selector: getPointSel arguments: args.
	setPoint := WeakMessageSend receiver: aMorph selector: setPointSel arguments: args.
	targetPoint := nil.
	self center: getPoint value.! !

!NCCurveHandleMorph methodsFor: 'initialization' stamp: 'nk 8/20/2003 16:16'!
vertices
	^getPoint arguments! !

Morph subclass: #NCHighlightMorph
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Connectors-Base'!

!NCHighlightMorph methodsFor: 'adjacency layout' stamp: 'nk 11/26/2002 11:23'!
beAdjacent
	^false! !

!NCHighlightMorph methodsFor: 'naming' stamp: 'cmm 12/4/2010 20:05'!
defaultNameStemForInstances
	^ self class name
		copyFrom: 3
		to: self class name size! !

!NCHighlightMorph methodsFor: 'layout-properties' stamp: 'nk 9/10/2004 11:45'!
disableTableLayout
	"Layout specific. Disable laying out the receiver in table layout"
	^true! !

!NCHighlightMorph methodsFor: 'classification' stamp: 'nk 8/9/2004 16:22'!
isConnectable
	^false
! !

!NCHighlightMorph methodsFor: 'classification' stamp: 'nk 11/26/2002 11:23'!
isHighlight
	^true! !

!NCHighlightMorph methodsFor: 'WiW support' stamp: 'nk 11/26/2002 11:24'!
morphicLayerNumber
	^10! !

!NCHighlightMorph methodsFor: 'classification' stamp: 'nk 8/9/2004 16:08'!
wantsAttachmentFromEnd: anEnd ofConnector: aConnector
	^false
! !

!PartsBin class methodsFor: '*connectors-instance creation' stamp: 'nk 9/1/2004 20:06'!
newPartsBinWithOrientation: aListDirection andColor: aColor from: quadList buttonClass: buttonClass
	"Answer a new PartBin object, to run horizontally or vertically, obtaining its elements from the list of tuples of the form:
		(<receiver> <selector> <label> <balloonHelp>)"

	^ (self new)
		color: aColor;
		listDirection: aListDirection quadList: quadList buttonClass: buttonClass! !

!PartsBin class methodsFor: '*connectors-instance creation' stamp: 'nk 3/20/2002 16:51'!
newPartsBinWithOrientation: aListDirection from: quadList buttonClass: buttonClass
	"Answer a new PartBin object, to run horizontally or vertically, obtaining its elements from the list of tuples of the form:
		(<receiver> <selector> <label> <balloonHelp>)"

	^ self new listDirection: aListDirection quadList: quadList buttonClass: buttonClass! !

!String methodsFor: '*connectors-comparing' stamp: 'nk 5/16/2002 07:39'!
beginsWith2: prefix
        "Answer whether the receiver begins with the given prefix string.
        The comparison is NOT case-sensitive."


        self size < prefix size ifTrue: [^ false].
        self first asLowercase == prefix first asLowercase ifFalse: [^false].
        ^ (self findSubstring: prefix in: self startingAt: 1
                        matchTable: CaseInsensitiveOrder) = 1
! !

!String methodsFor: '*connectors-converting' stamp: 'nk 3/27/2001 12:28'!
splitOnCapBoundaries
	| stream prev |
	stream := WriteStream on: (self species new: self size * 2).
	prev := self at: 1.
	self do: [ :c |
		(c = $:)
			ifTrue: [ stream nextPutAll: '...' ]
			ifFalse: [
				((c isUppercase) and: [ c isLowercase ~= prev isLowercase ])
					ifTrue: [ stream space; nextPut: c asLowercase ]
					ifFalse: [ stream nextPut: c ].
			].
		prev := c.
	].
	^stream contents! !

!LineSegment methodsFor: '*connectors-intersection' stamp: 'nk 8/19/2003 18:04'!
intersectionsWithLineFrom: e1 to: e2
	"Answer the set of intersections, if any"
	| intersection |
	intersection := self intersectionWith: (LineSegment from: e1 to: e2).
	^intersection ifNotNil: [ { intersection } ] ifNil: [ {} ].! !

!LineSegment methodsFor: '*connectors-vector functions' stamp: 'nk 8/18/2003 21:35'!
lineSegmentsWithParametricValueDo: aBlock
	"Evaluate aBlock with the receiver's line segments"
	aBlock value: start value: end value: 0.0 value: 1.0! !

!TTGlyph methodsFor: '*connectors' stamp: 'nk 8/17/2003 15:43'!
asFormWithExtent: extent fgColor: fgColor bgColor: bgColor depth: depth replaceColor: replaceColorFlag
	| form canvas transform |
	transform := self transformForExtent: extent.
	form := Form extent:  extent depth: depth.
	form fillColor: bgColor.
	canvas := BalloonCanvas on: form.
	canvas aaLevel: 4.
	canvas transformBy: transform.
	self drawOn: canvas fgColor: fgColor.
	^ form! !

!TTGlyph methodsFor: '*connectors' stamp: 'nk 8/19/2003 13:09'!
contoursShiftedToStartWith: aPoint
	"Answer a shifted version of my contours, in which the first and last points in the first contour are aPoint."
	| newContours shifted firstFound pivotFound |
	newContours := OrderedCollection new.
	firstFound := false.
	contours withIndexDo: [ :c :cIndex | | pointIndex |
		pivotFound := true.
		pointIndex := (1 to: c size by: 3) detect: [ :i | (c at: i) = aPoint ] ifNone: [ pivotFound := false. 1 ].
		pivotFound ifTrue: [
			shifted := ShortPointArray new: c size.
			shifted replaceFrom: 1 to: (c size - pointIndex) + 1
				with: c startingAt: pointIndex.
			shifted replaceFrom: (c size - pointIndex) + 2 to: shifted size
				with: c startingAt: 1.
			firstFound ifFalse: [ newContours addFirst: shifted. firstFound := true ]
				ifTrue: [ newContours addLast: shifted ].
		] ifFalse: [ newContours addLast: c ].
	].
	^newContours asArray! !

!TTGlyph methodsFor: '*connectors' stamp: 'nk 8/17/2003 17:01'!
drawContourEndDotsOn: canvas 
	| dotSize |
	dotSize := ((bounds extent x max: bounds extent y) / 20) asPoint asFloatPoint.
	contours withIndexDo: 
			[:a :i | 
			canvas 
				drawOval: (Rectangle center: a first extent: dotSize)
				color: (Color 
						h: i * 60 + 100.0
						s: 1.0
						v: 1.0 alpha: 0.3)
				borderWidth: 0
				borderColor: nil.
			canvas 
				drawOval: (Rectangle center: a last extent: dotSize)
				color: (Color 
						h: i * 60 + 200.0
						s: 1.0
						v: 1.0  alpha: 0.3)
				borderWidth: 0
				borderColor: nil]! !

!TTGlyph methodsFor: '*connectors' stamp: 'nk 8/17/2003 16:58'!
drawOn: aCanvas fgColor: fgColor 
	| canvas |
	canvas := aCanvas asBalloonCanvas.
	canvas aaLevel: 4.
	canvas 
		drawGeneralBezierShape: self contours
		color: fgColor
		borderWidth: 0
		borderColor: fgColor.
! !

!TTGlyph methodsFor: '*connectors' stamp: 'nk 8/17/2003 15:22'!
transformForExtent: extent
	| scale pointsRect newRect |
	contours do: [ :c | | rect | 
		rect := Rectangle encompassing: c.
		pointsRect := pointsRect ifNil: [ rect ] ifNotNil: [ pointsRect quickMerge: rect ]].
	newRect := pointsRect scaleAndTranslateToFitInside: (0@0 extent: extent asFloatPoint).
	scale := newRect extent asFloatPoint / pointsRect extent.
	^(MatrixTransform2x3 withScale: scale asPoint * (1 @ -1))
		composedWithLocal: (MatrixTransform2x3 withOffset: ((pointsRect origin x) @ (pointsRect corner y)) negated).! !

!TTGlyph methodsFor: '*connectors' stamp: 'nk 8/19/2003 12:49'!
transformForRect: outsideRect rotation: angle
	| scale pointsRect newRect |
	contours do: [ :c | | rect | 
		rect := Rectangle encompassing: c.
		pointsRect := pointsRect ifNil: [ rect ] ifNotNil: [ pointsRect quickMerge: rect ]].
	newRect := pointsRect scaleAndTranslateToFitInside: outsideRect.
	scale := newRect extent asFloatPoint / pointsRect extent.
	^((((MatrixTransform2x3 withOffset: (outsideRect center))
		composedWithLocal: (MatrixTransform2x3 withRotation: angle))
		composedWithLocal: (MatrixTransform2x3 withOffset: outsideRect extent / -2))
		composedWithLocal: (MatrixTransform2x3 withScale: scale asPoint * (1 @ -1)))
		composedWithLocal: (MatrixTransform2x3 withOffset: ((pointsRect origin x) @ (pointsRect corner y)) negated)
! !

!IconicButton methodsFor: '*connectors-buttons' stamp: 'nk 7/17/2004 09:10'!
updateImage
	! !

!IconicButton methodsFor: '*connectors-connection' stamp: 'nk 8/26/2003 15:58'!
wantsAttachmentFromEnd: anEnd ofConnector: aConnector
	(owner notNil and: [ owner isObjectsTool ]) ifTrue: [ ^false ].
	^super wantsAttachmentFromEnd: anEnd ofConnector: aConnector! !

!DisplayScreen methodsFor: '*connectors-displaying' stamp: 'nk 8/14/2004 12:38'!
flashPoints: points
	| rects |
	rects := points collect: [ :pt | Rectangle center: pt extent: 4@4 ].
	rects do: [ :rect | self reverse: rect ] .
	self forceDisplayUpdate.
	(Delay forMilliseconds: 100) wait.
	rects do: [ :rect | self reverse: rect ] .
	self forceDisplayUpdate.
! !

!Player methodsFor: '*connectors-commands' stamp: 'nk 7/3/2003 21:35'!
addLabel: aString
	(self costume respondsTo: #addLabel:near:)
		ifFalse: [^self ].
	^self costume addLabel: aString near: self costume midpoint
 ! !

!Player methodsFor: '*connectors-embeddings' stamp: 'nk 9/10/2004 12:38'!
createSlotFor: anEmbeddedMorph 
	"anEmbeddedMorph has just been added to my costume because of explicit user action.  
	Create an instance variable and accessors for it as a Player-valued slot."
	| initialAnswer originalString itsName itsPlayer openViewers setterSelector initialValue |
	self costume isInWorld ifFalse: [ ^self ].

	initialAnswer := (anEmbeddedMorph externalName withoutTrailingDigits), '1'.
	initialAnswer := ScriptingSystem
				acceptableSlotNameFrom: initialAnswer
				forSlotCurrentlyNamed: nil
				asSlotNameIn: self
				world: self costume world.
	originalString := FillInTheBlank request: 'name for new variable: ' translated initialAnswer: initialAnswer.
	originalString isEmptyOrNil
		ifTrue: [^ self].
	itsName := ScriptingSystem
				acceptableSlotNameFrom: originalString
				forSlotCurrentlyNamed: nil
				asSlotNameIn: self
				world: self costume world.
	itsName size == 0
		ifTrue: [^ self].
	self assureUniClass.
	self slotInfo
		at: itsName
		put: (SlotInformation new initialize type: #Player).
	self addInstanceVarNamed: itsName withValue: (itsPlayer := anEmbeddedMorph assuredPlayer).
	self class compileAccessorsFor: itsName.
	setterSelector := Utilities setterSelectorFor: itsName.
	initialValue :=  self costume presenter standardPlayer.
	self class allSubInstancesDo: [ :p | p ~~ self ifTrue: [ p perform: setterSelector with: initialValue ]].
	anEmbeddedMorph setNamePropertyTo: itsName.
	openViewers := self allOpenViewersOnReceiverAndSiblings.
	openViewers isEmpty
		ifTrue: [^ self]
		ifFalse: [| aPresenter | 
			(aPresenter := self costume presenter)
				ifNil: [^ self].
			itsPlayer beViewed.
			openViewers
				do: [:aViewer | aPresenter updateViewer: aViewer forceToShow: ScriptingSystem nameForInstanceVariablesCategory]]! !

!Player methodsFor: '*connectors-slot getters/setters' stamp: 'nk 8/21/2004 08:35'!
getAllowsConnectionToEmbeddedPins
	^(self getValueFromCostume: #allowsConnectionToSubmorphs) ifNil: [ true ]! !

!Player methodsFor: '*connectors-slot getters/setters' stamp: 'nk 8/26/2003 11:00'!
getDestinationArrowName
	^(costume respondsTo: #destinationArrowName)
		ifFalse: [ 'unsupported' ]
		ifTrue: [ costume destinationArrowName  ].
		! !

!Player methodsFor: '*connectors-slot getters/setters' stamp: 'nk 7/3/2003 15:10'!
getDestinationPlayer
	"Answer the player for the morph at my last end"
	| dest |
	dest := self getValueFromCostume: #destinationMorph.
	^dest ifNil: [ costume presenter standardPlayer ]
		ifNotNil: [ dest assuredPlayer ]! !

!Player methodsFor: '*connectors-slot getters/setters' stamp: 'nk 7/3/2003 23:36'!
getDestinationX
	| destination |
	destination := (self getValueFromCostume: #lastVertex) ifNil: [ ^0 ].
	^destination x! !

!Player methodsFor: '*connectors-slot getters/setters' stamp: 'nk 7/3/2003 23:37'!
getDestinationY
	| destination |
	destination := (self getValueFromCostume: #lastVertex) ifNil: [ ^0 ].
	^destination y! !

!Player methodsFor: '*connectors-slot getters/setters' stamp: 'nk 8/21/2004 08:13'!
getHasPins
	^(self getValueFromCostume: #pinCount) > 0! !

!Player methodsFor: '*connectors-slot getters/setters' stamp: 'nk 8/21/2004 09:23'!
getIncomingConnectionCount
	^(self getValueFromCostume: #incomingConnectionCount) ifNil: [ 0 ]! !

!Player methodsFor: '*connectors-slot getters/setters' stamp: 'nk 9/4/2004 11:29'!
getLineWidth
	^self getValueFromCostume: #lineWidth! !

!Player methodsFor: '*connectors-slot getters/setters' stamp: 'nk 7/3/2003 23:36'!
getMidpointX
	| midpoint |
	midpoint := (self getValueFromCostume: #midpoint) ifNil: [ ^0 ].
	^midpoint x! !

!Player methodsFor: '*connectors-slot getters/setters' stamp: 'nk 12/27/2003 11:48'!
getMidpointY
	| midpoint w aPlayfield |
	midpoint := (self getValueFromCostume: #midpoint) ifNil: [ ^0 ].
	w := costume world.
	w ifNil: [^costume bounds top].
	aPlayfield := costume referencePlayfield.
	^aPlayfield isNil 
		ifTrue: [w cartesianOrigin y - midpoint y]
		ifFalse: [aPlayfield cartesianOrigin y - midpoint y]
! !

!Player methodsFor: '*connectors-slot getters/setters' stamp: 'nk 7/3/2003 21:10'!
getOrthogonal
	^(self getValueFromCostume: #isOrthogonal) ifNil: [ false ].
! !

!Player methodsFor: '*connectors-slot getters/setters' stamp: 'nk 8/21/2004 09:23'!
getOutgoingConnectionCount
	^(self getValueFromCostume: #outgoingConnectionCount) ifNil: [ 0 ]! !

!Player methodsFor: '*connectors-slot getters/setters' stamp: 'nk 8/21/2004 08:35'!
getPinCount
	^(self getValueFromCostume: #pinCount) ifNil: [ 0 ]! !

!Player methodsFor: '*connectors-slot getters/setters' stamp: 'nk 8/21/2004 08:09'!
getRootOwner
	"Answer the player for my root morph."
	| source |
	source := self getValueFromCostume: #rootInPasteUp.
	^source ifNil: [ costume presenter standardPlayer ]
		ifNotNil: [ source assuredPlayer ]! !

!Player methodsFor: '*connectors-slot getters/setters' stamp: 'nk 7/3/2003 21:15'!
getSmoothCurve
	(self costume respondsTo: #isCurve)
		ifFalse: [^false ].
	^self costume isCurve
 ! !

!Player methodsFor: '*connectors-slot getters/setters' stamp: 'nk 8/26/2003 11:00'!
getSourceArrowName
	^(costume respondsTo: #sourceArrowName)
		ifFalse: [ 'unsupported' ]
		ifTrue: [ costume sourceArrowName  ].
		! !

!Player methodsFor: '*connectors-slot getters/setters' stamp: 'nk 7/3/2003 15:11'!
getSourcePlayer
	"Answer the player for the morph at my first end"
	| source |
	source := self getValueFromCostume: #sourceMorph.
	^source ifNil: [ costume presenter standardPlayer ]
		ifNotNil: [ source assuredPlayer ]! !

!Player methodsFor: '*connectors-slot getters/setters' stamp: 'nk 7/3/2003 23:43'!
getSourceX
	| source |
	source := (self getValueFromCostume: #firstVertex) ifNil: [ ^0 ].
	^source x! !

!Player methodsFor: '*connectors-slot getters/setters' stamp: 'nk 7/3/2003 23:43'!
getSourceY
	| source |
	source := (self getValueFromCostume: #firstVertex) ifNil: [ ^0 ].
	^source y! !

!Player methodsFor: '*connectors-slot getters/setters' stamp: 'nk 7/3/2003 23:36'!
getTotalLength
	^(self getValueFromCostume: #totalLength) ifNil: [ 0 ]! !

!Player methodsFor: '*connectors-slot getters/setters' stamp: 'nk 8/21/2004 08:36'!
getWantsConnectionWhenEmbedded
	^(self getValueFromCostume: #wantsConnectionWhenEmbedded) ifNil: [ false ]! !

!Player methodsFor: '*connectors-embeddings' stamp: 'nk 9/9/2004 14:05'!
noteDeletionOfSubmorph: aMorph fromCostume: aCostume

	"(self class instVarNames includes: aMorph knownName)
		ifTrue: [ self removeSlotNamed: aMorph knownName ]."
! !

!Player methodsFor: '*connectors-commands' stamp: 'nk 7/3/2003 21:37'!
removeLabels
	(self costume respondsTo: #deleteAllLabels)
		ifFalse: [^self ].
	^self costume deleteAllLabels
 ! !

!Player methodsFor: '*connectors-slot getters/setters' stamp: 'nk 8/21/2004 08:15'!
setAllowsConnectionToEmbeddedPins: aBoolean
	(self costume respondsTo: #allowsConnectionToSubmorphs:)
		ifTrue: [ self costume allowsConnectionToSubmorphs: aBoolean ].
! !

!Player methodsFor: '*connectors-slot getters/setters' stamp: 'nk 8/26/2003 10:58'!
setDestinationArrowName: aString
	(costume respondsTo: #destinationArrowName:)
		ifTrue: [ costume destinationArrowName: aString ].
		! !

!Player methodsFor: '*connectors-slot getters/setters' stamp: 'nk 7/3/2003 15:19'!
setDestinationPlayer: aPlayer
	"Answer the player for the morph at my first end"
	(costume respondsTo: #connectFinishTo:)
		ifTrue: [ costume connectFinishTo: aPlayer costume ].
		! !

!Player methodsFor: '*connectors-slot getters/setters' stamp: 'nk 9/4/2004 11:29'!
setLineWidth: aNumber
	(self costume respondsTo: #lineWidth:)
		ifTrue: [ self costume lineWidth: aNumber ].
! !

!Player methodsFor: '*connectors-slot getters/setters' stamp: 'nk 7/3/2003 21:11'!
setOrthogonal: aBoolean
	(self costume respondsTo: #beOrthogonal:)
		ifTrue: [ self costume beOrthogonal: aBoolean ].
! !

!Player methodsFor: '*connectors-slot getters/setters' stamp: 'nk 7/3/2003 21:15'!
setSmoothCurve: aBoolean
	(self costume respondsTo: #beSmoothCurve)
		ifFalse: [^self ].
	aBoolean ifTrue: [ self costume beSmoothCurve ]
		ifFalse: [ self costume beStraightSegments ].
! !

!Player methodsFor: '*connectors-slot getters/setters' stamp: 'nk 8/26/2003 10:58'!
setSourceArrowName: aString
	(costume respondsTo: #sourceArrowName:)
		ifTrue: [ costume sourceArrowName: aString ].
		! !

!Player methodsFor: '*connectors-slot getters/setters' stamp: 'nk 7/3/2003 15:19'!
setSourcePlayer: aPlayer
	"Answer the player for the morph at my first end"
	(costume respondsTo: #connectStartTo:)
		ifTrue: [ costume connectStartTo: aPlayer costume ].
		! !

!Player methodsFor: '*connectors-slot getters/setters' stamp: 'nk 8/21/2004 08:14'!
setWantsConnectionWhenEmbedded: aBoolean
	(self costume respondsTo: #wantsConnectionWhenEmbedded:)
		ifTrue: [ self costume wantsConnectionWhenEmbedded: aBoolean ].
! !

!Player methodsFor: '*connectors-commands' stamp: 'nk 7/3/2003 20:51'!
straighten
	(costume respondsTo: #straighten)
		ifTrue: [ costume straighten ].
		! !

!Player methodsFor: '*connectors-embeddings' stamp: 'nk 9/10/2004 11:27'!
tellAllEmbeddedObjects: aMessageSelector
	"Send the given message selector to all the players in my costume's embedded morphs.
	This is a deep version of tellAllContents:"

	self costume embeddedPlayersDo:
		[:p | p performScriptIfCan: aMessageSelector]! !

!Player methodsFor: '*connectors-commands' stamp: 'nk 8/21/2004 09:21'!
tellAllIncomingConnections: aMessageSelector
	"Send the given message selector to all the players of the receiver's costume's incoming connections"

	self costume incomingConnections do:
		[:m |
			m playerRepresented ifNotNilDo:
				[:p | p performScriptIfCan: aMessageSelector]]! !

!Player methodsFor: '*connectors-commands' stamp: 'nk 8/21/2004 09:21'!
tellAllOutgoingConnections: aMessageSelector
	"Send the given message selector to all the players of the receiver's costume's outgoing connections"

	self costume outgoingConnections do:
		[:m |
			m playerRepresented ifNotNilDo:
				[:p | p performScriptIfCan: aMessageSelector]]! !

!Player methodsFor: '*connectors-embeddings' stamp: 'nk 8/21/2004 08:18'!
tellAllPins: aMessageSelector
	"Send the given message selector to all the objects within the receiver that want connections when embedded"

	self costume embeddedPinsDo:
		[:m |
			m playerRepresented ifNotNilDo:
				[:p | p performScriptIfCan: aMessageSelector]]! !

!Player methodsFor: '*connectors-commands' stamp: 'nk 1/29/2004 13:43'!
tellAllPredecessors: aMessageSelector
	"Send the given message selector to all the objects within the receiver"

	self costume predecessors do:
		[:m |
			m playerRepresented ifNotNilDo:
				[:p | p performScriptIfCan: aMessageSelector]]! !

!Player methodsFor: '*connectors-commands' stamp: 'nk 1/29/2004 13:43'!
tellAllSuccessors: aMessageSelector
	"Send the given message selector to all the objects within the receiver"

	self costume successors do:
		[:m |
			m playerRepresented ifNotNilDo:
				[:p | p performScriptIfCan: aMessageSelector]]! !

!Point methodsFor: '*connectors-point functions' stamp: 'nk 6/21/2003 11:49'!
pointsLoopDo: aBlock X: xDelta Y: yDelta 
	"Implements the Bresenham plotting algorithm (IBM Systems 
	Journal, Vol. 4 No. 1, 1965). It chooses a principal direction, and 
	maintains a potential, P. When P's sign changes, it is time to move in 
	the minor direction as well. This particular version does not write the 
	first and last points, so that these can be called for as needed in client 
	code. "
	| dx dy px py P destX destY |
	dx := xDelta sign.
	dy := yDelta sign.
	px := yDelta abs.
	py := xDelta abs.
	destX := self x.
	destY := self y.
	py > px
		ifTrue: ["more horizontal"
			P := py // 2.
			1
				to: py
				do: [:i | 
					destX := destX + dx.
					(P := P - px) < 0
						ifTrue: [destY := destY + dy.
							P := P + py].
					aBlock value: destX @ destY]]
		ifFalse: ["more vertical"
			P := px // 2.
			1
				to: px
				do: [:i | 
					destY := destY + dy.
					(P := P - py) < 0
						ifTrue: [destX := destX + dx.
							P := P + px].
					aBlock value: destX @ destY]]! !

!Point methodsFor: '*connectors-point functions' stamp: 'nk 6/21/2003 11:44'!
pointsTo: aPoint do: aBlock
	"Using the Bresenham algorithm, evaluate aBlock with all the points from me to aPoint, inclusive"
	aBlock value: self.
	self pointsLoopDo: aBlock X: aPoint x - self x Y: aPoint y - self y.
	aBlock value: aPoint.! !

!Point methodsFor: '*connectors-point functions' stamp: 'nk 1/10/2004 14:46'!
scaleToNoGreaterThan: anotherPoint
	"Answer a Point with the same x/y ratio as me, but less than or equal to anotherPoint."
	| ratio |
	x isZero ifTrue: [ ^0 @ anotherPoint y ].
	y isZero ifTrue: [ ^anotherPoint x @ 0 ].
	ratio := anotherPoint / self.
	^ratio x > ratio y
		ifTrue: [ self * ratio y ]
		ifFalse: [ self * ratio x ].
! !

!TransformMorph methodsFor: '*connectors-accessing' stamp: 'nk 3/9/2001 09:23'!
isSmoothingOn
	^smoothing ~= 1! !

BorderedMorph subclass: #NCLineMorph
	instanceVariableNames: 'lineWidth fillStyle smoothCurve vertices segments contours arrows arrowForms arrowScales arrowJoints dashedBorderSpec'
	classVariableNames: 'CurveDebug'
	poolDictionaries: ''
	category: 'Connectors-Lines and Curves'!

!NCLineMorph class methodsFor: 'arrow forms' stamp: 'nk 8/25/2003 12:49'!
basicArrow
	"closed arrow"
	"glyph 54 from Arrows1"
^(NCLineArrowGlyph withContours: (
(Array new: 1)
	at: 1 put: { 
	0.0@0.0 . 0.0@0.0 . -0.5@0.3298 . 
	-0.5@0.3298 . -0.5@0.3298 . 0.0@-0.666 . 
	0.0@-0.666 . 0.0@-0.666 . 0.5@0.3298 . 
	0.5@0.3298 . 0.5@0.3298 . 0.0@0.0 };
	yourself
"clicks 0.0@0.0"
)) offset: 0@0.666;
defaultScale: 3.0;
yourself! !

!NCLineMorph class methodsFor: 'debugging' stamp: 'nk 8/19/2003 12:04'!
curveDebug
	^CurveDebug ifNil: [ CurveDebug := false ].! !

!NCLineMorph class methodsFor: 'debugging' stamp: 'nk 8/19/2003 12:04'!
curveDebug: aBoolean
	"NCLineMorph curveDebug: true"
	"NCLineMorph curveDebug: false"
	CurveDebug := aBoolean! !

!NCLineMorph class methodsFor: 'arrow forms' stamp: 'nk 8/20/2003 10:26'!
defaultArrow
	^NCLineArrowGlyph noArrow! !

!NCLineMorph class methodsFor: 'error handling' stamp: 'nk 8/20/2003 10:10'!
doesNotUnderstand: aMessage
	"| failed |
	failed := false.
	(PolygonMorph class includesSelector: aMessage selector) ifTrue: [
		[ self class compile: (PolygonMorph class sourceCodeAt: aMessage selector)
			classified: 'stolen from PM'
			notifying: nil ] on: Error do: [ :ex | failed := true ].
	].
	^failed
		ifFalse: [ aMessage sentTo: self ]
		ifTrue: [ super doesNotUnderstand: aMessage ]."
	
	^super doesNotUnderstand: aMessage! !

!NCLineMorph class methodsFor: 'class initialization' stamp: 'nk 8/19/2003 11:58'!
initialize
	"NCLineMorph initialize"
	CurveDebug ifNil: [ CurveDebug := false ]! !

!NCLineMorph class methodsFor: 'testing' stamp: 'nk 9/22/2003 14:42'!
showSamples
	"self showSamples"
	| pum line offset verts |
	"lineWidth borderWidth smooth curve"
	pum := PasteUpMorph new bounds: (0@0 corner: 560@900).
	pum openInWorld.
	0 to: 10 do: [ :lw |
		0 to: 4 do: [ :bw |
			{true. false} do: [ :sc |
				offset := ((bw * 100) + 20) @ ((lw * 80) + 20).
				sc ifTrue: [ offset := offset + (0@ 40) ].
				verts := { offset . offset + (40@20). offset + (80@0) }.
				line := self vertices: verts
					lineWidth: (lw/4) + 0.25
					lineColor: Color blue
					borderWidth: bw
					borderColor: Color black.
				sc ifTrue: [ line beSmoothCurve ].
				pum addMorph: line
			]
		]
	]
		! !

!NCLineMorph class methodsFor: 'instance creation' stamp: 'nk 8/20/2003 10:16'!
vertices: verts color: c borderWidth: bw borderColor: bc
	^ self new vertices: verts color: c borderWidth: bw borderColor: bc! !

!NCLineMorph class methodsFor: 'instance creation' stamp: 'nk 8/20/2003 10:16'!
vertices: verts lineWidth: w lineColor: c borderWidth: bw borderColor: bc
	^ self new
		vertices: verts lineWidth: w lineColor: c borderWidth: bw borderColor: bc! !

!NCLineMorph methodsFor: 'rotate scale and flex' stamp: 'nk 9/4/2004 07:34'!
addFlexShell
	"When scaling or rotating from a halo, I can do this without a flex shell"

	^ self
! !

!NCLineMorph methodsFor: 'rotate scale and flex' stamp: 'nk 7/11/2003 18:56'!
addFlexShellIfNecessary
	"When scaling or rotating from a halo, I can do this without a flex shell"

	^ self
! !

!NCLineMorph methodsFor: 'editing' stamp: 'nk 8/22/2003 17:29'!
addHandles
	self addAllMorphs: self handles! !

!NCLineMorph methodsFor: 'drawing' stamp: 'nk 8/4/2003 18:55'!
areasRemainingToFill: aRectangle
	^ { aRectangle }! !

!NCLineMorph methodsFor: 'arrows' stamp: 'nk 8/21/2003 17:11'!
arrowContours
	^{ contours fourth. contours second }! !

!NCLineMorph methodsFor: 'arrows' stamp: 'nk 8/23/2003 21:54'!
arrowForms
	arrowForms ifNotNil: [ ^arrowForms ].
	arrowForms := self arrowContours withIndexCollect: [ :ac :aci | | form rect ext offset ji |
		rect := (Rectangle encompassing: ac) rounded.
		ext := rect extent max: 4@4.
		form := Form extent: rect extent depth: 1.
		ji := (arrows at: aci) joinIndex.
		((BalloonCanvas on: form) asShadowDrawingCanvas: Color black)
			transformBy: (MatrixTransform2x3 withOffset: rect origin negated)
			during: [ :canvas |
				canvas aaLevel: 4; deferred: true.
				canvas 
					drawGeneralBezierShape: { { ac at: ji. ac at: ji. ac first }. ac}
					color: Color black
					borderWidth: 1
					borderColor: Color black.
				canvas flush.
			].
		form offset: rect origin
	].
	^arrowForms! !

!NCLineMorph methodsFor: 'arrows' stamp: 'nk 8/23/2003 10:36'!
arrowJoints
	^arrowJoints ifNil: [ arrowJoints := { vertices first. vertices last } ].! !

!NCLineMorph methodsFor: 'TODO' stamp: 'nk 7/11/2003 18:53'!
arrowLength: aLength
	"Assumes that I have exactly two vertices"

	| theta horizontalOffset verticalOffset newTip delta |
	delta := vertices second - vertices first.
	theta := delta theta.
	horizontalOffset := aLength * (theta cos).
	verticalOffset := aLength * (theta sin).
	newTip := vertices first + (horizontalOffset @ verticalOffset).
	self verticesAt: 2 put: newTip! !

!NCLineMorph methodsFor: 'arrows' stamp: 'nk 8/24/2003 17:31'!
arrowScales
	"Answer the arrow scales relative to my linewidth"
	^arrowScales ifNil: [ { 1.0@1.0. 1.0@1.0 }].! !

!NCLineMorph methodsFor: 'arrows' stamp: 'nk 10/21/2003 10:54'!
arrowScales: aScalarOrArray 
	"Change the arrow scales relative to my linewidth"
	
	arrowScales := (aScalarOrArray isCollection
		ifFalse: [ Array new: 2 withAll: aScalarOrArray ]
		ifTrue: [ Array withAll: aScalarOrArray ]) replaceAll: 0.0 with: 0.00001.
	self arrows: self arrows.! !

!NCLineMorph methodsFor: 'arrows' stamp: 'nk 8/23/2003 14:32'!
arrowVertices
	"Answer my vertices, but with the first and last point from arrowJoints"
	| retval |
	retval := vertices collect: [ :v | v asFloatPoint ].
	retval at: 1 put: self arrowJoints first.
	retval at: retval size put: self arrowJoints second.
	^retval! !

!NCLineMorph methodsFor: 'arrows' stamp: 'nk 8/24/2003 15:15'!
arrows
	^arrows ifNil: [ arrows := self arrows: {nil. nil} ].! !

!NCLineMorph methodsFor: 'arrows' stamp: 'nk 8/24/2003 18:14'!
arrows: anArray
	"Given a pair of NCLineArrowGlyph instances at any scale,
	scale them proportionally to my line width and save copies."
	| newArrows |
	newArrows := (anArray with: self arrowScales collect: [ :ea :scale | | arrow min |
		arrow := (ea ifNil: [ self class defaultArrow ]) veryDeepCopy.
		min := 2.0@2.0 / (arrow bounds extent  / arrow scale).
		arrow scale: ((scale asPoint * self fullLineWidth) max: min); yourself ]).
	arrowForms ifNotNilDo: [ :fs |
		fs do: [ :f | f ifNotNil: [ self invalidRect: (f offset extent: f extent) ] ]].
	arrows := newArrows.
	arrowForms := nil.
	self computeBounds.
	^arrows! !

!NCLineMorph methodsFor: 'geometry testing' stamp: 'nk 8/21/2003 17:41'!
arrowsContainPoint: aPoint
	"Answer an Array of two Booleans that indicate whether the given point is inside either arrow"
	^self arrowForms collect: [:f | (f pixelValueAt: aPoint - f offset) > 0 ].
! !

!NCLineMorph methodsFor: 'accessing' stamp: 'nk 7/18/2003 15:48'!
beSmoothCurve
	self isCurve ifFalse: [
		smoothCurve := true.
		self computeBounds]! !

!NCLineMorph methodsFor: 'accessing' stamp: 'nk 7/18/2003 15:47'!
beStraightSegments

	self isCurve ifTrue: [
		smoothCurve := false.
		self computeBounds]! !

!NCLineMorph methodsFor: 'geometry-curves' stamp: 'nk 8/19/2003 10:34'!
bezierPoints
	"Answer nVertices*3 points that are the control points for my bezier curves."

	| pts i |
	pts := Array new: self segments size * 3.
	i := 1.
	self bezierSegmentsDo: 
			[:seg | 
			pts
				at: i put: seg start;
				at: i + 1 put: seg via;
				at: i + 2 put: seg end.
			i := i + 3].
	^pts! !

!NCLineMorph methodsFor: 'geometry-curves' stamp: 'nk 8/18/2003 17:55'!
bezierSegmentsDo: segmentBlock 
	"Emit a sequence of LineSegment subinstances into segmentBlock"
	self segmentsDo: [ :s | segmentBlock value: s asBezier2Segment ]! !

!NCLineMorph methodsFor: 'accessing' stamp: 'nk 9/4/2004 11:51'!
borderWidth: anInteger 
	| ratio newWidth oldWidth |
	oldWidth := self fullLineWidth.
	anInteger > 0 ifTrue: [borderColor ifNil: [borderColor := Color black]].
	newWidth := ((anInteger max: 0) min: self lineWidth / 2) asInteger.
	borderWidth := newWidth.
	ratio := self fullLineWidth / oldWidth.
	self arrows do: [:a | a scale: (a scale * ratio) asPoint].
	self computeBounds! !

!NCLineMorph methodsFor: 'labels' stamp: 'nk 8/14/2004 13:58'!
boundsSignatureHash
	^(contours - (self positionInWorld)) hash! !

!NCLineMorph methodsFor: 'testing' stamp: 'nk 9/22/2003 14:49'!
canDrawBorder: aBorderStyle
	^aBorderStyle style == #simple! !

!NCLineMorph methodsFor: 'visual properties' stamp: 'nk 7/11/2003 18:58'!
canHaveFillStyles
	"Return true if the receiver can have general fill styles; not just colors.
	This method is for gradually converting old morphs."
	^true! !

!NCLineMorph methodsFor: 'arrows' stamp: 'nk 12/27/2003 11:24'!
changeArrowScaleInteractively: evt
	| handle origin aHand oldScales scaleDelta newScales |
	aHand := evt ifNil: [self primaryHand] ifNotNil: [evt hand].
	origin := aHand position.
	oldScales := self arrowScales.
	handle := HandleMorph new
		forEachPointDo:
			[:newPoint | handle removeAllMorphs.
			handle addMorph:
				(LineMorph from: origin to: newPoint color: Color black width: 1).
			scaleDelta := ((newPoint - origin) / 100.0).
			newScales := oldScales + scaleDelta.
			self arrowScales: newScales.
		]
		lastPointDo:
			[:newPoint | handle deleteBalloon.
			self halo ifNotNilDo: [:halo | halo addHandles].
			self rememberCommand:
				(Command new cmdWording: 'arrow scale change';
					undoTarget: self selector: #arrowScales: argument: oldScales;
					redoTarget: self selector: #arrowScales: argument: newScales)].
	aHand attachMorph: handle.
	handle setProperty: #helpAtCenter toValue: true.
	handle showBalloon:
'Move cursor farther from
this point to increase the size of the arrows.
Click when done.' hand: evt hand.
	handle startStepping! !

!NCLineMorph methodsFor: 'geometry testing' stamp: 'nk 8/21/2003 13:30'!
closestPointTo: aPoint
	| curveInfo |
	curveInfo := self curveInfoNearest: aPoint.
	^curveInfo closestPoint.! !

!NCLineMorph methodsFor: 'geometry testing' stamp: 'nk 8/19/2003 10:33'!
closestSegmentTo: aPoint
	"Answer the starting index of my (big) segment nearest to aPoint"
	^(self curveInfoNearest: aPoint) closestBigSegment! !

!NCLineMorph methodsFor: 'geometry' stamp: 'nk 12/28/2003 16:53'!
computeBounds
	| oldFullBounds oldBounds delta |
	vertices ifNil: [^self].
	oldFullBounds := (fullBounds ifNil: [bounds]) copy.
	contours := Array new: 4 withAll: #().
	self updateArrowContours.
	segments := self computeSegments.
	segments do: [:ea | ea roundTo: 0.001].
	arrowForms := nil.
	self computeContours.
	oldBounds := bounds.
	bounds := self curveBounds truncated.
	self arrowForms 
		do: [:af | bounds := bounds quickMerge: (af offset extent: af extent)].

	"since we are directly updating bounds,
	see if any ordinary submorphs exist and move them accordingly"
	submorphs isEmpty 
		ifFalse: 
			[(oldBounds notNil 
				and: [(delta := bounds origin - oldBounds origin) ~= (0 @ 0)]) 
					ifTrue: [self submorphsDo: [:each | each position: each position + delta]]].
	self invalidRect: oldFullBounds.
	self layoutChanged.
	self changed! !

!NCLineMorph methodsFor: 'geometry-curves' stamp: 'nk 12/28/2003 22:41'!
computeContours
	| sides csize |
	(vertices size = 2 and: [vertices first = vertices second]) 
		ifTrue: [^contours := #()].
	true "(self lineWidth > 2 or: [self isCurve]) "
		ifTrue: [sides := self outsidePoints	"sideA, sideB reversed"]
		ifFalse: 
			[| bpts |
			bpts := self bezierPoints.
			sides := { 
						bpts.
						bpts reversed}].
	contours
		at: 1 put: sides first;
		at: 3 put: sides second.
	csize := contours first size.
	(contours at: 3)
		at: csize put: contours fourth first;
		at: 1 put: (contours second at: arrows second joinIndex).
	(contours at: 1)
		at: 1 put: (contours fourth at: arrows first joinIndex);
		at: csize put: contours second first.
	contours := contours collect: [ :ctr | ctr collect: [ :pt | pt roundTo: 0.001 ]].
	^contours! !

!NCLineMorph methodsFor: 'geometry-curves' stamp: 'nk 8/23/2003 11:00'!
computeSegments
	| segs v3 l1 l2 m1 m2 lastUsed end arrowVertices |
	arrowVertices := self arrowVertices.
	arrowVertices size = 2 
		ifTrue: 
			[^{ LineSegment 
					from: arrowVertices first
					to: arrowVertices last }].

	smoothCurve ifFalse: [
		segs := Array new: arrowVertices size - 1.
		arrowVertices overlappingPairsWithIndexDo:  [ :v1 :v2 :i |
			segs at: i put: (LineSegment from: v1 to: v2)
		].
		^segs
	].

	segs := OrderedCollection new.
	lastUsed := arrowVertices first.
	2 to: arrowVertices size - 1 do:  [:i | | v2 symmetry |
			v2 := (arrowVertices at: i).
			v3 := (arrowVertices at: i + 1).
			l1 := lastUsed dist: v2.
			l2 := v2 dist: v3.
			symmetry := l1 - l2.
			symmetry < 0
				ifTrue: 
					[m1 := (lastUsed + v2) / 2.0.
					end := v2 interpolateTo: v3 at: l1 / l2 * 0.5.
					lastUsed ~= m1 
						ifTrue: 
							[segs add: (LineSegment 
										from: lastUsed
										to: m1)].
					segs add: (Bezier2Segment 
								from: m1
								to: end
								via: v2).
					lastUsed := end]
		ifFalse: [ m2 := (v2 + v3) / 2.0.
					end := v2 interpolateTo: lastUsed at: l2 / l1 * 0.5.
					lastUsed ~= end 
						ifTrue: 
							[segs add: (LineSegment 
										from: lastUsed
										to: end)].
					segs add: (Bezier2Segment 
								from: end
								to: m2
								via: v2).
					lastUsed := m2]
		].
	lastUsed ~= v3 
		ifTrue: 
			[segs add: (LineSegment 
						from: lastUsed
						to: v3)].
	^segs asArray ! !

!NCLineMorph methodsFor: 'geometry-curves' stamp: 'nk 8/23/2003 14:34'!
computeSegments2
	| segs arrowVertices midpoints bezierSegs fillSegs |
	arrowVertices := self arrowVertices.
	arrowVertices size = 2 
		ifTrue: 
			[^{ LineSegment 
					from: arrowVertices first
					to: arrowVertices last }].

	smoothCurve ifFalse: [
		segs := Array new: arrowVertices size - 1.
		arrowVertices overlappingPairsWithIndexDo:  [ :v1 :v2 :i |
			segs at: i put: (LineSegment from: v1 to: v2) ].
		^segs ].

	midpoints := Array new: arrowVertices size - 1.
	arrowVertices overlappingPairsWithIndexDo: [ :v1 :v2 : i | midpoints at: i put: (v1 + v2) / 2 ].
	midpoints
		at: 1 put: arrowVertices first;
		at: midpoints size put: arrowVertices last.

	bezierSegs := Array new: arrowVertices size - 2.
	2 to: arrowVertices size - 1 do: [ :i | | r1 r2 t1 t2 v1 v2  v3 preferredLength |
		v1 := midpoints at: i - 1.
		v2 := arrowVertices at: i.
		v3 := midpoints at: i.
		r1 := (v2 - v1) r.
		t1 := (v1 - v2) normalized.
		r2 := (v2 - v3) r.
		t2 := (v3 - v2) normalized.
		preferredLength := r1 min: r2.
		bezierSegs at: i - 1 put: (Bezier2Segment from: (v2 + (t1 * preferredLength))
									via: v2
									to: (v2 + (t2 * preferredLength))).
	].

	fillSegs := Array new: arrowVertices size - 2.	"fill before bezier"
	bezierSegs withIndexDo: [ :seg :bi | | priorEnd nextStart d1 d2 dp dn stretch fill |
		priorEnd := bi = 1 ifTrue: [ arrowVertices first ] ifFalse: [ (bezierSegs at: bi - 1) end ].
		nextStart := bi = bezierSegs size ifTrue: [ arrowVertices last ] ifFalse: [ (bezierSegs at: bi + 1) start ].
		dp := seg via dist: priorEnd.
		d1 := seg via dist: seg start.	"same as dist to end"
		dp < d1 ifTrue: [
			d1 := dp.
			seg start: priorEnd.
			seg end: (seg via + ((seg end - seg via) normalized * d1))
		].
		dn := seg via dist: nextStart.
		d2 := d1.
		stretch := d1 * 0.2.
		(dp - dn) abs <= stretch ifTrue: [	"can stretch both ends"
				seg start: priorEnd.
				seg end: nextStart.
			] ifFalse: [

		((dp - d1) abs <= stretch)
			ifTrue: [ seg start: priorEnd ]
			ifFalse: [ | t1 |
				t1 := (seg start - seg via) normalized.
				seg start: (seg start + (t1 * stretch)).
				fill := LineSegment from: priorEnd to: seg start.
				fillSegs at: bi put: fill.
			].
		(dn - d2) abs <= stretch
			ifTrue: [ seg end: nextStart ]
			ifFalse: [  | t2 |
				t2 := (nextStart - seg end) normalized.
				seg end: (seg end + (t2 * stretch)).].
		]
	].

	^Array streamContents: [ :s |
		fillSegs with: bezierSegs do: [ :fill :bezier |
			fill ifNotNil: [ fill length > 10 ifTrue: [ s nextPut: fill ]
				ifFalse: [ bezier start: fill start ]].
			bezier ifNotNil: [ s nextPut: bezier ].

	].
		(bezierSegs last end dist: arrowVertices last) > 2 ifTrue: [
			s nextPut: (LineSegment from: bezierSegs last end to: arrowVertices last)
			]]! !

!NCLineMorph methodsFor: 'geometry testing' stamp: 'nk 9/28/2003 17:46'!
containsPoint: aPoint
	(super containsPoint: aPoint) ifFalse: [^ false].
	(self hasProperty: #errorOnDraw) ifTrue: [^true].
	
	(self arrowForms anySatisfy: [:f |
		 "(f pixelValueAt: aPoint - f offset) > 0"
		(Rectangle origin: f offset extent: f extent) containsPoint: aPoint
	 ])
		ifTrue: [^true ].

	self lineSegmentsDo:
		[:p1 :p2 |
		(aPoint onLineFrom: p1 to: p2 within: (10 max: (self fullLineWidth + 4) / 2))
				ifTrue: [^ true]].
			
	^ false! !

!NCLineMorph methodsFor: 'geometry-curves' stamp: 'nk 8/23/2003 11:09'!
contourPoints
	"Answer my contours flattened as a series of points."

	^Array streamContents: [ :s |
		self contours do: [ :c | s nextPutAll: c ]]! !

!NCLineMorph methodsFor: 'geometry-curves' stamp: 'nk 8/23/2003 11:10'!
contours
	contours ifNotNil: [^contours].
	(vertices size = 2 and: [vertices first = vertices second]) 
		ifTrue: [^contours := #()].
	self error: 'should be set'.! !

!NCLineMorph methodsFor: 'accessing' stamp: 'nk 7/11/2003 18:45'!
couldHaveRoundedCorners
	^ false! !

!NCLineMorph methodsFor: 'geometry' stamp: 'nk 8/24/2003 19:16'!
curveBounds

	^(Rectangle encompassing: self contourPoints)
		expandBy: (borderWidth + 4.0)! !

!NCLineMorph methodsFor: 'geometry testing' stamp: 'nk 1/13/2004 17:28'!
curveInfoNearest: aPoint 
	"Answer the index of the segment nearest aPoint, along with the closest point, its distance, and the parametric value. "

	| closestPoint minDist minT segmentIndex |
	closestPoint := minDist := nil.
	self segments withIndexDo: 
			[:seg :i | 
			seg lineSegmentsWithParametricValueDo: 
					[:p1 :p2 :t1 :t2 | 
					| curvePoint dist |
					curvePoint := aPoint nearestPointOnLineFrom: p1 to: p2.
					dist := curvePoint dist: aPoint.
					(closestPoint isNil or: [dist < minDist]) 
						ifTrue: 
							[closestPoint := curvePoint.
							minDist := dist.
							segmentIndex := i.
							minT := (p1 = p2) ifTrue: [ t1 ] ifFalse: [ t1 interpolateTo: t2 at: (curvePoint dist: p1) / (p1 dist: p2)] ]]].
	^NCLineCurveInfo new 
		line: self
		targetPoint: aPoint
		segmentIndex: segmentIndex
		closestPoint: closestPoint
		minDist: minDist
		minT: minT! !

!NCLineMorph methodsFor: 'editing' stamp: 'nk 8/20/2003 23:36'!
curveSegmentAt: n 
	^(self vertices at: n) + (self vertices at: (n + 1 min: vertices size)) / 2! !

!NCLineMorph methodsFor: 'editing' stamp: 'nk 8/21/2003 09:24'!
curveSegmentAt: n put: aPoint
	| delta newVerts |
	(n between: 2 and: vertices size - 1) ifFalse: [ ^false ].
	delta := (self curveSegmentAt: n) - aPoint.
	newVerts := vertices copyFrom: n to: n + 1.
	delta := aPoint - newVerts average.
	newVerts := newVerts + delta.
	self setVertices: (vertices copyReplaceFrom: n to: n + 1 with: newVerts).
	^true
! !

!NCLineMorph methodsFor: 'accessing' stamp: 'nk 8/22/2003 19:24'!
curveVertexAt: n
	"Answer the point on the visible curve corresponding to the given vertex."
	| index seg |
	index := (n min: vertices size) max: 1.
	(smoothCurve not  or: [ index = 1 or: [ (index = vertices size) ]]) ifTrue: [ ^vertices at: index ].
	seg := (self segments select: [ :ea | ea isBezier2Segment ]) at: index - 1
		ifAbsent: [ ^vertices first ].
	^seg valueAt: (seg start dist: seg via) / seg length! !

!NCLineMorph methodsFor: 'editing' stamp: 'nk 8/22/2003 19:26'!
curveVertexAt: index put: aPoint
	"Set the point on the visible curve corresponding to the given vertex."
	| seg error actual attempts |
	(index between: 1 and: vertices size) ifFalse: [ ^false ].
	(smoothCurve not  or: [ index = 1 or: [ (index = vertices size) ]])
		ifTrue: [ self verticesAt: index put: aPoint. ^true ].
	attempts := 0.
	[attempts < 10] whileTrue: [
		seg := (self segments select: [ :ea | ea isBezier2Segment ]) at: index - 1 ifAbsent: [ ^false ].
		actual := seg valueAt: (seg start dist: seg via) / seg length.
		error := aPoint - actual.
		error r < 1 ifTrue: [ ^true ].
		self verticesAt: index put: (vertices at: index) + error.
		attempts := attempts + 1.
	].
	^false! !

!NCLineMorph methodsFor: 'accessing' stamp: 'ag 9/5/2011 12:52:44.301'!
dashedBorder
	^ dashedBorderSpec! !

!NCLineMorph methodsFor: 'accessing' stamp: 'ag 2/22/2012 15:03:24.705'!
dashedBorder: aDashedBorderSpec

	dashedBorderSpec := aDashedBorderSpec.
	self changed.! !

!NCLineMorph methodsFor: 'arrows' stamp: 'nk 8/20/2003 08:59'!
defaultArrow
	^NCLineArrowGlyph noArrow! !

!NCLineMorph methodsFor: 'visual properties' stamp: 'nk 8/28/2003 15:59'!
defaultBitmapFillForm
	^Imports default imports
		at: 'checker' ifAbsent: [ super defaultBitmapFillForm ]
! !

!NCLineMorph methodsFor: 'naming' stamp: 'cmm 12/4/2010 20:05'!
defaultNameStemForInstances
	^ self class name
		copyFrom: 3
		to: self class name size! !

!NCLineMorph methodsFor: 'error handling' stamp: 'nk 8/4/2003 18:48'!
doesNotUnderstand: aMessage
	"| failed |
	failed := false.
	(PolygonMorph includesSelector: aMessage selector) ifTrue: [
		[ NCLineMorph compile: (PolygonMorph sourceCodeAt: aMessage selector)
			classified: 'stolen from PM'
			notifying: nil ] on: Error do: [ :ex | failed := true ].
	].
	^failed
		ifFalse: [ aMessage sentTo: self ]
		ifTrue: [ super doesNotUnderstand: aMessage ]."
	^super doesNotUnderstand: aMessage
	! !

!NCLineMorph methodsFor: 'drawing' stamp: 'nk 8/18/2003 11:30'!
drawCurveDebugOn: aaCanvas 
	"debug"

	aaCanvas 
		drawPolygon: self contourPoints
		color: nil
		borderWidth: 1
		borderColor: (Color blue alpha: 0.5).
	self bezierSegmentsDo:  [:seg | 
		seg controlPointsDo: [ :v |
			aaCanvas 
				drawOval: (Rectangle center: v extent: 7 @ 7)
				color: Color red
				borderWidth: 0
				borderColor: Color black]
	]! !

!NCLineMorph methodsFor: 'drawing' stamp: 'nk 9/4/2004 17:34'!
drawOn: aCanvas 
	| aaCanvas isShadow |
	aaCanvas := aCanvas asBalloonCanvas.
	isShadow := aCanvas isShadowDrawing.
	(aaCanvas isBalloonCanvas or: [ aaCanvas isPostscriptCanvas ])
		ifTrue: 
			[aaCanvas
				aaLevel: 4;
				deferred: true.
			smoothCurve 
				ifTrue: 
					[self 
						drawSmoothCurveOnAACanvas: aaCanvas
						canvas: aCanvas
						isShadow: isShadow]
				ifFalse: 
					[self 
						drawSegmentedCurveOnAACanvas: aaCanvas
						canvas: aCanvas
						isShadow: isShadow].
			aaCanvas flush]
		ifFalse: 
			[aCanvas 
				drawPolygon: self contourPoints
				color: self fillStyle
				borderWidth: self lineWidth asInteger
				borderColor: (isShadow 
						ifTrue: [aCanvas shadowColor]
						ifFalse: [self borderColor])]! !

!NCLineMorph methodsFor: 'drawing' stamp: 'ag 9/7/2011 12:03:32.919'!
drawSegmentedCurveOnAACanvas: aaCanvas canvas: aCanvas isShadow: isShadow 
	| c bw bc |
	c := isShadow
				ifTrue: [aCanvas shadowColor]
				ifFalse: [self fillStyle].
	bw := self borderWidth.
	bc := isShadow
				ifTrue: [aCanvas shadowColor]
				ifFalse: [self borderColor].
	bw > (self lineWidth / 2)
		ifTrue: [c := nil.
			bw := self lineWidth / 2 max: 1].
	self contours.
	(contours
			anySatisfy: [:ctr | ctr isEmpty])
		ifTrue: [^ self].
	self dashedBorder
		ifNotNil: [^ self
				drawSegmentedDashedCurveOn: aCanvas
				color: c
				borderColor: bc].
	aaCanvas
		drawPolygon: contours first , contours third
		fillStyle: c
		borderWidth: bw
		borderColor: bc.
	aaCanvas
		drawGeneralBezierShape: {contours second. {contours third first. contours third first. contours first last}. contours fourth. {contours first first. contours first first. contours third last}}
		color: c
		borderWidth: bw
		borderColor: bc! !

!NCLineMorph methodsFor: 'drawing' stamp: 'ag 9/7/2011 12:03:23.289'!
drawSegmentedDashedCurveOn: aCanvas color: c borderColor: bc 
	| segmentOffset |
	segmentOffset := dashedBorderSpec size < 4 ifTrue: [0.0] ifFalse: [dashedBorderSpec fourth asFloat].
	self segmentsDo: [ :segment |
			segmentOffset := aCanvas
						line: segment start
						to: segment end
						width: self lineWidth
						color: c
						dashLength: dashedBorderSpec first
						secondColor: dashedBorderSpec third
						secondDashLength: dashedBorderSpec second
						startingOffset: segmentOffset].! !

!NCLineMorph methodsFor: 'drawing' stamp: 'nk 1/17/2004 15:38'!
drawSmoothCurveOnAACanvas: aaCanvas canvas: aCanvas isShadow: isShadow 
	self class curveDebug ifTrue: [
		isShadow ifFalse: 
			[ self drawCurveDebugOn: aaCanvas] ].
	aaCanvas 
		drawGeneralBezierShape: self contours
		color: (isShadow ifTrue: [aCanvas shadowColor] ifFalse: [self fillStyle])
		borderWidth: self borderWidth
		borderColor: (isShadow 
				ifTrue: [aCanvas shadowColor]
				ifFalse: [self borderColor])
! !

!NCLineMorph methodsFor: 'geometry' stamp: 'nk 7/18/2003 14:48'!
extent: newExtent 
	"Not really advisable, but we can preserve most of the geometry if we don't
	shrink things too small."
	| safeExtent center |
	self changed.
	center := self referencePosition.
	safeExtent := newExtent max: 20@20.
	self setVertices: (vertices collect:
		[:p | p - center * (safeExtent asFloatPoint / (bounds extent max: 1@1)) + center])! !

!NCLineMorph methodsFor: 'visual properties' stamp: 'nk 7/11/2003 19:04'!
fillStyle

	^fillStyle! !

!NCLineMorph methodsFor: 'visual properties' stamp: 'nk 7/11/2003 19:00'!
fillStyle: newColor

	fillStyle := newColor.
	color := newColor asColor.
	self changed.! !

!NCLineMorph methodsFor: 'accessing' stamp: 'nk 7/11/2003 17:49'!
firstVertex
	^vertices first! !

!NCLineMorph methodsFor: 'geometry' stamp: 'nk 8/22/2003 23:42'!
fullBounds
	^fullBounds ifNil: [
		self computeBounds.
		fullBounds := self outerBounds ].! !

!NCLineMorph methodsFor: 'arrows' stamp: 'nk 12/28/2003 15:42'!
fullLineWidth
	| bw |
	bw := 0.
	self borderColor 
		ifNotNilDo: [:bc | bc isTransparent ifFalse: [bw := borderWidth]].
	^self lineWidth + (2 * bw) + 2! !

!NCLineMorph methodsFor: 'editing' stamp: 'nk 8/22/2003 16:01'!
handleForEvent: anEvent 
	"Answer a NCCurveHandleMorph that can be used to adjust me at the 
	given point."
	^self handleForPosition: anEvent position addAllowed: anEvent shiftPressed! !

!NCLineMorph methodsFor: 'editing' stamp: 'nk 12/28/2003 15:42'!
handleForPosition: pos addAllowed: canAddVertex 
	"Answer a NCCurveHandleMorph that can be used to adjust me at the 
	given point."

	| v s curveInfo |
	curveInfo := self curveInfoNearest: pos.
	v := curveInfo vertexWithinProportion: 0.15
				andDistance: (40 max: self lineWidth / 2 + borderWidth + 2).
	s := curveInfo closestBigSegment.
	v isZero 
		ifTrue: 
			[(canAddVertex or: [s = 1 or: [s = (vertices size - 1)]]) 
				ifTrue: 
					["Make a new vertex to edit"

					v := s + 1.
					self setVertices: (self vertices 
								copyReplaceFrom: v
								to: v - 1
								with: (Array with: pos))]
				ifFalse: 
					[^NCCurveHandleMorph new 
						target: self
						arguments: { 
								s}
						getPointSelector: #curveSegmentAt:
						setPointSelector: #curveSegmentAt:put:]].
	^NCCurveHandleMorph new 
		target: self
		arguments: { 
				v}
		getPointSelector: #curveVertexAt:
		setPointSelector: #curveVertexAt:put:! !

!NCLineMorph methodsFor: 'arrows' stamp: 'nk 8/22/2003 21:42'!
handles
	"Answer a collection of NCCurveHandleMorphs that can be used to adjust my vertices"
	| handles |
	handles := submorphs select: [ :ea | ea isCurveHandle ].
	handles size = vertices size ifTrue: [ ^handles ].

	^(1 to: vertices size) collect: [ :v |
		NCCurveHandleMorph new 
		target: self
		arguments: {  v}
		getPointSelector: #curveVertexAt:
		setPointSelector: #curveVertexAt:put:
		]! !

!NCLineMorph methodsFor: 'accessing' stamp: 'nk 7/18/2003 16:30'!
hasTranslucentColor
	"Answer true if this any of this morph is translucent but not transparent."

	(fillStyle isTranslucent and: [ fillStyle isTransparent not]) ifTrue: [^ true].
	^super hasTranslucentColor! !

!NCLineMorph methodsFor: 'rotate scale and flex' stamp: 'nk 7/11/2003 18:48'!
heading: newHeading
	"Set the receiver's heading (in eToy terms).
	Note that polygons never use flex shells."
	self rotationDegrees: newHeading.! !

!NCLineMorph methodsFor: 'initialization' stamp: 'nk 8/24/2003 17:08'!
initialize
	super initialize.
	smoothCurve := false.
	self vertices: (Array 
				with: 5 @ 0
				with: 200 @ 100
				with: 0 @ 200)
			lineWidth: 2
			lineColor: (Color r: 0 g: 0.7 b: 0.7)
			borderWidth: 0
			borderColor: Color black.
	self makeNoArrows.
! !

!NCLineMorph methodsFor: 'geometry' stamp: 'nk 8/4/2003 18:06'!
innerBounds
	^self bounds! !

!NCLineMorph methodsFor: 'editing' stamp: 'nk 8/18/2003 11:10'!
insertVertexAfterEndAt: position
	"insert another vertex, before first"

	self setVertices: (self vertices copyWith: position) asArray.
! !

!NCLineMorph methodsFor: 'editing' stamp: 'nk 8/20/2003 15:48'!
insertVertexBeforeStartAt: position
	"insert another vertex, before first"

	self setVertices: ({ position }, self vertices) asArray.
! !

!NCLineMorph methodsFor: 'editing' stamp: 'nk 1/14/2004 21:53'!
interiorCurveSegmentAt: n put: aPoint
	^self curveSegmentAt: n put: aPoint
! !

!NCLineMorph methodsFor: 'editing' stamp: 'nk 1/14/2004 22:06'!
interiorCurveVertexAt: index put: aPoint
	(index = 1 or: [ index = vertices size ]) ifTrue: [ ^true ].
	^self curveVertexAt: index put: aPoint! !

!NCLineMorph methodsFor: 'editing' stamp: 'nk 1/14/2004 21:40'!
interiorHandleForEvent: anEvent 
	"Answer a NCCurveHandleMorph that can be used to adjust me at the 
	given point."
	^self interiorHandleForPosition: anEvent position addAllowed: anEvent shiftPressed! !

!NCLineMorph methodsFor: 'editing' stamp: 'nk 1/14/2004 22:03'!
interiorHandleForPosition: pos addAllowed: canAddVertex 
	| handle sel |
	handle := self handleForPosition: pos addAllowed: canAddVertex.
	sel := handle setPointSelector.
	handle setPointSelector: ('interiorC', (sel allButFirst)) asSymbol.
	^handle
! !

!NCLineMorph methodsFor: 'geometry testing' stamp: 'nk 8/22/2003 11:49'!
intersectionsWith: aRectangle
	"Answer a Set of points where the given Rectangle intersects with me.
	Ignores arrowForms."

	| retval |
	retval := IdentitySet new: 4.
	(self bounds intersects: aRectangle) ifFalse: [^ retval].

	self segmentsDo: [ :seg |
		aRectangle lineSegmentsDo: [ :rp1 :rp2 |
			retval addAll: (seg intersectionsWithLineFrom: rp1 to: rp2) ] ].

	^retval
! !

!NCLineMorph methodsFor: 'visual properties' stamp: 'nk 7/18/2003 17:38'!
isBordered
	^true! !

!NCLineMorph methodsFor: 'accessing' stamp: 'nk 7/11/2003 19:12'!
isCurve
	^smoothCurve ifNil: [ smoothCurve := false ]! !

!NCLineMorph methodsFor: 'testing' stamp: 'nk 10/13/2003 18:37'!
isLineMorph
	^true! !

!NCLineMorph methodsFor: 'accessing' stamp: 'nk 7/11/2003 17:49'!
lastVertex
	^vertices last! !

!NCLineMorph methodsFor: 'accessing' stamp: 'nk 7/18/2003 17:09'!
lineBorderColor
	^self borderColor! !

!NCLineMorph methodsFor: 'accessing' stamp: 'nk 7/18/2003 17:09'!
lineBorderColor: aColor
	self borderColor: aColor! !

!NCLineMorph methodsFor: 'accessing' stamp: 'nk 7/18/2003 17:08'!
lineBorderWidth
	^self borderWidth! !

!NCLineMorph methodsFor: 'accessing' stamp: 'nk 7/18/2003 17:08'!
lineBorderWidth: anInteger
	self borderWidth: anInteger! !

!NCLineMorph methodsFor: 'visual properties' stamp: 'nk 7/18/2003 16:49'!
lineColor

	^fillStyle! !

!NCLineMorph methodsFor: 'visual properties' stamp: 'nk 7/18/2003 16:49'!
lineColor: newColor
	self fillStyle: newColor
! !

!NCLineMorph methodsFor: 'geometry-curves' stamp: 'nk 8/18/2003 15:44'!
lineSegments
	| segs |
	segs := OrderedCollection new.
	self lineSegmentsDo: [ :p1 :p2 | segs add: (LineSegment from: p1 to: p2) ].
	^segs! !

!NCLineMorph methodsFor: 'geometry-curves' stamp: 'nk 7/11/2003 18:19'!
lineSegmentsDo: aBlock
	self segmentsDo: [ :seg | seg lineSegmentsDo: aBlock ]! !

!NCLineMorph methodsFor: 'accessing' stamp: 'nk 7/18/2003 14:30'!
lineWidth
	^lineWidth! !

!NCLineMorph methodsFor: 'accessing' stamp: 'nk 12/28/2003 15:42'!
lineWidth1: anObject
	lineWidth := anObject! !

!NCLineMorph methodsFor: 'accessing' stamp: 'nk 12/28/2003 15:42'!
lineWidth: aNumber 
	| ratio oldWidth |
	oldWidth := self fullLineWidth.
	self lineWidth1: aNumber.
	ratio := self fullLineWidth / oldWidth.
	self arrows do: [:a | a scale: (a scale * ratio) asPoint].
	self computeBounds! !

!NCLineMorph methodsFor: 'arrows' stamp: 'nk 8/24/2003 15:18'!
makeBackArrow
	self arrows: { self class basicArrow. self class defaultArrow }.! !

!NCLineMorph methodsFor: 'arrows' stamp: 'nk 8/24/2003 15:19'!
makeBothArrows
	self arrows: { self class basicArrow. self class basicArrow }.! !

!NCLineMorph methodsFor: 'arrows' stamp: 'nk 8/24/2003 15:19'!
makeForwardArrow
	self arrows: { self class defaultArrow. self class basicArrow }.! !

!NCLineMorph methodsFor: 'arrows' stamp: 'nk 8/20/2003 10:38'!
makeNoArrows
	self arrows: { self class defaultArrow.  self class defaultArrow }.! !

!NCLineMorph methodsFor: 'geometry' stamp: 'nk 7/12/2003 00:09'!
midpoint
	"Answer the midpoint along my segments"
	| middle |
	middle := 0.
	self lineSegmentsDo: [ :a :b | middle := middle + (a dist: b) ].
	middle < 2 ifTrue: [ ^ self center ].
	middle := middle / 2.
	self lineSegmentsDo: [ :a :b | | dist |
		dist := (a dist: b).
		middle < dist
			ifTrue: [ ^(a + ((b - a) * (middle / dist))) asIntegerPoint ].
		middle := middle - dist.
	].
	self error: 'can''t happen'! !

!NCLineMorph methodsFor: 'private' stamp: 'nk 7/11/2003 19:53'!
nextDuplicateVertexIndex
	vertices
		doWithIndex: [:vert :index | ((index between: 2 and: vertices size - 1)
					and: [| epsilon v1 v2 | 
						v1 := vertices at: index - 1.
						v2 := vertices at: index + 1.
						epsilon := ((v1 x - v2 x) abs max: (v1 y - v2 y) abs)
									// 32 max: 1.
						vert
							onLineFrom: v1
							to: v2
							within: epsilon])
				ifTrue: [^ index]].
	^ 0! !

!NCLineMorph methodsFor: 'geometry' stamp: 'nk 7/11/2003 17:49'!
nextToFirstPoint
	^vertices second! !

!NCLineMorph methodsFor: 'geometry' stamp: 'nk 7/11/2003 17:48'!
nextToLastPoint
	^vertices at: vertices size - 1! !

!NCLineMorph methodsFor: 'geometry testing' stamp: 'nk 12/28/2003 14:32'!
nudgeForLabel: aRectangle
	"Try to move the label off me. Prefer labels on the top and right."

	| nudge |
	(self bounds intersects: aRectangle) ifFalse: [^ 0@0 ].
	nudge := 0@0.
	(aRectangle insetBy: 1) lineSegmentsDo: [ :rp1 :rp2 | | info bpts |
		info := self curveInfoNearest: rp1.
		bpts := info borderPoints.
		(aRectangle containsPoint: bpts first)
			ifTrue: [ nudge := nudge + (info closestPoint - bpts first) ]
			ifFalse: [
		(aRectangle containsPoint: bpts second)
			ifTrue: [  nudge := nudge - (info closestPoint - bpts second)  ]].
	].
	nudge isZero ifTrue: [ ^nudge ].
	^nudge / (nudge r)! !

!NCLineMorph methodsFor: 'geometry' stamp: 'nk 8/22/2003 23:41'!
outerBounds
	^self curveBounds! !

!NCLineMorph methodsFor: 'geometry-curves' stamp: 'nk 12/28/2003 15:42'!
outsideBezierPoints
	"Answer the nVertices + 1 *6 points that are the boundary points for two sets of bezier curves that define my outer bounds."

	| npts a b halfWidth normal ptsA ptsB tan1 tan2 angle1 angle2 halfAngle midAngle r delta |
	npts := self segments size * 3.
	ptsA := Array new: npts.
	ptsB := Array new: npts.
	a := 1.
	b := npts.
	halfWidth := self lineWidth * 0.5.
	self segmentsDo: 
			[:seg | 
			tan1 ifNil: 
					[tan1 := seg tangentAtStart.
					tan1 isZero ifTrue: [tan1 := 0.0 @ 1.0]].
			angle1 := tan1 negated theta.
			tan2 := seg tangentAtEnd.
			tan2 isZero ifTrue: [tan2 := 0.0 @ 1.0].
			angle2 := tan2 theta.

			"from end"
			normal := tan1 normal * halfWidth.
			ptsA at: a put: seg start - normal.
			ptsB at: b put: seg start + normal.
			a := a + 1.
			b := b - 1.
			seg isBezier2Segment 
				ifTrue: 
					["via"

					halfAngle := (angle1 - angle2) / 2.0.
					halfAngle abs < 4.0e-5 ifTrue: [halfAngle := 4.0e-5].
					midAngle := angle2 + halfAngle.
					r := halfWidth / halfAngle sin.
					r := r min: halfWidth * 6.0 max: halfWidth * -6.0.
					delta := (r * midAngle cos) @ (r * midAngle sin).
					ptsA at: a put: seg via - delta.
					ptsB at: b put: seg via + delta.
					a := a + 1.
					b := b - 1].

			"to end"
			normal := tan2 normal * halfWidth.
			ptsA at: a put: seg end - normal.
			ptsB at: b put: seg end + normal.
			a := a + 1.
			b := b - 1.
			seg isBezier2Segment 
				ifFalse: 
					[ptsA at: a put: seg end - normal.
					ptsB at: b put: seg end + normal.
					a := a + 1.
					b := b - 1].
			tan1 := tan2].
	segments last isBezier2Segment 
		ifFalse: 
			[| csize |
			csize := ptsA size.
			ptsA at: csize - 1 put: (ptsA at: csize - 2).
			ptsB at: 2 put: (ptsB at: 3)].
	^{ 
		ptsA.
		ptsB}! !

!NCLineMorph methodsFor: 'geometry-curves' stamp: 'nk 8/16/2004 13:42'!
outsidePoints
	"Answer the nVertices *2 points that are the boundary points for two sets of lines that define my outer bounds."

	| npts a b halfWidth normal ptsA ptsB tan1 tan2 angle1 angle2 halfAngle midAngle r delta arrowVertices |
	smoothCurve ifTrue: [^self outsideBezierPoints].
	arrowVertices := self arrowVertices.
	npts := arrowVertices size.
	ptsA := Array new: npts.
	ptsB := Array new: npts.
	a := 1.
	b := npts.
	halfWidth := self lineWidth * 0.5.
	2 to: arrowVertices size
		do: 
			[:i | 
			| v1 v2 v3 |
			v1 := arrowVertices at: i - 1.
			v2 := arrowVertices at: i.
			tan1 := v2 - v1.
			tan1 isZero 
				ifTrue: 
					[tan1 := i + 1 > arrowVertices size 
								ifTrue: [v2 - (arrowVertices at: i - 1)]
								ifFalse: [(arrowVertices at: i + 1) - v1]].

			tan1 isZero ifTrue: [ tan1 := 0.0 @ 1.0 ].

			angle1 := tan1 theta.

			"from end"
			a = 1 
				ifTrue: 
					[normal := tan1 normal * halfWidth.
					ptsA at: a put: v1 - normal.
					ptsB at: b put: v1 + normal.
					a := a + 1.
					b := b - 1].
			b = 1 
				ifTrue: 
					["to end"

					normal := tan1 normal * halfWidth.
					ptsA at: a put: v2 - normal.
					ptsB at: b put: v2 + normal.
					a := a + 1.
					b := b - 1]
				ifFalse: 
					[v3 := arrowVertices at: i + 1.
					tan2 := v2 - v3.
					angle2 := tan2 theta.

					"via"
					halfAngle := (angle1 - angle2) / 2.0.
					halfAngle abs < 0.004 ifTrue: [halfAngle := 0.004].
					midAngle := angle2 + halfAngle.
					r := halfWidth / halfAngle sin.
					r := r min: halfWidth * 6 max: halfWidth * -6.
					delta := (r * midAngle cos) @ (r * midAngle sin).
					ptsA at: a put: v2 + delta.
					ptsB at: b put: v2 - delta.
					a := a + 1.
					b := b - 1]].
	^{ 
		ptsA.
		ptsB}! !

!NCLineMorph methodsFor: 'private' stamp: 'nk 12/27/2003 11:23'!
privateMoveBy: delta
	super privateMoveBy: delta.
	(self valueOfProperty: #referencePosition) ifNotNilDo:
		[:oldPos | self setProperty: #referencePosition toValue: oldPos + delta].
	self setVertices: (vertices collect: [:p | p + delta]).
! !

!NCLineMorph methodsFor: 'menus' stamp: 'nk 8/28/2003 15:54'!
reasonableForms
	| reasonableForms |
	reasonableForms := Imports default images asOrderedCollection.
	reasonableForms addAll: (BitmapFillStyle allSubInstances collect:[:f| f form]).
	^reasonableForms! !

!NCLineMorph methodsFor: 'geometry' stamp: 'nk 7/11/2003 19:53'!
reduceVertices
	"Reduces the vertices size, when 3 vertices are on the same line with a 
	little epsilon. Based on code by Steffen Mueller"
	| dup |
	[ (dup := self nextDuplicateVertexIndex) > 0 ] whileTrue: [
		self setVertices: (vertices copyWithoutIndex: dup)
	].
	^vertices size.! !

!NCLineMorph methodsFor: 'rotate scale and flex' stamp: 'nk 12/27/2003 11:23'!
referencePosition 
	"Return the current reference position of the receiver"
	^ self valueOfProperty: #referencePosition ifAbsent: [ self midpoint ]
! !

!NCLineMorph methodsFor: 'editing' stamp: 'nk 8/22/2003 17:33'!
removeHandles
	| handles |
	handles := submorphs select: [ :ea | ea isCurveHandle ].
	self deleteMorphs: handles.! !

!NCLineMorph methodsFor: 'geometry' stamp: 'nk 10/19/2003 16:52'!
removeShortEndSegmentsStepping: stepBlock
	| v1 v2 n minLength |
	minLength := 30.
	
	(n := self vertices size) > 2 ifTrue: [
		stepBlock value.
		v1 := self vertexAt: 1.
		v2 := self vertexAt: 2.
		(v1 dist: v2) < minLength ifTrue: [ self removeVertex: v2 ].
	].
	(n := self vertices size) > 2 ifTrue: [
		stepBlock value.
		v1 := self vertexAt: n.
		v2 := self vertexAt: n-1.
		(v1 dist: v2) < minLength ifTrue: [ self removeVertex: v2 ].
	].
	self vertices size > 2 ifTrue: [ self reduceVertices ].
	stepBlock value.
! !

!NCLineMorph methodsFor: 'editing' stamp: 'nk 10/19/2003 16:57'!
removeVertex: aVertex
	^self setVertices: (self vertices copyWithout: aVertex)! !

!NCLineMorph methodsFor: 'rotate scale and flex' stamp: 'nk 7/11/2003 18:49'!
rotationCenter
	"Return the rotation center of the receiver. The rotation center defines the relative offset inside the receiver's bounds for locating the reference position."
	| refPos |
	refPos := self valueOfProperty: #referencePosition
		ifAbsent: [^ 0.5@0.5].
	^ (refPos - self bounds origin) / self bounds extent asFloatPoint! !

!NCLineMorph methodsFor: 'rotate scale and flex' stamp: 'nk 7/11/2003 18:49'!
rotationCenter: aPointOrNil
	"Set the new rotation center of the receiver. The rotation center defines the relative offset inside the receiver's bounds for locating the reference position."
	| box |
	aPointOrNil isNil
		ifTrue: [self removeProperty: #referencePosition]
		ifFalse: [box := self bounds.
				self setProperty: #referencePosition
					toValue: box origin + (aPointOrNil * box extent)]
! !

!NCLineMorph methodsFor: 'rotate scale and flex' stamp: 'nk 7/11/2003 18:56'!
rotationDegrees

	^ self forwardDirection! !

!NCLineMorph methodsFor: 'rotate scale and flex' stamp: 'nk 12/27/2003 11:24'!
rotationDegrees: degrees 
	| flex center |
	center := self referencePosition.
	flex := (MorphicTransform offset: center negated)
			withAngle: (degrees - self forwardDirection) degreesToRadians.
	self setVertices: (vertices collect: [:v | (flex transform: v) - flex offset]).
	self forwardDirection: degrees.

! !

!NCLineMorph methodsFor: 'rotate scale and flex' stamp: 'nk 9/4/2004 07:41'!
scale
	^self valueOfProperty: #scaleFactor ifAbsent: [ 1.0 ]
! !

!NCLineMorph methodsFor: 'rotate scale and flex' stamp: 'nk 9/4/2004 11:54'!
scale: scaleFactor 
	| flex center ratio |
	ratio := self scaleFactor / scaleFactor.
	self lineWidth: self lineWidth / ratio.
	self borderWidth: self borderWidth / ratio.
	center := self referencePosition.
	flex := (MorphicTransform offset: center negated)
				withScale: ratio.
	self
		setVertices: (vertices
				collect: [:v | (flex transform: v)
						- flex offset]).
	super scale: scaleFactor.! !

!NCLineMorph methodsFor: 'accessing' stamp: 'nk 12/28/2003 16:53'!
segments
	^segments ifNil: 
			[segments := self computeSegments segments do: [:ea | ea roundTo: 0.001]]! !

!NCLineMorph methodsFor: 'geometry-curves' stamp: 'nk 8/19/2003 10:34'!
segmentsDo: segmentBlock 
	"Emit a sequence of LineSegment subinstances into segmentBlock"

	self segments do: segmentBlock! !

!NCLineMorph methodsFor: 'editing' stamp: 'nk 1/14/2004 17:39'!
setVertices: newVertices

	| proposedVertices |
	vertices = newVertices ifTrue: [ ^self ].

	proposedVertices := Array streamContents: [ :s |
		s nextPut: newVertices first asFloatPoint.
		newVertices overlappingPairsDo: [ :v1 :v2 |
			(v1 dist: v2) >= 2 ifTrue: [ s nextPut: v2 asFloatPoint ] ] ].
	
	proposedVertices := newVertices.
	
	proposedVertices size < 2 ifTrue: [ self halt ].

	proposedVertices size = 2 ifTrue: [
		(newVertices first dist: newVertices last) < 2 ifTrue: [
			proposedVertices := { newVertices first. newVertices first + (2@2) }] ].
	vertices = proposedVertices ifTrue: [ ^self ].
	vertices := proposedVertices.
	self computeBounds! !

!NCLineMorph methodsFor: 'editing' stamp: 'nk 7/11/2003 21:57'!
straighten
	self setVertices: { vertices first . vertices last }! !

!NCLineMorph methodsFor: 'accessing' stamp: 'nk 7/11/2003 18:50'!
toggleSmoothing

	smoothCurve := smoothCurve not.
	self computeBounds! !

!NCLineMorph methodsFor: 'accessing' stamp: 'nk 8/25/2003 16:21'!
totalLength
	"Answer the full length of my segments. Can take a long time if I'm curved."
	^self segments detectSum: [ :seg | seg length ]! !

!NCLineMorph methodsFor: 'rotate scale and flex' stamp: 'nk 1/23/2004 14:57'!
transformedBy: aTransform
	self setVertices: (self vertices collect:[:v| aTransform localPointToGlobal: v])! !

!NCLineMorph methodsFor: 'TODO' stamp: 'nk 7/11/2003 18:51'!
unrotatedLength
	"If the receiver bears rotation without a transformation morph, answer what its length in the direction it is headed is"

	vertices size == 2 ifTrue:
		[^ (vertices second - vertices first) r].

	^ ((self class new setVertices: vertices) rotationDegrees: self rotationDegrees negated) height! !

!NCLineMorph methodsFor: 'TODO' stamp: 'nk 7/11/2003 18:51'!
unrotatedLength: aLength
	"If the receiver bears rotation without a transformation morph, answer what its length in the direction it is headed is"

	vertices size == 2 ifTrue: [^ self arrowLength: aLength].

	self setVertices: ((((self class new setVertices: vertices) rotationDegrees: self rotationDegrees negated) height: aLength) rotationDegrees: 0) vertices! !

!NCLineMorph methodsFor: 'geometry' stamp: 'nk 9/4/2004 07:19'!
unrotatedWidth
	^self lineWidth! !

!NCLineMorph methodsFor: 'arrows' stamp: 'nk 12/28/2003 16:13'!
updateArrowContours
	"Open the arrows for my line width.
	Set contours[2] and contours[4] with arrow contours.
	Put the arrow join points into arrowJoints."

	| transform lw |
	lw := (smoothCurve not and: [self lineWidth <= 0]) 
				ifTrue: [0]
				ifFalse: [self lineWidth].
	self arrows first openForLineWidth: lw.
	arrows second openForLineWidth: lw.
	transform := (MatrixTransform2x3 
				withAngle: (vertices second - vertices first) negated theta 
						radiansToDegrees + 90.0) 
					offset: vertices first.
	contours at: 4 put: (arrows first contourPointsTransformedBy: transform).
	self arrowJoints at: 1
		put: (transform transformPoint: arrows first joinPoint).
	transform := (MatrixTransform2x3 
				withAngle: (vertices last - vertices nextToLast) theta radiansToDegrees 
						+ 90.0) 
					offset: vertices last.
	contours at: 2 put: (arrows second contourPointsTransformedBy: transform).
	self arrowJoints at: 2
		put: (transform transformPoint: arrows second joinPoint)! !

!NCLineMorph methodsFor: 'accessing' stamp: 'nk 7/11/2003 19:50'!
vertexAt: n
	^vertices at: (n min: vertices size).! !

!NCLineMorph methodsFor: 'drawing' stamp: 'nk 7/17/2003 17:54'!
vertexPairsDo: aBlock 
	"Emit pairs of start/end points into aBlock"
	vertices overlappingPairsDo: aBlock! !

!NCLineMorph methodsFor: 'accessing' stamp: 'nk 7/11/2003 15:08'!
vertices
	^ vertices! !

!NCLineMorph methodsFor: 'initialization' stamp: 'nk 12/28/2003 15:42'!
vertices: verts color: aColor borderWidth: borderWidthInteger borderColor: anotherColor 
	self fillStyle: anotherColor.
	self borderColor: anotherColor.
	self lineWidth1: borderWidthInteger.
	borderWidth := 0.
	self setVertices: verts! !

!NCLineMorph methodsFor: 'initialization' stamp: 'nk 12/28/2003 15:42'!
vertices: verts lineWidth: w lineColor: c borderWidth: bw borderColor: bc 
	fillStyle := c.
	color := c asColor.
	borderColor := bc.
	self lineWidth1: w.
	borderWidth := bw asInteger.
	self setVertices: verts! !

!NCLineMorph methodsFor: 'editing' stamp: 'nk 8/19/2003 10:52'!
verticesAt: n put: newPoint
	| index |
	index := (n min: vertices size) max: 1.
	self setVertices: (vertices copy at: index put: newPoint; yourself).
! !

TextMorph subclass: #NCTextMorph
	instanceVariableNames: 'acceptOnCR acceptOnFocusLoss'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Connectors-Text-Base'!

!NCTextMorph commentStamp: '<historical>' prior: 0!
I am a subclass of TextMorph that can be told to accept on either focus loss and/or CR.!

!NCTextMorph class methodsFor: 'defaults' stamp: 'nk 10/19/2003 18:41'!
editorClass
	^NCTextMorphEditor! !

!NCTextMorph class methodsFor: 'class initialization' stamp: 'nk 11/1/2004 07:52'!
initialize
	"NCTextMorph initialize"
	ScriptingSystem addCustomEventFor: self named: #acceptedTextContents help: 'when text is accepted into a text morph by hitting Enter or losing focus' targetMorphClass: self.
! !

!NCTextMorph class methodsFor: 'class initialization' stamp: 'nk 11/1/2004 07:37'!
unload
	"NCTextMorph unload"
	ScriptingSystem removeCustomEventNamed: #acceptedTextContents for: self.
! !

!NCTextMorph methodsFor: 'editing' stamp: 'nk 11/1/2004 10:59'!
acceptContents
	"Tell our interested parties about our new contents"
	super acceptContents.
	self triggerEvent: #accept withArguments: {self contents. self}.
	self triggerEtoyEvent: #acceptedTextContents.
	ActiveHand releaseKeyboardFocus: self! !

!NCTextMorph methodsFor: 'accessing' stamp: 'nk 5/19/2003 18:37'!
acceptOnCR
	^acceptOnCR ifNil: [ acceptOnCR := false ]. ! !

!NCTextMorph methodsFor: 'accessing' stamp: 'nk 5/20/2003 15:46'!
acceptOnCR: aBoolean
	acceptOnCR := aBoolean.
	self crAction: (aBoolean
		ifTrue: [ MessageSend receiver: self selector: #acceptContents ]
		ifFalse: [ ]).
! !

!NCTextMorph methodsFor: 'menus' stamp: 'nk 10/13/2003 17:21'!
acceptOnCROnOff
	self acceptOnCR: self acceptOnCR not! !

!NCTextMorph methodsFor: 'menus' stamp: 'dgd 8/26/2004 14:40'!
acceptOnCRString

	^ (self acceptOnCR
		ifTrue: ['<yes>']
		ifFalse: ['<no>'])
		, 'accept on CR' translated
! !

!NCTextMorph methodsFor: 'accessing' stamp: 'nk 5/20/2003 15:43'!
acceptOnFocusLoss
	^acceptOnFocusLoss ifNil: [ acceptOnFocusLoss := false ]! !

!NCTextMorph methodsFor: 'accessing' stamp: 'nk 5/20/2003 15:39'!
acceptOnFocusLoss: anObject
	acceptOnFocusLoss := anObject! !

!NCTextMorph methodsFor: 'menus' stamp: 'nk 10/13/2003 17:21'!
acceptOnFocusLossOnOff
	self acceptOnFocusLoss: self acceptOnFocusLoss not! !

!NCTextMorph methodsFor: 'menus' stamp: 'dgd 8/26/2004 14:40'!
acceptOnFocusLossString

	^ (self acceptOnFocusLoss
		ifTrue: ['<yes>']
		ifFalse: ['<no>'])
		, 'accept on focus loss' translated
! !

!NCTextMorph methodsFor: 'menus' stamp: 'nk 10/13/2003 20:52'!
addAcceptMenuItemsTo: aCustomMenu 
	aCustomMenu 
		addUpdating: #acceptOnCRString
		target: self
		action: #acceptOnCROnOff.
	aCustomMenu 
		addUpdating: #acceptOnFocusLossString
		target: self
		action: #acceptOnFocusLossOnOff! !

!NCTextMorph methodsFor: 'menus' stamp: 'nk 10/13/2003 20:52'!
addCustomMenuItems: aCustomMenu hand: aHandMorph 
	"Include our modest command set in the ctrl-menu"

	super addCustomMenuItems: aCustomMenu hand: aHandMorph.
	self addAcceptMenuItemsTo: aCustomMenu! !

!NCTextMorph methodsFor: 'menus' stamp: 'nk 10/13/2003 20:55'!
addFitAndWrapItemsTo: aCustomMenu
	aCustomMenu 
		addUpdating: #autoFitString
		target: self
		action: #autoFitOnOff.
	aCustomMenu 
		addUpdating: #wrapString
		target: self
		action: #wrapOnOff.! !

!NCTextMorph methodsFor: 'text links' stamp: 'nk 10/21/2002 14:36'!
addItem: classAndMethod
	"Make a linked message list and put this method in it"

	Model new addItem: classAndMethod	"let him do all the work"! !

!NCTextMorph methodsFor: 'menus' stamp: 'nk 8/6/2004 18:15'!
chooseEmphasisOrAlignment
	| wasEmptyOrNil |
	self isLocked ifTrue: [ ^self ].
	wasEmptyOrNil :=  self selection isEmptyOrNil.
	wasEmptyOrNil ifTrue: [ self selectAll ].
	super chooseEmphasisOrAlignment.
	wasEmptyOrNil ifTrue: [ self selectFrom: 1 to: 0 ].
! !

!NCTextMorph methodsFor: 'naming' stamp: 'cmm 12/4/2010 20:06'!
defaultNameStemForInstances
	^ self class name
		copyFrom: 3
		to: self class name size! !

!NCTextMorph methodsFor: 'drawing' stamp: 'nk 8/31/2004 11:01'!
drawOn: aCanvas 
	super drawOn: aCanvas.
	aCanvas isShadowDrawing
		ifTrue: [^ self].
	self hasFocus
		ifTrue: [aCanvas
				frameRectangle: (self fullBounds insetBy: 0)
				color: Preferences keyboardFocusColor]! !

!NCTextMorph methodsFor: 'event handling' stamp: 'nk 8/22/2004 17:24'!
handlesKeyboard: evt
	"Avoid stray keystrokes, especially from gestures."
	self trackFocusFromEvent: evt.
	^self hasFocus! !

!NCTextMorph methodsFor: 'event handling' stamp: 'nk 8/22/2004 17:24'!
handlesMouseDown: evt
	self trackFocusFromEvent: evt.
	self isPartsDonor ifTrue: [^ false].
	self eventHandler ifNotNil: [^ self eventHandler handlesMouseDown: evt].
	^evt redButtonPressed or: [ self allowsGestureStart: evt].
! !

!NCTextMorph methodsFor: 'editing' stamp: 'nk 8/22/2004 17:17'!
highlightRectChanged

	(self outerBounds areasOutside: (self innerBounds insetBy: 3))
		do: [ :rect | self invalidRect: rect ]! !

!NCTextMorph methodsFor: 'private' stamp: 'nk 8/22/2004 15:53'!
installEditorToReplace: priorEditor
	"Install an editor for my paragraph.  This constitutes 'hasFocus'.
	If priorEditor is not nil, then initialize the new editor from its state.
	We may want to rework this so it actually uses the prior editor."

	| stateArray |

	priorEditor ifNotNil: [stateArray := priorEditor stateArray].
	editor := self class editorClass new morph: self.
	editor changeParagraph: self paragraph.
	priorEditor ifNotNil: [editor stateArrayPut: stateArray].
	self highlightRectChanged.
	self selectionChanged.
	^ editor! !

!NCTextMorph methodsFor: 'editing' stamp: 'nk 8/22/2004 17:29'!
keyboardFocusChange: aBoolean

	aBoolean ifFalse: [
		(self hasFocus and: [ self acceptOnFocusLoss ])
			ifTrue: [ self acceptContents ].
		self releaseEditor ].
	super keyboardFocusChange: aBoolean.
! !

!NCTextMorph methodsFor: 'accessing' stamp: 'nk 6/17/2002 20:31'!
margins: newMargins
	super margins: newMargins.
	(owner respondsTo: #forceRedraw) ifTrue: [ owner forceRedraw ]! !

!NCTextMorph methodsFor: 'event handling' stamp: 'nk 8/22/2004 17:25'!
rejectsEvent: anEvent
	self trackFocusFromEvent: anEvent.
	^(super rejectsEvent: anEvent)
		or: [ anEvent isKeyboard and: [ (self handlesKeyboard: anEvent) not ] ]! !

!NCTextMorph methodsFor: 'private' stamp: 'nk 8/22/2004 16:02'!
releaseEditor
	super releaseEditor.
	self highlightRectChanged
! !

!NCTextMorph methodsFor: 'event handling' stamp: 'nk 8/22/2004 17:24'!
trackFocusFromEvent: evt
	evt hand keyboardFocus == self ifFalse: [
		self hasFocus ifTrue: [ self highlightRectChanged. self releaseEditor ]].! !

QHsmMorphicEventHandler subclass: #NCAAEventHandler
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Connectors-Base'!

!NCAAEventHandler commentStamp: '<historical>' prior: 0!
This is the state machine handler for the NCAAConnectorMorph instances.!

!NCAAEventHandler methodsFor: 'state handlers' stamp: 'nk 1/14/2004 15:40'!
determineConnectionState
	| connections |
	connections := morph validConnections.
	^#(Disconnected OneEndConnected Connected) at: connections size + 1! !

!NCAAEventHandler methodsFor: 'state handlers' stamp: 'nk 8/16/2003 15:25'!
metaMouseButtonOr: aBlock
	"If the current event is a meta-operation, then handle it and return nil.
	Otherwise, evaluate aBlock."

	event blueButtonChanged ifTrue: [ morph blueButtonDown: event. ^nil ].
	event controlKeyPressed ifTrue: [ morph invokeMetaMenu: event. ^nil ].
	event yellowButtonPressed ifTrue: [ 
		morph yellowButtonDown: event onEnd: (morph findEndNear: event position).
		^nil ].
	morph isPartsDonor ifTrue: [ ^nil ].
	^aBlock value
! !

!NCAAEventHandler methodsFor: 'state handlers' stamp: 'nk 1/15/2004 11:08'!
stateConnected: evt
	^evt caseOf: {
		[#entry] -> [ morph enterConnected; connectionsChanged. nil ].
		[#mouseDown] -> [ self metaMouseButtonOr: [ morph mouseDownInConnected: event. nil] ].
		[#startEditing] -> [ self newState: #Editing ].
		[#step] -> [ morph stepConnected. nil ].
	} otherwise: [ self state: #Idle ].! !

!NCAAEventHandler methodsFor: 'state handlers' stamp: 'nk 8/21/2003 11:32'!
stateDisconnected: evt

	^evt caseOf: {
		[#entry] -> [ morph connectionsChanged. nil ].
		[#exit] -> [ nil ].

		[#mouseDown] -> [ self metaMouseButtonOr: [morph mouseDownInDisconnected: event. nil] ].
	} otherwise: [ self state: #Idle ].! !

!NCAAEventHandler methodsFor: 'state handlers' stamp: 'nk 1/14/2004 17:45'!
stateDraggingEnd: evt 
	evt caseOf: {
		[#entry] -> 
			[morph enterDraggingEnd ].
		[#exit] -> 
			[morph exitDraggingEnd; enterHandlingDrop ].
		[#step] -> 
			[morph monitorDestinationForEnds ifTrue: [ ^self newState: #DraggingEndOver ]].
		[#mouseUp] -> 
			[^(morph mouseUpInDraggingEnd: event)
				ifTrue: [self newState: #Idle]].
		[#mouseDown] -> 
			[morph mouseDownInDraggingEnd: event. ^nil ]}
		 otherwise: [].
	^self global! !

!NCAAEventHandler methodsFor: 'state handlers' stamp: 'nk 1/17/2004 10:59'!
stateDraggingEndOver: evt 
	evt caseOf: {
		[#entry] -> [morph enterDraggingEndOver].
		[#exit] -> [morph exitDraggingEndOver].
		[#step] -> 
			[ ^morph monitorDestinationForEnds ifFalse: [ 
				morph constraints do: [ :ea | ea startStepping ].
				self newState: #DraggingEnd ]].
}
		 otherwise: [].

	^ self state: #DraggingEnd! !

!NCAAEventHandler methodsFor: 'state handlers' stamp: 'nk 1/14/2004 17:07'!
stateDraggingMorph: evt 
	evt caseOf: {
		[#entry] -> [morph enterDraggingMorph].
		[#exit] -> [morph exitDraggingMorph; enterHandlingDrop].
		[#step] -> [morph monitorDestinationForEnds ].
		[#wireMorph] -> [ ^self newState: #WiringMorph ].
		[#ownerChangedToPasteUp] -> [^ self newState: #Idle]}
		 otherwise: [].
	^ self global! !

!NCAAEventHandler methodsFor: 'state handlers' stamp: 'nk 2/28/2006 10:40'!
stateEditing: evt
	^evt caseOf: {
		[#dragStart] -> [ morph enterEditing. nil ].
		[#dragStop] -> [
			|  buttonEvent |
			morph exitEditing.
			buttonEvent := event arguments first mouseUpEvent.
			event arguments first delete.
			self handleMouseUp: buttonEvent.
			self newState: #Connected ].
		[#step] -> [	 morph stepEditing. nil ].
	} otherwise: [ self state: #Connected ]! !

!NCAAEventHandler methodsFor: 'state handlers' stamp: 'nk 1/14/2004 15:44'!
stateGlobal: evt 
	evt caseOf: {
		[#ownerChangedToPasteUp] -> 
			[morph pasteUpChanged: event.
			^ nil].
		[#step] -> [^ nil].
		[#suspendEventHandler] -> 
			[Cursor normal show.
			^ nil]}
		 otherwise: [].
	event isMorphicEvent
		ifTrue: [morph addedSubmorphs
				do: [:h | ((evt == #mouseLeave) or: [ (h containsPoint: event position) ])
						ifTrue: [h handleEvent: event.
							^ nil]]].
	^ evt caseOf: {
		[#mouseEnter] -> [nil].
		[#mouseLeave] -> [nil].
		[#mouseDown] -> [self
			metaMouseButtonOr: [self top]].
		[#keystroke] -> 
			[event keyValue == 27
				ifTrue: [morph
						yellowButtonDown: event
						onEnd: (morph findEndNear: event position).
					nil]]}
		 otherwise: [super stateGlobal: evt]! !

!NCAAEventHandler methodsFor: 'state handlers' stamp: 'nk 1/14/2004 20:36'!
stateIdle: evt 
	evt caseOf: { 
				[#init] -> [^ self newInitialState: self determineConnectionState ].
				[#entry]->[ morph fixDegenerateLine ].
				[#ownerChangedToHand] -> [^self newState: #DraggingMorph].
				[#mouseDownNearEnd] -> [^self newState: #DraggingEnd].
				[#wireFrom] -> [^self newState: #WiringEnd].
				[#wireFinish] -> [ ^self newState: #WiringEnd ].
				[#keystroke] -> [ morph keyStroke: event. ^nil ].
	}
		otherwise: [].

	"determine connection state"
	(#(#connectEndTo #createdConnected #endGone #bothEndsGone) includes: evt) 
		ifTrue: [^ self newState: self determineConnectionState].
	^self global! !

!NCAAEventHandler methodsFor: 'state handlers' stamp: 'nk 8/22/2004 06:25'!
stateOneEndConnected: evt
	^evt caseOf: {
		[#entry] -> [ morph connectionsChanged. nil ].
		[#mouseDown] -> [ self metaMouseButtonOr: [morph mouseDownInOneEndConnected: event. nil] ].
		[#step] -> [ morph stepOneEndConnected. nil ].
	} otherwise: [ self state: #Idle ].! !

!NCAAEventHandler methodsFor: 'state handlers' stamp: 'nk 1/14/2004 16:12'!
stateWiringEnd: evt 
	evt caseOf: {
		[#entry] -> 
			[morph enterWiringEnd.].
		[#exit] -> 
			[morph exitWiringEnd.]}
		 otherwise: [].
	^self state: #DraggingEnd.! !

!NCAAEventHandler methodsFor: 'state handlers' stamp: 'nk 1/14/2004 16:10'!
stateWiringEndOver: evt 
	evt caseOf: {
		[#entry] -> [morph enterWiringEndOver].
		[#exit] -> [morph exitWiringEndOver]}
		 otherwise: [].

	^ self state: #WiringEnd! !

!NCAAEventHandler methodsFor: 'state handlers' stamp: 'nk 1/14/2004 16:59'!
stateWiringMorph: evt 
	evt caseOf: {
		[#entry] -> [morph enterWiringMorph].
		[#exit] -> [morph exitWiringMorph]}
		 otherwise: [].
	^ self state: #DraggingMorph! !

!SimpleButtonMorph methodsFor: '*connectors-labels' stamp: 'nk 12/4/2002 13:20'!
beLabeled: aBoolean
	"nothing to do here"! !

!SimpleButtonMorph methodsFor: '*connectors-labels' stamp: 'nk 12/26/2003 17:33'!
isButton
	^true! !

GraphicalDictionaryMenu subclass: #NCArrowDictionaryMenu
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Connectors-Tools'!

!NCArrowDictionaryMenu class methodsFor: 'as yet unclassified' stamp: 'dgd 8/26/2004 15:08'!
chooseArrowGlyphFor: target selector: selector
	"NCArrowDictionaryMenu chooseArrowGlyphFor: ObjectExplorer new selector: #openExplorerFor:"
	| inst |
	inst := self new chooseArrowGlyphFor: target selector: selector.
     	HandMorph attach: (inst wrappedInWindowWithTitle: 'Arrows' translated).
	^ inst! !

!NCArrowDictionaryMenu methodsFor: 'menu commands' stamp: 'nk 6/30/2004 16:08'!
buildFormDict
	| formDict |
	formDict := Dictionary new.
	self glyphDictionary keysAndValuesDo: [ :k :v |
		v scale: 12.
		formDict at: k put: (v sampleFormForLineWidth: 10 borderWidth: 1). ].
	self baseDictionary: formDict.
	^formDict.
! !

!NCArrowDictionaryMenu methodsFor: 'menu commands' stamp: 'dgd 8/26/2004 15:09'!
chooseArrowGlyphFor: client selector: sel
	| formDict bb |
	self glyphDictionary addDependent: self.
	entryNames := #().
	currentIndex := 0.
	formDict := self buildFormDict.
	super initializeFor: nil fromDictionary: formDict.
	target := client.
	self selector: sel.
	bb := SimpleButtonMorph new target: self;
				 borderColor: Color black.
	self addMorph: (bb label: 'OK' translated;
			 actionSelector: #okay)
		fullFrame: (LayoutFrame fractions: (0.5 @ 0 extent: 0 @ 0)
				offsets: (-80 @ 4 extent: bb extent)).
! !

!NCArrowDictionaryMenu methodsFor: 'naming' stamp: 'cmm 12/4/2010 20:06'!
defaultNameStemForInstances
	^ self class name
		copyFrom: 3
		to: self class name size! !

!NCArrowDictionaryMenu methodsFor: 'submorphs-add/remove' stamp: 'nk 6/30/2004 15:45'!
delete
	self glyphDictionary removeDependent: self.
	super delete.
! !

!NCArrowDictionaryMenu methodsFor: 'accessing' stamp: 'nk 8/23/2003 20:43'!
glyphDictionary
	^NCAAConnectorMorph arrowDictionary ! !

!NCArrowDictionaryMenu methodsFor: 'accessing' stamp: 'nk 6/30/2004 16:06'!
nameOfGraphic
	^ entryNames ifNotNilDo: [ :en | en at: currentIndex ifAbsent: []]
! !

!NCArrowDictionaryMenu methodsFor: 'event handling' stamp: 'nk 8/23/2003 20:41'!
okay
	| nArgs glyph |
	glyph := self glyphDictionary at: self nameOfGraphic ifAbsent: [ ].
	target ifNotNil:[
		nArgs := selector numArgs.
		nArgs = 1 ifTrue:[target perform: selector with: glyph ].
		nArgs = 2 ifTrue:[target perform: selector with: glyph with: argument]].

	owner isSystemWindow ifTrue: [ owner delete ]! !

!NCArrowDictionaryMenu methodsFor: 'menu commands' stamp: 'nk 6/30/2004 16:25'!
openGlyphEditor
	Smalltalk at: #NCGlyphEditor ifPresent: [ :ncge | | oldKeys |
		self setProperty: #oldGlyphKeys toValue: entryNames asOrderedCollection.
		ncge openInWindow. ].
! !

!NCArrowDictionaryMenu methodsFor: 'menu commands' stamp: 'nk 6/30/2004 15:54'!
removeEntry
	self glyphDictionary
		removeKey: self nameOfGraphic;
		changed: #arrowsChanged.
	super removeEntry.
! !

!NCArrowDictionaryMenu methodsFor: 'menu commands' stamp: 'nk 8/31/2004 16:19'!
renameEntry
	| reply currentName |
	reply := FillInTheBlank 
				request: 'New key? ' translated
				initialAnswer: (currentName := self nameOfGraphic)
				centerAt: self center.
	(reply isEmptyOrNil or: [reply = currentName]) ifTrue: [^Beeper beep].
	(self glyphDictionary includesKey: reply) 
		ifTrue: 
			[^self 
				inform: 'sorry, that conflicts with
the name of another
entry in this dictionary' translated ].
	self glyphDictionary at: reply put: (self glyphDictionary at: currentName).
	(self glyphDictionary)
		removeKey: currentName;
		changed: #arrowsChanged! !

!NCArrowDictionaryMenu methodsFor: 'menu commands' stamp: 'nk 8/31/2004 16:22'!
showMenu
	"Show the receiver's menu"

	| aMenu |
	aMenu := MenuMorph new defaultTarget: self.
	aMenu addTranslatedList: #(
		('remove'			removeEntry			'Remove this arrow from the dictionary')
		('rename'			renameEntry			'Rename this arrow')
		-
		('copy name'		copyName				'Copy the name of this graphic to the clipboard')
		-
		('find...'			findEntry				'Find an entry by name')).
	
	Smalltalk at: #NCGlyphEditor ifPresent: [ :ncge |
		aMenu
			addLine;
			add: 'make new arrow' translated target: self selector: #openGlyphEditor.
		aMenu balloonTextForLastItem: 'Open an Arrow editor' translated. 
	].

	aMenu popUpInWorld
! !

!NCArrowDictionaryMenu methodsFor: 'updating' stamp: 'nk 6/30/2004 16:31'!
update: anAspect
	| currentName newKeys |
	currentName := self nameOfGraphic.
	super update: anAspect.
	(anAspect == #arrowsChanged) ifFalse: [ ^self ].
	self updateFromGlyphDictionary.
	currentIndex := entryNames indexOf: currentName ifAbsent: [ 1].
	self valueOfProperty: #oldGlyphKeys ifPresentDo: [ :oldKeys |
		newKeys := entryNames difference: oldKeys.
		self removeProperty: #oldGlyphKeys.
		newKeys isEmpty ifTrue: [ ^self ].
		currentIndex := entryNames indexOf: newKeys first.
		self updateThumbnail].
! !

!NCArrowDictionaryMenu methodsFor: 'menu commands' stamp: 'nk 6/30/2004 16:14'!
updateFromGlyphDictionary
	| currentName formDict |
	currentName := self nameOfGraphic.
	formDict := self buildFormDict.
	currentIndex := entryNames indexOf: currentName ifAbsent: [ 1].
	self updateThumbnail.
	^formDict.
! !

!NCArrowDictionaryMenu methodsFor: 'private' stamp: 'nk 1/11/2004 17:43'!
updateThumbnail
	| f newBounds |
	"Maintain aspect ratio by resizing the Thumbnail"
	f := formChoices at: currentIndex.
	newBounds := (0@0 extent: f extent) scaleAndTranslateToFitInside: ((8@72) extent: 100@100).
	formDisplayMorph layoutFrame rightOffset: newBounds right; bottomOffset: newBounds bottom.
	super updateThumbnail.! !

!SelectionMorph methodsFor: '*connectors-dropping/grabbing' stamp: 'nk 6/14/2004 21:12'!
dismissMorph
	selectedItems do: [:m | m dismissMorph ].
	super dismissMorph! !

!ObjectPropertiesMorph methodsFor: '*connectors-visual properties' stamp: 'nk 1/18/2004 16:43'!
addOptionalButtonsTo: aCollection 
	aCollection
		addLast: (self transparentSpacerOfSize: 20 @ 3).
	aCollection
		addLast: (self
				buttonNamed: 'Button' translated
				action: #doButtonProperties
				color: color lighter
				help: 'open a button properties panel for the morph' translated).
	myTarget wantsTextPropertiesSheet
		ifTrue: [aCollection
				addLast: (self
						buttonNamed: 'Text' translated
						action: #doTextProperties
						color: color lighter
						help: 'open a text properties panel for the morph' translated)].
	myTarget wantsConnectorPropertiesSheet
		ifTrue: [aCollection
				addLast: (self
						buttonNamed: 'Connector' translated
						action: #doConnectorProperties
						color: color lighter
						help: 'open a Connector properties panel for the morph' translated)].
	^ aCollection! !

!ObjectsTool methodsFor: '*connectors-initialization' stamp: 'nk 8/26/2003 15:57'!
isObjectsTool
	^true! !

!SystemWindow methodsFor: '*connectors-embeddings' stamp: 'nk 8/14/2004 12:59'!
allowsConnectionToSubmorphs
	"Answer true if I will allow a connector to connect to any of my submorphs.
	If not, the search for attachment points in embedded morphs will not be made."

	^self valueOfProperty: #allowsConnectionToSubmorphs ifAbsent: [ false ]! !

!Collection methodsFor: '*connectors-truncation and round-off' stamp: 'nk 12/30/2003 15:47'!
roundTo: quantum
	^self collect: [ :ea | ea roundTo: quantum ]! !

!TransformationMorph methodsFor: '*connectors-connection' stamp: 'nk 6/14/2002 16:18'!
connectionTarget
	^submorphs at: 1 ifAbsent: [ self ]! !

!TransformationMorph methodsFor: '*connectors-private' stamp: 'nk 6/13/2002 09:34'!
isConnectable
	"Answer whether I will accept connections from ConnectorMorphs"
	^false! !

!TransformationMorph methodsFor: '*connectors-private' stamp: 'nk 6/17/2002 20:56'!
lineAttachmentOffset
	^(submorphs at: 1 ifAbsent: [ ^0@0 ]) lineAttachmentOffset! !

!Vocabulary class methodsFor: '*connectors' stamp: 'nk 8/21/2004 15:40'!
addArrowNamesVocabulary
	| vocab |
	self addStandardVocabulary: (vocab := ConnectorArrowNameType new).
	SymbolListTile updateAllTilesForVocabularyNamed: #ArrowNames.
	^vocab! !

!Vocabulary class methodsFor: '*connectors' stamp: 'nk 8/21/2004 19:57'!
arrowNamesVocabulary
	"Vocabulary arrowNamesVocabulary"
	^(self vocabularyNamed: #ArrowNames)
			ifNil: [ self addArrowNamesVocabulary ]
! !

!MorphExtension methodsFor: '*connectors-copying-override' stamp: 'nk 5/1/2004 17:20'!
copyWeakly
	"list of names of properties whose values should be weak-copied when veryDeepCopying a morph.  See DeepCopier."

	^ #(formerOwner newPermanentPlayer logger graphModel gestureDictionaryOrName)
	"add yours to this list" 

	"formerOwner should really be nil at the time of the copy, but this will work just fine."! !

!MorphExtension methodsFor: '*connectors-copying-override' stamp: 'nk 5/1/2004 17:23'!
propertyNamesNotCopied
	"list of names of properties whose values should be deleted when veryDeepCopying a morph.
	See DeepCopier."

	^ #(connectedConstraints connectionHighlights highlightedTargets)
	"add yours to this list" 
! !

!MorphExtension methodsFor: '*connectors-copying-override' stamp: 'nk 5/1/2004 17:39'!
veryDeepFixupWith: deepCopier 
	"If target and arguments fields were weakly copied, fix them here.
	If they were in the tree being copied, fix them up, otherwise point to the originals!!!!"

	super veryDeepFixupWith: deepCopier.
	otherProperties ifNil: [ ^self ].

	"Properties whose values are only copied weakly replace those values if they were copied via another path"
	self copyWeakly do: [ :propertyName |
		otherProperties at: propertyName ifPresent: [ :property |
			otherProperties at: propertyName
				put: (deepCopier references at: property ifAbsent: [ property ])]].
! !

!Rectangle methodsFor: '*connectors-geometry' stamp: 'nk 6/21/2003 13:14'!
extrapolatedIntersectionsWithLineFrom: start to: end
	| intersections side ex ey delta dx dy |
	intersections := self intersectionsWithLineFrom: start to: end.
	intersections size = 1 ifFalse: [ ^intersections ].
	side := self sideNearestTo: intersections anyOne.
	delta := end - start.
	side == #left ifTrue: [ ex := self right. ].
	side == #right ifTrue: [ ex := self left ].
	side == #top ifTrue: [ ey := self bottom ].
	side == #bottom ifTrue: [ ey := self top ].

	ex ifNil: [
		delta y isZero ifTrue: [ ex := start x ] ifFalse: [ 
			dx := (ey - start y) * delta x / delta y.
			ex := start x + dx.
		]
	].
	ey ifNil: [
		delta x isZero ifTrue: [ ey := start y ] ifFalse: [ 
			dy := (ex - start x) * delta y / delta x.
			ey := start y + dy
		].
	].
	intersections add: ex@ey.
	^intersections! !

!Rectangle methodsFor: '*connectors-geometry' stamp: 'nk 5/14/2003 23:10'!
intersectionWithLineSegmentFromCenterTo: aPoint 
	"Answer the point, if any, along my border that intersects the  
	line segment between my center and aPoint. Return aPoint if  
	aPoint is inside me or is on my border, else a computed point  
	somewhere along my border."

	| seg |
	(self containsPoint: aPoint)
		ifTrue: [^ aPoint].
	"Answer the set of intersections, if any"
	seg := LineSegment from: self center to: aPoint.
	self lineSegmentsDo: [ :p1 :p2 | | side int |
		side := LineSegment from: p1 to: p2.
		int := side intersectionWith: seg.
		int ifNotNil: [ ^int ].
	].

	^(self center + aPoint) // 2
! !

!Rectangle methodsFor: '*connectors-geometry' stamp: 'nk 6/21/2003 13:13'!
intersectionsWithLineFrom: start to: end
	"Answer the set of intersections, if any"
	| retval l i |
	retval := Set new: 4.
	l := LineSegment from: start to: end.
	self lineSegmentsDo: [ :p1 :p2 | | s |
		s := LineSegment from: p1 to: p2.
		i := s intersectionWith: l.
		i ifNotNil: [ retval add: i ].
	].
	^retval! !

!Rectangle methodsFor: '*connectors-testing' stamp: 'nk 2/27/2001 11:12'!
intersectsLineFrom: start to: finish 
	^ (self topLeft
			to: self topRight
			intersects: start
			to: finish)
		or: [(self topRight
					to: self bottomRight
					intersects: start
					to: finish)
				or: [(self bottomRight
							to: self bottomLeft
							intersects: start
							to: finish)
						or: [self bottomLeft
								to: self topLeft
								intersects: start
								to: finish]]]! !

!Rectangle methodsFor: '*connectors-rectangle functions' stamp: 'nk 3/29/2002 22:34'!
lineSegmentsDo: aBlock
	"Evaluate aBlock with the receiver's line segments"
	| p |
	aBlock value: origin value: (p := self topRight).
	aBlock value: p value: corner.
	aBlock value: corner value: (p := self bottomLeft).
	aBlock value: p value: origin.! !

!Rectangle methodsFor: '*connectors-transforming' stamp: 'nk 3/27/2001 18:19'!
scaleAndTranslateToFitInside: anotherRectangle
	"Answer a Rectangle with the same aspect ratio as me, but that will fit inside
	anotherRectangle."
	| newExtent |
	newExtent := self extent scaleToNoGreaterThan: anotherRectangle extent.
	^(Rectangle origin: self origin extent: newExtent) translatedToBeWithin: anotherRectangle.
! !

QHsmMorph subclass: #NCAAConnectorMorph
	instanceVariableNames: 'constraints line graphModel'
	classVariableNames: 'ArrowGlyphs'
	poolDictionaries: ''
	category: 'Connectors-Base'!

!NCAAConnectorMorph commentStamp: '<historical>' prior: 0!
This is the Anti-Aliased ConnectorMorph, part of Connectors2.!

!NCAAConnectorMorph class methodsFor: 'scripting' stamp: 'nk 9/4/2004 11:24'!
additionsToViewerCategoryColorAndBorder
	"Answer viewer additions for the 'color & border' category"

	^#(
		#'color & border' 
		(
			(slot lineWidth 'The width of the main part of the Connector' Number readWrite Player getLineWidth Player setLineWidth:)
		)
	)
! !

!NCAAConnectorMorph class methodsFor: 'scripting' stamp: 'nk 9/10/2004 11:35'!
additionsToViewerCategoryConnection
	"Answer viewer additions for the 'connection' category"
	"Vocabulary initialize"

	^#(
		connector
		(
			(slot source 'The Player at my first end' Player readWrite Player getSourcePlayer Player setSourcePlayer:)
			(slot destination 'The Player at my second end' Player readWrite Player getDestinationPlayer Player setDestinationPlayer:)
			(slot orthogonal 'Whether I am limited to horizontal and vertical segments' Boolean readWrite Player getOrthogonal Player setOrthogonal:)
			(slot smoothCurve 'Whether I am drawn as a smooth curve' Boolean readWrite Player getSmoothCurve Player setSmoothCurve:)
			(slot sourceArrowName 'The name of the arrow at my first end' ArrowNames readWrite Player getSourceArrowName  Player setSourceArrowName:)
			(slot destinationArrowName 'The name of the arrow at my second end' ArrowNames readWrite Player getDestinationArrowName  Player setDestinationArrowName:)
			(command straighten 'remove all my bends' unused)
			(command addLabel: 'add a label at my middle' String)
			(command removeLabels 'remove all my labels' unused)
		)

	)
! !

!NCAAConnectorMorph class methodsFor: 'scripting' stamp: 'nk 8/25/2003 16:16'!
additionsToViewerCategoryGeometry
	"Answer viewer additions for the 'geometry' category"
	"Vocabulary initialize"

	^#(

		geometry 
		(
			(slot totalLength  'My total length' Number readOnly Player getTotalLength unused unused)
			(slot midpointX  'The X coordinate of my midpoint' Number readOnly Player getMidpointX unused unused)
			(slot midpointY  'The Y coordinate of my midpoint' Number readOnly Player getMidpointY unused unused)
			(slot sourceX  'The X coordinate of my first end' Number readOnly Player getSourceX unused unused)
			(slot sourceY  'The Y coordinate of my first end' Number readOnly Player getSourceY unused unused)
			(slot destinationX  'The X coordinate of my second end' Number readOnly Player getDestinationX unused unused)
			(slot destinationY  'The Y coordinate of my second end' Number readOnly Player getDestinationY unused unused)
		)
	

	)
! !

!NCAAConnectorMorph class methodsFor: 'arrows' stamp: 'nk 8/23/2003 18:08'!
arrowDictionary
	^ArrowGlyphs ifNil: [
		ArrowGlyphs := Dictionary new.
		self initializeArrowDictionary ].! !

!NCAAConnectorMorph class methodsFor: 'arrow forms' stamp: 'nk 8/24/2003 17:17'!
basicArrow
	"glyph 54 from Arrows1"
^(NCLineArrowGlyph withContours: (
(Array new: 1)
	at: 1 put: { 
	0.0@0.0 . 0.0@0.0 . -0.5@0.3298 . 
	-0.5@0.3298 . -0.5@0.3298 . 0.0@-0.666 . 
	0.0@-0.666 . 0.0@-0.666 . 0.5@0.3298 . 
	0.5@0.3298 . 0.5@0.3298 . 0.0@0.0 };
	yourself
"clicks 0.0@0.0"
)) offset: 0@0.666;
defaultScale: 3.0;
yourself! !

!NCAAConnectorMorph class methodsFor: 'new-morph participation' stamp: 'nk 9/2/2004 08:17'!
buttonFlapImage
	"NCAAConnectorMorph buttonFlapImage openAsMorph"
	"(Flaps newNCPartsBinFlap imageFormDepth: 8) storeString"

	^(Form
	extent: 19@115
	depth: 8
	fromArray: #( 3755991007 3755991007 3755991007 3755933696 0 3756384229 3756384229 3756384229 3756384229 0 3755991007 3755991007 3755991007 3755991007 3741319168 3756384223 3755991007 3755991007 3755991013 3756326912 3755991007 3755991007 3755991007 3755991007 3755933696 3756384223 3755991007 3755991007 3755991007 3756384000 3755991007 3755991007 3755991007 3755991007 3755981568 3756384223 3755991007 3755991007 3755991007 3753631488 3755991007 3755991007 3755991007 3755991007 3755981568 3756384223 31449055 3755991007 16900063 3753631488 3755991007 16900063 3755990785 16900063 3755981568 3756384223 16843231 3755933953 16900063 3753631488 3755991007 16843009 3741384961 16900063 3755981568 3756384223 31392001 16843231 16900063 3753631488 3755991007 31448833 16900063 16900063 3755981568 3756384223 31449055 31449055 16900063 3753631488 3755991007 31449055 3755991007 16900063 3755981568 3756384223 31449055 3755991007 16900063 3753631488 3755991007 3755991007 3755991007 3755991007 3755981568 3756384223 3755991007 3755991007 3755991007 3753631488 3755991007 3755991007 3755991007 3755991007 3755981568 3756384223 3755991007 3755991007 3755991007 3753631488 3755991007 3755991007 3755991007 3755991007 3755981568 3756384223 3755991007 3755991007 3755991007 3753631488 3755991007 3755991007 3755991007 3755991007 3755981568 3756384223 3755991007 3755991007 3755991007 3753631488 3755991007 3755933953 3755933953 3755991007 3755981568 3756384223 3755933953 3755933953 3755991007 3753631488 3755991007 3755933953 3755933953 3755991007 3755981568 3756384223 3755933953 3755933953 3755991007 3753631488 3755991007 3755933953 3755933953 3755991007 3755981568 3756384223 3755933953 3755933953 3755991007 3753631488 3755991007 3755990785 16843009 3755991007 3755981568 3756384223 3755991007 3755933953 3755991007 3753631488 3755991007 3755934175 3755933953 3755991007 3755981568 3756384223 3755990785 16843231 3755991007 3753631488 3755991007 3755991007 3755991007 3755991007 3755981568 3756384223 3755991007 3755991007 3755991007 3753631488 3755991007 3755991007 3755991007 3755991007 3755981568 3756384223 3755991007 3755991007 3755991007 3753631488 3755991007 3755991007 3755991007 3755991007 3755981568 3756384223 3755991007 3755991007 3755991007 3753631488 3755991007 3755991007 3755991007 3755991007 3755981568 3756384223 3755991007 3755991007 3755991007 3753631488 3755991007 3755991007 3755991007 3755991007 3755981568 3756384223 3755991007 3755991007 3755991007 3753631488 3755991007 3755991007 3755991007 3755991007 3755981568 3756384223 3755991007 3755991007 3755991007 3753631488 3755991007 3755991007 3755991007 3755991007 3755981568 3756384223 3755991007 3755991007 3755991007 3753631488 3755991007 3755991007 3755991007 3755991007 3755981568 3756384223 3755991007 3755991007 3755991007 3753631488 3755991007 3755991007 3755991007 3755991007 3755981568 3756384223 3755991007 3755991007 3755991007 3753631488 3755991007 3755933953 16843231 3755991007 3755981568 3756384223 3755933953 3755991007 3755991007 3753631488 3755991007 3755933953 3755991007 3755991007 3755981568 3756384223 3755933953 3755991007 3755991007 3753631488 3755991007 3755933953 16900063 3755991007 3755981568 3756384223 3755933953 3755991007 3755991007 3753631488 3755991007 3755933953 3755991007 3755991007 3755981568 3756384223 3755933953 3755991007 3755991007 3753631488 3755991007 3755933953 3755991007 3755991007 3755981568 3756384223 3755991007 3755991007 3755991007 3753631488 3755991007 3755991007 3755991007 3755991007 3755981568 3756384223 3755991007 3755991007 3755991007 3753631488 3755991007 3755991007 3755991007 3755991007 3755981568 3756384223 3755991007 3755991007 3755991007 3753631488 3755991007 3755991007 3755991007 3755991007 3755981568 3756384223 3755991007 16900063 3755991007 3753631488 3755991007 3755991007 16900063 3755991007 3755981568 3756384223 3755991007 16900063 3755991007 3753631488 3755991007 3755991007 16900063 3755991007 3755981568 3756384223 3755991007 16900063 3755991007 3753631488 3755991007 3755991007 16900063 3755991007 3755981568 3756384223 3755991007 16900063 3755991007 3753631488 3755991007 3755991007 16900063 3755991007 3755981568 3756384223 3755991007 16900063 3755991007 3753631488 3755991007 3755991007 3755991007 3755991007 3755981568 3756384223 3755991007 3755991007 3755991007 3753631488 3755991007 3755991007 3755991007 3755991007 3755981568 3756384223 3755991007 3755991007 3755991007 3753631488 3755991007 3755991007 3755991007 3755991007 3755981568 3756384223 3755991007 3755991007 3755991007 3753631488 3755991007 3755991007 3755991007 3755991007 3755981568 3756384223 3755991007 3755991007 3755991007 3753631488 3755991007 3755990785 16843231 3755991007 3755981568 3756384223 3755934175 3755933953 3755991007 3753631488 3755991007 3755990785 16843009 3755991007 3755981568 3756384223 3755933953 3755933953 3755991007 3753631488 3755991007 3755933953 3755933953 3755991007 3755981568 3756384223 3755933953 3755933953 3755991007 3753631488 3755991007 3755990785 16843009 3755991007 3755981568 3756384223 3755991007 3755991007 3755991007 3753631488 3755991007 3755991007 3755991007 3755991007 3755981568 3756384223 3755991007 3755991007 3755991007 3753631488 3755991007 3755991007 3755991007 3755991007 3755981568 3756384223 3755991007 3755991007 3755991007 3753631488 3755991007 3755991007 3755991007 3755991007 3755981568 3756384223 3755991007 3755991007 3755991007 3753631488 3755991007 3755991007 3755991007 3755991007 3755981568 3756384223 3755933953 16843231 3755991007 3753631488 3755991007 3755933953 3755933953 3755991007 3755981568 3756384223 3755933953 3755933953 3755991007 3753631488 3755991007 3755933953 3755933953 3755991007 3755981568 3756384223 3755933953 3755933953 3755991007 3753631488 3755991007 3755933953 3755933953 3755991007 3755981568 3756384223 3755933953 16843231 3755991007 3753631488 3755991007 3755933953 3755991007 3755991007 3755981568 3756384223 3755933953 3755991007 3755991007 3753631488 3755991007 3755991007 3755991007 3755991007 3151953920 3756384223 3755991007 3755991007 3755990971 3753574400 3755991007 3755991007 3755991007 3755981791 3137339392 3756384187 3753631675 3753631675 3753631675 0 3755981791 3152002015 3152002015 3151953920 0)
	offset: 2533@581)! !

!NCAAConnectorMorph class methodsFor: 'arrows' stamp: 'nk 8/24/2003 14:56'!
chooseArrowGlyphFor: client selector: selector argument: arg
	"Bring up an arrow catalog.
	When the OK button is pressed, perform the selector on the client with the optional argument."
	| menu |
	menu := NCArrowDictionaryMenu chooseArrowGlyphFor: client selector: #selector.
	menu argument: arg
	! !

!NCAAConnectorMorph class methodsFor: 'arrows' stamp: 'nk 6/30/2004 16:03'!
clearArrowDictionary
	ArrowGlyphs ifNil: [ ^ArrowGlyphs := Dictionary new ].
	ArrowGlyphs keysAndValuesRemove: [ :k :v | true ];
		changed: #arrowsChanged.
! !

!NCAAConnectorMorph class methodsFor: 'new-morph participation' stamp: 'nk 9/2/2004 08:15'!
connectorsFlapImage
	"NCAAConnectorMorph connectorsFlapImage openAsMorph"
	"(Flaps newConnectorsFlap imageFormDepth: 8) storeString"

^(Form
	extent: 116@20
	depth: 8
	fromArray: #( 0 14342874 3671775962 3671775962 3671775962 3671775962 3671775962 3671775962 3671775962 3671775962 3671775962 3671775962 3671775962 3671775962 3671775962 3671775962 3671775962 3671775962 3671775962 3671775962 3671775962 3671775962 3671775962 3671775962 3671775962 3671775962 3671775962 3671775744 0 218 3671775962 3671775962 3671775962 3671775962 3671775962 3671775962 3671775962 3671775962 3671775962 3671775962 3671775962 3671775962 3671775962 3671775962 3671775962 3671775962 3671775962 3671775962 3671775962 3671775962 3671775962 3671775962 3671775962 3671775962 3671775962 3671775962 3671775962 3657433088 56026 3671775962 3671775962 3671775962 3671775962 3671775962 3671775962 3671775962 3671775962 3671775962 3671775962 3671775962 3671775962 3671775962 3671775962 3671775962 3671775962 3671775962 3671775962 3671775962 3671775962 3671775962 3671775962 3671775962 3671775962 3671775962 3671775962 3671775962 3671719936 14342874 3671775962 3671775962 3671775962 3671775962 3671775962 3671775962 3671775962 3671775962 3671775962 3671775962 3671775962 3671775962 3671775962 3671775962 3671775962 3671775962 3671775962 3671775962 3671775962 3671775962 3671775962 3671775962 3671775962 3671775962 3671775962 3671775962 3671775962 3671775744 14342874 3671775962 3671775962 3671775962 3671775962 3671775962 3671775962 3671775962 3671775962 3671775962 3671775962 3671775962 3671775962 3671775962 3671775962 3671775962 3671775962 3671775962 3671775962 3671775962 3671775962 3671775962 3671775962 3671775962 3671775962 3671775962 3671775962 3671775962 3671775744 3671775962 3671775962 3671775962 3671775962 3671775962 3671775962 3671775962 3671775962 3671775962 3671775962 3671775962 3671775962 3671775962 3671775962 3671775962 3671775962 3671775962 3671775962 3671775962 3671775962 3671775962 3671775962 3671775962 3671775962 3671775962 3671775962 3671775962 3671775962 3671775962 3671775962 3671775962 3657498881 31120090 3671775962 3671775962 3671775962 3671775962 3671775962 3671775962 3671775962 3671775962 3671775962 3671775962 3671775962 3671775962 3671775962 3671775962 3671720193 3671775962 3671775962 3671775962 3671775962 3671775962 3671775962 3671775962 3671775962 3671775962 3671764954 3671775962 3671775962 16898778 3657554650 3671775962 3671775962 3671775962 3671775962 3671775962 3671775962 3671775962 3671775962 3671775962 3671775962 3671775962 3671775962 3671775962 3671775962 3671720193 3671775962 3671775962 3671775962 3671775962 3671775962 3671775962 3671775962 3671775962 3671775962 3671775919 3671775962 3671775962 16898778 3671775962 3671775962 16843009 3671775962 3671720193 16843226 3671775962 3657498881 16898778 3671775962 3657498881 31120090 3671775962 16843226 3671775962 3657498881 31120090 3671775745 16843226 3671775962 3657499098 16898778 3671775745 16898778 3671775962 3671764954 3671775962 3671775962 16898778 3671775962 3671775745 31119873 31120090 3671720193 3671720193 3671775962 3657499098 3657499098 3671775962 16898778 16898778 3671775745 31119873 3671775962 3671720193 3671775962 3671720193 3671720193 3671775962 3657498881 16898778 3671720193 3671720410 3671775962 3671775919 3671775962 3671775962 16898778 3671775962 3671775745 31119873 31120090 3671720193 3671720193 3671775962 3657499098 3657499098 3671775962 16898778 16898778 3671775745 31120090 3671775962 3671720193 3671775962 3671720193 3671720193 3671775962 3657499098 3671775962 3671720193 31120090 3671775962 3671764954 3671775962 3671775962 16898778 3671775962 3671775745 31119873 31120090 3671720193 3671720193 3671775962 3657499098 3657499098 3671775962 16843009 16898778 3671775745 31120090 3671775962 3671720193 3671775962 3671720193 3671720193 3671775962 3657499098 3671775962 3671775745 16898778 3671775962 3671775919 3671775962 3671775962 16898778 3671775962 3671775745 31119873 31120090 3671720193 3671720193 3671775962 3657499098 3657499098 3671775962 16898778 3671775962 3671775745 31120090 3671775962 3671720193 3671775962 3671720193 3671720193 3671775962 3657499098 3671775962 3671775962 16843226 3671775962 3671764954 3671775962 3671775962 16898778 3657554650 3671775745 31119873 31120090 3671720193 3671720193 3671775962 3657499098 3657499098 3671775962 16898778 3657554650 3671775745 31119873 3671775962 3671720193 3671775962 3671720193 3671720193 3671775962 3657499098 3671775962 3671720410 3657499098 3671775962 3671775919 3671775962 3671775962 3657498881 31120090 3671775962 16843009 3671775962 3671720193 3671720193 3671775962 3657499098 3657499098 3671775962 3657498881 31120090 3671775962 16843226 3671775962 3671775745 31120090 3671775745 16843226 3671775962 3657499098 3671775962 3671775745 16898778 3671775962 3671764954 3671775962 3671775962 3671775962 3671775962 3671775962 3671775962 3671775962 3671775962 3671775962 3671775962 3671775962 3671775962 3671775962 3671775962 3671775962 3671775962 3671775962 3671775962 3671775962 3671775962 3671775962 3671775962 3671775962 3671775962 3671775962 3671775962 3671775962 3671775962 3671775919 3671775962 3671775962 3671775962 3671775962 3671775962 3671775962 3671775962 3671775962 3671775962 3671775962 3671775962 3671775962 3671775962 3671775962 3671775962 3671775962 3671775962 3671775962 3671775962 3671775962 3671775962 3671775962 3671775962 3671775962 3671775962 3671775962 3671775962 3671775962 3671764954 3671775962 3671775962 3671775962 3671775962 3671775962 3671775962 3671775962 3671775962 3671775962 3671775962 3671775962 3671775962 3671775962 3671775962 3671775962 3671775962 3671775962 3671775962 3671775962 3671775962 3671775962 3671775962 3671775962 3671775962 3671775962 3671775962 3671775962 3671775962 3671775919 3671764954 2950344666 2950344666 2950344666 2950344666 2950344666 2950344666 2950344666 2950344666 2950344666 2950344666 2950344666 2950344666 2950344666 2950344666 2950344666 2950344666 2950344666 2950344666 2950344666 2950344666 2950344666 2950344666 2950344666 2950344666 2950344666 2950344666 2950344666 2950344666 3668957871 3668957871 3668957871 3668957871 3668957871 3668957871 3668957871 3668957871 3668957871 3668957871 3668957871 3668957871 3668957871 3668957871 3668957871 3668957871 3668957871 3668957871 3668957871 3668957871 3668957871 3668957871 3668957871 3668957871 3668957871 3668957871 3668957871 3668957871 3668957871)
	offset: 0@0)! !

!NCAAConnectorMorph class methodsFor: 'defaults' stamp: 'nk 7/11/2003 14:01'!
defaultLineClass
	^NCLineMorph! !

!NCAAConnectorMorph class methodsFor: 'defaults' stamp: 'nk 8/15/2003 09:47'!
defaultLineEndConstraintClass
	^NCLineEndConstraintMorph! !

!NCAAConnectorMorph class methodsFor: 'defaults' stamp: 'nk 8/15/2003 09:27'!
eventHandlerClass
	^NCAAEventHandler ! !

!NCAAConnectorMorph class methodsFor: 'instance creation' stamp: 'nk 8/24/2003 14:49'!
fromMorph: startMorph toMorph: endMorph

	^self new fromMorph: startMorph toMorph: endMorph
! !

!NCAAConnectorMorph class methodsFor: 'instance creation' stamp: 'nk 8/24/2003 14:49'!
fromMorph: startMorph toMorph: endMorph color: lineColor width: lineWidth

	^self new fromMorph: startMorph toMorph: endMorph color: lineColor width: lineWidth
! !

!NCAAConnectorMorph class methodsFor: 'new-morph participation' stamp: 'nk 8/24/2003 14:49'!
includeInNewMorphMenu
	^true! !

!NCAAConnectorMorph class methodsFor: 'class initialization' stamp: 'nk 11/1/2004 08:29'!
initialize
	"NCAAConnectorMorph initialize"

	Preferences addPreference: #useSmartLabels
		category:  'connectors'
		default:  false
		balloonHelp: 'if set, Connectors will use labels that try to re-position themselves to stay out of the way.'.

	Preferences addPreference: #alwaysShowConnectionVocabulary
		category:  'connectors'
		default:  false
		balloonHelp: 'if set, viewers will display the ''connection'' vocabulary even for Morphs that aren''t connected to by Connectors.'.

	ScriptingSystem addCustomEventFor: self named: #connectedTo help: 'immediately after a connector has connected to me' targetMorphClass: Morph.
	ScriptingSystem addCustomEventFor: self named: #disconnectedFrom help: 'immediately before a connector is going to disconnect from me' targetMorphClass: Morph.
	ScriptingSystem addCustomEventFor: self named: #sourceConnected help: 'immediately after my first end has been connected' targetMorphClass: self.
	ScriptingSystem addCustomEventFor: self named: #sourceDisconnecting help: 'immediately before my first end is going to be disconnected' targetMorphClass: self.
	ScriptingSystem addCustomEventFor: self named: #destinationConnected help: 'immediately after my second end has been connected' targetMorphClass: self.
	ScriptingSystem addCustomEventFor: self named: #destinationDisconnecting help: 'immediately before my second end is going to be disconnected' targetMorphClass: self.
	ScriptingSystem addCustomEventFor: self named: #attachmentOwnerChanged help: 'when either of the morphs to which I''m connected have changed their owner or been grabbed or dropped with the hand.' targetMorphClass: self.
	Smalltalk addToStartUpList: self.
	self initializeArrowDictionary.! !

!NCAAConnectorMorph class methodsFor: 'arrows' stamp: 'nk 6/30/2004 15:59'!
initializeArrowDictionary
	"NCAAConnectorMorph initializeArrowDictionary"

	ArrowGlyphs ifNil: [ ArrowGlyphs := Dictionary new ].

	(self class organization listAtCategoryNamed: 'arrow forms') do: [ :sym |
		ArrowGlyphs at: sym put: (self perform: sym) ].

	(NCLineArrowGlyph class organization  listAtCategoryNamed: 'arrow forms') do: [ :sym |
		ArrowGlyphs at: sym put: (NCLineArrowGlyph perform: sym) ].
	
	ArrowGlyphs changed: #arrowsChanged.

	^ArrowGlyphs
! !

!NCAAConnectorMorph class methodsFor: 'arrows' stamp: 'nk 8/26/2003 14:41'!
nameForArrowGlyph: anArrow
	self arrowDictionary keysAndValuesDo: [ :key :value |
		anArrow originalFirstContour = value originalFirstContour
			ifTrue: [ ^key ].
	].
	^#unknown! !

!NCAAConnectorMorph class methodsFor: 'instance creation-UML' stamp: 'nk 8/24/2003 14:53'!
newAssociation
	"Answer a new one of me ready for use in a class diagram"
	"NCAAConnectorMorph newAssociation openInHand"
	^self new
		name: 'Association'! !

!NCAAConnectorMorph class methodsFor: 'instance creation' stamp: 'nk 7/17/2004 16:57'!
newCurvyArrow
	"NCAAConnectorMorph newCurvyArrow openInHand"
	^ self new
		beSmoothCurve;
		name: 'Curvy ConnectorArrow';
		arrowNames: #(#noArrow #basicArrow )! !

!NCAAConnectorMorph class methodsFor: 'instance creation' stamp: 'nk 5/13/2004 20:07'!
newCurvyConnector
	"NCAAConnectorMorph newCurvyConnector openInHand"
	^ self new
		beSmoothCurve;
		name: 'Curvy Connector';
		arrowNames: #(#noArrow #noArrow )! !

!NCAAConnectorMorph class methodsFor: 'instance creation-UML' stamp: 'nk 8/25/2003 13:47'!
newDirectionalAssociation
	"Answer a new one of me with an arrow at the 'to' end ready for use in a class diagram"

	"NCAAConnectorMorph newDirectionalAssociation openInHand"

	^(self new)
		name: 'Association';
		arrowNames: #(noArrow umlArrow)! !

!NCAAConnectorMorph class methodsFor: 'instance creation-UML' stamp: 'nk 8/25/2003 13:29'!
newGeneralizationRelationship
	"Answer a new one of me ready for use in a class diagram"

	"NCAAConnectorMorph newGeneralizationRelationship openInHand"

	^(self new) 
		arrowNames: #(noArrow openArrow);
		name: 'Generalization'! !

!NCAAConnectorMorph class methodsFor: 'instance creation-UML' stamp: 'nk 8/24/2003 14:52'!
newImplementationRelationship
	"Answer a new one of me ready for use in a class diagram"
	"NCAAConnectorMorph newImplementationRelationship openInHand"
	^self newGeneralizationRelationship
		dashedLine: true;
		name: 'Implements'! !

!NCAAConnectorMorph class methodsFor: 'instance creation' stamp: 'nk 7/17/2004 17:38'!
newMakerButton

	^Smalltalk at: #NCMakerButton ifPresent: [ :ncmb |
		^(ncmb fromPrototype: self newWithArrow)
			actionSelector: #startWiring;
			cornerStyle: #rounded;
			color: Color blue muchLighter;
			changePixelsOfColor: Color red toColor: Color transparent;
			yourself ].! !

!NCAAConnectorMorph class methodsFor: 'instance creation-UML' stamp: 'nk 8/24/2003 14:52'!
newNoteConnector
	"Answer a new one of me ready for use in a class diagram"
	"NCAAConnectorMorph newNoteConnector openInHand"
	^self new
		dashedLine: true;
		lineWidth: 1;
		name: 'Note Connector';
		connectToNearestPointToCenter.! !

!NCAAConnectorMorph class methodsFor: 'instance creation' stamp: 'nk 9/2/2004 08:42'!
newRandomButton
	"NCAAConnectorMorph newRandomButton openInHand"
	
	^Smalltalk at: #NCMakerButton ifPresent: [ :ncmb |
		^(ncmb fromPrototype: (self newWithArrow nameForButton: 'Random' translated; yourself))
			beLabeled: true;
			actionSelector: #startWiringRandomly;
			cornerStyle: #rounded;
			color: Color blue muchLighter;
			changePixelsOfColor: Color red toColor: Color transparent;
			yourself ].! !

!NCAAConnectorMorph class methodsFor: 'instance creation-UML' stamp: 'nk 8/25/2003 13:31'!
newStateTransition
	"Answer a new one of me ready for use in a state diagram"
	"NCAAConnectorMorph newStateTransition openInHand"
	^self new
		arrowNames: #(noArrow basicArrow);
		name: 'Transition'! !

!NCAAConnectorMorph class methodsFor: 'instance creation' stamp: 'nk 7/17/2004 16:56'!
newWithArrow
	"NCAAConnectorMorph newWithArrow openInHand"
	^ self new
		name: 'Arrow';
		arrowNames: #(#noArrow #basicArrow )! !

!NCAAConnectorMorph class methodsFor: 'arrow forms' stamp: 'nk 8/25/2003 13:23'!
noArrow
	^NCLineArrowGlyph noArrow! !

!NCAAConnectorMorph class methodsFor: 'arrow forms' stamp: 'nk 8/25/2003 13:18'!
openArrow
	| a |
	a := { 0.0@0.0 . 0.0@0.0 . -0.5@0.0 . 
	-0.5@0.0 . -0.5@0.0 . 0.0@-1.0 . 
	0.0@-1.0 . 0.0@-1.0 . 0.5@0.0 . 
	0.5@0.0 . 0.5@0.0 . 0.0@0.0 }.
^(NCLineArrowGlyph withContours: (
(Array new: 2)
	at: 1 put: a; 
	at: 2 put:  (a + (0.0@0.33) * 0.6 - (0.0@0.33));
	yourself
"clicks 0.0@0.0"
)) offset: 0@1.0;
defaultScale: 2.5;
yourself! !

!NCAAConnectorMorph class methodsFor: 'new-morph participation' stamp: 'nk 7/17/2004 17:18'!
questionMarkForm
	 ^(Form
	extent: 17@27
	depth: 16
	fromArray: #( 0 0 1246898506 484907206 484913548 1731788800 0 0 0 0 1177621669 65537 65537 65537 80334 0 0 0 10570 69271553 69271553 69271553 69271553 69271553 762014554 0 0 1593245697 65537 67650 415636711 346357761 65537 71878 1801060352 0 1454702593 69271553 415656794 0 22197 138543105 69271553 900530176 0 14798 484925175 0 0 0 1662522533 65537 87733 0 0 0 0 0 0 25368 69271553 69281065 0 0 0 0 0 0 0 761987073 66593 0 0 0 0 0 0 0 1108344833 69271553 1801060352 0 0 0 0 0 0 969801729 66593 0 0 0 0 0 0 27482 277086209 69275780 0 0 0 0 0 0 1801065637 65537 79277 0 0 0 0 0 0 623443969 69271553 138570586 0 0 0 0 0 1662522533 65537 67650 1454702592 0 0 0 0 20083 138543105 69271553 138567415 0 0 0 0 27482 692715521 65537 70821 1662517248 0 0 0 0 5285 69271553 69272609 761987072 0 0 0 0 0 1108344833 65537 69291635 0 0 0 0 0 0 1039073281 69273666 1593245696 0 0 0 0 0 0 1801070922 554194613 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 25368 207814657 900530176 0 0 0 0 0 0 9513 65537 346357760 0 0 0 0 0 0 7399 69271553 554172416 0 0 0 0 0 0 13741 65537 761987072 0 0 0 0 0 0 0 1731810997 0 0 0 0 0 0)
	offset: 0@0)! !

!NCAAConnectorMorph class methodsFor: 'instance creation' stamp: 'nk 8/28/2003 15:12'!
random
	"NCAAConnectorMorph random startWiring"
	"((SimpleButtonMorph newWithLabel: 'Random Connector')
		target: NCAAConnectorMorph;
		actionSelector: #wireRandom) openInHand"
	^(self new)
		color: Color random;
		lineWidth: (1 to: 20) atRandom;
		borderWidth: (1 to: 4) atRandom;
		arrowNames: {
			self arrowDictionary keys atRandom.
			self arrowDictionary keys atRandom };
		yourself! !

!NCAAConnectorMorph class methodsFor: 'class initialization' stamp: 'nk 1/14/2004 18:24'!
startUp: resuming
	"My instances no longer step the #stepCleanup method"
	resuming ifTrue: [ self allSubInstancesDo: [ :i | i stopSteppingSelector: #stepCleanup ]]! !

!NCAAConnectorMorph class methodsFor: 'instance creation' stamp: 'nk 8/24/2003 14:49'!
startWiringVia: selector label: aString
	| newMe |
	newMe := self perform: selector.
	newMe name: aString.
	newMe startWiring.! !

!NCAAConnectorMorph class methodsFor: 'new-morph participation' stamp: 'nk 9/2/2004 08:37'!
supplementaryPartsDescriptions
	| descriptions |
	"PartsBin clearThumbnailCache; cacheAllThumbnails"
	descriptions := OrderedCollection new.
	descriptions addAll: {
	DescriptionForPartsBin
		formalName: 'Connector'
		categoryList: #('Connectors' )
		documentation: 'A basic Connector'
		globalReceiverSymbol: #NCAAConnectorMorph
		nativitySelector: #newStandAlone.

	DescriptionForPartsBin
		formalName: 'ConnectorArrow'
		categoryList: #('Connectors' )
		documentation: 'A basic Connector with an arrowhead'
		globalReceiverSymbol: #NCAAConnectorMorph
		nativitySelector: #newWithArrow.
		
	DescriptionForPartsBin
		formalName: 'Curvy Connector'
		categoryList: #('Connectors' )
		documentation: 'A basic Connector that bends smoothly'
		globalReceiverSymbol: #NCAAConnectorMorph
		nativitySelector: #newCurvyConnector.

	DescriptionForPartsBin
		formalName: 'Curvy ConnectorArrow'
		categoryList: #('Connectors' )
		documentation: 'A basic smooth Connector with an arrowhead'
		globalReceiverSymbol: #NCAAConnectorMorph
		nativitySelector: #newCurvyArrow.
	}.

	(Flaps respondsTo: #newNCPartsBinFlap) ifTrue: [
		descriptions add: (
	(DescriptionForPartsBin
		formalName: 'Button Flap'
		categoryList: #('Connectors' )
		documentation: 'A blank buttton flap'
		globalReceiverSymbol: #Flaps
		nativitySelector: #newNCPartsBinFlap)
			sampleImageForm: self buttonFlapImage; yourself) ].

	(Flaps respondsTo: #newConnectorsFlap) ifTrue: [
		descriptions add: (
	(DescriptionForPartsBin
		formalName: 'Connectors Flap'
		categoryList: #('Connectors' )
		documentation: 'A pre-loaded Connectors flap'
		globalReceiverSymbol: #Flaps
		nativitySelector: #newConnectorsFlap)
			sampleImageForm: self connectorsFlapImage; yourself) ].

	Smalltalk at: #NCMakerButton ifPresent: [ :ncmb |
		descriptions addAll: {
	DescriptionForPartsBin
		formalName: 'Connector Button'
		categoryList: #('Connectors' )
		documentation: 'A Maker Button with a Connector loaded'
		globalReceiverSymbol: #NCAAConnectorMorph
		nativitySelector: #newMakerButton.
	DescriptionForPartsBin
		formalName: 'Random Connectors'
		categoryList: #('Connectors' )
		documentation: 'A button that makes fancy Connectors at random'
		globalReceiverSymbol: #NCAAConnectorMorph
		nativitySelector: #newRandomButton.
		}
	].
	^descriptions
! !

!NCAAConnectorMorph class methodsFor: 'testing' stamp: 'nk 5/2/2004 15:15'!
testPerformanceWithLotsOfConnectors
	"
	self testPerformanceWithLotsOfConnectors
	"

	| center r1 conn newConnectors newMorphs |
	newConnectors := OrderedCollection new.
	newMorphs := OrderedCollection new.

	1 to: 400 do: [ :i |
		center := Display width atRandom @ Display height atRandom.
		r1 := Morph new bounds: (Rectangle center: center extent: 10@10); openInWorld; yourself.
		newMorphs add: r1 ].

	World changed; doOneCycle.

	(self confirm: 'MessageTally will start now,
and stop when the cursor goes
to the top of the screen') ifTrue: [
	TimeProfileBrowser spyOn: [ 
		1 to: 400 do: [ :i |
			conn := self fromMorph: newMorphs atRandom toMorph: newMorphs atRandom.
			newConnectors add: conn.
			World doOneCycle.
		].
		[ ActiveHand position y > 0 ] whileTrue: [ World doOneCycle ]]].

	World removeAllMorphsIn: newConnectors.
	World removeAllMorphsIn: newMorphs.
! !

!NCAAConnectorMorph class methodsFor: 'arrow forms' stamp: 'nk 8/25/2003 13:47'!
umlArrow
	| a |
	a := { 0.0@0.0 . 0.0@0.0 . -0.5@0.0 . 
	-0.5@0.0 . -0.5@0.0 . 0.0@-1.0 . 
	0.0@-1.0 . 0.0@-1.0 . 0.5@0.0 . 
	0.5@0.0 . 0.5@0.0 . 0.0@0.0 }.
^(NCLineArrowGlyph withContours: (
(Array new: 1)
	at: 1 put: a; 
	yourself
"clicks 0.0@0.0"
)) offset: 0@1.0;
defaultScale: 2.5;
yourself! !

!NCAAConnectorMorph class methodsFor: 'class initialization' stamp: 'nk 10/19/2003 17:37'!
unload
	Smalltalk removeFromStartUpList: self.

	ScriptingSystem removeCustomEventNamed: #connectedTo for: self ;
		removeCustomEventNamed: #disconnectedFrom for: self;
		removeCustomEventNamed: #sourceConnected for: self;
		removeCustomEventNamed: #sourceDisconnecting for: self;
		removeCustomEventNamed: #destinationConnected for: self;
		removeCustomEventNamed: #destinationDisconnecting for: self.! !

!NCAAConnectorMorph class methodsFor: 'instance creation' stamp: 'nk 8/24/2003 14:49'!
vertices: aCollection
	| newMe |
	newMe := self new.
	newMe setVertices: aCollection.
	newMe constraints first center: aCollection first.
	newMe constraints second center: aCollection last.
	^newMe! !

!NCAAConnectorMorph class methodsFor: 'instance creation' stamp: 'nk 8/28/2003 15:13'!
wireRandom
	"((SimpleButtonMorph newWithLabel: 'Random Connector')
		target: NCAAConnectorMorph;
		actionSelector: #wireRandom) openInHand"
	self random startWiring.! !

!NCAAConnectorMorph methodsFor: 'menus' stamp: 'dgd 8/26/2004 14:35'!
addBasicArrowMenuItemsTo: aMenu event: evt

	aMenu addLine.
	aMenu add: '---' action:  #makeNoArrows.
	aMenu add: '-->' action:  #makeForwardArrow.
	aMenu add: '<--'  action:  #makeBackArrow.
	aMenu add: '<->' action:  #makeBothArrows.
	aMenu add: 'arrow sizes' translated action: #changeArrowScaleInteractively:
! !

!NCAAConnectorMorph methodsFor: 'menus' stamp: 'dgd 8/26/2004 14:44'!
addBasicMenuItemsTo: aMenu event: evt 
	owner
		ifNotNil: [aMenu add: 'delete' translated action: #dismissViaHalo].
	aMenu addLine.
	owner
		ifNotNil: [aMenu add: 'add label' translated action: #addLabel:].
	(Smalltalk includesKey: #NCMakerButton) ifTrue: [
		aMenu add: 'make button' translated action: #makeButton: ].
	aMenu
		 	addLine.
	aMenu
		add: 'line color...' translated
		target: self
		selector: #changeLineColor:;
		
		add: 'line width...' translated
		target: self
		selector: #changeLineWidth:.
	line isBordered
		ifTrue: [aMenu
				add: 'line border color...' translated
				target: self
				selector: #changeBorderColor:;
				
				add: 'line border width...' translated
				target: self
				selector: #changeBorderWidth:].
	line addFillStyleMenuItems: aMenu hand: evt hand.
	aMenu addLine.
	aMenu
		addUpdating: #handlesShowingPhrase
		target: self
		action: #showOrHideHandles;
		
		addUpdating: #orthogonalityPhrase
		target: self
		action: #toggleOrthogonality;
		 addUpdating: #makeSegmentedOrSmoothLinePhrase action: #toggleSmoothing.
	aMenu
		addWithLabel: 'straighten' translated
		enablement: [self vertices size > 2]
		action: #straighten! !

!NCAAConnectorMorph methodsFor: 'connectors-queries' stamp: 'nk 8/15/2003 09:56'!
addConnectedConnectorsTo: aCollection
	"Add myself and all my connected connectors (recursively) to aCollection"
	aCollection add: self.
	self connections do: [ :ea |
		(ea notNil and: [ ea isConnector ]) ifTrue: [
			(aCollection includes: ea)
				ifFalse: [ ea addConnectedConnectorsTo: aCollection. ]
		]
	].
	self connectedMorphs do: [ :ea |
		(ea notNil and: [ ea isConnector ]) ifTrue: [
			(aCollection includes: ea)
				ifFalse: [ ea addConnectedConnectorsTo: aCollection. ]
		]
	].
	^aCollection
! !

!NCAAConnectorMorph methodsFor: 'menus' stamp: 'dgd 8/26/2004 14:35'!
addCustomMenuItems: aMenu hand: aHandMorph
	"Use my line's menu additions"
	super addCustomMenuItems: aMenu hand: aHandMorph.
	aMenu addLine.


	line addFillStyleMenuItems: aMenu hand: aHandMorph.
	aMenu addLine.

	aMenu
		addUpdating: #handlesShowingPhrase target: self action: #showOrHideHandles;
		addUpdating: #orthogonalityPhrase target: self action: #toggleOrthogonality.

	self vertices size > 2 ifTrue: [ 
		line isCurve
			ifTrue: [aMenu add: 'make segmented line' translated action: #toggleSmoothing]
			ifFalse: [aMenu add: 'make smooth line' translated action: #toggleSmoothing].
		aMenu add: 'straighten' translated target: self selector: #straighten.
	]. 
	
	self addBasicArrowMenuItemsTo: aMenu event: nil.

	aMenu 
		addLine;
		add: 'add label' translated action: #addLabel:.
! !

!NCAAConnectorMorph methodsFor: 'drop shadows' stamp: 'nk 8/21/2003 10:27'!
addDragShadow
	"I'm starting to be dragged but I'm not in the hand yet.
	So use the hand's drop shadow information to set my own drop shadow"

self flag: #ned.

	self changed.
	self setProperty: #dragShadowOffset toValue: ActiveHand shadowOffset.
	self setProperty: #dragShadowColor toValue: Color gray darker "ActiveHand shadowColor".
	self layoutChanged.
	self changed.! !

!NCAAConnectorMorph methodsFor: 'menus' stamp: 'nk 8/24/2003 23:03'!
addEndItemsFor: aConstraintOrNil to: subMenu event: evt
	subMenu add: 'new arrow...' target: self selector: #chooseArrowGlyphFor: argumentList: { constraints indexOf: aConstraintOrNil }.
! !

!NCAAConnectorMorph methodsFor: 'rotate scale and flex' stamp: 'nk 8/16/2003 13:03'!
addFlexShell
	"Behave like a line"
	^self! !

!NCAAConnectorMorph methodsFor: 'menus' stamp: 'dgd 8/26/2004 14:59'!
addGraphModelYellowButtonItemsTo: aCustomMenu event: evt
	self graphModel
		ifNil: [aCustomMenu add: 'duplicate' translated action: #duplicateMorph:]
		ifNotNil: [
			self graphModel addModelYellowButtonMenuItemsTo: aCustomMenu
				forMorph: self
				hand: evt hand.
			aCustomMenu add: 'duplicate model and view' translated action: #duplicateMorph:.
			aCustomMenu addLine ].
! !

!NCAAConnectorMorph methodsFor: 'submorphs-add/remove' stamp: 'nk 8/18/2003 09:49'!
addHandles
	self setProperty: #showingHandles toValue: true.
	self deleteMorphs: self handles.
	line handles withIndexDo: [ :h :i |
		h setProperty: #connectorHandle toValue: i.
		self addMorphFront: h.
	].
! !

!NCAAConnectorMorph methodsFor: 'labeling' stamp: 'nk 1/20/2004 11:13'!
addLabel: anEvent
	| label |
	label := self labelClass new string: '  ' wrap: false minWidth: 20.
	label
		align: label center with: anEvent position;
		attachUnobnoxiouslyTo: self at: anEvent position;
		dieWithInput: true;
		acceptOnCR: true;
		acceptOnFocusLoss: true;
		selectAll.
	self pasteUpMorphOrWorld addMorph: label.
	anEvent hand newKeyboardFocus: label label.
	^label! !

!NCAAConnectorMorph methodsFor: 'labeling' stamp: 'nk 8/15/2003 10:35'!
addLabel: text at: aPoint
	| label |
	label := (self labelClass new string: text).
	label
		attachTo: self at: aPoint;
		align: label center with: aPoint;
		dieWithInput: true.
	self pasteUpMorphOrWorld addMorph: label.
	^label! !

!NCAAConnectorMorph methodsFor: 'labeling' stamp: 'nk 8/15/2003 10:35'!
addLabel: text near: aPoint
	| label |
	label := (self labelClass new string: text).
	label
		color: self color;
		align: label center with: aPoint;
		attachUnobnoxiouslyTo: self at: aPoint;
		dieWithInput: true.
	self pasteUpMorphOrWorld addMorph: label.
	^label! !

!NCAAConnectorMorph methodsFor: 'menus' stamp: 'nk 6/15/2004 07:33'!
addModelYellowButtonItemsTo: aCustomMenu event: evt
	super addModelYellowButtonItemsTo: aCustomMenu event: evt.
	self addGraphModelYellowButtonItemsTo: aCustomMenu event: evt.
	^aCustomMenu! !

!NCAAConnectorMorph methodsFor: 'labeling' stamp: 'nk 8/15/2003 10:35'!
addShapes: anArray
	constraints with: anArray do: [ :end :shape | end addShape: shape; step ].! !

!NCAAConnectorMorph methodsFor: 'labeling' stamp: 'nk 10/21/2003 11:22'!
addUnwrappedLabel: text at: aPoint
	| label |
	label := (self labelClass new string: text wrap: false).
	label
		attachTo: self at: aPoint;
		align: label center with: aPoint;
		dieWithInput: true.
	self pasteUpMorphOrWorld addMorph: label.
	^label! !

!NCAAConnectorMorph methodsFor: 'labeling' stamp: 'nk 10/28/2003 09:30'!
addUnwrappedLabel: text near: aPoint
	| label |
	label := (self labelClass new string: text wrap: false).
	label
		color: self color;
		align: label center with: aPoint;
		attachUnobnoxiouslyTo: self at: aPoint;
		dieWithInput: true.
	self pasteUpMorphOrWorld addMorph: label.
	^label! !

!NCAAConnectorMorph methodsFor: 'menus' stamp: 'nk 1/23/2004 16:15'!
addYellowButtonMenuItemsTo: aMenu event: evt
	super addYellowButtonMenuItemsTo: aMenu event: evt.

	self addBasicMenuItemsTo: aMenu event: evt.

	self addBasicArrowMenuItemsTo: aMenu event: evt.
	^aMenu! !

!NCAAConnectorMorph methodsFor: 'connectors-queries' stamp: 'nk 10/25/2003 17:25'!
addedSubmorphs
	"Answer my submorphs other than my line end constraints and line."
	^submorphs reject: [ :ea  | ea == line or: [ ea isConstraint ]].! !

!NCAAConnectorMorph methodsFor: 'scripting' stamp: 'nk 8/21/2004 09:26'!
allowsConnectionToSubmorphs
	"Answer true if I will allow a connector to connect to any of my submorphs.
	If not, the search for attachment points in embedded morphs will not be made."

	^self valueOfProperty: #allowsConnectionToSubmorphs ifAbsent: [ false ]! !

!NCAAConnectorMorph methodsFor: 'event handling' stamp: 'nk 8/15/2003 10:31'!
allowsDetachmentOfConstraint: constraint
	"Answer true if the given constraint is either already disconnected, or can be disconnected after checking with the attached Morph and myself"
	| aMorph index |
	aMorph := constraint input.
	index := constraints indexOf: constraint.
	^aMorph isNil or: [
		(self wantsToDetachEnd: index fromMorph: aMorph)
			and: [ aMorph allowsDetachmentFromEnd: index ofConnector: self ]]
! !

!NCAAConnectorMorph methodsFor: 'arrows' stamp: 'nk 8/25/2003 13:20'!
arrowNames: anArray
	"anArray has the keys into my arrow dictionary for the desired arrows"
	line arrows: (anArray collect: [ :ea | self class arrowDictionary at: ea ifAbsent: [ nil ]])! !

!NCAAConnectorMorph methodsFor: 'arrows' stamp: 'nk 9/20/2003 17:07'!
arrowScales
	^line arrowScales! !

!NCAAConnectorMorph methodsFor: 'arrows' stamp: 'nk 9/20/2003 17:03'!
arrowScales: anArray 
	line arrowScales: anArray ! !

!NCAAConnectorMorph methodsFor: 'arrows' stamp: 'nk 8/15/2003 10:35'!
arrowSpec: aPoint
	"The standard arrow is equivalent to arrowSpec: 5@4."
	line arrowSpec: aPoint! !

!NCAAConnectorMorph methodsFor: 'dropping/grabbing' stamp: 'nk 12/26/2003 17:22'!
asButtonPrototype
	| logger proto |
	logger := self valueOfProperty: #logger ifAbsent: [].
	self removeProperty: #logger.
	proto := super asButtonPrototype
		removeProperty: #suspendedEventHandler;
		detach;
		straighten;
		setVertices: { 0@0. 50@0 };
		yourself.
	self setProperty: #logger toValue: logger.
	proto setProperty: #logger toValue: logger.	"share my logger if any but copy the types"
	^proto.! !

!NCAAConnectorMorph methodsFor: 'geometry' stamp: 'nk 8/16/2003 13:02'!
asVector
	"Answer a Point representing me as a vector. Ignore intermediate vertices"
	^line vertices last - line vertices first! !

!NCAAConnectorMorph methodsFor: 'geometry' stamp: 'nk 8/16/2003 13:02'!
asVectorFrom: aMorph 
	"Answer a Point representing me as a vector. Ignore 
	intermediate vertices"
	^ aMorph == self startConnection
		ifTrue: [line vertices last - line vertices first]
		ifFalse: [line vertices first - line vertices last]! !

!NCAAConnectorMorph methodsFor: 'protected-notifications' stamp: 'nk 8/16/2003 13:03'!
attachFrom: aConstraintMorph at: aPoint
	(constraints includes: aConstraintMorph) ifTrue: [ self error: 'can''t happen' ].
	"I don't want my labels to wander"
	(aConstraintMorph isLabelConstraint)
		ifTrue: [ ^self attachFrom: aConstraintMorph atNearestSpecTo: aPoint ].
	(aConstraintMorph isLineConstraint)
		ifTrue: [ aConstraintMorph connectToNearestPoint ]
		ifFalse: [ super attachFrom: aConstraintMorph at: aPoint ]
! !

!NCAAConnectorMorph methodsFor: 'visual properties' stamp: 'nk 8/15/2003 10:23'!
beAdjacent
	^false! !

!NCAAConnectorMorph methodsFor: 'orthogonality' stamp: 'nk 8/15/2003 10:34'!
beOrthogonal: aBoolean
	aBoolean ifTrue: [ self setProperty: #orthogonal toValue: true]
		ifFalse: [ self removeProperty: #orthogonal ]
! !

!NCAAConnectorMorph methodsFor: 'appearance' stamp: 'nk 8/15/2003 10:35'!
beSmoothCurve
	line beSmoothCurve.
	constraints do: [ :ea | ea forceRedraw; step ].
	self changed.! !

!NCAAConnectorMorph methodsFor: 'appearance' stamp: 'nk 1/18/2004 16:54'!
beSmoothCurve: aBoolean
	aBoolean ifTrue: [ self beSmoothCurve ]
		ifFalse: [ self beStraightSegments ]
! !

!NCAAConnectorMorph methodsFor: 'appearance' stamp: 'nk 8/15/2003 10:35'!
beStraightSegments
	line beStraightSegments.
	constraints do: [ :ea | ea forceRedraw; step ].
	self changed.! !

!NCAAConnectorMorph methodsFor: 'appearance' stamp: 'nk 8/16/2003 13:03'!
borderColor
	^line ifNotNil: [ line lineBorderColor  ]
		ifNil: [ super borderColor ]! !

!NCAAConnectorMorph methodsFor: 'appearance' stamp: 'nk 8/16/2003 13:03'!
borderColor: aColor
	constraints do: [ :ea | ea endShapeBorderColor: aColor ].
	^line ifNotNil: [ line lineBorderColor: aColor ]
		ifNil: [ super borderColor: aColor ]! !

!NCAAConnectorMorph methodsFor: 'appearance' stamp: 'nk 1/18/2004 17:30'!
borderStyle
	"for compatibility with properties sheets"
	^ ((BorderStyle simple)
		width: self borderWidth;
		baseColor: self borderColor asColor;
		color: self borderColor asColor;
		yourself)! !

!NCAAConnectorMorph methodsFor: 'appearance' stamp: 'nk 1/18/2004 17:20'!
borderStyle: aBorderStyle
	"for compatibility with properties sheets"
	self borderWidth: aBorderStyle width;
		borderColor: aBorderStyle color asColor.! !

!NCAAConnectorMorph methodsFor: 'appearance' stamp: 'nk 8/16/2003 13:03'!
borderWidth
	^line ifNotNil: [ line lineBorderWidth ] ifNil: [ 0 ]! !

!NCAAConnectorMorph methodsFor: 'appearance' stamp: 'nk 8/16/2003 13:03'!
borderWidth: aFloat
	constraints do: [ :ea | ea endShapeBorderWidth: aFloat rounded ].
	line
		ifNotNil: [ line lineBorderWidth: aFloat rounded ]
		ifNil: [ super borderWidth: aFloat rounded ].
! !

!NCAAConnectorMorph methodsFor: 'connectors-labels' stamp: 'nk 8/16/2003 13:03'!
boundsSignatureHash
	| hash |
	hash := line boundsSignatureHash.
	self validConnections do: [ :ea |
		hash := (hash + ea boundsSignatureHash + ea owner identityHash) hashMultiply
	].
	^hash
! !

!NCAAConnectorMorph methodsFor: 'menus' stamp: 'nk 8/22/2003 15:07'!
changeArrowScaleInteractively: evt
	
	line changeArrowScaleInteractively: evt! !

!NCAAConnectorMorph methodsFor: 'menus' stamp: 'nk 8/15/2003 10:24'!
changeBorderColor: evt
	| aHand |
	aHand := evt ifNotNil: [evt hand] ifNil: [self primaryHand].
	self changeColorTarget: self selector: #borderColor: originalColor: self borderColor hand: aHand! !

!NCAAConnectorMorph methodsFor: 'menus' stamp: 'nk 8/15/2003 10:24'!
changeBorderWidth: evt 
	"Copied from BorderedMorph"

	| newWidth |
	(owner notNil and: [self visible]) 
		ifTrue: [^self changeBorderWidthInteractively: evt].
	newWidth := FillInTheBlank request: 'New line border width?'
				initialAnswer: self borderWidth asString.
	newWidth isEmpty ifTrue: [ ^self ].
	newWidth := newWidth asNumber.
	self borderWidth: newWidth! !

!NCAAConnectorMorph methodsFor: 'menus' stamp: 'nk 12/27/2003 11:23'!
changeBorderWidthInteractively: evt
	"Copied from BorderedMorph"
	| handle origin aHand newWidth oldWidth |
	aHand := evt ifNil: [self primaryHand] ifNotNil: [evt hand].
	origin := aHand position.
	oldWidth := self borderWidth.
	handle := HandleMorph new
		forEachPointDo:
			[:newPoint | handle removeAllMorphs.
			handle addMorph:
				(LineMorph from: origin to: newPoint color: Color black width: 1).
			newWidth := (newPoint - origin) r / 10.0.
			self borderWidth: (newWidth max: 0)]
		lastPointDo:
			[:newPoint | handle deleteBalloon.
			self halo ifNotNilDo: [:halo | halo addHandles].
			self rememberCommand:
				(Command new cmdWording: 'border change';
					undoTarget: self selector: #borderWidth: argument: oldWidth;
					redoTarget: self selector: #borderWidth: argument: newWidth)].
	aHand attachMorph: handle.
	handle setProperty: #helpAtCenter toValue: true.
	handle showBalloon:
'Move cursor farther from
this point to increase border width.
Click when done.' hand: evt hand.
	handle startStepping! !

!NCAAConnectorMorph methodsFor: 'menus' stamp: 'nk 8/15/2003 10:24'!
changeLineColor: evt
	| aHand |
	aHand := evt ifNotNil: [evt hand] ifNil: [self primaryHand].
	self changeColorTarget: self selector: #lineColor: originalColor: self lineColor hand: aHand! !

!NCAAConnectorMorph methodsFor: 'menus' stamp: 'nk 8/15/2003 10:24'!
changeLineWidth: evt 
	"Copied from BorderedMorph"

	| newWidth |
	(owner notNil and: [self visible]) 
		ifTrue: [^self changeLineWidthInteractively: evt].
	newWidth := FillInTheBlank request: 'New line width?'
				initialAnswer: self lineWidth asString.
	newWidth isEmpty ifTrue: [ ^self ].
	newWidth := newWidth asNumber.
	self lineWidth: newWidth! !

!NCAAConnectorMorph methodsFor: 'menus' stamp: 'nk 12/27/2003 11:24'!
changeLineWidthInteractively: evt
	"Copied from BorderedMorph"
	| handle origin aHand newWidth oldWidth |
	aHand := evt ifNil: [self primaryHand] ifNotNil: [evt hand].
	origin := aHand position.
	oldWidth := self lineWidth.
	handle := HandleMorph new
		forEachPointDo:
			[:newPoint | handle removeAllMorphs.
			handle addMorph:
				(LineMorph from: origin to: newPoint color: Color black width: 1).
			newWidth := (newPoint - origin) r / 15.0.
			self lineWidth: (newWidth max: 1)]
		lastPointDo:
			[:newPoint | handle deleteBalloon.
			self halo ifNotNilDo: [:halo | halo addHandles].
			self rememberCommand:
				(Command new cmdWording: 'line width change';
					undoTarget: self selector: #lineWidth: argument: oldWidth;
					redoTarget: self selector: #lineWidth: argument: newWidth)].
	aHand attachMorph: handle.
	handle setProperty: #helpAtCenter toValue: true.
	handle showBalloon:
'Move cursor farther from
this point to increase line width.
Click when done.' hand: evt hand.
	handle startStepping! !

!NCAAConnectorMorph methodsFor: 'menus' stamp: 'nk 8/23/2003 20:48'!
chooseArrowGlyphFor: endIndex 
	| menu |
	menu := NCArrowDictionaryMenu chooseArrowGlyphFor: self selector: #setGlyph:forEnd:.
	menu argument: endIndex
	! !

!NCAAConnectorMorph methodsFor: 'geometry' stamp: 'nk 5/1/2004 13:00'!
closestOrthogonalPointTo: aPoint
	"first, find an intersection"
	|  intersections |
	intersections := OrderedCollection new.
	intersections addAll: (self line intersectionsWith: (0@0 corner: aPoint)).
	intersections addAll: (self line intersectionsWith: (aPoint corner: ActiveWorld bounds extent)).	
	intersections isEmpty ifTrue: [ ^self closestPointTo: aPoint ].
	^intersections detectMin: [ :ea | ea dist: aPoint ] ! !

!NCAAConnectorMorph methodsFor: 'geometry' stamp: 'nk 8/15/2003 09:53'!
closestPointTo: aPoint
	^line closestPointTo: aPoint! !

!NCAAConnectorMorph methodsFor: 'appearance' stamp: 'nk 8/16/2003 13:03'!
color
	^line ifNotNil: [ line fillStyle ]
		ifNil: [ super color]! !

!NCAAConnectorMorph methodsFor: 'appearance' stamp: 'nk 8/16/2003 13:03'!
color: aColor
	constraints do: [ :ea | ea endShapeColor: aColor ].
	^line ifNotNil: [ line lineColor: aColor ]
		ifNil: [ super color: aColor ]! !

!NCAAConnectorMorph methodsFor: 'connection' stamp: 'nk 8/18/2003 11:34'!
connectEnds
	"I've been created and given an owner. If appropriate, connect my ends to Morphs that are under or on top of them."

	constraints do: 
			[:constraint | 
			| targets localPoint |
			constraint inputs isEmpty 
				ifTrue: 
					[localPoint := self griddedPoint: constraint targetPoint.
					targets := self connectionTargetsAt: localPoint for: constraint.
					targets isEmpty 
						ifFalse: 
							[constraint attachTo: targets first at: localPoint.
							self trigger: #connectEndTo with: targets first]]].
	self
		changed;
		layoutChanged! !

!NCAAConnectorMorph methodsFor: 'connection' stamp: 'nk 1/7/2004 13:44'!
connectFinishTo: aMorph 
	| old |
	old := constraints last input.
	aMorph
		ifNil: [constraints last input: nil]
		ifNotNil: [constraints last attachTo: aMorph at: aMorph center.
			self trigger: #connectEndTo with: aMorph].
	^ old! !

!NCAAConnectorMorph methodsFor: 'connection' stamp: 'nk 1/7/2004 13:44'!
connectFinishTo: aMorph at: aPoint 
	| end old |
	old := (end := constraints last) input.
	aMorph
		ifNil: [end input: nil]
		ifNotNil: [end attachTo: aMorph at: aPoint.
			self trigger: #connectEndTo with: aMorph].
	^ old! !

!NCAAConnectorMorph methodsFor: 'connection' stamp: 'nk 1/7/2004 13:45'!
connectStartTo: aMorph 
	| old |
	old := constraints first input.
	aMorph
		ifNil: [constraints first input: nil]
		ifNotNil: [
			constraints first attachTo: aMorph at: aMorph center.
			self trigger: #connectEndTo with: aMorph ].
	^ old! !

!NCAAConnectorMorph methodsFor: 'connection' stamp: 'nk 1/7/2004 13:45'!
connectStartTo: aMorph at: aPoint 
	| old |
	old := constraints first input.
	aMorph
		ifNil: [constraints first input: nil]
		ifNotNil: [ constraints first attachTo: aMorph at: aPoint.
	self trigger: #connectEndTo with: aMorph ].
	^ old! !

!NCAAConnectorMorph methodsFor: 'connection' stamp: 'nk 8/15/2003 10:21'!
connectToNearestAttachmentPoint
	"Connect both ends to wherever the target Morphs prefer"
	constraints do: [ :ea | ea connectToNearestAttachmentPoint ]! !

!NCAAConnectorMorph methodsFor: 'connection' stamp: 'nk 8/15/2003 10:21'!
connectToNearestOrthogonalPoint
	"Connect both ends to the nearest point orthogonally, or the nearest point if none"
	constraints do: [ :ea | ea connectToNearestOrthogonalPoint ]
! !

!NCAAConnectorMorph methodsFor: 'connection' stamp: 'nk 8/15/2003 10:21'!
connectToNearestPoint
	constraints do: [ :ea | ea connectToNearestPoint ]! !

!NCAAConnectorMorph methodsFor: 'connection' stamp: 'nk 8/15/2003 10:21'!
connectToNearestPointToCenter
	constraints do: [ :ea | ea connectToNearestPointToCenter ]! !

!NCAAConnectorMorph methodsFor: 'connectors-queries' stamp: 'nk 8/15/2003 09:56'!
connectedMorphs
	"Answer a 2 element Array with my connections. Either or both could be nil"
	^(constraints collect: [ :ea | ea input ])! !

!NCAAConnectorMorph methodsFor: 'protected-policy' stamp: 'nk 1/13/2004 15:46'!
connectionTargetsAt: localPoint for: constraint
	"Answer a collection of prospective connection targets.
	Search up to the first playfield."
	| targets globalPoint index firstPlayfield |
	index := constraints indexOf: constraint.
	globalPoint := (constraint transformFrom: self world) localPointToGlobal: localPoint.
	targets := self world morphsAt: globalPoint.
	firstPlayfield := targets detect: [ :ea | ea isPlayfieldLike ] ifNone: [ ].
	targets := targets copyUpTo: firstPlayfield.
	targets := targets select: [ :ea |
		(ea wantsAttachmentFromEnd: index ofConnector: self)
			and: [ (self wantsToAttachEnd: index toMorph: ea) 
				"and: [ ea isShadowVisibleIn: (Rectangle center: globalPoint extent: 30@30) ]" ] ].

	"The following logic is to avoid connecting to other NCConnectors
	when there is something else there to connect to"
	^targets asSortedCollection: [ :a :b |
		a isConnector not and: [ a owner isPlayfieldLike ] ]! !

!NCAAConnectorMorph methodsFor: 'protected-notifications' stamp: 'nk 8/15/2003 10:31'!
connectionsChanged
	"The number of my connections has changed.
	This is a hook for subclasses"
! !

!NCAAConnectorMorph methodsFor: 'connectors-queries' stamp: 'nk 8/15/2003 09:56'!
constraints
	^constraints! !

!NCAAConnectorMorph methodsFor: 'geometry' stamp: 'nk 10/26/2003 10:13'!
containsPoint: aPoint 
	"first, bounds test (simple)"

	(super containsPoint: aPoint) ifFalse: [^false].
	constraints do: [:ea | (ea containsPoint: aPoint) ifTrue: [^true]].
	line ifNil: [^true].
	^(line containsPoint: aPoint)
		or: [ self addedSubmorphs anySatisfy: [ :m | m containsPoint: aPoint ]] ! !

!NCAAConnectorMorph methodsFor: 'protected-notifications' stamp: 'nk 10/28/2003 09:04'!
couldConnectMorphs: morphs
	"Answer true if I am willing to connect to the (1 or 2) morphs given,
	and they're willing to be connected to me."

	morphs isEmpty ifTrue: [ ^false ].

	morphs withIndexDo: [ :end :i |
		i > 2 ifTrue: [ ^false ].
		(self wantsToAttachEnd: i toMorph: end)
			ifFalse: [ ^false ].
		end ifNotNil: [
			(end wantsAttachmentFromEnd: i ofConnector: self)
				ifFalse: [ ^false ].
		]
	].
	^true.! !

!NCAAConnectorMorph methodsFor: 'accessing' stamp: 'nk 2/28/2006 11:19'!
curveSegmentAt: n
	^line curveSegmentAt: n! !

!NCAAConnectorMorph methodsFor: 'accessing' stamp: 'nk 2/28/2006 11:18'!
curveSegmentAt: n put: aPoint
	^line curveSegmentAt: n  put: aPoint! !

!NCAAConnectorMorph methodsFor: 'accessing' stamp: 'nk 8/21/2003 16:36'!
curveVertexAt: n
	^line curveVertexAt: n! !

!NCAAConnectorMorph methodsFor: 'accessing' stamp: 'nk 8/21/2003 16:36'!
curveVertexAt: n put: aPoint
	^line curveVertexAt: n  put: aPoint! !

!NCAAConnectorMorph methodsFor: 'appearance' stamp: 'nk 8/15/2003 10:35'!
dashedLine: aBoolean
	line dashedBorder: (aBoolean
		ifTrue: [ { line lineWidth * 4 . line lineWidth * 3 . Color transparent } ]
		ifFalse: [ nil ])! !

!NCAAConnectorMorph methodsFor: 'attachments' stamp: 'nk 8/16/2003 13:03'!
defaultAttachmentPointSpecs
	"Where would labels attach to me? (note that I normally let Connectors slide along my length)"
	^{ 
		{ #firstVertex } .
		{ #midpoint  } .
		{ #lastVertex }
	}
! !

!NCAAConnectorMorph methodsFor: 'naming' stamp: 'cmm 12/4/2010 20:00'!
defaultNameStemForInstances
	^ self class name
		copyFrom: 3
		to: self class name size! !

!NCAAConnectorMorph methodsFor: 'submorphs-add/remove' stamp: 'nk 9/22/2003 15:31'!
delete
	self removeHandles.
	self deleteAllLabels.
	self stopMonitoringEnd.
	self detach.
	super delete.! !

!NCAAConnectorMorph methodsFor: 'arrows' stamp: 'nk 8/26/2003 14:43'!
destinationArrowName
	| arrows |
	arrows := line arrows.
	^self class nameForArrowGlyph: arrows second
! !

!NCAAConnectorMorph methodsFor: 'arrows' stamp: 'nk 8/26/2003 10:55'!
destinationArrowName: aSymbol
	| arrows |
	arrows := line arrows.
	arrows at: 2 put: (self class arrowDictionary at: aSymbol ifAbsent: [ ^nil ]).
	line arrows: arrows! !

!NCAAConnectorMorph methodsFor: 'connectors-queries' stamp: 'nk 8/15/2003 09:56'!
destinationMorph
	"Answer the morph connected to my first vertex, if any"
	^constraints last input! !

!NCAAConnectorMorph methodsFor: 'connection' stamp: 'nk 8/15/2003 10:21'!
detach
	constraints do: [ :const | const input: nil ]
! !

!NCAAConnectorMorph methodsFor: 'connection' stamp: 'nk 8/15/2003 10:21'!
detachFrom: aMorph
	constraints do: [ :const |
		(const input == aMorph)
			ifTrue: [ const input: nil ]
	].! !

!NCAAConnectorMorph methodsFor: 'meta-actions' stamp: 'nk 1/20/2004 19:46'!
dismissMorph
	self detach.
	super dismissMorph.! !

!NCAAConnectorMorph methodsFor: 'appearance' stamp: 'nk 8/15/2003 10:35'!
doesBevels
	^false! !

!NCAAConnectorMorph methodsFor: 'drop shadows' stamp: 'nk 8/15/2003 10:36'!
dragShadowColor
	^self valueOfProperty: #dragShadowColor! !

!NCAAConnectorMorph methodsFor: 'drop shadows' stamp: 'nk 8/20/2003 17:58'!
dragShadowOffset
	^self valueOfProperty: #dragShadowOffset ifAbsent: [ 0@0 ].! !

!NCAAConnectorMorph methodsFor: 'drop shadows' stamp: 'nk 9/22/2003 18:22'!
drawDragShadowOn: aCanvas 
	aCanvas translateBy: self dragShadowOffset
		during: 
			[:shadowCanvas | 
			shadowCanvas shadowColor: self dragShadowColor.
			(shadowCanvas isVisible: self outerBounds) 
				ifTrue: [
					self withTransparentConstraintsDo: [
						self drawOn: shadowCanvas.
						self drawSubmorphsOn: shadowCanvas ] ]]! !

!NCAAConnectorMorph methodsFor: 'drawing' stamp: 'nk 1/18/2004 17:32'!
drawOn: aCanvas 
	"Nothing to do, since I'm invisible..."! !

!NCAAConnectorMorph methodsFor: 'meta-actions' stamp: 'nk 9/27/2003 15:29'!
duplicate
	| newMorph |
	newMorph := super duplicate.
	newMorph detach.
	newMorph setProperty: #justDuplicated toValue: true.
	^newMorph! !

!NCAAConnectorMorph methodsFor: 'dropping/grabbing' stamp: 'nk 10/21/2003 14:13'!
duplicateForEmbedding
	| logger proto |
	logger := self valueOfProperty: #logger ifAbsent: [].
	self removeProperty: #logger.
	proto := super duplicateForEmbedding.
	self setProperty: #logger toValue: logger.
	proto setProperty: #logger toValue: logger.	"share my logger if any but copy the types"
	^proto.! !

!NCAAConnectorMorph methodsFor: 'meta-actions' stamp: 'nk 8/16/2003 13:03'!
duplicateMorph: evt
	| newMorph |
	newMorph := super duplicateMorph: evt.
	newMorph setProperty: #justDuplicated toValue: true.
	^newMorph! !

!NCAAConnectorMorph methodsFor: 'connectors-queries' stamp: 'nk 8/15/2003 09:56'!
endConnection
	"Answer my ending connection or nil"
	^constraints second input! !

!NCAAConnectorMorph methodsFor: 'connectors-queries' stamp: 'nk 8/15/2003 09:56'!
endConstraint
	"Answer my ending constraint"
	^constraints second
! !

!NCAAConnectorMorph methodsFor: 'event handling' stamp: 'nk 1/14/2004 20:38'!
enterConnected
	self removeShortEndSegments.
	self constraints do: [ :ea | ea startStepping ]! !

!NCAAConnectorMorph methodsFor: 'event handling' stamp: 'nk 1/17/2004 10:51'!
enterDraggingEnd
	| constraint evt args label |
	self log: #debug items: {'enterDraggingEnd'}.
	self comeToFront.
	args := eventHandler lastEventArguments.
	constraint := args first.
	evt := args second.
	(self isWiring not
			and: [evt shiftPressed
					or: [evt yellowButtonPressed]])
		ifTrue: [self
				insertVertexNextTo: constraint
				at: (self griddedPoint: evt position)].

	owner isWorldMorph
		ifFalse: [self jumpToWorld].
	"make end follow hand"
	constraint input: evt hand;
		 connectToCursorPoint;
		 step.
	evt hand newMouseFocus: self.
	self
		setProperty: #startDraggingEvent
		toValue: (evt copy setHand: nil).
	self visible: true.
	label := self valueOfProperty: #wiringLabel.
	label
		ifNotNil: [self removeProperty: #wiringLabel.
			self pasteUpMorphOrWorld addMorph: label.
			label visible: true].
	self addDragShadow.
! !

!NCAAConnectorMorph methodsFor: 'event handling' stamp: 'nk 1/17/2004 10:41'!
enterDraggingEndOver
	self log: #debug items: {'enterDraggingEndOver'}.
	self valueOfProperty: #highlightedTargets ifPresentDo: [ :targets |
		(targets includes: self sourceMorph) ifTrue: [ constraints first stopStepping; step ].
		(targets includes: self destinationMorph) ifTrue: [ constraints second stopStepping; step ].
	].! !

!NCAAConnectorMorph methodsFor: 'event handling' stamp: 'nk 1/14/2004 16:25'!
enterDraggingMorph
	self log: #debug items: {'enterDraggingMorph'}.
	fullBounds := nil.
	(self
			valueOfProperty: #justDuplicated
			ifAbsent: [false])
		ifTrue: [self removeProperty: #justDuplicated]! !

!NCAAConnectorMorph methodsFor: 'event handling' stamp: 'nk 1/14/2004 18:08'!
enterDraggingSegment
	"lastEventArguments has anEvent . anEvent position . vertexIndex"

	| args |
	args := eventHandler lastEventArguments.
	self comeToFront.
	self setProperty: #draggingSegmentStart
		toValue: (self localPointToGlobal: args first position).
	owner isWorldMorph ifFalse: [self jumpToWorld].
	self addDragShadow.
	args first hand newMouseFocus: self.
	self setProperty: #draggingSegmentIndex toValue: args third.
	self setProperty: #draggingSegmentVertices toValue: line vertices copy

	"constraints do: [ :ea | ea stopStepping ]."! !

!NCAAConnectorMorph methodsFor: 'event handling' stamp: 'nk 1/15/2004 11:09'!
enterEditing
	"lastArgs are { handle }"

	| args handle |
	args := eventHandler lastEventArguments.
	handle := args first.
	self comeToFront.
	owner isWorldMorph 
		ifFalse: 
			[self jumpToWorld ].
	self addDragShadow.
	self newCursor: Cursor crossHair.
	self setProperty: #editingStartPosition toValue: handle dragStartPoint.
	self setProperty: #editingHandle toValue: handle.
! !

!NCAAConnectorMorph methodsFor: 'event handling' stamp: 'nk 6/14/2004 15:11'!
enterHandlingDrop
	| potentialConnections nextConnection potentialTargets firstTarget |
	self log: #debug items: { 'enterHandlingDrop' }.
	
	"Find which ends we should potentially try to connect"
	potentialConnections := constraints select: [:ea | ea input isNil].
	potentialTargets := potentialConnections collect: [ :constraint |
		| localPoint targets |
		localPoint := self griddedPoint: constraint targetPoint.
		targets := self connectionTargetsAt: localPoint for: constraint.
		targets notEmpty 
			ifTrue: [ { (Sensor shiftPressed 
						ifFalse: [targets first ]
						ifTrue: [ targets last]). localPoint } ]
				ifFalse: [ { nil. nil } ]].

	self isWiring ifTrue: [
		potentialConnections size = 2 ifTrue: [
			firstTarget := potentialTargets findFirst: [ :arr | arr first notNil ].
			firstTarget = 0 ifTrue: [ firstTarget := 1 ].
			nextConnection := potentialConnections atWrap: firstTarget + 1.
			potentialConnections := potentialConnections copyFrom: firstTarget to: firstTarget.
			potentialTargets := potentialTargets copyFrom: firstTarget to: firstTarget
		]].

	potentialConnections with: potentialTargets do: 
			[:constraint :targetAndPoint | 
			| localPoint target |
			target := targetAndPoint first.
			localPoint := targetAndPoint second.
			target isNil
				ifFalse: [ constraint attachTo: target at: localPoint ]
				ifTrue: [ constraint connectToNothing; forceRedraw]].
	self
		changed;
		layoutChanged.

	(self isWiring and: [nextConnection notNil]) 
		ifTrue: 
			[| evt pos |
			evt := ActiveHand lastEvent.
			pos := self griddedPoint: (self globalPointToLocal: evt position).
			potentialConnections first lastTargetPoint: pos.
			nextConnection center: pos.
			self setProperty: #triggeredWireFinish toValue: pos.
			self trigger: #wireFinish
				withArguments: { 
						nextConnection.
						evt}].
! !

!NCAAConnectorMorph methodsFor: 'event handling' stamp: 'nk 1/14/2004 16:20'!
enterWiringEnd
	self log: #debug items: {'enterWiringEnd'}.
	self isWiring: true.
	self newCursor: Cursor crossHair! !

!NCAAConnectorMorph methodsFor: 'event handling' stamp: 'nk 1/14/2004 16:10'!
enterWiringEndOver
	self log: #debug items: {'enterWiringEndOver'}! !

!NCAAConnectorMorph methodsFor: 'event handling' stamp: 'nk 1/14/2004 16:20'!
enterWiringMorph
	self log: #debug items: {'enterWiringMorph'}.
	self isWiring: true.
	Cursor crossHair show! !

!NCAAConnectorMorph methodsFor: 'event handling' stamp: 'nk 1/16/2004 17:50'!
exitDraggingEnd
	| hand |

	self removeDragShadow.
	self stopMonitoringEnd.
	"First, detach any ends that were following the hand."
	constraints
		do: [:ea | ea input
				ifNotNilDo: [:m | m isHandMorph
						ifTrue: [hand := m.
							ea connectToNothing]]].
	hand
		ifNotNil: [hand releaseMouseFocus: self].
	self newCursor: Cursor normal.

	self log: #debug items: {'exitDraggingEnd'. self vertices}.! !

!NCAAConnectorMorph methodsFor: 'event handling' stamp: 'nk 1/17/2004 11:01'!
exitDraggingEndOver
	self log: #debug items: {'exitDraggingEndOver'. self vertices }.
! !

!NCAAConnectorMorph methodsFor: 'event handling' stamp: 'nk 1/14/2004 16:27'!
exitDraggingMorph
	self log: #debug items: {'exitDraggingMorph'}.
	self removeShortEndSegments.
	self stopMonitoringEnd.
	Cursor normal show! !

!NCAAConnectorMorph methodsFor: 'event handling' stamp: 'nk 1/15/2004 12:10'!
exitEditing
	self removeDragShadow.
	self newCursor: Cursor normal.
	self removeShortEndSegments.
	self
		valueOfProperty: #editingHandle
		ifPresentDo: [:handle | self isOrthogonal
				ifTrue: [handle vertices
						do: [:vert | self orthogonalizeVertex: vert]]].
	self removeProperty: #editingStartPosition.
	self removeProperty: #editingHandle.
	constraints
		do: [:ea | ea forceRedraw]! !

!NCAAConnectorMorph methodsFor: 'event handling' stamp: 'nk 1/16/2004 16:19'!
exitWiringEnd
	self log: #debug items: {'exitWiringEnd'. self vertices }.
	self isWiring: false.
	self newCursor: Cursor normal! !

!NCAAConnectorMorph methodsFor: 'event handling' stamp: 'nk 1/14/2004 16:10'!
exitWiringEndOver
	self log: #debug items: {'exitWiringEndOver'}.
! !

!NCAAConnectorMorph methodsFor: 'event handling' stamp: 'nk 1/14/2004 16:54'!
exitWiringMorph
	self log: #debug items: {'exitWiringMorph'}.
	self visible: true.
! !

!NCAAConnectorMorph methodsFor: 'drop shadows' stamp: 'nk 9/22/2003 17:33'!
expandFullBoundsForDragShadow: aRectangle
	"Return an expanded rectangle for an eventual drag shadow"
	| offset |
	offset := self dragShadowOffset.
	offset isZero ifTrue: [ ^aRectangle ].
	^aRectangle quickMerge: (aRectangle translateBy: offset)! !

!NCAAConnectorMorph methodsFor: 'appearance' stamp: 'nk 9/4/2004 16:49'!
fillStyle
	^line ifNotNil: [ line fillStyle ]
		ifNil: [ super fillStyle ]! !

!NCAAConnectorMorph methodsFor: 'appearance' stamp: 'nk 8/16/2003 13:03'!
fillStyle: aFillStyle
	^line ifNotNil: [ line fillStyle: aFillStyle ]
		ifNil: [ super fillStyle: aFillStyle ]! !

!NCAAConnectorMorph methodsFor: 'event handling' stamp: 'nk 8/15/2003 10:31'!
findEndNear: localPoint
	"Answer a constraint near localPoint, or nil.
	If more than one near, prefer disconnected one."
	| possible margin |
	possible := constraints detect: [ :ea | ea bounds containsPoint: localPoint ] ifNone: [].
	possible ifNotNil: [ ^possible ].
	margin := 20 min: ((constraints first position dist: constraints second position) / 3).
	possible := SortedCollection sortBlock: [ :a :b | a input isNil ].
	constraints
		with: (line arrowsContainPoint: localPoint)
		do: [ :constraint :flag | (flag or: [ (localPoint dist: constraint center) < margin ])
			ifTrue: [ possible add: constraint ] ].
	^possible at: 1 ifAbsent: [ ]! !

!NCAAConnectorMorph methodsFor: 'event handling' stamp: 'nk 8/16/2003 15:42'!
findEndToDetachNear: position
	"If there is a connected constraint near the given position, and I will allow its detachment, then answer that constraint.
	Otherwise answer nil."
	| constraint |
	constraint := self findEndNear: position.
	^constraint ifNotNil:  [ (self allowsDetachmentOfConstraint: constraint) ifTrue: [ constraint ] ifFalse: [ nil ]].
! !

!NCAAConnectorMorph methodsFor: 'accessing' stamp: 'nk 8/15/2003 10:36'!
firstVertex
	^line firstVertex! !

!NCAAConnectorMorph methodsFor: 'submorphs-add/remove' stamp: 'nk 2/28/2006 13:20'!
fixDegenerateLine
	"If I have two vertices, and both are connected to the same object, 
	and they're both attached to the same side of the object, or are only a 
	few pixels apart, 
	then add a couple more vertices and space them so I can be grabbed. 
	Two policy settings affect this behavior: 
	minimumVertexSpacing -- minimum distance between endpoints of 
	a 2-vertex connector 
	minimumSameTargetOutsetSpacing -- minimum distance to offset 
	vertices added for 2-vertex connectors connected to the same side of 
	the same object."
	| sourceMorph v1 side1 v2 side2 minVertexSpacing spacingToMorph midpoint vertexSpacing box newPoints rotate flip |
	flip := false.
	line vertices size = 2
		ifFalse: [^ self].
	sourceMorph := self sourceMorph.
	(sourceMorph isNil
			or: [sourceMorph ~~ self destinationMorph])
		ifTrue: [^ self].
	v1 := self firstVertex.
	side1 := sourceMorph bounds sideNearestTo: v1.
	v2 := self lastVertex.
	side2 := sourceMorph bounds sideNearestTo: v2.
	minVertexSpacing := self minimumVertexSpacing.
	spacingToMorph := self minimumSameTargetOutsetSpacing.
	midpoint := v1 + v2 / 2.
	vertexSpacing := v1 dist: v2.
	side1 = side2
		ifTrue: [box := 0 @ 0 extent: (vertexSpacing max: minVertexSpacing)
							@ spacingToMorph.
			box := box align: box topCenter with: midpoint.
			newPoints := box corners.
			side1 caseOf: {
				[#left] -> 
					[rotate := #right.
					flip := v1 y > v2 y].
				[#right] -> 
					[rotate := #left.
					flip := v1 y < v2 y].
				[#bottom] -> 
					[rotate := nil.
					flip := v1 x > v2 x].
				[#top] -> 
					[rotate := #pi.
					flip := v1 x < v2 x]}]
		ifFalse: [vertexSpacing < minVertexSpacing
				ifFalse: [^ self].
			newPoints := {midpoint + (0 @ minVertexSpacing). midpoint + (0 @ (minVertexSpacing + spacingToMorph)). midpoint + (minVertexSpacing + spacingToMorph @ (minVertexSpacing + spacingToMorph)). midpoint + (minVertexSpacing + spacingToMorph @ 0). midpoint + (minVertexSpacing @ 0)}.
			side1 , side2 caseOf: {
				['bottomright'] -> 
					[rotate := nil.
					flip := false].
				['righttop'] -> 
					[rotate := #left.
					flip := false].
				['topleft'] -> 
					[rotate := #pi.
					flip := false].
				['leftbottom'] -> 
					[rotate := #right.
					flip := false].
				['rightbottom'] -> 
					[rotate := nil.
					flip := true].
				['topright'] -> 
					[rotate := #left.
					flip := true].
				['lefttop'] -> 
					[rotate := #pi.
					flip := true].
				['bottomleft'] -> 
					[rotate := #right.
					flip := true]  } otherwise: [] ] .
	rotate
		ifNotNil: [newPoints := newPoints
						collect: [:p | p rotateBy: rotate centerAt: midpoint]].
	flip
		ifTrue: [newPoints := newPoints reversed ].
		
"	self log: #debug items: { v1. side1. v2. side2. newPoints. rotate. flip }."

	self setVertices: newPoints! !

!NCAAConnectorMorph methodsFor: 'initialization' stamp: 'nk 1/21/2004 12:55'!
fromMorph: startMorph toMorph: endMorph 
	| morphs |
	morphs :=  { 
				startMorph.
				endMorph}.
	constraints with: morphs
		do: [:c :m | c attachTo: m at: m center ].
	self setCorrectOwnerForConnections: morphs.
	self trigger: #createdConnected! !

!NCAAConnectorMorph methodsFor: 'initialization' stamp: 'nk 8/15/2003 10:26'!
fromMorph: startMorph toMorph: endMorph color: lineColor width: lineWidth

	self fromMorph: startMorph toMorph: endMorph.
	line lineBorderWidth: 0; lineColor: lineColor; lineWidth: lineWidth.
! !

!NCAAConnectorMorph methodsFor: 'drawing' stamp: 'nk 8/16/2003 13:03'!
fullDrawOn: aCanvas 
	"Draw the full Morphic structure on the given Canvas"

	self visible ifFalse: [^self].
	(aCanvas isVisible: self fullBounds) ifFalse: [^self].
	(self hasProperty: #errorOnDraw) ifTrue: [^self drawErrorOn: aCanvas].
	self hasDragShadow 
		ifTrue: [self drawDragShadowOn: aCanvas]
		ifFalse: [self hasDropShadow ifTrue: [self drawDropShadowOn: aCanvas]].
	(aCanvas isVisible: self bounds) ifTrue: [aCanvas drawMorph: self].
	self drawSubmorphsOn: aCanvas.
	self drawDropHighlightOn: aCanvas! !

!NCAAConnectorMorph methodsFor: 'event handling' stamp: 'nk 11/10/2003 23:25'!
gesture: aGesture
	self log: #debug items: { #gesture. aGesture }.
	^super gesture: aGesture
	"^self handleGesture: aGesture"
! !

!NCAAConnectorMorph methodsFor: 'graph edge view' stamp: 'nk 1/20/2004 16:57'!
graphModel
	^graphModel! !

!NCAAConnectorMorph methodsFor: 'submorphs-accessing' stamp: 'nk 8/18/2003 09:49'!
handles
	^self submorphsSatisfying: [ :m | m isCurveHandle ]! !

!NCAAConnectorMorph methodsFor: 'menus' stamp: 'dgd 8/26/2004 14:35'!
handlesShowingPhrase
	^ self showingHandles
		ifTrue:
			['hide handles' translated]
		ifFalse:
			['show handles' translated]! !

!NCAAConnectorMorph methodsFor: 'drop shadows' stamp: 'nk 8/15/2003 09:55'!
hasDragShadow
	^self hasProperty: #dragShadowOffset! !

!NCAAConnectorMorph methodsFor: 'menus' stamp: 'nk 1/23/2004 14:25'!
hasYellowButtonMenu	
	^true! !

!NCAAConnectorMorph methodsFor: 'geometry eToy' stamp: 'nk 9/4/2004 07:14'!
heading
	^line heading! !

!NCAAConnectorMorph methodsFor: 'geometry eToy' stamp: 'nk 9/4/2004 07:13'!
heading: newHeading
	"Set the receiver's heading (in eToy terms).
	Note that polygons never use flex shells."
	self rotationDegrees: newHeading.! !

!NCAAConnectorMorph methodsFor: 'drawing' stamp: 'nk 9/1/2004 17:48'!
imageForm: depth backgroundColor: aColor forRectangle: aRectangle
	^self withTransparentConstraintsDo: [ super imageForm: depth backgroundColor: aColor forRectangle: aRectangle ]! !

!NCAAConnectorMorph methodsFor: 'drawing' stamp: 'nk 9/1/2004 17:50'!
imageForm: depth forRectangle: rect

	^self withTransparentConstraintsDo: [ super imageForm: depth forRectangle: rect ]! !

!NCAAConnectorMorph methodsFor: 'initialization' stamp: 'nk 8/20/2003 10:07'!
initialize
	super initialize.
	super color: Color transparent.
	line := self class defaultLineClass vertices: { 0@0 . 50@0 } color: Color blue borderWidth: 2 borderColor: Color black.
	constraints := Array new: 2.
	constraints at: 1 put: ((self class defaultLineEndConstraintClass  new line: line firstVertex: true otherMorph: nil)
		center: 0@0;
		lastTargetPoint: 0@0;
		yourself).
	constraints at: 2 put: ((self class defaultLineEndConstraintClass new line: line firstVertex: false otherMorph: nil)
		center: line vertices second;
		lastTargetPoint: line vertices second;
		yourself).
	self addAllMorphs: constraints.
	self addMorphBack: line.
	self preferredEndConnection: #connectToNearestPointToCenter.
	self initializeFSM.
! !

!NCAAConnectorMorph methodsFor: 'naming' stamp: 'nk 1/27/2004 18:18'!
innocuousName 
	| newName |
	newName := super innocuousName.
	^(newName beginsWith: 'AA')
	 ifTrue: [ newName copyFrom: 3 to: newName size ]
		ifFalse: [ newName ]! !

!NCAAConnectorMorph methodsFor: 'private' stamp: 'nk 8/22/2004 06:04'!
inputOwnerChanged
	"A notification from an end that its input's owner has changed.
	Must by asynchronous to avoid recursion, as this is called from a step message."

	self setProperty: #inputOwnerChanged toValue: true! !

!NCAAConnectorMorph methodsFor: 'connection' stamp: 'nk 8/18/2003 11:12'!
insertVertexNextTo: constraint at: position
	"insert another vertex, before first or after last"
	constraint == constraints first
		ifTrue: [ line insertVertexBeforeStartAt: position ]
		ifFalse: [ line insertVertexAfterEndAt: position ].
	self changed.! !

!NCAAConnectorMorph methodsFor: 'accessing' stamp: 'nk 2/28/2006 11:12'!
interiorCurveSegmentAt: n put: aPoint
	^line interiorCurveSegmentAt: n  put: aPoint! !

!NCAAConnectorMorph methodsFor: 'accessing' stamp: 'nk 1/14/2004 21:38'!
interiorCurveVertexAt: n put: aPoint
	^line interiorCurveVertexAt: n  put: aPoint! !

!NCAAConnectorMorph methodsFor: 'geometry' stamp: 'nk 8/15/2003 09:54'!
intersectionWithLineSegmentFromCenterTo: aPoint 
	^line closestPointTo: aPoint! !

!NCAAConnectorMorph methodsFor: 'geometry' stamp: 'nk 8/15/2003 09:54'!
intersects: aRectangle
	"Answer whether aRectangle, which is in World coordinates, intersects any of my segments."
	^line ifNil: [ super intersects: aRectangle ]
		ifNotNil: [ line intersects: aRectangle ]! !

!NCAAConnectorMorph methodsFor: 'change reporting' stamp: 'nk 8/24/2003 20:25'!
invalidRect: aRectangle from: aMorph
	super invalidRect: (aMorph = line
		ifTrue: [ self expandFullBoundsForDragShadow: aRectangle ]
		ifFalse: [ aRectangle ]) from: aMorph! !

!NCAAConnectorMorph methodsFor: 'testing' stamp: 'nk 8/16/2003 13:03'!
isConnector
	^true! !

!NCAAConnectorMorph methodsFor: 'appearance' stamp: 'nk 8/15/2003 10:35'!
isCurve
	^line isCurve! !

!NCAAConnectorMorph methodsFor: 'orthogonality' stamp: 'nk 8/15/2003 10:34'!
isOrthogonal
	^self valueOfProperty: #orthogonal ifAbsent: [ false ]
! !

!NCAAConnectorMorph methodsFor: 'wiring' stamp: 'nk 8/15/2003 10:35'!
isWiring

	^self valueOfProperty: #wiring ifAbsent: [ false ]! !

!NCAAConnectorMorph methodsFor: 'wiring' stamp: 'nk 8/15/2003 10:35'!
isWiring: aBoolean

	aBoolean
		ifTrue: [ self setProperty: #wiring toValue: true ]
		ifFalse: [ self removeProperty: #wiring ]! !

!NCAAConnectorMorph methodsFor: 'structure' stamp: 'nk 1/21/2004 12:57'!
jumpToOwner: newOwner
	| transform |
	transform := (owner ifNil: [ self currentWorld ]) transformFromWorld composedWithLocal:
		newOwner transformFromWorld inverseTransformation.
	self transformedBy: transform.
	newOwner addMorphFront: self.
! !

!NCAAConnectorMorph methodsFor: 'structure' stamp: 'nk 8/15/2003 10:37'!
jumpToWorld
	"Change from my owner to the World, but maintain my original global position."
	self jumpToOwner: World! !

!NCAAConnectorMorph methodsFor: 'event handling' stamp: 'nk 10/21/2003 13:04'!
keyStroke: anEvent
	| labels |
	labels := self labels.
	labels isEmpty ifTrue: [ (self addUnwrappedLabel: (String with: anEvent keyCharacter) at: anEvent position) openForEditingFrom: anEvent ]
		ifFalse: [ labels size = 1 ifTrue: [ labels anyOne appendKeyStroke: anEvent ]].! !

!NCAAConnectorMorph methodsFor: 'connectors-labels' stamp: 'nk 4/30/2004 18:19'!
labelClass
	^(Preferences valueOfFlag: #useSmartLabels)
		ifTrue: [ NCSmartLabelMorph ]
		ifFalse: [ NCLabelMorph ]
! !

!NCAAConnectorMorph methodsFor: 'accessing' stamp: 'nk 8/15/2003 10:36'!
lastVertex
	^line lastVertex! !

!NCAAConnectorMorph methodsFor: 'layout' stamp: 'nk 8/24/2003 20:12'!
layoutChanged
	| oldBounds |
	oldBounds := self outerBounds.
	self needsOwnerSet: true.
	constraints do: [ :ea | ea forceRedraw ].
	bounds := self submorphBounds.
	fullBounds := self outerBounds.
	fullBounds ~= oldBounds ifTrue: [
		oldBounds ifNotNil: [ self invalidRect: oldBounds ].
		self invalidRect: fullBounds ].
	owner ifNotNil: [ owner layoutChanged ].! !

!NCAAConnectorMorph methodsFor: 'layout-properties' stamp: 'nk 9/4/2004 08:17'!
layoutPolicy: aLayoutPolicy
	"I do my own layout!!"
	^self! !

!NCAAConnectorMorph methodsFor: 'accessing' stamp: 'nk 8/15/2003 10:36'!
line
	^line! !

!NCAAConnectorMorph methodsFor: 'visual properties' stamp: 'nk 8/16/2003 15:21'!
lineColor
	^line lineColor! !

!NCAAConnectorMorph methodsFor: 'visual properties' stamp: 'nk 8/16/2003 15:21'!
lineColor: aColor 
	line lineColor: aColor! !

!NCAAConnectorMorph methodsFor: 'appearance' stamp: 'nk 8/15/2003 10:35'!
lineWidth
	^line ifNotNil: [ line lineWidth ] ifNil: [ 0 ]! !

!NCAAConnectorMorph methodsFor: 'appearance' stamp: 'nk 8/15/2003 10:35'!
lineWidth: aFloat
	constraints do: [ :ea | ea endShapeWidth: aFloat ].
	line
		ifNotNil: [ line lineWidth: aFloat ]
		ifNil: [ super borderWidth: aFloat rounded ].
! !

!NCAAConnectorMorph methodsFor: 'arrows' stamp: 'nk 8/25/2003 13:21'!
makeBackArrow
	self arrowNames: #(basicArrow noArrow)! !

!NCAAConnectorMorph methodsFor: 'arrows' stamp: 'nk 8/25/2003 13:22'!
makeBothArrows
	self arrowNames: #( basicArrow basicArrow)! !

!NCAAConnectorMorph methodsFor: 'menus' stamp: 'nk 8/15/2003 10:24'!
makeDashedOrSolidLinePhrase
	^line dashedBorder
		ifNil: [ 'make dashed line' ]
		ifNotNil: [ 'make solid line' ].
! !

!NCAAConnectorMorph methodsFor: 'arrows' stamp: 'nk 8/25/2003 13:21'!
makeForwardArrow
	self arrowNames: #( noArrow basicArrow)! !

!NCAAConnectorMorph methodsFor: 'arrows' stamp: 'nk 8/25/2003 13:22'!
makeNoArrows
	self arrowNames: #( noArrow noArrow)! !

!NCAAConnectorMorph methodsFor: 'menus' stamp: 'dgd 8/26/2004 14:45'!
makeSegmentedOrSmoothLinePhrase
	^ line isCurve
		ifTrue: ['make segmented line' translated]
		ifFalse: ['make smooth line' translated]! !

!NCAAConnectorMorph methodsFor: 'buttons' stamp: 'nk 9/1/2004 20:01'!
makerButtonImageOfColor: aColor
	^self withTransparentConstraintsDo: [ super makerButtonImageOfColor: aColor ]! !

!NCAAConnectorMorph methodsFor: 'geometry' stamp: 'nk 8/15/2003 10:36'!
midpoint
	^line midpoint! !

!NCAAConnectorMorph methodsFor: 'submorphs-add/remove' stamp: 'nk 1/16/2004 14:28'!
minimumSameTargetOutsetSpacing
	^40! !

!NCAAConnectorMorph methodsFor: 'submorphs-add/remove' stamp: 'nk 1/16/2004 14:28'!
minimumVertexSpacing
	^20! !

!NCAAConnectorMorph methodsFor: 'graph edge view' stamp: 'nk 3/29/2004 16:42'!
models
	^Array streamContents: [ :s |
		self modelOrNil ifNotNilDo: [ :m | s nextPut: m ].
		self graphModel ifNotNilDo: [ :m | s nextPut: m ]]
! !

!NCAAConnectorMorph methodsFor: 'event handling' stamp: 'nk 6/14/2004 14:59'!
monitorDestinationForEnds
	| targets cTargets |
	targets := IdentitySet new.
	constraints
		do: [:aConstraint | 
			| constraintInput | 
			constraintInput := aConstraint input.
			(constraintInput isNil
					or: [constraintInput isHandMorph])
				ifTrue: [cTargets := self
								connectionTargetsAt: (self visible
										ifTrue: [aConstraint center]
										ifFalse: [ActiveHand position])
								for: aConstraint.
					cTargets isEmpty
						ifFalse: [targets
								add: (ActiveHand lastEvent shiftPressed
										ifTrue: [cTargets last]
										ifFalse: [cTargets first])]]].
	self
		newCursor: (targets isEmpty
				ifTrue: [Cursor crossHair]
				ifFalse: [Cursor webLink]).
	self
		valueOfProperty: #highlightedTargets
		ifPresentDo: [:ht | (ht difference: targets)
				do: [:ea | ea highlightForConnection: false]].
	targets
		do: [:ea | ea highlightedForConnection
				ifFalse: [ea highlightForConnection: true]].
	self setProperty: #highlightedTargets toValue: targets.
	^ targets isEmpty not! !

!NCAAConnectorMorph methodsFor: 'event handling' stamp: 'nk 1/15/2004 11:42'!
mouseDownInConnected: anEvent 
	| constraint pos handle |
	pos := anEvent position.	"self griddedPoint:"

	self addedSubmorphs do: [ :h | (h containsPoint: pos)
		ifTrue: [ ^h handleEvent: anEvent  ]].

	constraint := self findEndToDetachNear: pos.
	constraint ifNotNil: [ ^self trigger: #mouseDownNearEnd withArguments: {  constraint. anEvent}].

	handle := line handleForEvent: anEvent.
	handle ifNotNil: [ 
		handle visible: false.
		handle when: #dragStart send: #trigger:with: to: self withArguments: {#dragStart. handle}.
		handle when: #dragStop send: #trigger:with: to: self withArguments: {#dragStop. handle}.
		self addMorphFront: handle.
		self trigger: #startEditing withArguments: { anEvent. handle }.
		anEvent sentTo: handle.	"causes #dragStart"
	 ].
! !

!NCAAConnectorMorph methodsFor: 'event handling' stamp: 'nk 10/26/2003 06:19'!
mouseDownInDisconnected: anEvent
	| constraint pos |
	constraint := self findEndNear: (pos := anEvent position).
	"self addedSubmorphs do: [ :h | (h containsPoint: pos)
		ifTrue: [ ^h processEvent: anEvent  ]]."

	constraint
		ifNil: [
			self straighten.
			anEvent hand grabMorph: self.	"will generate ownerChangedToHand event" ]
		ifNotNil: [ self trigger: #mouseDownNearEnd withArguments: { constraint . anEvent } ].
! !

!NCAAConnectorMorph methodsFor: 'event handling' stamp: 'nk 10/20/2003 18:35'!
mouseDownInDraggingEnd: anEvent
	| constraint |
	constraint := constraints detect: [ :ea | ea input == anEvent hand ] ifNone: [ ^self ].

	(anEvent shiftPressed or: [anEvent yellowButtonPressed])
		ifTrue: [
			self insertVertexNextTo: constraint at: (self griddedPoint: anEvent position).
			self setProperty: #startDraggingEvent toValue: (anEvent copy setHand: nil).
			self log: #debug items: { 'add '. self vertices }
		]! !

!NCAAConnectorMorph methodsFor: 'event handling' stamp: 'nk 10/26/2003 07:59'!
mouseDownInOneEndConnected: anEvent 
	| constraint pos |
	constraint := self findEndNear: (pos := anEvent position).
	self addedSubmorphs do: [ :h | (h containsPoint: pos)
		ifTrue: [ ^h handleEvent: anEvent  ]].

	constraint ifNotNil: 
			[(self allowsDetachmentOfConstraint: constraint) 
				ifTrue: 
					[constraint connectToNothing.
					^self trigger: #mouseDownNearEnd
						withArguments: { 
								constraint.
								anEvent}]
				ifFalse: [^self]].
	constraint := constraints detect: [:ea | ea input isNil].
	constraint ifNil: [self error: 'can''t happen']
		ifNotNil: 
			[constraint center: anEvent position.
			self trigger: #mouseDownNearEnd
				withArguments: { 
						constraint.
						anEvent}]! !

!NCAAConnectorMorph methodsFor: 'event handling' stamp: 'nk 10/20/2003 20:41'!
mouseUpInDraggingEnd: anEvent
	"Answer true if the mouse up means that we're done dragging"
	| pt dist |

	self valueOfProperty: #startDraggingEvent ifPresentDo: [ :started |
		anEvent timeStamp - started timeStamp < HandMorph doubleClickTime
			ifTrue: [ ^false ].
		"(anEvent position dist: started position) < 10 ifTrue: [ ^false ]."
		self removeProperty: #startDraggingEvent
	].

	pt := self valueOfProperty: #triggeredWireFinish ifAbsent: [ 0@0 ].
	dist := anEvent position dist: pt.
	dist < 10 ifTrue: [ ^false ].

	anEvent hand releaseMouseFocus: self.
	^true! !

!NCAAConnectorMorph methodsFor: 'layout' stamp: 'nk 8/15/2003 10:27'!
needsOwnerSet
	^(self valueOfProperty: #needsOwnerSet) ifNil: [ false ]! !

!NCAAConnectorMorph methodsFor: 'layout' stamp: 'nk 8/15/2003 10:27'!
needsOwnerSet: aBoolean
	self setProperty: #needsOwnerSet toValue: (aBoolean or: [ nil ])! !

!NCAAConnectorMorph methodsFor: 'connectors-queries' stamp: 'nk 8/15/2003 09:56'!
network
	"Answer a set of myself plus all of the connectors that are either connected directly to me or that I'm connected to (recursively)"
	| retval |
	retval := Set new.
	self addConnectedConnectorsTo: retval.
	^retval
! !

!NCAAConnectorMorph methodsFor: 'private' stamp: 'nk 8/21/2003 15:10'!
newCursor: aCursor
	Cursor currentCursor ~= aCursor
		ifTrue: [ Cursor currentCursor: aCursor ]! !

!NCAAConnectorMorph methodsFor: 'submorphs-accessing' stamp: 'nk 1/24/2004 19:19'!
noteNewOwner: aMorph 
	super noteNewOwner: aMorph.
	aMorph ifNil: [^self].
	self wantsGraphModel ifTrue: [
		aMorph graphContext ifNotNilDo: [ :gc | self ensureGraphModelInOwner: aMorph andContext:  gc ]].
	self needsOwnerSet: true.
	aMorph isHandMorph 
		ifTrue: [self trigger: #ownerChangedToHand with: aMorph]
		ifFalse: 
			[aMorph isPlayfieldLike 
				ifTrue: [self trigger: #ownerChangedToPasteUp with: aMorph]
				ifFalse: [self trigger: #ownerChanged with: aMorph]]! !

!NCAAConnectorMorph methodsFor: 'connectors-labels' stamp: 'nk 8/16/2003 13:03'!
nudgeForLabel: labelBounds
	"First, clear my connected morphs, then clear my line"
	| nudge |
	"self labels do: [ :ea |
		nudge := ea nudgeForLabel: labelBounds.
		nudge isZero ifFalse: [ ^nudge ]
	]."
	self validConnections do: [ :ea |
		nudge := ea nudgeForLabel: labelBounds.
		nudge isZero ifFalse: [ ^nudge ].
	].
	^line nudgeForLabel: labelBounds
! !

!NCAAConnectorMorph methodsFor: 'halos and balloon help' stamp: 'nk 8/16/2003 13:03'!
okayToResizeEasily
	"Answer whether it is appropriate for a sizing handle to be shown for the receiver"

	^ false! !

!NCAAConnectorMorph methodsFor: 'halos and balloon help' stamp: 'nk 8/16/2003 13:03'!
okayToRotateEasily
	"Answer whether it is appropriate for a rotation handle to be shown for the receiver"

	^ false! !

!NCAAConnectorMorph methodsFor: 'visual properties' stamp: 'nk 1/18/2004 16:48'!
openAConnectorPropertySheet

	ConnectorPropertiesMorph basicNew
		targetMorph: self;
		initialize;
		openNearTarget! !

!NCAAConnectorMorph methodsFor: 'menus' stamp: 'dgd 8/26/2004 14:35'!
orthogonalityPhrase
	^ self isOrthogonal
		ifTrue:
			['stop being orthogonal' translated]
		ifFalse:
			['be orthogonal' translated]
! !

!NCAAConnectorMorph methodsFor: 'orthogonality' stamp: 'nk 1/14/2004 17:41'!
orthogonalizeVertex: index
	| rect newPoint oldVertices |
	oldVertices := line vertices.
	(index <= 1 or: [ index >= line vertices size ]) ifTrue: [ ^self  ].
	rect := (oldVertices at: index - 1) corner: (oldVertices at: index + 1).
	newPoint := { rect bottomLeft . rect topRight } detectMin: [ :pt | pt dist: (oldVertices at: index) ].
	(oldVertices at: index) ~= newPoint ifTrue: [
	self setVertices: (oldVertices copyReplaceFrom: index to: index with: { self griddedPoint: newPoint } ) ]
! !

!NCAAConnectorMorph methodsFor: 'layout' stamp: 'nk 8/21/2003 16:31'!
outerBounds
	"Return the 'outer' bounds of the receiver, e.g., the bounds that need to be invalidated when the receiver changes."
	| box |
	box := self bounds.
	self hasDragShadow ifTrue:[box := self expandFullBoundsForDragShadow: box]
		ifFalse: [ self hasDropShadow ifTrue:[box := self expandFullBoundsForDropShadow: box]].
	^box! !

!NCAAConnectorMorph methodsFor: 'structure' stamp: 'nk 8/20/2003 17:32'!
ownerChanged

	self needsOwnerSet: true.
	super ownerChanged! !

!NCAAConnectorMorph methodsFor: 'event handling' stamp: 'nk 8/15/2003 10:31'!
pasteUpChanged: newPasteUp
	self needsOwnerSet: true.
	constraints do: [ :ea | ea forceRedraw ]
! !

!NCAAConnectorMorph methodsFor: 'connection' stamp: 'nk 8/15/2003 10:21'!
preferredEndConnection: anArrayOrSelector
	constraints do: [ :ea | ea preferredConnection: anArrayOrSelector ]! !

!NCAAConnectorMorph methodsFor: 'graph edge view' stamp: 'nk 1/20/2004 19:31'!
pvtGraphModel: anObject 
	graphModel := anObject.
! !

!NCAAConnectorMorph methodsFor: 'menus' stamp: 'nk 8/31/2004 08:09'!
reasonableBitmapFillForms
	| reasonableForms |
	reasonableForms := OrderedCollection new.
	reasonableForms 
		addAll: (BitmapFillStyle allSubInstances collect: [:f | f form]).
	reasonableForms addAll: Imports default images.
	^reasonableForms! !

!NCAAConnectorMorph methodsFor: 'geometry' stamp: 'nk 5/2/2004 14:29'!
reduceVertices
	| oldVerts |
	(oldVerts := line vertices size) > 2 ifTrue: [ 
		(line reduceVertices ~= oldVerts) ifTrue: [ constraints do: [ :ea | ea step ] ]
	].
! !

!NCAAConnectorMorph methodsFor: 'geometry' stamp: 'nk 12/27/2003 11:22'!
referencePosition
	^self valueOfProperty: #referencePosition ifAbsent: [ line referencePosition ]! !

!NCAAConnectorMorph methodsFor: 'dropping/grabbing' stamp: 'nk 12/28/2003 13:05'!
rejectDropMorphEvent: evt
	"Make sure that I restore the cursor"
	self isWiring ifTrue: [ Cursor normal show ].
	super rejectDropMorphEvent: evt.
! !

!NCAAConnectorMorph methodsFor: 'connectors-labels' stamp: 'nk 8/16/2003 13:03'!
relocateLabelFrom: aRectangle
	"Given a prospective label location, answer a new rectangle which doesn't overlap me"
	| newBounds |
	newBounds := aRectangle.
	self validConnections do: [ :ea |
		newBounds := ea relocateLabelFrom: newBounds.
	].
	newBounds := line relocateLabelFrom: newBounds.
	^ newBounds
! !

!NCAAConnectorMorph methodsFor: 'drop shadows' stamp: 'nk 9/22/2003 17:34'!
removeDragShadow
	"I'm starting to be dragged but I'm not in the hand yet.
	So use the hand's drop shadow information to set my own drop shadow"

	self changed.
	self
		removeProperty: #dragShadowOffset;
		removeProperty: #dragShadowColor.
	self layoutChanged.
	self changed.! !

!NCAAConnectorMorph methodsFor: 'submorphs-add/remove' stamp: 'nk 8/18/2003 09:50'!
removeHandles
	self deleteMorphs: self handles.
	self removeProperty: #showingHandles! !

!NCAAConnectorMorph methodsFor: 'geometry' stamp: 'nk 8/28/2003 14:09'!
removeShortEndSegments
	"First remove overlaps"
	| inputs vertices |
	inputs := self connectedMorphs.
	inputs first ifNotNilDo: [ :m |
		(vertices := self vertices) size > 2 ifTrue: [
			((m fullContainsPoint: vertices second)
				and: [ (m fullContainsPoint: vertices third) not ])
					ifTrue: [ self setVertices: (vertices copyWithout: vertices second) ]]].
	inputs last ifNotNilDo: [ :m |
		(vertices := self vertices) size > 2 ifTrue: [
			((m fullContainsPoint: vertices nextToLast)
				and: [ (m fullContainsPoint: (vertices at: vertices size - 2)) not ])
					ifTrue: [ self setVertices: (vertices copyWithout: vertices nextToLast) ]]].
	line removeShortEndSegmentsStepping: [ constraints do: [ :ea | ea step ]].
! !

!NCAAConnectorMorph methodsFor: 'geometry' stamp: 'nk 9/4/2004 08:02'!
rotationDegrees: degrees
	self changed.
	line rotationDegrees: degrees.
	constraints first center: line vertices first.
	constraints last center: line vertices last.
	self changed.! !

!NCAAConnectorMorph methodsFor: 'geometry eToy' stamp: 'nk 9/4/2004 07:44'!
scale
	^line scale! !

!NCAAConnectorMorph methodsFor: 'geometry eToy' stamp: 'nk 9/4/2004 11:12'!
scale: scaleFactor
	super scale: scaleFactor.
	line scaleFactor: scaleFactor.
	constraints first center: line vertices first.
	constraints last center: line vertices last.
! !

!NCAAConnectorMorph methodsFor: 'geometry eToy' stamp: 'nk 9/4/2004 11:12'!
scaleFactor
	^line scaleFactor! !

!NCAAConnectorMorph methodsFor: 'geometry' stamp: 'nk 8/15/2003 09:52'!
selectionSlop
	"Answer the distance from the line centerline that I will accept as a valid selection point"
	"This is set class-wide; the accessor is on the instance side so that subclasses can override on a per-instance basis"
	^4! !

!NCAAConnectorMorph methodsFor: 'protected-policy' stamp: 'nk 8/18/2003 17:04'!
setCorrectOwnerForConnections: connections
	| numberOfConnections owners1 owners2 n innermostOwner |

	self needsOwnerSet: false.

	numberOfConnections := connections size.
	numberOfConnections = 0 ifTrue: [ ^self ].

	numberOfConnections = 1 ifTrue: [ innermostOwner := connections first pasteUpMorphOrWorld ].

	"Make sure that I am owned by the innermost of my connections' common owners"
	numberOfConnections = 2 ifTrue: [

	innermostOwner := connections first pasteUpMorphOrWorld.
	innermostOwner ifNil: [ ^self ].

	(connections second pasteUpMorphOrWorld ~~ innermostOwner) ifTrue: [

		owners1 := innermostOwner withAllOwners select: [ :ea | ea isPlayfieldLike ].
		owners2 := connections second allOwners  select: [ :ea | ea isPlayfieldLike ] .

		owners1 size + (owners2 size) == 3 ifTrue: [ innermostOwner := owners1 last ]
		ifFalse: [
			n := (owners1 size min: owners2 size) - 1.
			(owners1 size to: owners1 size - n by: -1)
				with: (owners2 size to: owners2 size - n by: -1)
					do: [ :i1 :i2 | | o1 o2 |
				o1 := (owners1 at: i1).
				o2 := (owners2 at: i2).
				o1 == o2 ifTrue: [ innermostOwner := o1 ]
			].
		].
	].
	].

	innermostOwner ifNotNil: [
		self owner == innermostOwner
			ifFalse: [
				^self jumpToOwner: innermostOwner.
			]
	].

	connections do: [ :conn | | hisOwner |
		hisOwner := conn orOwnerSuchThat: [ :ea | ea owner == owner ].
		hisOwner ifNotNil: [
			(line bounds intersects: (hisOwner bounds insetBy: 1))
				ifTrue: [ "owner ensureMorph: hisOwner behind: self" ]
		]
	]! !

!NCAAConnectorMorph methodsFor: 'menus' stamp: 'nk 8/24/2003 15:20'!
setGlyph: anArrowGlyph forEnd: endIndex
	| newArrows |
	newArrows := line arrows copy.
	newArrows at: endIndex put: anArrowGlyph.
	line arrows: newArrows! !

!NCAAConnectorMorph methodsFor: 'geometry' stamp: 'nk 1/17/2004 10:45'!
setVertices: newVertices 
	| verts |
	verts := Array streamContents: [ :s |
		s nextPut: (newVertices at: 1 ifAbsent: [ self halt ]).
		newVertices overlappingPairsDo: [ :v1 :v2 |
			v1 rounded = v2 rounded ifFalse: [ s nextPut: v2 rounded ]]].

	verts size >= 2 ifTrue: [
		self changed.
		line setVertices: verts.
		constraints first center: verts first.
		constraints second center: verts last.
		self changed; layoutChanged ].! !

!NCAAConnectorMorph methodsFor: 'menus' stamp: 'nk 8/15/2003 10:24'!
showOrHideHandles
	self showingHandles
		ifTrue:	[self removeHandles]
		ifFalse:	[self addHandles]! !

!NCAAConnectorMorph methodsFor: 'submorphs-add/remove' stamp: 'nk 8/15/2003 10:34'!
showingHandles
	^self valueOfProperty: #showingHandles ifAbsent: [ false ]! !

!NCAAConnectorMorph methodsFor: 'arrows' stamp: 'nk 8/26/2003 14:42'!
sourceArrowName
	| arrows |
	arrows := line arrows.
	^self class nameForArrowGlyph: arrows first
! !

!NCAAConnectorMorph methodsFor: 'arrows' stamp: 'nk 8/26/2003 10:55'!
sourceArrowName: aSymbol
	| arrows |
	arrows := line arrows.
	arrows at: 1 put: (self class arrowDictionary at: aSymbol ifAbsent: [ ^nil ]).
	line arrows: arrows! !

!NCAAConnectorMorph methodsFor: 'connectors-queries' stamp: 'nk 8/15/2003 09:56'!
sourceMorph
	"Answer the morph connected to my first vertex, if any"
	^constraints first input! !

!NCAAConnectorMorph methodsFor: 'connectors-queries' stamp: 'nk 8/15/2003 09:56'!
startConnection
	"Answer my beginning connection or nil"
	^constraints first input! !

!NCAAConnectorMorph methodsFor: 'connectors-queries' stamp: 'nk 8/15/2003 09:56'!
startConstraint
	"Answer my beginning constraint"
	^constraints first
! !

!NCAAConnectorMorph methodsFor: 'wiring' stamp: 'nk 1/14/2004 17:05'!
startWiring
	"NCAAConnectorMorph new startWiring"
	| hand |
	hand := World currentHand.
	self isWiring: true.
	self visible: false.
	hand attachMorph: self.
	self position: self position + hand position - constraints first center.
	self trigger: #wireMorph withArguments: { hand lastEvent }.
! !

!NCAAConnectorMorph methodsFor: 'wiring' stamp: 'nk 1/14/2004 15:48'!
startWiringFrom: aMorph
	"NCConnectorMorph new startWiringFrom: (Morph new openInWorld)"

	| hand |
	hand := World currentHand.
	self isWiring: true.
	self connectStartTo: aMorph at: hand position.
	constraints second position: hand position.
	self openInWorld.
	self trigger: #wireFrom withArguments: { constraints last . hand lastEvent }.
! !

!NCAAConnectorMorph methodsFor: 'wiring' stamp: 'nk 11/4/2003 13:51'!
startWiringFrom: aMorph event: anEvent
	"NCAAConnectorMorph new startWiringFrom: (Morph new openInWorld)"

	| hand |
	self isWiring: true.
	hand := anEvent hand.
	self connectStartTo: aMorph at: anEvent position.
	constraints second position: hand cursorPoint.
	self openInWorld.
	self trigger: #wireFrom withArguments: { constraints last . hand lastEvent }.
! !

!NCAAConnectorMorph methodsFor: 'wiring' stamp: 'nk 11/4/2003 13:42'!
startWiringFromEvent: anEvent
	"NCAAConnectorMorph new startWiringFromEvent: ActiveHand lastMouseEvent"

	self isWiring: true.
	self position: anEvent position.
	constraints first position: anEvent position; input: nil.
	constraints second position: anEvent position.
	self openInWorld.
	self trigger: #wireFrom withArguments: { constraints last . anEvent }.
! !

!NCAAConnectorMorph methodsFor: 'wiring' stamp: 'nk 11/4/2003 13:35'!
startWiringFromPoint: aPoint
	"NCAAConnectorMorph new startWiringFromPoint: 200@200"

	| hand |
	hand := World currentHand.
	self isWiring: true.
	constraints first position: aPoint; input: nil.
	constraints second position: hand lastEvent cursorPoint.
	self openInWorld.
	self trigger: #wireFrom withArguments: { constraints last . hand lastEvent }.
! !

!NCAAConnectorMorph methodsFor: 'wiring' stamp: 'nk 8/15/2003 10:35'!
startWiringLabeled
	self startWiringLabeled: 'label'! !

!NCAAConnectorMorph methodsFor: 'wiring' stamp: 'nk 10/19/2003 21:43'!
startWiringLabeled: labelString
	| label |
	self startWiring.
	label := (self addUnwrappedLabel: labelString at: self midpoint) visible: false.
	label offset: 0@10.
	label textWrap: false.
	self setProperty: #wiringLabel toValue: label.
! !

!NCAAConnectorMorph methodsFor: 'buttonActions' stamp: 'nk 7/17/2004 17:28'!
startWiringRandomly
	self delete.
	^(self class random)
		startWiring;
		yourself! !

!NCAAConnectorMorph methodsFor: 'wiring' stamp: 'nk 8/15/2003 10:35'!
startWiringVisibly
	"NCConnectorMorph new startWiringVisibly"

	self startWiring.
	self visible: true.
! !

!NCAAConnectorMorph methodsFor: 'stepping and presenter' stamp: 'nk 1/15/2004 11:28'!
stepCleanup
	"backwards compatibility."
	self stopSteppingSelector: #stepCleanup! !

!NCAAConnectorMorph methodsFor: 'stepping and presenter' stamp: 'nk 11/1/2004 10:55'!
stepConnected
	"Called upon step in Connected state."

	self stepEditing.
	self isOrthogonal
		ifTrue: [
			line vertices size > 2
				ifTrue: [self orthogonalizeVertex: 2].
			line vertices size > 3
				ifTrue: [self orthogonalizeVertex: line vertices size - 1].
			line vertices size = 2 ifTrue: [ | rect |
				rect := Rectangle origin: line vertices first corner: line vertices second.
				(rect width ~= 0 and: [rect height ~= 0 ]) ifTrue: [ self setVertices: (rect corners copyFrom: 1 to: 3) ].
			].
	].
	self reduceVertices.
	self fixDegenerateLine.

	(self hasProperty: #inputOwnerChanged)
		ifTrue: [ self removeProperty: #inputOwnerChanged.
			WorldState addDeferredUIMessage: [ self triggerEtoyEvent: #attachmentOwnerChanged ]].! !

!NCAAConnectorMorph methodsFor: 'stepping and presenter' stamp: 'nk 1/15/2004 11:48'!
stepEditing
	"Called upon step in Editing state."
	| numberOfConnections connections |
	owner
		ifNil: [^ self].
	"don't bother doing anything if unowned"
	self needsOwnerSet
		ifFalse: [^ self].
	connections := self validConnections.
	numberOfConnections := connections size.
	owner isHandMorph
		ifFalse: [self setCorrectOwnerForConnections: connections].
	numberOfConnections == 0
		ifTrue: [^ self trigger: #bothEndsGone].
	numberOfConnections == 1
		ifTrue: [^ self trigger: #endGone].
! !

!NCAAConnectorMorph methodsFor: 'stepping and presenter' stamp: 'nk 11/1/2004 10:56'!
stepOneEndConnected
	"Called upon step in OneEndConnected state."

	(self hasProperty: #inputOwnerChanged)
		ifTrue: [ self removeProperty: #inputOwnerChanged.
			WorldState addDeferredUIMessage: [ self triggerEtoyEvent: #attachmentOwnerChanged ]].! !

!NCAAConnectorMorph methodsFor: 'stepping and presenter' stamp: 'nk 5/2/2004 22:07'!
stepTime
	^200! !

!NCAAConnectorMorph methodsFor: 'event handling' stamp: 'nk 1/14/2004 16:26'!
stopMonitoringEnd
	self
		valueOfProperty: #highlightedTargets
		ifPresentDo: [:targets | 
			targets
				do: [:ea | ea highlightForConnection: false].
			self removeProperty: #highlightedTargets].
! !

!NCAAConnectorMorph methodsFor: 'appearance' stamp: 'nk 5/13/2004 20:57'!
straighten
	"Remove all my intermediate vertices."
	constraints first stopSteppingSelector: #stepAt:.
	constraints second stopSteppingSelector: #stepAt:.
	[line straighten.
	self fixDegenerateLine]
		ensure: [constraints first
				startStepping: #stepAt:
				at: Time millisecondClockValue
				arguments: nil
				stepTime: nil;
				 forceRedraw;
				 step.
			constraints second
				startStepping: #stepAt:
				at: Time millisecondClockValue
				arguments: nil
				stepTime: nil;
				 forceRedraw;
				 step]! !

!NCAAConnectorMorph methodsFor: 'connectors-queries' stamp: 'nk 8/15/2003 09:56'!
subNet
	"Answer a set of myself plus all of the connectors that are either connected directly to me or that I'm connected to (recursively)"
	| retval |
	retval := Set new.
	self addConnectedConnectorsTo: retval.
	^retval
! !

!NCAAConnectorMorph methodsFor: 'layout' stamp: 'nk 10/25/2003 17:30'!
submorphBounds
	| rect |
	^line
		ifNotNil: [ rect := (line bounds outsetBy: self selectionSlop).
			submorphs do: [ :m | m == line ifFalse: [ rect := rect quickMerge: m fullBounds ]].
			rect ]
		ifNil: [ super bounds ]! !

!NCAAConnectorMorph methodsFor: 'connectors-queries' stamp: 'nk 8/15/2003 09:56'!
subnet
	"Answer a set of myself plus all of the connectors that are either connected directly to me or that I'm connected to (recursively)"
	| retval |
	retval := Set new.
	self addConnectedConnectorsTo: retval.
	^retval
! !

!NCAAConnectorMorph methodsFor: 'menus' stamp: 'nk 8/15/2003 10:24'!
toggleDashedLine
	self dashedLine: line dashedBorder isNil.! !

!NCAAConnectorMorph methodsFor: 'submorphs-add/remove' stamp: 'nk 8/15/2003 10:34'!
toggleHandles
	self showingHandles
		ifTrue: [ self removeHandles ]
		ifFalse: [ self addHandles ]! !

!NCAAConnectorMorph methodsFor: 'menus' stamp: 'nk 8/15/2003 10:24'!
toggleOrthogonality
	^ self beOrthogonal: self isOrthogonal not
! !

!NCAAConnectorMorph methodsFor: 'menus' stamp: 'nk 8/15/2003 10:24'!
toggleSmoothing
	line toggleSmoothing
! !

!NCAAConnectorMorph methodsFor: 'accessing' stamp: 'nk 8/15/2003 10:36'!
totalLength
	^line totalLength! !

!NCAAConnectorMorph methodsFor: 'rotate scale and flex' stamp: 'nk 8/16/2003 13:03'!
transformedBy: aTransform
	"Since I don't really add a flex shell, just behave like a PolygonMorph"
	line transformedBy: aTransform.
	"constraints do: [ :ea | ea transformedBy: aTransform ]."
	^self.! !

!NCAAConnectorMorph methodsFor: 'scripting' stamp: 'nk 9/4/2004 11:47'!
understandsBorderVocabulary
	"Replace the 'isKindOf: BorderedMorph' so that (for instance) Connectors can have their border vocabulary visible in viewers."
	^true! !

!NCAAConnectorMorph methodsFor: 'copying' stamp: 'nk 9/27/2003 15:26'!
usableSiblingInstance
	^(super usableSiblingInstance) detach! !

!NCAAConnectorMorph methodsFor: 'copying' stamp: 'nk 6/14/2004 16:37'!
usableSiblingInstance2
	^(super usableSiblingInstance2) detach! !

!NCAAConnectorMorph methodsFor: 'connectors-queries' stamp: 'nk 8/15/2003 09:56'!
validConnections
	"Answer an Array with all of my valid connections. This will be from 0 to 2 long."
	| f s |
	f := constraints at: 1.
	s := constraints at: 2.
	^f input ifNotNil: [
		s input ifNotNil: [ { f input . s input } ]
		ifNil: [ { f input } ]
	]
	ifNil: [
		constraints second input ifNotNil: [ { s input } ]
		ifNil: [ { } ]
	]! !

!NCAAConnectorMorph methodsFor: 'accessing' stamp: 'nk 8/21/2003 16:35'!
vertexAt: n
	^line vertexAt: n! !

!NCAAConnectorMorph methodsFor: 'accessing' stamp: 'nk 8/21/2003 16:35'!
vertexAt: n put: aPoint
	^line vertexAt: n  put: aPoint! !

!NCAAConnectorMorph methodsFor: 'accessing' stamp: 'nk 8/18/2003 11:44'!
vertices
	^line vertices! !

!NCAAConnectorMorph methodsFor: 'scripting' stamp: 'nk 9/10/2004 11:36'!
wantsConnectorVocabulary
	^true! !

!NCAAConnectorMorph methodsFor: 'scripting' stamp: 'nk 8/21/2004 09:27'!
wantsEmbeddingsVocabulary
	^self allowsConnectionToSubmorphs or: [ self wantsConnectionWhenEmbedded ]! !

!NCAAConnectorMorph methodsFor: 'graph edge view' stamp: 'nk 1/24/2004 19:19'!
wantsGraphModel
	^true! !

!NCAAConnectorMorph methodsFor: 'stepping and presenter' stamp: 'nk 1/14/2004 15:49'!
wantsSteps
	^true! !

!NCAAConnectorMorph methodsFor: 'protected-policy' stamp: 'nk 9/4/2004 17:35'!
wantsToAttachEnd: index toMorph: aMorph
	"Answer true if I want my end with index 'index' (1 or 2) to attach to the given Morph."
	^(aMorph inPartsBin or: [ aMorph owner isPartsBin ]) not! !

!NCAAConnectorMorph methodsFor: 'protected-policy' stamp: 'nk 8/15/2003 10:21'!
wantsToAttachEndsToMorphs: morphs
	"Answer true if I am willing to connect to the (1 or 2) morphs given.
	Either can be nil."

	morphs isEmpty ifTrue: [ ^false ].

	morphs withIndexDo: [ :end :i |
		i > 2 ifTrue: [ ^false ].
		(self wantsToAttachEnd: i toMorph: end)
			ifFalse: [ ^false ].
	].
	^true.! !

!NCAAConnectorMorph methodsFor: 'protected-policy' stamp: 'nk 8/15/2003 10:21'!
wantsToDetachEnd: index fromMorph: aMorph
	"Answer true if I want my end with index 'index' (1 or 2) to be detached from the given Morph."
	^aMorph isNil or: [ aMorph allowsDetachmentFromEnd: index ofConnector: self ]! !

!NCAAConnectorMorph methodsFor: 'wiring' stamp: 'nk 1/14/2004 15:52'!
wireFreeEnd
	"Answer true if we have a free end and it has been attached to the Hand"
	self currentState ~~ #OneEndConnected ifTrue: [ ^false ].
	self trigger: #wireFrom withArguments: { (constraints detect: [ :ea | ea input isNil ]). ActiveHand lastEvent }.
	^true! !

!NCAAConnectorMorph methodsFor: 'drawing' stamp: 'nk 8/15/2003 10:34'!
withTransparentConstraintsDo: aBlock
	"Make my constraints transparent and then evaluate the block."
	| oldColors retval |
	oldColors := constraints collect: [ :ea | ea fillStyle ].
	constraints do: [ :ea | ea fillStyle: Color transparent; computeLineAttachmentPoint; forceRedraw ].
	retval := aBlock ensure: [
		constraints with: oldColors do: [ :constraint :ccolor | constraint fillStyle: ccolor ]].
	^retval! !

!NCAAConnectorMorph methodsFor: 'event handling' stamp: 'nk 8/23/2003 15:47'!
yellowButtonDown: evt onEnd: aConstraintOrNil 
	| aMenu |
	aMenu := MenuMorph new defaultTarget: self.
	aMenu addTitle: self externalName.
	self addYellowButtonMenuItemsTo: aMenu event: evt.

	aConstraintOrNil ifNotNil: [ | subMenu |
		subMenu := MenuMorph new defaultTarget: aConstraintOrNil.
		aMenu addLine.
		aMenu add: 'end...' subMenu: subMenu.
		self addEndItemsFor: aConstraintOrNil to: aMenu event: evt.
		aConstraintOrNil addYellowButtonMenuItemsTo: subMenu event: evt.
			#('delete' 'resize' 'properties...' )
				do: [:ea | (subMenu itemWithWording: ea)
						ifNotNilDo: [:item | item target = aConstraintOrNil
								ifTrue: [item delete]]]
	].

	aMenu popUpInWorld: World! !

NCAAConnectorMorph subclass: #NCSchematicConnectorMorph
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Connectors-Base'!

!NCSchematicConnectorMorph commentStamp: '<historical>' prior: 0!
This is a Connector that automatically sets its connection mode as appropriate for use in electronic schematic diagrams.

It also automatically decorates its ends that are attached to other Connectors with junction dots.
!

!NCSchematicConnectorMorph class methodsFor: 'parts bin' stamp: 'nk 6/30/2004 14:06'!
descriptionForPartsBin
	^ self partName:	'Schematic Connector'
		categories:		#('Connectors')
		documentation:	'A Connector whose endpoints automatically become dots when connected.'
! !

!NCSchematicConnectorMorph methodsFor: 'initialization' stamp: 'nk 12/27/2003 12:35'!
initialize
	super initialize.
	self preferredEndConnection: nil.! !

!NCSchematicConnectorMorph methodsFor: 'stepping and presenter' stamp: 'nk 6/30/2004 14:13'!
step
	"Make a junction dot if I'm tied to another Connector,
	otherwise get rid of my end decorations"
	| newArrowNames |
	super step.
	newArrowNames := constraints collect: [ :const |
		(const input notNil and: [ const input isConnector ])
			ifTrue: [ const connectToNearestOrthogonalPoint.
				const preferredConnection: #(#connectToNearestOrthogonalPoint).
				 #junctionDot ]
			ifFalse: [ const preferredConnection: nil.
				#noArrow  ] ].
	self arrowNames: newArrowNames.
! !

!NCSchematicConnectorMorph methodsFor: 'stepping and presenter' stamp: 'nk 12/27/2003 12:38'!
wantsSteps
	^true! !

!NCSchematicConnectorMorph methodsFor: 'event handling' stamp: 'nk 7/1/2002 17:39'!
yellowButtonDown: evt onEnd: aConstraintOrNil 
	| aMenu |
	aMenu := MenuMorph new defaultTarget: self.
	aMenu addTitle: self externalName.
	self addYellowButtonMenuItemsTo: aMenu event: evt.

	aConstraintOrNil ifNotNil: [ | subMenu |
		subMenu := MenuMorph new defaultTarget: aConstraintOrNil.
		aMenu addLine.
		aMenu add: 'end...' subMenu: subMenu.
		aConstraintOrNil addYellowButtonMenuItemsTo: subMenu event: evt.
			#('delete' 'resize' 'properties...' )
				do: [:ea | (subMenu itemWithWording: ea)
						ifNotNilDo: [:item | item target = aConstraintOrNil
								ifTrue: [item delete]]].
		subMenu deleteMorphs: (subMenu submorphs copyFrom: 3 to: subMenu submorphs size).
	].

	aMenu popUpInWorld: World
! !

!Preferences class methodsFor: '*connectors-preferences' stamp: 'nk 12/28/2003 13:42'!
chooseConnectorsLabelFont
	self
		chooseFontWithPrompt: 'Connectors label font'
		andSendTo: self
		withSelector: #setConnectorsLabelFontTo:! !

!Preferences class methodsFor: '*connectors-preferences' stamp: 'nk 12/28/2003 13:44'!
connectorsLabelFont
	"Answer the font to be used to label Connectors"
	^ Parameters
		at: #connectorsLabelFont
		ifAbsent: [Parameters at: #connectorsLabelFont put: TextStyle defaultFont]! !

!Preferences class methodsFor: '*connectors-fonts-override' stamp: 'nk 4/30/2004 18:31'!
presentMvcFontConfigurationMenu
	| aMenu result |
	aMenu := CustomMenu new.
	aMenu title: 'Standard System Fonts'.
	aMenu add: 'default text font...' translated action: #chooseSystemFont.
	aMenu add: 'list font...' translated action: #chooseListFont.
	aMenu add: 'flaps font...' translated action: #chooseFlapsFont.
	aMenu add: 'menu font...' translated action: #chooseMenuFont.
	aMenu add: 'window-title font...' translated action: #chooseWindowTitleFont.
	aMenu add: 'balloon-help font...' translated action: #chooseBalloonFont.
	aMenu add: 'code font...' translated action: 	#chooseCodeFont.
	aMenu add: 'Connectors label font...' translated action: #chooseConnectorsLabelFont. 
	aMenu addLine.
	aMenu add: 'restore default font choices' action: #restoreDefaultFonts.

	(result := aMenu startUp) ifNotNil:
		[self perform: result]
! !

!Preferences class methodsFor: '*connectors-preferences' stamp: 'nk 12/28/2003 13:43'!
setConnectorsLabelFontTo: aFont 
	Parameters at: #connectorsLabelFont put: aFont.
! !

!BookMorph methodsFor: '*connectors-attachments-nk' stamp: 'nk 6/13/2002 09:33'!
isConnectable
	"Answer whether I will accept connections from ConnectorMorphs"
	^false! !

SmalltalkEditor subclass: #NCTextMorphEditor
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Connectors-Text-Base'!

!NCTextMorphEditor methodsFor: 'menu commands' stamp: 'nk 6/24/2004 13:28'!
raiseContextMenu: characterStream 
	(morph orOwnerSuchThat: [ :ea | ea hasYellowButtonMenu ])
		ifNotNilDo: [ :m | m yellowButtonActivity: ActiveEvent shiftPressed ].
	^true
! !

SymbolListType subclass: #ConnectorArrowNameType
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Connectors-EToys'!

!ConnectorArrowNameType class methodsFor: 'as yet unclassified' stamp: 'nk 8/26/2003 08:46'!
initialize
	Vocabulary embraceAddedTypeVocabularies ! !

!ConnectorArrowNameType class methodsFor: 'as yet unclassified' stamp: 'nk 8/26/2003 14:30'!
updateNames
	Vocabulary arrowNamesVocabulary.
	SymbolListTile updateAllTilesForVocabularyNamed: #ArrowNames! !

!ConnectorArrowNameType methodsFor: 'queries' stamp: 'nk 8/26/2003 08:47'!
choices
	"Answer an alphabetized list of known user custom event selectors"

	| choices |
	choices := NCAAConnectorMorph arrowDictionary keys asSortedCollection.
	^choices isEmpty ifTrue: [ #('no arrows') ] ifFalse: [ choices ]! !

!ConnectorArrowNameType methodsFor: 'initialization' stamp: 'nk 6/30/2004 15:29'!
initialize
	"Initialize the ArrowNames vocabulary"

	super initialize.
	self vocabularyName: #ArrowNames.
	NCAAConnectorMorph arrowDictionary addDependent: self.
! !

!ConnectorArrowNameType methodsFor: 'initialization' stamp: 'nk 6/30/2004 15:31'!
update: anAspect
	anAspect == #arrowsChanged
		ifTrue: [ self class updateNames ].
	super update: anAspect.
! !

Object subclass: #NCCurve
	instanceVariableNames: 'contours bounds scale offset'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Connectors-Lines and Curves'!

!NCCurve class methodsFor: 'arrow forms' stamp: 'nk 2/3/2004 17:14'!
appleWithBite
	self subclassResponsibility! !

!NCCurve class methodsFor: 'arrow forms' stamp: 'nk 2/3/2004 17:14'!
basicArrow1
	self subclassResponsibility! !

!NCCurve class methodsFor: 'arrow forms' stamp: 'nk 2/3/2004 17:14'!
basicArrow2
	self subclassResponsibility! !

!NCCurve class methodsFor: 'arrow forms' stamp: 'nk 2/3/2004 17:14'!
doubleArrow
	self subclassResponsibility! !

!NCCurve class methodsFor: 'arrow forms' stamp: 'nk 2/3/2004 17:14'!
epsilonFork
	self subclassResponsibility! !

!NCCurve class methodsFor: 'arrow forms' stamp: 'nk 2/3/2004 17:14'!
filledCircle
	self subclassResponsibility! !

!NCCurve class methodsFor: 'arrow forms' stamp: 'nk 2/3/2004 17:14'!
fred
	self subclassResponsibility! !

!NCCurve class methodsFor: 'arrow forms' stamp: 'nk 2/3/2004 17:14'!
heavyDoubleArrow
	self subclassResponsibility! !

!NCCurve class methodsFor: 'arrow forms' stamp: 'nk 2/3/2004 17:14'!
junctionDot
	self subclassResponsibility! !

!NCCurve class methodsFor: 'arrow forms' stamp: 'nk 2/3/2004 17:14'!
noArrow
	self subclassResponsibility! !

!NCCurve class methodsFor: 'arrow forms' stamp: 'nk 2/3/2004 17:14'!
numeroDos
	self subclassResponsibility! !

!NCCurve class methodsFor: 'arrow forms' stamp: 'nk 2/3/2004 17:14'!
numeroUno
	self subclassResponsibility! !

!NCCurve class methodsFor: 'arrow forms' stamp: 'nk 2/3/2004 17:14'!
openArrow
	self subclassResponsibility! !

!NCCurve class methodsFor: 'arrow forms' stamp: 'nk 2/3/2004 17:14'!
openCircleWithPlus
	self subclassResponsibility! !

!NCCurve class methodsFor: 'arrow forms' stamp: 'nk 2/3/2004 17:14'!
openCircleWithSquareDot
	self subclassResponsibility! !

!NCCurve class methodsFor: 'arrow forms' stamp: 'nk 2/3/2004 17:14'!
openCircleWithX
	self subclassResponsibility! !

!NCCurve class methodsFor: 'arrow forms' stamp: 'nk 2/3/2004 17:14'!
openDiamond
	self subclassResponsibility! !

!NCCurve class methodsFor: 'arrow forms' stamp: 'nk 2/3/2004 17:14'!
pi
	self subclassResponsibility! !

!NCCurve class methodsFor: 'arrow forms' stamp: 'nk 2/3/2004 17:14'!
pointingHand
	self subclassResponsibility! !

!NCCurve class methodsFor: 'arrow forms' stamp: 'nk 2/3/2004 17:14'!
quadDiamond
	self subclassResponsibility! !

!NCCurve class methodsFor: 'arrow forms' stamp: 'nk 2/3/2004 17:14'!
roundFork
	self subclassResponsibility! !

!NCCurve class methodsFor: 'arrow forms' stamp: 'nk 2/3/2004 17:14'!
roundTrident
	self subclassResponsibility! !

!NCCurve class methodsFor: 'testing' stamp: 'nk 2/3/2004 17:14'!
sampleAllArrows
	self subclassResponsibility! !

!NCCurve class methodsFor: 'testing' stamp: 'nk 2/3/2004 17:14'!
sampleAllArrowsWithLineWidth: lw borderWidth: bw
	self subclassResponsibility! !

!NCCurve class methodsFor: 'arrow forms' stamp: 'nk 2/3/2004 17:14'!
starInCircle
	self subclassResponsibility! !

!NCCurve class methodsFor: 'arrow forms' stamp: 'nk 2/3/2004 17:14'!
test2
	self subclassResponsibility! !

!NCCurve class methodsFor: 'arrow forms' stamp: 'nk 2/3/2004 17:14'!
thinDoubleArrow
	self subclassResponsibility! !

!NCCurve class methodsFor: 'instance creation' stamp: 'nk 8/17/2004 14:59'!
withContours: anArray
	"anArray is expected to have lists of points already scaled to fit in a unit box"
	^self new withContours: anArray! !

!NCCurve class methodsFor: 'arrow forms' stamp: 'nk 2/3/2004 17:14'!
zap
	self subclassResponsibility! !

!NCCurve methodsFor: 'drawing' stamp: 'nk 8/17/2004 15:35'!
asFormWithExtent: extent depth: depth bgColor: bgColor fill: fillColor borderWidth: bw borderColor: bc 
	| canvas |
	canvas := BalloonCanvas extent: extent depth: depth.
	self drawOn: canvas
		in:  (Rectangle origin: 0@0 corner: extent)
		fill: fillColor
		borderWidth: bw
		borderColor: bc. 
	^canvas form! !

!NCCurve methodsFor: 'accessing contours' stamp: 'nk 8/17/2004 15:57'!
bezierSegments
	^Array streamContents: [ :s | self bezierSegmentsDo: [ :seg | s nextPut: seg ]]! !

!NCCurve methodsFor: 'accessing contours' stamp: 'nk 8/17/2004 15:56'!
bezierSegmentsDo: aBlock
	| seg |
	contours do: [ :c |
		1 to: c size by: 3 do: [ :i |
			seg := Bezier2Segment from: (c at: i) to: (c at: i + 2) via: (c at: i+1).
			aBlock value: seg ]]! !

!NCCurve methodsFor: 'accessing contours' stamp: 'nk 8/17/2004 17:42'!
bezierSegmentsTransformedBy: aTransform do: aBlock
	| seg |
	contours do: [ :c |
		1 to: c size by: 3 do: [ :i |
			seg := Bezier2Segment from: (c at: i) to: (c at: i + 2) via: (c at: i+1).
			aBlock value: seg ]]! !

!NCCurve methodsFor: 'accessing contours' stamp: 'nk 8/17/2004 17:45'!
bezierSegmentsTransformedFor: aRectangle rotation: rotation do: aBlock
	| transform seg c |
	transform := self transformForRect: aRectangle rotation: rotation.
	contours do: [ :ct |
		c := transform localPointsToGlobal: ct.
		1 to: c size by: 3 do: [ :i |
			seg := Bezier2Segment from: (c at: i) to: (c at: i + 2) via: (c at: i+1).
			aBlock value: seg ]]
! !

!NCCurve methodsFor: 'accessing contours' stamp: 'nk 8/18/2004 10:26'!
boundingBox
	"return my actual bounding box"
	| rect |
	rect := Rectangle encompassing: (Array streamContents: [ :s |
		self bezierSegmentsDo: [ :seg | seg lineSegmentsDo: [ :p1 :p2 | s nextPut: p1; nextPut: p2 ]]]).
	^rect! !

!NCCurve methodsFor: 'accessing' stamp: 'nk 2/3/2004 17:24'!
bounds
	^bounds! !

!NCCurve methodsFor: 'accessing' stamp: 'nk 9/2/2004 09:22'!
bounds: aRectangle
	self updateBounds.
	self scale: (self scale * ((aRectangle width / (bounds width max: 1)) @ (aRectangle height / (bounds height max: 1))))! !

!NCCurve methodsFor: 'accessing contours' stamp: 'nk 2/4/2004 08:40'!
contourPoints
	"Answer my contours flattened as a series of points."
	^self contourPointsCollect: [ :ea | ea ].! !

!NCCurve methodsFor: 'accessing contours' stamp: 'nk 8/17/2004 15:18'!
contourPointsCollect: aBlock 
	^ Array
		streamContents: [:s | self contours
				do: [:c | c
						do: [:p | s
								nextPut: (aBlock value: p)]]]! !

!NCCurve methodsFor: 'accessing contours' stamp: 'nk 2/4/2004 08:38'!
contourPointsDo: aBlock 
	self contours do: [:c | c do: aBlock]! !

!NCCurve methodsFor: 'accessing contours' stamp: 'nk 2/4/2004 08:38'!
contourPointsTransformedBy: aTransform
	^self contourPointsCollect: [ :p | aTransform transformPoint: p ].! !

!NCCurve methodsFor: 'accessing' stamp: 'nk 2/3/2004 17:25'!
contours
	^contours! !

!NCCurve methodsFor: 'accessing' stamp: 'nk 2/3/2004 17:20'!
contours: anObject
	contours := anObject! !

!NCCurve methodsFor: 'accessing contours' stamp: 'nk 8/17/2004 15:15'!
contoursTransformedFor: aRectangle rotation: rotation
	| transform |
	transform := self transformForRect: aRectangle rotation: rotation.
	^contours collect: [ :c | transform localPointsToGlobal: c ]! !

!NCCurve methodsFor: 'as yet unclassified' stamp: 'nk 8/18/2004 13:28'!
curveInfoNearest: aPoint 
	"Answer the index of the segment nearest aPoint, along with the closest point, its distance, and the parametric value. "

	| closestPoint minDist minT segmentIndex i |
	closestPoint := minDist := nil.
	i := 0.
	self bezierSegmentsDo:  
			[:seg |
			i := i + 1. 
			seg lineSegmentsWithParametricValueDo: 
					[:p1 :p2 :t1 :t2 | 
					| curvePoint dist |
					curvePoint := aPoint nearestPointOnLineFrom: p1 to: p2.
					dist := curvePoint dist: aPoint.
					(closestPoint isNil or: [dist < minDist]) 
						ifTrue: 
							[closestPoint := curvePoint.
							minDist := dist.
							segmentIndex := i.
							minT := (p1 = p2) ifTrue: [ t1 ] ifFalse: [ t1 interpolateTo: t2 at: (curvePoint dist: p1) / (p1 dist: p2)] ]]].
	^{ closestPoint. segmentIndex. minDist. minT }! !

!NCCurve methodsFor: 'accessing' stamp: 'nk 2/4/2004 08:40'!
defaultScale: floatOrPoint
	"scale me as indicated, then call my current scale the default"
	self scale: floatOrPoint; setAsDefaultScale.! !

!NCCurve methodsFor: 'drawing' stamp: 'nk 2/4/2004 08:38'!
drawOn: aCanvas fgColor: fgColor 
	| canvas |
	canvas := aCanvas asBalloonCanvas.
	canvas aaLevel: 4.
	canvas 
		drawGeneralBezierShape: self contours
		color: nil
		borderWidth: 1
		borderColor: fgColor.
! !

!NCCurve methodsFor: 'drawing' stamp: 'nk 8/17/2004 15:34'!
drawOn: aCanvas in: aRectangle bgColor: bgColor fill: fillColor borderWidth: bw borderColor: bc 
	self drawOn: aCanvas in: aRectangle rotatedBy: 0.0 fill: fillColor borderWidth: bw borderColor: bc ! !

!NCCurve methodsFor: 'drawing' stamp: 'nk 8/17/2004 15:35'!
drawOn: aCanvas in: aRectangle fill: fillColor borderWidth: bw borderColor: bc 
	self drawOn: aCanvas in: aRectangle rotatedBy: 0.0 fill: fillColor borderWidth: bw borderColor: bc ! !

!NCCurve methodsFor: 'drawing' stamp: 'nk 8/17/2004 18:47'!
drawOn: aCanvas in: aRectangle rotatedBy: rotation fill: fillColor borderWidth: bw borderColor: bc 
	aCanvas
		drawGeneralBezierShape: (self contoursTransformedFor: aRectangle rotation: rotation)
		color: fillColor
		borderWidth: bw
		borderColor: bc! !

!NCCurve methodsFor: 'accessing' stamp: 'nk 2/4/2004 08:40'!
extent
	^self bounds extent! !

!NCCurve methodsFor: 'accessing' stamp: 'nk 2/4/2004 08:40'!
extent: aPoint 
	self scale: self scale * (aPoint asFloatPoint / self bounds extent)! !

!NCCurve methodsFor: 'accessing' stamp: 'nk 2/3/2004 17:14'!
fixEndContours
	self subclassResponsibility! !

!NCCurve methodsFor: 'accessing' stamp: 'nk 2/3/2004 17:30'!
offset
	^offset ifNil: [ offset := 0.0 @ 0.0]! !

!NCCurve methodsFor: 'accessing' stamp: 'nk 2/3/2004 17:30'!
offset: aPoint 
	| newOffset delta |
	newOffset := aPoint asPoint asFloatPoint.
	delta := newOffset - self offset.
	self contours: (self contours collect: [:c | c + delta]).
	offset  := newOffset.
	self updateBounds! !

!NCCurve methodsFor: 'accessing' stamp: 'nk 2/3/2004 17:14'!
originalFirstContour
	self subclassResponsibility! !

!NCCurve methodsFor: 'drawing' stamp: 'nk 2/4/2004 08:43'!
sampleMorph
	(self asFormWithExtent: 200@200 depth: 32
	 bgColor: Color white fill: (Color blue alpha: 0.3)
	 borderWidth: 1 borderColor: Color black) openAsMorph! !

!NCCurve methodsFor: 'accessing' stamp: 'nk 2/3/2004 17:35'!
scale
	^scale ifNil: [ scale := 1.0@1.0 ]! !

!NCCurve methodsFor: 'accessing' stamp: 'nk 9/2/2004 09:28'!
scale: aFloatOrPoint 
	| newScale ratio newOffset |
	newScale := (aFloatOrPoint asPoint asFloatPoint) max: 0.001 @ 0.001.
	newScale = scale ifTrue: [^self].
	ratio := newScale / scale.
	newOffset := self offset * ratio.
	self contours: (self contours 
				collect: [:c | c collect: [:p | (p - self offset) * ratio + newOffset]]).
	scale := newScale.
	offset := newOffset.
	self updateBounds! !

!NCCurve methodsFor: 'accessing' stamp: 'nk 2/4/2004 08:44'!
setAsDefaultScale
	"Call my current scale the default"
	scale := 1.0 @ 1.0! !

!NCCurve methodsFor: 'printing' stamp: 'nk 2/3/2004 17:36'!
storeOn: aStream 
	aStream
		nextPutAll: '((';
		nextPutAll: self class name;
		nextPutAll: ' withContours: ((Array new: ';
		print: self contours size;
		nextPutAll: ')'.
	self contours withIndexDo: 
			[:c :ci | 
			aStream
				cr;
				nextPutAll: 'at: ';
				print: ci;
				nextPutAll: ' put: '.
			(c collect: [:p | (p - self offset) / self scale roundTo: 0.001]) 
				storeOn: aStream.
			aStream nextPut: $;].
	aStream
		nextPutAll: ' yourself))';
		cr.
	aStream
		nextPutAll: 'scale: ';
		print: (self scale roundTo: 0.01);
		nextPutAll: ';';
		cr.
	aStream
		nextPutAll: 'offset: ';
		print: (self offset roundTo: 0.001);
		nextPutAll: ';';
		cr.
	aStream
		nextPutAll: 'yourself)';
		cr! !

!NCCurve methodsFor: 'drawing' stamp: 'nk 2/4/2004 08:49'!
transformForRect: outsideRect rotation: angle 
	| requiredScale newRect |
	newRect := self bounds scaleAndTranslateToFitInside: outsideRect.
	requiredScale := newRect extent asFloatPoint / self bounds extent.
	^((((MatrixTransform2x3 withOffset: outsideRect center) 
		composedWithLocal: (MatrixTransform2x3 withRotation: angle)) 
			composedWithLocal: (MatrixTransform2x3 withOffset: outsideRect extent / -2)) 
			composedWithLocal: (MatrixTransform2x3 withScale: requiredScale asPoint)) 
			composedWithLocal: (MatrixTransform2x3 
					withOffset: self bounds origin negated)! !

!NCCurve methodsFor: 'accessing' stamp: 'nk 8/18/2004 10:28'!
updateBounds
	| boundsRect |
	boundsRect := self boundingBox.

	boundsRect width isZero 
		ifTrue: [boundsRect := boundsRect expandBy: (self scale x / 20.0) @ 0.0].
	boundsRect height isZero 
		ifTrue: [boundsRect := boundsRect expandBy: 0.0 @ (self scale y / 20.0)].

	bounds := boundsRect! !

!NCCurve methodsFor: 'initialize-release' stamp: 'nk 2/4/2004 08:43'!
withContours: anArray
	scale := 1.0 @ 1.0.
	offset := 0.0 @ 0.0.
	self contours: (anArray collect: [:a | a collect: [:p | p asFloatPoint]]).
	self updateBounds! !

NCCurve subclass: #NCLineArrowGlyph
	instanceVariableNames: 'originalFirstContour'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Connectors-Lines and Curves'!

!NCLineArrowGlyph class methodsFor: 'arrow forms' stamp: 'nk 8/23/2003 18:42'!
appleWithBite
	"glyph 210 from BitstreamVeraSans"
	^((NCLineArrowGlyph withContours: ((Array new: 2)
at: 1 put: ((Array new: 66) at: 1 put: 0.0@2.059; at: 2 put: -0.082@2.059; at: 3 put: -0.238@2.119; at: 4 put: -0.238@2.119; at: 5 put: -0.395@2.179; at: 6 put: -0.469@2.179; at: 7 put: -0.469@2.179; at: 8 put: -0.586@2.179; at: 9 put: -0.704@2.098; at: 10 put: -0.704@2.098; at: 11 put: -0.82@2.018; at: 12 put: -0.938@1.858; at: 13 put: -0.938@1.858; at: 14 put: -1.123@1.603; at: 15 put: -1.225@1.321; at: 16 put: -1.225@1.321; at: 17 put: -1.324@1.036; at: 18 put: -1.324@0.773; at: 19 put: -1.324@0.773; at: 20 put: -1.323@0.374; at: 21 put: -1.095@0.125; at: 22 put: -1.095@0.125; at: 23 put: -0.868@-0.125; at: 24 put: -0.507@-0.125; at: 25 put: -0.507@-0.125; at: 26 put: -0.384@-0.124; at: 27 put: -0.226@-0.062; at: 28 put: -0.226@-0.062; at: 29 put: -0.067@0.0; at: 30 put: 0.0@0.0; at: 31 put: 0.0@0.0; at: 32 put: 0.062@0.0; at: 33 put: 0.22@-0.063; at: 34 put: 0.22@-0.063; at: 35 put: 0.38@-0.127; at: 36 put: 0.499@-0.126; at: 37 put: 0.499@-0.126; at: 38 put: 0.699@-0.126; at: 39 put: 0.859@-0.053; at: 40 put: 0.859@-0.053; at: 41 put: 1.019@0.02; at: 42 put: 1.151@0.173; at: 43 put: 1.151@0.173; at: 44 put: 0.965@0.291; at: 45 put: 0.874@0.438; at: 46 put: 0.874@0.438; at: 47 put: 0.785@0.583; at: 48 put: 0.785@0.769; at: 49 put: 0.785@0.769; at: 50 put: 0.784@0.975; at: 51 put: 0.893@1.133; at: 52 put: 0.893@1.133; at: 53 put: 1.002@1.292; at: 54 put: 1.208@1.385; at: 55 put: 1.208@1.385; at: 56 put: 0.999@1.834; at: 57 put: 0.833@2.008; at: 58 put: 0.833@2.008; at: 59 put: 0.668@2.181; at: 60 put: 0.462@2.18; at: 61 put: 0.462@2.18; at: 62 put: 0.371@2.18; at: 63 put: 0.222@2.12; at: 64 put: 0.222@2.12; at: 65 put: 0.075@2.059; at: 66 put: 0.0@2.059; yourself);
at: 2 put: ((Array new: 33) at: 1 put: -0.007@-0.096; at: 2 put: -0.009@-0.107; at: 3 put: -0.011@-0.122; at: 4 put: -0.011@-0.122; at: 5 put: -0.011@-0.137; at: 6 put: -0.01@-0.164; at: 7 put: -0.01@-0.164; at: 8 put: -0.01@-0.389; at: 9 put: 0.163@-0.58; at: 10 put: 0.163@-0.58; at: 11 put: 0.336@-0.773; at: 12 put: 0.593@-0.819; at: 13 put: 0.593@-0.819; at: 14 put: 0.595@-0.801; at: 15 put: 0.595@-0.778; at: 16 put: 0.595@-0.778; at: 17 put: 0.596@-0.754; at: 18 put: 0.596@-0.707; at: 19 put: 0.596@-0.707; at: 20 put: 0.596@-0.468; at: 21 put: 0.406@-0.279; at: 22 put: 0.406@-0.279; at: 23 put: 0.218@-0.092; at: 24 put: 0.037@-0.093; at: 25 put: 0.037@-0.093; at: 26 put: 0.037@-0.093; at: 27 put: 0.022@-0.093; at: 28 put: 0.022@-0.093; at: 29 put: 0.017@-0.093; at: 30 put: 0.008@-0.095; at: 31 put: 0.008@-0.095; at: 32 put: 0.0@-0.096; at: 33 put: -0.007@-0.096; yourself); yourself))
offset: 0@0;
yourself)
! !

!NCLineArrowGlyph class methodsFor: 'arrow forms' stamp: 'nk 8/23/2003 17:06'!
basicArrow1
	"glyph 54 from Arrows1"
^(NCLineArrowGlyph withContours: (
(Array new: 1)
	at: 1 put: { 
	0.0@0.0 . 0.0@0.0 . -0.5@0.3298 . 
	-0.5@0.3298 . -0.5@0.3298 . 0.0@-0.666 . 
	0.0@-0.666 . 0.0@-0.666 . 0.5@0.3298 . 
	0.5@0.3298 . 0.5@0.3298 . 0.0@0.0 };
	yourself
"clicks 0.0@0.0"
)) offset: 0@0.666; defaultScale: 2.5; yourself! !

!NCLineArrowGlyph class methodsFor: 'arrow forms' stamp: 'nk 8/23/2003 18:36'!
basicArrow2
	"glyph 54 from Arrows1, modified"
^(NCLineArrowGlyph withContours: (
(Array new: 1)
	at: 1 put: { 
	0.0@0.0 . 0.0@0.0 . -0.5@0.0 . 
	-0.5@0.0 . -0.5@0.0 . 0.0@-0.666 . 
	0.0@-0.666 . 0.0@-0.666 . 0.5@0.0 . 
	0.5@0.0 . 0.5@0.0 . 0.0@0.0 };
	yourself
"clicks 0.0@0.0"
)) offset: 0@0.666; defaultScale: 2; yourself! !

!NCLineArrowGlyph class methodsFor: 'arrow forms' stamp: 'nk 8/25/2003 18:30'!
doubleArrow
	"glyph 154 from StarBats"
	^((NCLineArrowGlyph withContours: ((Array new: 2)
at: 1 put: ((Array new: 18) at: 1 put: 0.0@0.0; at: 2 put: -0.499@0.29; at: 3 put: -0.499@0.29; at: 4 put: -0.499@0.29; at: 5 put: -0.499@0.124; at: 6 put: -0.499@0.124; at: 7 put: -0.499@0.124; at: 8 put: 0.0@-0.22; at: 9 put: 0.0@-0.22; at: 10 put: 0.0@-0.22; at: 11 put: 0.501@0.124; at: 12 put: 0.501@0.124; at: 13 put: 0.501@0.124; at: 14 put: 0.501@0.29; at: 15 put: 0.501@0.29; at: 16 put: 0.501@0.29; at: 17 put: 0.0@0.0; at: 18 put: 0.0@0.0; yourself);
at: 2 put: ((Array new: 18) at: 1 put: 0.0@-0.425; at: 2 put: -0.499@-0.137; at: 3 put: -0.499@-0.137; at: 4 put: -0.499@-0.137; at: 5 put: -0.499@-0.301; at: 6 put: -0.499@-0.301; at: 7 put: -0.499@-0.301; at: 8 put: 0.0@-0.652; at: 9 put: 0.0@-0.652; at: 10 put: 0.0@-0.652; at: 11 put: 0.501@-0.301; at: 12 put: 0.501@-0.301; at: 13 put: 0.501@-0.301; at: 14 put: 0.501@-0.139; at: 15 put: 0.501@-0.139; at: 16 put: 0.501@-0.139; at: 17 put: 0.0@-0.425; at: 18 put: 0.0@-0.425; yourself); yourself))
scale: 1.85@2.49;
offset: 0.0@1.623;
yourself)
"clicks: 0.01@0.01 0.0@-0.65 0.0@0.01 "! !

!NCLineArrowGlyph class methodsFor: 'arrow forms' stamp: 'nk 8/17/2004 13:59'!
doubleBubble
	"glyph 65 from Webdings"
	^((NCLineArrowGlyph withContours: ((Array new: 2)
at: 1 put: ((Array new: 39) at: 1 put: 0.0@0.0; at: 2 put: 0.042@-0.005; at: 3 put: 0.067@-0.035; at: 4 put: 0.067@-0.035; at: 5 put: 0.093@-0.066; at: 6 put: 0.093@-0.113; at: 7 put: 0.093@-0.113; at: 8 put: 0.093@-0.12; at: 9 put: 0.093@-0.128; at: 10 put: 0.093@-0.128; at: 11 put: 0.013@-0.192; at: 12 put: 0.013@-0.281; at: 13 put: 0.013@-0.281; at: 14 put: 0.013@-0.373; at: 15 put: 0.099@-0.435; at: 16 put: 0.099@-0.435; at: 17 put: 0.183@-0.497; at: 18 put: 0.306@-0.497; at: 19 put: 0.306@-0.497; at: 20 put: 0.429@-0.497; at: 21 put: 0.515@-0.435; at: 22 put: 0.515@-0.435; at: 23 put: 0.601@-0.373; at: 24 put: 0.601@-0.281; at: 25 put: 0.601@-0.281; at: 26 put: 0.601@-0.188; at: 27 put: 0.515@-0.123; at: 28 put: 0.515@-0.123; at: 29 put: 0.429@-0.057; at: 30 put: 0.307@-0.057; at: 31 put: 0.307@-0.057; at: 32 put: 0.248@-0.057; at: 33 put: 0.194@-0.074; at: 34 put: 0.194@-0.074; at: 35 put: 0.178@0.027; at: 36 put: 0.079@0.027; at: 37 put: 0.079@0.027; at: 38 put: 0.046@0.027; at: 39 put: 0.0@0.0; yourself);
at: 2 put: ((Array new: 45) at: 1 put: 1.0@-0.523; at: 2 put: 1.0@-0.418; at: 3 put: 0.894@-0.351; at: 4 put: 0.894@-0.351; at: 5 put: 0.892@-0.334; at: 6 put: 0.892@-0.319; at: 7 put: 0.892@-0.319; at: 8 put: 0.892@-0.274; at: 9 put: 0.919@-0.244; at: 10 put: 0.919@-0.244; at: 11 put: 0.922@-0.24; at: 12 put: 0.975@-0.196; at: 13 put: 0.975@-0.196; at: 14 put: 0.887@-0.194; at: 15 put: 0.838@-0.227; at: 16 put: 0.838@-0.227; at: 17 put: 0.801@-0.252; at: 18 put: 0.778@-0.306; at: 19 put: 0.778@-0.306; at: 20 put: 0.743@-0.299; at: 21 put: 0.706@-0.299; at: 22 put: 0.706@-0.299; at: 23 put: 0.666@-0.299; at: 24 put: 0.621@-0.312; at: 25 put: 0.621@-0.312; at: 26 put: 0.603@-0.377; at: 27 put: 0.547@-0.43; at: 28 put: 0.547@-0.43; at: 29 put: 0.486@-0.487; at: 30 put: 0.417@-0.492; at: 31 put: 0.417@-0.492; at: 32 put: 0.413@-0.512; at: 33 put: 0.413@-0.523; at: 34 put: 0.413@-0.523; at: 35 put: 0.413@-0.614; at: 36 put: 0.499@-0.674; at: 37 put: 0.499@-0.674; at: 38 put: 0.583@-0.733; at: 39 put: 0.706@-0.733; at: 40 put: 0.706@-0.733; at: 41 put: 0.83@-0.733; at: 42 put: 0.914@-0.674; at: 43 put: 0.914@-0.674; at: 44 put: 1.0@-0.614; at: 45 put: 1.0@-0.523; yourself); yourself))
scale: 2.0@2.0;
offset: 0.0@0.0;
yourself)
"clicks: 0.02@0.0 "! !

!NCLineArrowGlyph class methodsFor: 'arrow forms' stamp: 'nk 8/23/2003 18:58'!
epsilonFork
	"glyph 90 from StarMath"
	^((NCLineArrowGlyph withContours: ((Array new: 1)
at: 1 put: ((Array new: 51) at: 1 put: 0.0@1.988; at: 2 put: -0.717@1.97; at: 3 put: -1.126@1.507; at: 4 put: -1.126@1.507; at: 5 put: -1.53@1.043; at: 6 put: -1.512@0.32; at: 7 put: -1.512@0.32; at: 8 put: -1.506@0.094; at: 9 put: -1.355@-0.412; at: 10 put: -1.355@-0.412; at: 11 put: -1.355@-0.412; at: 12 put: -1.166@-0.407; at: 13 put: -1.166@-0.407; at: 14 put: -1.294@-0.071; at: 15 put: -1.301@0.219; at: 16 put: -1.301@0.219; at: 17 put: -1.314@0.703; at: 18 put: -0.976@1.026; at: 19 put: -0.976@1.026; at: 20 put: -0.663@1.323; at: 21 put: -0.167@1.393; at: 22 put: -0.167@1.393; at: 23 put: -0.167@1.393; at: 24 put: -0.132@-0.022; at: 25 put: -0.132@-0.022; at: 26 put: -0.132@-0.022; at: 27 put: 0.095@-0.017; at: 28 put: 0.095@-0.017; at: 29 put: 0.095@-0.017; at: 30 put: 0.059@1.398; at: 31 put: 0.059@1.398; at: 32 put: 0.608@1.343; at: 33 put: 0.937@1.018; at: 34 put: 0.937@1.018; at: 35 put: 1.266@0.693; at: 36 put: 1.28@0.146; at: 37 put: 1.28@0.146; at: 38 put: 1.282@0.051; at: 39 put: 1.238@-0.182; at: 40 put: 1.238@-0.182; at: 41 put: 1.238@-0.182; at: 42 put: 1.414@-0.178; at: 43 put: 1.414@-0.178; at: 44 put: 1.493@0.189; at: 45 put: 1.485@0.472; at: 46 put: 1.485@0.472; at: 47 put: 1.469@1.144; at: 48 put: 1.074@1.562; at: 49 put: 1.074@1.562; at: 50 put: 0.66@2.005; at: 51 put: 0.0@1.988; yourself); yourself))
scale: 1.0@1.0;
offset: 0@0;
yourself)
! !

!NCLineArrowGlyph class methodsFor: 'arrow forms' stamp: 'nk 8/22/2003 21:15'!
filledCircle
	"glyph 34 from StarBats"
	^((NCLineArrowGlyph withContours: ((Array new: 1)
at: 1 put: {
 0.0@0.0 . -0.655@0.0 . -1.079@-0.419 .
 -1.079@-0.419 . -1.502@-0.842 . -1.502@-1.498 .
 -1.502@-1.498 . -1.502@-2.162 . -1.083@-2.581 .
 -1.083@-2.581 . -0.664@-3.0 . 0.0@-3.0 .
 0.0@-3.0 . 0.66@-3.0 . 1.079@-2.576 .
 1.079@-2.576 . 1.498@-2.158 . 1.498@-1.498 .
 1.498@-1.498 . 1.498@-0.834 . 1.079@-0.415 .
 1.079@-0.415 . 0.66@0.0 . 0.0@0.0 .};
yourself))
scale: 1.0@1.0;
offset: -0.01337295665871352@2.942050464916974;
yourself)
! !

!NCLineArrowGlyph class methodsFor: 'arrow forms' stamp: 'nk 8/23/2003 18:59'!
fred
	"glyph 85 from StarBats"
^(NCLineArrowGlyph withContours: (
(Array new: 5)
	at: 1 put: { 
	0.0@0.0 . -0.2224@0.0 . -0.3605@-0.1381 . 
	-0.3605@-0.1381 . -0.5@-0.2776 . -0.5@-0.5 . 
	-0.5@-0.5 . -0.5@-0.7196 . -0.3591@-0.8591 . 
	-0.3591@-0.8591 . -0.2196@-0.9986 . 0.0@-0.9986 . 
	0.0@-0.9986 . 0.221@-0.9986 . 0.3605@-0.8591 . 
	0.3605@-0.8591 . 0.5@-0.721 . 0.5@-0.5 . 
	0.5@-0.5 . 0.5@-0.2776 . 0.3619@-0.1381 . 
	0.3619@-0.1381 . 0.2224@0.0 . 0.0@0.0 };
	at: 2 put: { 
	0.0@-0.0677 . 0.1906@-0.0677 . 0.3122@-0.1892 . 
	0.3122@-0.1892 . 0.4337@-0.3108 . 0.4337@-0.5 . 
	0.4337@-0.5 . 0.4337@-0.6934 . 0.3135@-0.8122 . 
	0.3135@-0.8122 . 0.1934@-0.9309 . 0.0@-0.9309 . 
	0.0@-0.9309 . -0.1934@-0.9309 . -0.3135@-0.8108 . 
	-0.3135@-0.8108 . -0.4337@-0.692 . -0.4337@-0.5 . 
	-0.4337@-0.5 . -0.4337@-0.3094 . -0.3122@-0.1878 . 
	-0.3122@-0.1878 . -0.192@-0.0677 . 0.0@-0.0677 };
	at: 3 put: { 
	-0.0249@-0.4075 . -0.0884@-0.3163 . -0.1671@-0.3163 . 
	-0.1671@-0.3163 . -0.2431@-0.3163 . -0.2873@-0.3743 . 
	-0.2873@-0.3743 . -0.3273@-0.4268 . -0.3273@-0.5055 . 
	-0.3273@-0.5055 . -0.3273@-0.576 . -0.2956@-0.6243 . 
	-0.2956@-0.6243 . -0.2597@-0.6796 . -0.192@-0.6796 . 
	-0.192@-0.6796 . -0.1008@-0.6796 . -0.0249@-0.5622 . 
	-0.0249@-0.5622 . 0.0428@-0.7044 . 0.1547@-0.7044 . 
	0.1547@-0.7044 . 0.239@-0.7044 . 0.2887@-0.6381 . 
	0.2887@-0.6381 . 0.3329@-0.5787 . 0.3329@-0.4917 . 
	0.3329@-0.4917 . 0.3329@-0.4061 . 0.2914@-0.3481 . 
	0.2914@-0.3481 . 0.2445@-0.2831 . 0.1616@-0.2831 . 
	0.1616@-0.2831 . 0.0497@-0.2831 . -0.0249@-0.4075 };
	at: 4 put: { 
	-0.0511@-0.5152 . -0.1091@-0.6036 . -0.1809@-0.6036 . 
	-0.1809@-0.6036 . -0.2652@-0.6036 . -0.2652@-0.4972 . 
	-0.2652@-0.4972 . -0.2652@-0.4558 . -0.2445@-0.424 . 
	-0.2445@-0.424 . -0.2238@-0.3923 . -0.1837@-0.3923 . 
	-0.1837@-0.3923 . -0.1215@-0.3923 . -0.0511@-0.5152 };
	at: 5 put: { 
	0.0083@-0.4544 . 0.0746@-0.3674 . 0.1492@-0.3674 . 
	0.1492@-0.3674 . 0.2058@-0.3674 . 0.239@-0.4019 . 
	0.239@-0.4019 . 0.2707@-0.4365 . 0.2707@-0.4931 . 
	0.2707@-0.4931 . 0.2707@-0.5428 . 0.2431@-0.5801 . 
	0.2431@-0.5801 . 0.2155@-0.6188 . 0.1671@-0.6188 . 
	0.1671@-0.6188 . 0.1215@-0.6188 . 0.0732@-0.5566 . 
	0.0732@-0.5566 . 0.0539@-0.5318 . 0.0083@-0.4544 };
	yourself
"clicks -0.0152@-0.9903 -0.0055@-0.0055"
)) offset: 0.0@1.0;
defaultScale: 3.0;
yourself! !

!NCLineArrowGlyph class methodsFor: 'arrow forms' stamp: 'nk 8/23/2003 19:02'!
heavyDoubleArrow
	"glyph 170 from BitstreamVeraSans"
	^((NCLineArrowGlyph withContours: ((Array new: 2)
at: 1 put: ((Array new: 21) at: 1 put: 0.0@1.971; at: 2 put: 0.0@1.971; at: 3 put: -0.868@2.962; at: 4 put: -0.868@2.962; at: 5 put: -0.868@2.962; at: 6 put: -1.492@2.967; at: 7 put: -1.492@2.967; at: 8 put: -1.492@2.967; at: 9 put: -0.139@1.423; at: 10 put: -0.139@1.423; at: 11 put: -0.139@1.423; at: 12 put: 0.129@1.42; at: 13 put: 0.129@1.42; at: 14 put: 0.129@1.42; at: 15 put: 1.508@2.942; at: 16 put: 1.508@2.942; at: 17 put: 1.508@2.942; at: 18 put: 0.884@2.947; at: 19 put: 0.884@2.947; at: 20 put: 0.884@2.947; at: 21 put: 0.0@1.971; yourself);
at: 2 put: ((Array new: 21) at: 1 put: -1.504@1.549; at: 2 put: -0.879@1.543; at: 3 put: -0.879@1.543; at: 4 put: -0.879@1.543; at: 5 put: -0.012@0.552; at: 6 put: -0.012@0.552; at: 7 put: -0.012@0.552; at: 8 put: 0.872@1.529; at: 9 put: 0.872@1.529; at: 10 put: 0.872@1.529; at: 11 put: 1.496@1.524; at: 12 put: 1.496@1.524; at: 13 put: 1.496@1.524; at: 14 put: 0.118@0.002; at: 15 put: 0.118@0.002; at: 16 put: 0.118@0.002; at: 17 put: -0.15@0.004; at: 18 put: -0.15@0.004; at: 19 put: -0.15@0.004; at: 20 put: -1.504@1.549; at: 21 put: -1.504@1.549; yourself); yourself))
offset: 0@0;
defaultScale: 2.0;
yourself)
! !

!NCLineArrowGlyph class methodsFor: 'arrow forms' stamp: 'nk 6/30/2004 14:54'!
junctionDot
	"glyph 135 from BitstreamVeraSans"
	^((NCLineArrowGlyph withContours: ((Array new: 1)
at: 1 put: ((Array new: 24) at: 1 put: 0.0@0.0; at: 2 put: -0.209@0.004; at: 3 put: -0.354@-0.139; at: 4 put: -0.354@-0.139; at: 5 put: -0.502@-0.283; at: 6 put: -0.505@-0.495; at: 7 put: -0.505@-0.495; at: 8 put: -0.509@-0.704; at: 9 put: -0.366@-0.849; at: 10 put: -0.366@-0.849; at: 11 put: -0.226@-0.996; at: 12 put: -0.017@-1.0; at: 13 put: -0.017@-1.0; at: 14 put: 0.193@-1.003; at: 15 put: 0.341@-0.861; at: 16 put: 0.341@-0.861; at: 17 put: 0.488@-0.719; at: 18 put: 0.491@-0.508; at: 19 put: 0.491@-0.508; at: 20 put: 0.495@-0.298; at: 21 put: 0.353@-0.151; at: 22 put: 0.353@-0.151; at: 23 put: 0.21@-0.004; at: 24 put: 0.0@0.0; yourself); yourself))
scale: 1.0@1.0;
offset: 0.0@0.5;
defaultScale: 2.0;
yourself)
"clicks: 0.0@-0.5 0.0@0.0 "
! !

!NCLineArrowGlyph class methodsFor: 'arrow forms' stamp: 'nk 8/25/2003 15:07'!
noArrow
	^NCLineArrowGlyph withContours: { Array new: 6 withAll: 0.0@0.0 }! !

!NCLineArrowGlyph class methodsFor: 'arrow forms' stamp: 'nk 8/22/2003 21:14'!
numeroDos
	"glyph 79 from StarBats"
	^((NCLineArrowGlyph withContours: ((Array new: 3)
at: 1 put: {
 0.0@0.0 . -0.667@0.0 . -1.086@-0.414 .
 -1.086@-0.414 . -1.5@-0.833 . -1.5@-1.5 .
 -1.5@-1.5 . -1.5@-2.167 . -1.086@-2.582 .
 -1.086@-2.582 . -0.667@-3.0 . 0.0@-3.0 .
 0.0@-3.0 . 0.659@-3.0 . 1.077@-2.577 .
 1.077@-2.577 . 1.496@-2.159 . 1.496@-1.5 .
 1.496@-1.5 . 1.496@-0.837 . 1.077@-0.418 .
 1.077@-0.418 . 0.663@0.0 . 0.0@0.0 .};

at: 2 put: {
 -1.297@-1.5 . -1.297@-2.076 . -0.937@-2.437 .
 -0.937@-2.437 . -0.572@-2.801 . 0.0@-2.801 .
 0.0@-2.801 . 0.576@-2.801 . 0.932@-2.441 .
 0.932@-2.441 . 1.293@-2.08 . 1.293@-1.5 .
 1.293@-1.5 . 1.293@-0.92 . 0.937@-0.559 .
 0.937@-0.559 . 0.58@-0.199 . 0.0@-0.199 .
 0.0@-0.199 . -0.568@-0.199 . -0.932@-0.563 .
 -0.932@-0.563 . -1.297@-0.928 . -1.297@-1.5 .};

at: 3 put: {
 -0.551@-0.58 . -0.551@-0.58 . -0.551@-0.796 .
 -0.551@-0.796 . -0.481@-0.941 . -0.249@-1.181 .
 -0.249@-1.181 . 0.071@-1.508 . 0.128@-1.583 .
 0.128@-1.583 . 0.315@-1.811 . 0.315@-1.981 .
 0.315@-1.981 . 0.315@-2.321 . -0.041@-2.321 .
 -0.041@-2.321 . -0.236@-2.321 . -0.501@-2.175 .
 -0.501@-2.175 . -0.501@-2.175 . -0.501@-2.399 .
 -0.501@-2.399 . -0.236@-2.519 . -0.004@-2.519 .
 -0.004@-2.519 . 0.253@-2.519 . 0.414@-2.378 .
 0.414@-2.378 . 0.58@-2.238 . 0.58@-1.985 .
 0.58@-1.985 . 0.58@-1.753 . 0.386@-1.546 .
 0.386@-1.546 . 0.199@-1.38 . 0.008@-1.21 .
 0.008@-1.21 . -0.207@-1.003 . -0.245@-0.796 .
 -0.245@-0.796 . -0.245@-0.796 . 0.572@-0.796 .
 0.572@-0.796 . 0.572@-0.796 . 0.572@-0.58 .
 0.572@-0.58 . 0.572@-0.58 . -0.551@-0.58 .};
yourself))
scale: 1.0@1.0;
offset: 0.0580110519658774@2.970994589966722;
yourself)
! !

!NCLineArrowGlyph class methodsFor: 'arrow forms' stamp: 'nk 8/22/2003 21:13'!
numeroUno
	"glyph 78 from StarBats"
	^((NCLineArrowGlyph withContours: ((Array new: 3)
at: 1 put: {
 0.0@0.0 . -0.667@0.0 . -1.086@-0.414 .
 -1.086@-0.414 . -1.5@-0.833 . -1.5@-1.5 .
 -1.5@-1.5 . -1.5@-2.167 . -1.086@-2.582 .
 -1.086@-2.582 . -0.667@-3.0 . 0.0@-3.0 .
 0.0@-3.0 . 0.659@-3.0 . 1.077@-2.577 .
 1.077@-2.577 . 1.496@-2.159 . 1.496@-1.5 .
 1.496@-1.5 . 1.496@-0.837 . 1.077@-0.418 .
 1.077@-0.418 . 0.663@0.0 . 0.0@0.0 .};

at: 2 put: {
 -1.297@-1.5 . -1.297@-2.076 . -0.937@-2.437 .
 -0.937@-2.437 . -0.572@-2.801 . 0.0@-2.801 .
 0.0@-2.801 . 0.576@-2.801 . 0.932@-2.441 .
 0.932@-2.441 . 1.293@-2.08 . 1.293@-1.5 .
 1.293@-1.5 . 1.293@-0.92 . 0.937@-0.559 .
 0.937@-0.559 . 0.58@-0.199 . 0.0@-0.199 .
 0.0@-0.199 . -0.568@-0.199 . -0.932@-0.563 .
 -0.932@-0.563 . -1.297@-0.928 . -1.297@-1.5 .};

at: 3 put: {
 -0.099@-0.547 . -0.099@-0.547 . -0.099@-2.246 .
 -0.099@-2.246 . -0.099@-2.246 . -0.398@-2.246 .
 -0.398@-2.246 . -0.398@-2.246 . -0.398@-2.407 .
 -0.398@-2.407 . -0.398@-2.407 . 0.153@-2.453 .
 0.153@-2.453 . 0.153@-2.453 . 0.153@-0.547 .
 0.153@-0.547 . 0.153@-0.547 . -0.099@-0.547 .};
yourself))
scale: 1.0@1.0;
offset: -0.01657458627596497@2.925414477707818;
yourself)
! !

!NCLineArrowGlyph class methodsFor: 'arrow forms' stamp: 'nk 8/25/2003 13:17'!
openArrow
	| a |
	a := { 0.0@0.0 . 0.0@0.0 . -0.5@0.0 . 
	-0.5@0.0 . -0.5@0.0 . 0.0@-1.0 . 
	0.0@-1.0 . 0.0@-1.0 . 0.5@0.0 . 
	0.5@0.0 . 0.5@0.0 . 0.0@0.0 }.
^(NCLineArrowGlyph withContours: (
(Array new: 2)
	at: 1 put: a; 
	at: 2 put:  (a + (0.0@0.33) * 0.6 - (0.0@0.33));
	yourself
"clicks 0.0@0.0"
)) offset: 0@1.0;
defaultScale: 2.5;
yourself! !

!NCLineArrowGlyph class methodsFor: 'arrow forms' stamp: 'nk 8/23/2003 16:05'!
openCircleWithPlus
	"glyph 56 from StarMath"
	^((NCLineArrowGlyph withContours: ((Array new: 5)
at: 1 put: ((Array new: 24) at: 1 put: 0.0@1.485; at: 2 put: -0.68@1.499; at: 3 put: -1.098@1.098; at: 4 put: -1.098@1.098; at: 5 put: -1.517@0.696; at: 6 put: -1.531@0.017; at: 7 put: -1.531@0.017; at: 8 put: -1.546@-0.668; at: 9 put: -1.149@-1.082; at: 10 put: -1.149@-1.082; at: 11 put: -0.748@-1.5; at: 12 put: -0.063@-1.515; at: 13 put: -0.063@-1.515; at: 14 put: 0.617@-1.529; at: 15 put: 1.035@-1.128; at: 16 put: 1.035@-1.128; at: 17 put: 1.454@-0.726; at: 18 put: 1.468@-0.046; at: 19 put: 1.468@-0.046; at: 20 put: 1.483@0.638; at: 21 put: 1.081@1.057; at: 22 put: 1.081@1.057; at: 23 put: 0.685@1.47; at: 24 put: 0.0@1.485; yourself);
at: 2 put: ((Array new: 12) at: 1 put: 0.091@-0.142; at: 2 put: 0.067@-1.263; at: 3 put: 0.067@-1.263; at: 4 put: 0.067@-1.263; at: 5 put: 0.535@-1.21; at: 6 put: 0.837@-0.921; at: 7 put: 0.837@-0.921; at: 8 put: 1.139@-0.631; at: 9 put: 1.211@-0.166; at: 10 put: 1.211@-0.166; at: 11 put: 0.091@-0.142; at: 12 put: 0.091@-0.142; yourself);
at: 3 put: ((Array new: 9) at: 1 put: -0.148@0.112; at: 2 put: -0.148@0.112; at: 3 put: -1.274@0.136; at: 4 put: -1.274@0.136; at: 5 put: -1.129@1.098; at: 6 put: -0.13@1.233; at: 7 put: -0.13@1.233; at: 8 put: -0.13@1.233; at: 9 put: -0.148@0.112; yourself);
at: 4 put: ((Array new: 12) at: 1 put: -0.159@-0.137; at: 2 put: -1.279@-0.113; at: 3 put: -1.279@-0.113; at: 4 put: -1.279@-0.113; at: 5 put: -1.232@-0.576; at: 6 put: -0.937@-0.878; at: 7 put: -0.937@-0.878; at: 8 put: -0.637@-1.186; at: 9 put: -0.177@-1.258; at: 10 put: -0.177@-1.258; at: 11 put: -0.159@-0.137; at: 12 put: -0.159@-0.137; yourself);
at: 5 put: ((Array new: 12) at: 1 put: 0.096@0.107; at: 2 put: 0.096@0.107; at: 3 put: 0.119@1.228; at: 4 put: 0.119@1.228; at: 5 put: 0.585@1.156; at: 6 put: 0.875@0.854; at: 7 put: 0.875@0.854; at: 8 put: 1.164@0.552; at: 9 put: 1.217@0.083; at: 10 put: 1.217@0.083; at: 11 put: 1.217@0.083; at: 12 put: 0.096@0.107; yourself); yourself))
scale: 1.0@1.0;
offset: 9.16770659387112e-10@1.484755672790925;
yourself)
! !

!NCLineArrowGlyph class methodsFor: 'arrow forms' stamp: 'nk 8/23/2003 16:06'!
openCircleWithSquareDot
	"glyph 60 from StarMath"
	^((NCLineArrowGlyph withContours: ((Array new: 3)
at: 1 put: ((Array new: 24) at: 1 put: 0.0@1.516; at: 2 put: -0.685@1.518; at: 3 put: -1.091@1.114; at: 4 put: -1.091@1.114; at: 5 put: -1.498@0.706; at: 6 put: -1.5@0.021; at: 7 put: -1.5@0.021; at: 8 put: -1.502@-0.665; at: 9 put: -1.099@-1.071; at: 10 put: -1.099@-1.071; at: 11 put: -0.695@-1.482; at: 12 put: -0.01@-1.484; at: 13 put: -0.01@-1.484; at: 14 put: 0.675@-1.487; at: 15 put: 1.086@-1.078; at: 16 put: 1.086@-1.078; at: 17 put: 1.498@-0.675; at: 18 put: 1.5@0.01; at: 19 put: 1.5@0.01; at: 20 put: 1.502@0.696; at: 21 put: 1.094@1.107; at: 22 put: 1.094@1.107; at: 23 put: 0.685@1.513; at: 24 put: 0.0@1.516; yourself);
at: 2 put: ((Array new: 24) at: 1 put: -0.001@1.267; at: 2 put: -0.561@1.268; at: 3 put: -0.905@0.927; at: 4 put: -0.905@0.927; at: 5 put: -1.249@0.58; at: 6 put: -1.251@0.02; at: 7 put: -1.251@0.02; at: 8 put: -1.253@-0.541; at: 9 put: -0.911@-0.884; at: 10 put: -0.911@-0.884; at: 11 put: -0.57@-1.233; at: 12 put: -0.009@-1.235; at: 13 put: -0.009@-1.235; at: 14 put: 0.551@-1.237; at: 15 put: 0.9@-0.891; at: 16 put: 0.9@-0.891; at: 17 put: 1.249@-0.549; at: 18 put: 1.251@0.011; at: 19 put: 1.251@0.011; at: 20 put: 1.253@0.572; at: 21 put: 0.906@0.921; at: 22 put: 0.906@0.921; at: 23 put: 0.56@1.264; at: 24 put: -0.001@1.267; yourself);
at: 3 put: ((Array new: 12) at: 1 put: -0.284@0.297; at: 2 put: -0.284@0.297; at: 3 put: -0.286@-0.264; at: 4 put: -0.286@-0.264; at: 5 put: -0.286@-0.264; at: 6 put: 0.274@-0.266; at: 7 put: 0.274@-0.266; at: 8 put: 0.274@-0.266; at: 9 put: 0.276@0.295; at: 10 put: 0.276@0.295; at: 11 put: 0.276@0.295; at: 12 put: -0.284@0.297; yourself); yourself))
scale: 1.0@1.0;
offset: 1.63709046319127e-10@1.51557988458535;
yourself)
! !

!NCLineArrowGlyph class methodsFor: 'arrow forms' stamp: 'nk 8/23/2003 16:05'!
openCircleWithX
	"glyph 58 from StarMath"
	^((NCLineArrowGlyph withContours: ((Array new: 5)
at: 1 put: ((Array new: 24) at: 1 put: 0.0@1.469; at: 2 put: -0.685@1.464; at: 3 put: -1.087@1.056; at: 4 put: -1.087@1.056; at: 5 put: -1.489@0.644; at: 6 put: -1.484@-0.042; at: 7 put: -1.484@-0.042; at: 8 put: -1.479@-0.727; at: 9 put: -1.072@-1.129; at: 10 put: -1.072@-1.129; at: 11 put: -0.664@-1.536; at: 12 put: 0.021@-1.531; at: 13 put: 0.021@-1.531; at: 14 put: 0.706@-1.526; at: 15 put: 1.113@-1.113; at: 16 put: 1.113@-1.113; at: 17 put: 1.521@-0.706; at: 18 put: 1.516@-0.02; at: 19 put: 1.516@-0.02; at: 20 put: 1.511@0.665; at: 21 put: 1.098@1.072; at: 22 put: 1.098@1.072; at: 23 put: 0.685@1.474; at: 24 put: 0.0@1.469; yourself);
at: 2 put: ((Array new: 12) at: 1 put: 0.192@-0.03; at: 2 put: 0.192@-0.03; at: 3 put: 0.976@0.765; at: 4 put: 0.976@0.765; at: 5 put: 1.264@0.403; at: 6 put: 1.267@-0.022; at: 7 put: 1.267@-0.022; at: 8 put: 1.27@-0.448; at: 9 put: 0.987@-0.813; at: 10 put: 0.987@-0.813; at: 11 put: 0.987@-0.813; at: 12 put: 0.192@-0.03; yourself);
at: 3 put: ((Array new: 12) at: 1 put: -0.166@-0.032; at: 2 put: -0.96@0.751; at: 3 put: -0.96@0.751; at: 4 put: -0.96@0.751; at: 5 put: -1.238@0.365; at: 6 put: -1.235@-0.04; at: 7 put: -1.235@-0.04; at: 8 put: -1.232@-0.466; at: 9 put: -0.949@-0.827; at: 10 put: -0.949@-0.827; at: 11 put: -0.166@-0.032; at: 12 put: -0.166@-0.032; yourself);
at: 4 put: ((Array new: 12) at: 1 put: 0.012@-0.208; at: 2 put: -0.771@-1.007; at: 3 put: -0.771@-1.007; at: 4 put: -0.771@-1.007; at: 5 put: -0.416@-1.285; at: 6 put: 0.025@-1.282; at: 7 put: 0.025@-1.282; at: 8 put: 0.45@-1.279; at: 9 put: 0.811@-0.996; at: 10 put: 0.811@-0.996; at: 11 put: 0.012@-0.208; at: 12 put: 0.012@-0.208; yourself);
at: 5 put: ((Array new: 12) at: 1 put: 0.009@0.145; at: 2 put: 0.009@0.145; at: 3 put: -0.785@0.929; at: 4 put: -0.785@0.929; at: 5 put: -0.424@1.217; at: 6 put: 0.007@1.22; at: 7 put: 0.007@1.22; at: 8 put: 0.433@1.223; at: 9 put: 0.798@0.94; at: 10 put: 0.798@0.94; at: 11 put: 0.798@0.94; at: 12 put: 0.009@0.145; yourself); yourself))
scale: 1.0@1.0;
offset: -1.746229827404022e-10@1.46889486873988;
yourself)
! !

!NCLineArrowGlyph class methodsFor: 'arrow forms' stamp: 'nk 8/25/2003 12:34'!
openDiamond
	"glyph 185 from BitstreamVeraSans"
	^((NCLineArrowGlyph withContours: ((Array new: 2)
at: 1 put: ((Array new: 12) at: 1 put: 0.0@0.0; at: 2 put: 0.0@0.0; at: 3 put: -0.235@-0.498; at: 4 put: -0.235@-0.498; at: 5 put: -0.235@-0.498; at: 6 put: 0.0@-1.0; at: 7 put: 0.0@-1.0; at: 8 put: 0.0@-1.0; at: 9 put: 0.235@-0.498; at: 10 put: 0.235@-0.498; at: 11 put: 0.235@-0.498; at: 12 put: 0.0@0.0; yourself);
at: 2 put: ((Array new: 12) at: 1 put: 0.0@-0.115; at: 2 put: -0.181@-0.498; at: 3 put: -0.181@-0.498; at: 4 put: -0.181@-0.498; at: 5 put: 0.0@-0.885; at: 6 put: 0.0@-0.885; at: 7 put: 0.0@-0.885; at: 8 put: 0.181@-0.498; at: 9 put: 0.181@-0.498; at: 10 put: 0.181@-0.498; at: 11 put: 0.0@-0.115; at: 12 put: 0.0@-0.115; yourself); yourself))
scale: 4.71@3.86;
offset: 0.0@3.86;
yourself)
"clicks: 0.0@-0.12 0.0@-0.98 0.0@-0.01 "! !

!NCLineArrowGlyph class methodsFor: 'arrow forms' stamp: 'nk 12/17/2003 14:02'!
pi
	"glyph 155 from BitstreamVeraSans"
	^((NCLineArrowGlyph withContours: ((Array new: 1)
at: 1 put: ((Array new: 84) at: 1 put: 0.0@0.0; at: 2 put: 0.0@0.0; at: 3 put: -0.158@0.745; at: 4 put: -0.158@0.745; at: 5 put: -0.158@0.745; at: 6 put: -0.306@0.745; at: 7 put: -0.306@0.745; at: 8 put: -0.306@0.745; at: 9 put: -0.147@0.0; at: 10 put: -0.147@0.0; at: 11 put: -0.147@0.0; at: 12 put: -0.18@0.0; at: 13 put: -0.18@0.0; at: 14 put: -0.225@0.0; at: 15 put: -0.249@0.026; at: 16 put: -0.249@0.026; at: 17 put: -0.274@0.052; at: 18 put: -0.281@0.109; at: 19 put: -0.281@0.109; at: 20 put: -0.281@0.109; at: 21 put: -0.411@0.109; at: 22 put: -0.411@0.109; at: 23 put: -0.389@-0.027; at: 24 put: -0.331@-0.078; at: 25 put: -0.331@-0.078; at: 26 put: -0.272@-0.129; at: 27 put: -0.137@-0.129; at: 28 put: -0.137@-0.129; at: 29 put: -0.137@-0.129; at: 30 put: 0.589@-0.129; at: 31 put: 0.589@-0.129; at: 32 put: 0.589@-0.129; at: 33 put: 0.56@0.0; at: 34 put: 0.56@0.0; at: 35 put: 0.56@0.0; at: 36 put: 0.411@0.0; at: 37 put: 0.411@0.0; at: 38 put: 0.411@0.0; at: 39 put: 0.326@0.398; at: 40 put: 0.326@0.398; at: 41 put: 0.314@0.458; at: 42 put: 0.307@0.496; at: 43 put: 0.307@0.496; at: 44 put: 0.301@0.533; at: 45 put: 0.301@0.551; at: 46 put: 0.301@0.551; at: 47 put: 0.301@0.598; at: 48 put: 0.32@0.619; at: 49 put: 0.32@0.619; at: 50 put: 0.34@0.638; at: 51 put: 0.385@0.638; at: 52 put: 0.385@0.638; at: 53 put: 0.398@0.638; at: 54 put: 0.417@0.637; at: 55 put: 0.417@0.637; at: 56 put: 0.436@0.634; at: 57 put: 0.466@0.63; at: 58 put: 0.466@0.63; at: 59 put: 0.466@0.63; at: 60 put: 0.442@0.745; at: 61 put: 0.442@0.745; at: 62 put: 0.417@0.752; at: 63 put: 0.394@0.755; at: 64 put: 0.394@0.755; at: 65 put: 0.372@0.758; at: 66 put: 0.351@0.758; at: 67 put: 0.351@0.758; at: 68 put: 0.255@0.758; at: 69 put: 0.205@0.717; at: 70 put: 0.205@0.717; at: 71 put: 0.156@0.675; at: 72 put: 0.156@0.594; at: 73 put: 0.156@0.594; at: 74 put: 0.156@0.556; at: 75 put: 0.164@0.497; at: 76 put: 0.164@0.497; at: 77 put: 0.173@0.438; at: 78 put: 0.201@0.307; at: 79 put: 0.201@0.307; at: 80 put: 0.201@0.307; at: 81 put: 0.266@0.0; at: 82 put: 0.266@0.0; at: 83 put: 0.266@0.0; at: 84 put: 0.0@0.0; yourself); yourself))
scale: 2.0@2.0;
offset: 0.0@0.0;
yourself)
! !

!NCLineArrowGlyph class methodsFor: 'arrow forms' stamp: 'nk 8/23/2003 19:05'!
pointingHand
	"glyph 29 from StarBats"
	^((NCLineArrowGlyph withContours: ((Array new: 2)
at: 1 put: ((Array new: 30) at: 1 put: 0.0@3.02; at: 2 put: -0.178@2.999; at: 3 put: -0.37@2.913; at: 4 put: -0.37@2.913; at: 5 put: -0.623@2.802; at: 6 put: -0.608@2.671; at: 7 put: -0.608@2.671; at: 8 put: -0.602@2.615; at: 9 put: -0.542@2.584; at: 10 put: -0.542@2.584; at: 11 put: -0.546@2.725; at: 12 put: -0.316@2.82; at: 13 put: -0.316@2.82; at: 14 put: -0.153@2.891; at: 15 put: 0.042@2.913; at: 16 put: 0.042@2.913; at: 17 put: 0.221@2.933; at: 18 put: 0.365@2.907; at: 19 put: 0.365@2.907; at: 20 put: 0.571@2.871; at: 21 put: 0.587@2.739; at: 22 put: 0.587@2.739; at: 23 put: 0.651@2.773; at: 24 put: 0.644@2.84; at: 25 put: 0.644@2.84; at: 26 put: 0.628@2.976; at: 27 put: 0.374@3.016; at: 28 put: 0.374@3.016; at: 29 put: 0.203@3.043; at: 30 put: 0.0@3.02; yourself);
at: 2 put: ((Array new: 51) at: 1 put: 0.839@1.858; at: 2 put: 0.82@2.019; at: 3 put: 0.576@2.386; at: 4 put: 0.576@2.386; at: 5 put: 0.576@2.386; at: 6 put: 0.513@2.705; at: 7 put: 0.513@2.705; at: 8 put: 0.402@2.851; at: 9 put: -0.01@2.804; at: 10 put: -0.01@2.804; at: 11 put: -0.319@2.768; at: 12 put: -0.382@2.645; at: 13 put: -0.382@2.645; at: 14 put: -0.424@2.447; at: 15 put: -0.469@2.248; at: 16 put: -0.469@2.248; at: 17 put: -0.469@2.248; at: 18 put: -0.66@1.552; at: 19 put: -0.66@1.552; at: 20 put: -0.62@1.209; at: 21 put: -0.438@0.667; at: 22 put: -0.438@0.667; at: 23 put: -0.203@-0.023; at: 24 put: 0.0@0.0; at: 25 put: 0.0@0.0; at: 26 put: 0.055@0.006; at: 27 put: 0.083@0.095; at: 28 put: 0.083@0.095; at: 29 put: 0.11@0.167; at: 30 put: 0.102@0.231; at: 31 put: 0.102@0.231; at: 32 put: 0.086@0.371; at: 33 put: -0.077@0.932; at: 34 put: -0.077@0.932; at: 35 put: 0.023@0.806; at: 36 put: 0.142@0.82; at: 37 put: 0.142@0.82; at: 38 put: 0.299@0.838; at: 39 put: 0.306@1.036; at: 40 put: 0.306@1.036; at: 41 put: 0.35@0.955; at: 42 put: 0.46@0.968; at: 43 put: 0.46@0.968; at: 44 put: 0.574@0.981; at: 45 put: 0.615@1.115; at: 46 put: 0.615@1.115; at: 47 put: 0.646@1.105; at: 48 put: 0.676@1.109; at: 49 put: 0.676@1.109; at: 50 put: 0.921@1.137; at: 51 put: 0.839@1.858; yourself); yourself))
offset: 0@0;
defaultScale: 1.5;
yourself)
! !

!NCLineArrowGlyph class methodsFor: 'arrow forms' stamp: 'nk 8/23/2003 16:03'!
quadDiamond
	"glyph 19 from StarBats"
	^((NCLineArrowGlyph withContours: ((Array new: 4)
at: 1 put: ((Array new: 12) at: 1 put: 0.0@1.452; at: 2 put: 0.0@1.452; at: 3 put: -0.603@0.84; at: 4 put: -0.603@0.84; at: 5 put: -0.603@0.84; at: 6 put: 0.014@0.237; at: 7 put: 0.014@0.237; at: 8 put: 0.014@0.237; at: 9 put: 0.613@0.854; at: 10 put: 0.613@0.854; at: 11 put: 0.613@0.854; at: 12 put: 0.0@1.452; yourself);
at: 2 put: ((Array new: 12) at: 1 put: -0.882@0.55; at: 2 put: -0.882@0.55; at: 3 put: -1.485@-0.063; at: 4 put: -1.485@-0.063; at: 5 put: -1.485@-0.063; at: 6 put: -0.868@-0.666; at: 7 put: -0.868@-0.666; at: 8 put: -0.868@-0.666; at: 9 put: -0.269@-0.049; at: 10 put: -0.269@-0.049; at: 11 put: -0.269@-0.049; at: 12 put: -0.882@0.55; yourself);
at: 3 put: ((Array new: 12) at: 1 put: -0.583@-0.945; at: 2 put: 0.02@-0.328; at: 3 put: 0.02@-0.328; at: 4 put: 0.02@-0.328; at: 5 put: 0.637@-0.931; at: 6 put: 0.637@-0.931; at: 7 put: 0.637@-0.931; at: 8 put: 0.034@-1.547; at: 9 put: 0.034@-1.547; at: 10 put: 0.034@-1.547; at: 11 put: -0.583@-0.945; at: 12 put: -0.583@-0.945; yourself);
at: 4 put: ((Array new: 12) at: 1 put: 0.916@-0.645; at: 2 put: 0.299@-0.042; at: 3 put: 0.299@-0.042; at: 4 put: 0.299@-0.042; at: 5 put: 0.902@0.57; at: 6 put: 0.902@0.57; at: 7 put: 0.902@0.57; at: 8 put: 1.515@-0.028; at: 9 put: 1.515@-0.028; at: 10 put: 1.515@-0.028; at: 11 put: 0.916@-0.645; at: 12 put: 0.916@-0.645; yourself); yourself))
scale: 1.0@1.0;
offset: -4.91127138957381e-10@1.45237701159931;
yourself)
! !

!NCLineArrowGlyph class methodsFor: 'arrow forms' stamp: 'nk 8/23/2003 16:06'!
roundFork
	"glyph 61 from StarMath"
	^((NCLineArrowGlyph withContours: ((Array new: 1)
at: 1 put: ((Array new: 42) at: 1 put: 0.0@3.016; at: 2 put: -0.685@3.0; at: 3 put: -1.08@2.585; at: 4 put: -1.08@2.585; at: 5 put: -1.475@2.166; at: 6 put: -1.458@1.481; at: 7 put: -1.458@1.481; at: 8 put: -1.458@1.481; at: 9 put: -1.422@-0.019; at: 10 put: -1.422@-0.019; at: 11 put: -1.422@-0.019; at: 12 put: -1.173@-0.013; at: 13 put: -1.173@-0.013; at: 14 put: -1.173@-0.013; at: 15 put: -1.209@1.487; at: 16 put: -1.209@1.487; at: 17 put: -1.223@2.047; at: 18 put: -0.889@2.403; at: 19 put: -0.889@2.403; at: 20 put: -0.554@2.754; at: 21 put: 0.006@2.767; at: 22 put: 0.006@2.767; at: 23 put: 0.566@2.781; at: 24 put: 0.922@2.447; at: 25 put: 0.922@2.447; at: 26 put: 1.278@2.107; at: 27 put: 1.292@1.547; at: 28 put: 1.292@1.547; at: 29 put: 1.292@1.547; at: 30 put: 1.328@0.048; at: 31 put: 1.328@0.048; at: 32 put: 1.328@0.048; at: 33 put: 1.577@0.054; at: 34 put: 1.577@0.054; at: 35 put: 1.577@0.054; at: 36 put: 1.541@1.553; at: 37 put: 1.541@1.553; at: 38 put: 1.524@2.238; at: 39 put: 1.105@2.638; at: 40 put: 1.105@2.638; at: 41 put: 0.685@3.033; at: 42 put: 0.0@3.016; yourself); yourself))
scale: 1.0@1.0;
offset: 0.0@3.01644631032832;
yourself)
! !

!NCLineArrowGlyph class methodsFor: 'arrow forms' stamp: 'nk 8/23/2003 16:07'!
roundTrident
	"glyph 88 from StarMath"
	^((NCLineArrowGlyph withContours: ((Array new: 1)
at: 1 put: ((Array new: 45) at: 1 put: 0.0@2.967; at: 2 put: -0.714@2.971; at: 3 put: -1.113@2.462; at: 4 put: -1.113@2.462; at: 5 put: -1.472@2.003; at: 6 put: -1.477@1.266; at: 7 put: -1.477@1.266; at: 8 put: -1.477@1.266; at: 9 put: -1.484@-0.025; at: 10 put: -1.484@-0.025; at: 11 put: -1.484@-0.025; at: 12 put: -1.176@-0.026; at: 13 put: -1.176@-0.026; at: 14 put: -1.176@-0.026; at: 15 put: -1.168@1.303; at: 16 put: -1.168@1.303; at: 17 put: -1.162@2.462; at: 18 put: -0.161@2.649; at: 19 put: -0.161@2.649; at: 20 put: -0.161@2.649; at: 21 put: -0.176@-0.032; at: 22 put: -0.176@-0.032; at: 23 put: -0.176@-0.032; at: 24 put: 0.148@-0.034; at: 25 put: 0.148@-0.034; at: 26 put: 0.148@-0.034; at: 27 put: 0.163@2.642; at: 28 put: 0.163@2.642; at: 29 put: 1.167@2.373; at: 30 put: 1.161@1.285; at: 31 put: 1.161@1.285; at: 32 put: 1.161@1.285; at: 33 put: 1.154@-0.04; at: 34 put: 1.154@-0.04; at: 35 put: 1.154@-0.04; at: 36 put: 1.461@-0.041; at: 37 put: 1.461@-0.041; at: 38 put: 1.461@-0.041; at: 39 put: 1.469@1.239; at: 40 put: 1.469@1.239; at: 41 put: 1.473@1.986; at: 42 put: 1.118@2.45; at: 43 put: 1.118@2.45; at: 44 put: 0.725@2.963; at: 45 put: 0.0@2.967; yourself); yourself))
scale: 1.0@1.0;
offset: 1.113221514970064e-9@2.967078724404928;
yourself)
! !

!NCLineArrowGlyph class methodsFor: 'testing' stamp: 'nk 8/19/2003 19:04'!
sampleAllArrows
	"NCLineArrowGlyph sampleAllArrows"
	^self sampleAllArrowsWithLineWidth: 20 borderWidth: 1
! !

!NCLineArrowGlyph class methodsFor: 'testing' stamp: 'nk 8/19/2003 19:04'!
sampleAllArrowsWithLineWidth: lw borderWidth: bw
	"NCLineArrowGlyph sampleAllArrows"

	| pos g samples button |
	pos := 0@0.
	g := nil.
	samples := OrderedCollection new.
	(self class organization  listAtCategoryNamed: 'arrow forms') do: [ :sym |
		g := self perform: sym.
		g scale: 100@100.
		samples add: (((g sampleFormForLineWidth: lw borderWidth: bw) asMorph) position: pos; openInWorld).
		pos := pos + (50@20).
	].

	button := (SimpleButtonMorph newWithLabel: 'kill them').
	samples add: button.
	button target: World; actionSelector: #deleteMorphs:; arguments: { samples }; openInHand.
	
	^samples! !

!NCLineArrowGlyph class methodsFor: 'arrow forms' stamp: 'nk 8/23/2003 15:56'!
starInCircle
	"glyph 91 from StarBats"
	^((NCLineArrowGlyph withContours: ((Array new: 10)
at: 1 put: ((Array new: 12) at: 1 put: 0.0@1.483; at: 2 put: -0.281@0.871; at: 3 put: -0.281@0.871; at: 4 put: -0.281@0.871; at: 5 put: -0.111@0.925; at: 6 put: 0.02@0.929; at: 7 put: 0.02@0.929; at: 8 put: 0.15@0.932; at: 9 put: 0.323@0.888; at: 10 put: 0.323@0.888; at: 11 put: 0.0@1.483; at: 12 put: 0.0@1.483; yourself);
at: 2 put: ((Array new: 24) at: 1 put: 1.544@0.074; at: 2 put: 1.526@0.727; at: 3 put: 1.091@1.138; at: 4 put: 1.091@1.138; at: 5 put: 0.656@1.55; at: 6 put: 0.003@1.532; at: 7 put: 0.003@1.532; at: 8 put: -0.654@1.514; at: 9 put: -1.066@1.084; at: 10 put: -1.066@1.084; at: 11 put: -1.473@0.649; at: 12 put: -1.455@-0.008; at: 13 put: -1.455@-0.008; at: 14 put: -1.437@-0.666; at: 15 put: -1.007@-1.073; at: 16 put: -1.007@-1.073; at: 17 put: -0.572@-1.485; at: 18 put: 0.085@-1.467; at: 19 put: 0.085@-1.467; at: 20 put: 0.738@-1.449; at: 21 put: 1.15@-1.014; at: 22 put: 1.15@-1.014; at: 23 put: 1.562@-0.579; at: 24 put: 1.544@0.074; yourself);
at: 3 put: ((Array new: 12) at: 1 put: 0.369@-0.797; at: 2 put: 0.24@-0.859; at: 3 put: 0.069@-0.863; at: 4 put: 0.069@-0.863; at: 5 put: -0.111@-0.868; at: 6 put: -0.235@-0.813; at: 7 put: -0.235@-0.813; at: 8 put: 0.088@-1.417; at: 9 put: 0.088@-1.417; at: 10 put: 0.088@-1.417; at: 11 put: 0.369@-0.797; at: 12 put: 0.369@-0.797; yourself);
at: 4 put: ((Array new: 9) at: 1 put: 1.103@-0.966; at: 2 put: 1.103@-0.966; at: 3 put: 0.445@-0.776; at: 4 put: 0.445@-0.776; at: 5 put: 0.76@-0.592; at: 6 put: 0.865@-0.328; at: 7 put: 0.865@-0.328; at: 8 put: 0.865@-0.328; at: 9 put: 1.103@-0.966; yourself);
at: 5 put: ((Array new: 9) at: 1 put: -0.316@-0.797; at: 2 put: -0.316@-0.797; at: 3 put: -0.959@-1.027; at: 4 put: -0.959@-1.027; at: 5 put: -0.959@-1.027; at: 6 put: -0.756@-0.372; at: 7 put: -0.756@-0.372; at: 8 put: -0.631@-0.635; at: 9 put: -0.316@-0.797; yourself);
at: 6 put: ((Array new: 12) at: 1 put: 1.49@0.077; at: 2 put: 0.878@0.358; at: 3 put: 0.878@0.358; at: 4 put: 0.878@0.358; at: 5 put: 0.931@0.219; at: 6 put: 0.936@0.057; at: 7 put: 0.936@0.057; at: 8 put: 0.94@-0.105; at: 9 put: 0.894@-0.246; at: 10 put: 0.894@-0.246; at: 11 put: 1.49@0.077; at: 12 put: 1.49@0.077; yourself);
at: 7 put: ((Array new: 12) at: 1 put: -0.806@0.311; at: 2 put: -1.405@-0.011; at: 3 put: -1.405@-0.011; at: 4 put: -1.405@-0.011; at: 5 put: -0.79@-0.292; at: 6 put: -0.79@-0.292; at: 7 put: -0.79@-0.292; at: 8 put: -0.851@-0.172; at: 9 put: -0.856@0.008; at: 10 put: -0.856@0.008; at: 11 put: -0.861@0.188; at: 12 put: -0.806@0.311; yourself);
at: 8 put: ((Array new: 9) at: 1 put: -0.361@0.842; at: 2 put: -1.02@1.031; at: 3 put: -1.02@1.031; at: 4 put: -1.02@1.031; at: 5 put: -0.786@0.384; at: 6 put: -0.786@0.384; at: 7 put: -0.786@0.384; at: 8 put: -0.649@0.681; at: 9 put: -0.361@0.842; yourself);
at: 9 put: ((Array new: 9) at: 1 put: 1.038@1.096; at: 2 put: 0.401@0.849; at: 3 put: 0.401@0.849; at: 4 put: 0.401@0.849; at: 5 put: 0.697@0.731; at: 6 put: 0.844@0.438; at: 7 put: 0.844@0.438; at: 8 put: 1.038@1.096; at: 9 put: 1.038@1.096; yourself);
at: 10 put: ((Array new: 24) at: 1 put: -0.378@-0.006; at: 2 put: -0.383@0.156; at: 3 put: -0.264@0.268; at: 4 put: -0.264@0.268; at: 5 put: -0.141@0.375; at: 6 put: 0.026@0.379; at: 7 put: 0.026@0.379; at: 8 put: 0.188@0.384; at: 9 put: 0.312@0.279; at: 10 put: 0.312@0.279; at: 11 put: 0.441@0.174; at: 12 put: 0.446@0.017; at: 13 put: 0.446@0.017; at: 14 put: 0.45@-0.145; at: 15 put: 0.331@-0.252; at: 16 put: 0.331@-0.252; at: 17 put: 0.213@-0.364; at: 18 put: 0.046@-0.368; at: 19 put: 0.046@-0.368; at: 20 put: -0.121@-0.373; at: 21 put: -0.25@-0.268; at: 22 put: -0.25@-0.268; at: 23 put: -0.374@-0.168; at: 24 put: -0.378@-0.006; yourself); yourself))
scale: 1.0@1.0;
offset: 1.517037162557244e-9@1.482536442923447;
yourself)
! !

!NCLineArrowGlyph class methodsFor: 'arrow forms' stamp: 'nk 8/23/2003 19:17'!
test2
	"glyph 22 from StarBats"
	^((NCLineArrowGlyph withContours: ((Array new: 2)
at: 1 put: ((Array new: 12) at: 1 put: 0.0@0.0; at: 2 put: 0.0@0.0; at: 3 put: -1.502@0.867; at: 4 put: -1.502@0.867; at: 5 put: -1.502@0.867; at: 6 put: 0.0@-1.73; at: 7 put: 0.0@-1.73; at: 8 put: 0.0@-1.73; at: 9 put: 1.498@0.867; at: 10 put: 1.498@0.867; at: 11 put: 1.498@0.867; at: 12 put: 0.0@0.0; yourself);
at: 2 put: ((Array new: 9) at: 1 put: 0.0@-1.531; at: 2 put: 0.0@-1.531; at: 3 put: -1.228@0.593; at: 4 put: -1.228@0.593; at: 5 put: -1.228@0.593; at: 6 put: 0.0@-0.116; at: 7 put: 0.0@-0.116; at: 8 put: 0.0@-0.116; at: 9 put: 0.0@-1.531; yourself); yourself))
scale: 1.0@1.0;
offset: 0.0@1.730290498933755;
yourself)
! !

!NCLineArrowGlyph class methodsFor: 'arrow forms' stamp: 'nk 8/23/2003 16:07'!
thinDoubleArrow
	"glyph 70 from StarMath"
	^((NCLineArrowGlyph withContours: ((Array new: 2)
at: 1 put: ((Array new: 18) at: 1 put: 0.0@1.438; at: 2 put: 0.0@1.438; at: 3 put: -0.775@3.002; at: 4 put: -0.775@3.002; at: 5 put: -0.775@3.002; at: 6 put: -0.955@3.002; at: 7 put: -0.955@3.002; at: 8 put: -0.955@3.002; at: 9 put: 0.002@1.075; at: 10 put: 0.002@1.075; at: 11 put: 0.002@1.075; at: 12 put: 0.969@2.998; at: 13 put: 0.969@2.998; at: 14 put: 0.969@2.998; at: 15 put: 0.789@2.998; at: 16 put: 0.789@2.998; at: 17 put: 0.789@2.998; at: 18 put: 0.0@1.438; yourself);
at: 2 put: ((Array new: 18) at: 1 put: 0.967@1.922; at: 2 put: 0.0@0.0; at: 3 put: 0.0@0.0; at: 4 put: 0.0@0.0; at: 5 put: -0.958@1.927; at: 6 put: -0.958@1.927; at: 7 put: -0.958@1.927; at: 8 put: -0.778@1.926; at: 9 put: -0.778@1.926; at: 10 put: -0.778@1.926; at: 11 put: -0.002@0.363; at: 12 put: -0.002@0.363; at: 13 put: -0.002@0.363; at: 14 put: 0.787@1.923; at: 15 put: 0.787@1.923; at: 16 put: 0.787@1.923; at: 17 put: 0.967@1.922; at: 18 put: 0.967@1.922; yourself); yourself))
scale: 1.0@1.0;
offset: 5.45696821063757e-11@1.438405608597577;
yourself)
! !

!NCLineArrowGlyph class methodsFor: 'instance creation' stamp: 'nk 8/19/2003 14:58'!
withContours: anArray
	"anArray is expected to have lists of points already scaled to fit in a unit box"
	^self new withContours: anArray! !

!NCLineArrowGlyph class methodsFor: 'arrow forms' stamp: 'nk 8/23/2003 16:12'!
zap
	"glyph 97 from Webdings"
	^((NCLineArrowGlyph withContours: ((Array new: 1)
at: 1 put: ((Array new: 18) at: 1 put: 0.0@2.148; at: 2 put: 0.0@2.148; at: 3 put: -0.74@-0.429; at: 4 put: -0.74@-0.429; at: 5 put: -0.74@-0.429; at: 6 put: -0.052@-0.275; at: 7 put: -0.052@-0.275; at: 8 put: -0.052@-0.275; at: 9 put: -0.06@-2.077; at: 10 put: -0.06@-2.077; at: 11 put: -0.06@-2.077; at: 12 put: 0.761@0.545; at: 13 put: 0.761@0.545; at: 14 put: 0.761@0.545; at: 15 put: 0.01@0.326; at: 16 put: 0.01@0.326; at: 17 put: 0.01@0.326; at: 18 put: 0.0@2.148; yourself); yourself))
scale: 1.0@1.0;
offset: 4.67989593744278e-8@2.147536450764164;
yourself)
! !

!NCLineArrowGlyph methodsFor: 'line attachment' stamp: 'nk 2/3/2004 17:25'!
drawSampleFormForLineWidth: lw borderWidth: bw onCanvas: aCanvas 
	| lineShape p1 p2 bby halfLw |
	self openForLineWidth: lw.
	halfLw := (lw + bw) * 2.0 / 2.0.
	p1 := self contours first first.
	p2 := self contours first last.
	bby := halfLw * 4 + self bounds bottom.
	lineShape := { 
				{ 
					p1.
					p1.
					p1 x @ bby.
					p1 x @ bby.
					p1 x @ bby.
					p2 x @ bby.
					p2 x @ bby.
					p2.
					p2}}.
	aCanvas transformBy: (MatrixTransform2x3 
				withOffset: (self bounds encompass: 0 @ 0) topCenter negated)
		during: 
			[:canvas | 
			canvas 
				drawGeneralBezierShape: lineShape , self contours
				color: (Color blue alpha: 0.4)
				borderWidth: bw
				borderColor: Color black.
			canvas 
				drawOval: (Rectangle center: 0 @ 0 extent: halfLw)
				color: (Color red alpha: 0.6)
				borderWidth: 1
				borderColor: Color black]! !

!NCLineArrowGlyph methodsFor: 'accessing' stamp: 'nk 2/3/2004 17:25'!
fixEndContours
	"Re-order first so that first seg ends to left of last seg."

	| reversed |
	reversed := self contours collect: 
					[:c | 
					| p1 p2 |
					p1 := c at: 3.
					p2 := c at: c size - 2.
					p1 x > p2 x].
	self contours: (self contours with: reversed
				collect: 
					[:c :rev | 
					| newC |
					newC := rev ifTrue: [c reversed] ifFalse: [c].
					newC first ~= newC last 
						ifTrue: 
							[newC := newC asOrderedCollection.
							newC addAllLast: { 
										newC last.
										newC last.
										newC first}].
					newC asArray]).
	^self contours! !

!NCLineArrowGlyph methodsFor: 'line attachment' stamp: 'nk 2/3/2004 17:25'!
joinIndex
	"Answer the point index of the end of the first contour"

	^self contours first size! !

!NCLineArrowGlyph methodsFor: 'line attachment' stamp: 'nk 2/3/2004 17:36'!
joinPoint
	^originalFirstContour first * self scale + self offset! !

!NCLineArrowGlyph methodsFor: 'line attachment' stamp: 'nk 2/3/2004 17:36'!
openForLineWidth: width 
	| newContour xOffset seg intersections i1 realBounds |
	width isZero 
		ifTrue: 
			[self contours at: 1 put: originalFirstContour * self scale + self offset.
			self updateBounds.
			^self].
	newContour := OrderedCollection 
				withAll: originalFirstContour * self scale + self offset.
	realBounds := Rectangle encompassing: newContour.
	xOffset := width / -2.0.
	seg := LineSegment fromPoints: (newContour copyFrom: 1 to: 3).
	intersections := seg 
				intersectionsWithLineFrom: xOffset @ realBounds bottom
				to: xOffset @ realBounds top.
	i1 := intersections at: 1 ifAbsent: [xOffset @ realBounds bottom].
	seg isBezier2Segment 
		ifFalse: [newContour := newContour copyFrom: 4 to: newContour size].
	newContour addAllFirst: { 
				i1.
				i1.
				newContour first}.
	xOffset := xOffset negated.
	seg := LineSegment 
				from: newContour last
				to: (newContour at: newContour size - 2)
				via: (newContour at: newContour size - 1).
	intersections := seg 
				intersectionsWithLineFrom: xOffset @ realBounds bottom
				to: xOffset @ realBounds top.
	i1 := intersections at: 1 ifAbsent: [xOffset @ realBounds bottom].
	seg isBezier2Segment 
		ifFalse: [newContour := newContour copyFrom: 1 to: newContour size - 3].
	newContour addAll: { 
				newContour last.
				i1.
				i1}.
	self contours at: 1 put: newContour asArray.
	self updateBounds! !

!NCLineArrowGlyph methodsFor: 'accessing' stamp: 'nk 8/26/2003 14:41'!
originalFirstContour
	^originalFirstContour! !

!NCLineArrowGlyph methodsFor: 'line attachment' stamp: 'nk 2/3/2004 17:24'!
sampleFormForLineWidth: lw borderWidth: bw 
	| form rect canvas halfLw |
	halfLw := (lw + bw) * 2.0 / 2.0.
	rect := (self bounds extendBy: 0.0 @ (halfLw * 4)) outsetBy: halfLw.
	form := Form extent: rect extent depth: 32.
	form fillColor: Color white.
	canvas := BalloonCanvas on: form.
	canvas aaLevel: 4.
	canvas 
		transformBy: (MatrixTransform2x3 withOffset: (rect width / 2) @ halfLw)
		during: 
			[:tcanvas | 
			self 
				drawSampleFormForLineWidth: lw
				borderWidth: bw
				onCanvas: tcanvas].
	canvas flush.
	^form! !

!NCLineArrowGlyph methodsFor: 'accessing' stamp: 'nk 2/4/2004 08:44'!
setAsDefaultScale
	"Call my current scale the default"

	originalFirstContour := originalFirstContour 
				collect: [:ea | ea * scale ].
	super setAsDefaultScale.! !

!NCLineArrowGlyph methodsFor: 'accessing' stamp: 'nk 2/4/2004 08:42'!
withContours: anArray 
	"anArray is expected to have lists of points already scaled to fit in a unit box"

	super withContours: anArray.
	self fixEndContours.
	originalFirstContour := self contours first copy.
	self updateBounds! !

Object subclass: #NCLineCurveInfo
	instanceVariableNames: 'line targetPoint segmentIndex closestPoint minDist minT'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Connectors-Lines and Curves'!

!NCLineCurveInfo methodsFor: 'geometry' stamp: 'nk 8/23/2003 13:04'!
bezierBorderPoints
	| segOffs b1 b2 cpts |
	segOffs := segmentIndex - 1 * 3.
	cpts := line contours first.
	b1 := Bezier2Segment from: (cpts at: segOffs + 1)
		via: (cpts at: segOffs + 2)
		to: (cpts at: segOffs + 3).
	cpts := line contours third.
	segOffs := cpts size - segOffs.
	b2 := Bezier2Segment from: (cpts at: segOffs)
		via: (cpts at: segOffs - 1)
		to: (cpts at: segOffs - 2).
	^{ b2 valueAt: minT. b1 valueAt: minT }
! !

!NCLineCurveInfo methodsFor: 'geometry' stamp: 'nk 8/25/2003 14:46'!
borderPoints
	| segOffs b1 b2 cpts |
	line isCurve ifTrue: [ ^self bezierBorderPoints ].
	cpts := line contours first.
	b1 := LineSegment from: (cpts at: segmentIndex)
		to: (cpts at: segmentIndex + 1).
	cpts := line contours third.
	segOffs := cpts size - segmentIndex.
	b2 := LineSegment from: (cpts at: segOffs + 1)
		to: (cpts at: segOffs).
	^{ b2 valueAt: minT. b1 valueAt: minT }
! !

!NCLineCurveInfo methodsFor: 'geometry' stamp: 'nk 8/21/2003 10:22'!
closestBigSegment
	"Answer the starting index of the line's (big) segment nearest to aPoint"
	| bigSegment |
	line isCurve ifFalse: [ ^segmentIndex ].
	bigSegment := 1.
	1 to: segmentIndex do: [ :i |
		(line segments at: i) isBezier2Segment ifTrue: [
			bigSegment := bigSegment + 1.
			i = segmentIndex ifTrue: [ minT < 0.5 ifTrue: [ bigSegment := bigSegment - 1 ]].
		]].
	^bigSegment! !

!NCLineCurveInfo methodsFor: 'accessing' stamp: 'nk 8/19/2003 10:27'!
closestPoint
	^closestPoint! !

!NCLineCurveInfo methodsFor: 'accessing' stamp: 'nk 8/19/2003 10:27'!
line
	^line! !

!NCLineCurveInfo methodsFor: 'initialization' stamp: 'nk 8/19/2003 10:26'!
line: aNCLine targetPoint: tPoint segmentIndex: index closestPoint: cPoint minDist: mDist minT: mT
	line:=aNCLine.
	targetPoint:=tPoint.
	segmentIndex:=index.
	closestPoint:=cPoint.
	minDist:=mDist.
	minT:=mT! !

!NCLineCurveInfo methodsFor: 'accessing' stamp: 'nk 8/19/2003 10:27'!
minDist
	^minDist! !

!NCLineCurveInfo methodsFor: 'accessing' stamp: 'nk 8/19/2003 10:27'!
minT
	^minT! !

!NCLineCurveInfo methodsFor: 'accessing' stamp: 'nk 8/22/2003 12:05'!
normal
	^self tangent normal! !

!NCLineCurveInfo methodsFor: 'geometry' stamp: 'nk 8/19/2003 10:31'!
offset
	"Answer the offset between the target and the point on the curve"
	^targetPoint - closestPoint! !

!NCLineCurveInfo methodsFor: 'as yet unclassified' stamp: 'nk 8/19/2003 12:12'!
printOn: aStream
	aStream nextPutAll: 'NCLineCurveInfo(';
		print: line;
		nextPutAll: ' target='; print: (targetPoint truncateTo: 0.01);
		nextPutAll: ' seg='; print: segmentIndex;
		nextPutAll: ' closest='; print: (closestPoint truncateTo: 0.01);
		nextPutAll: ' minDist='; print: (minDist truncateTo: 0.01);
		nextPutAll: ' minT='; print: ( minT truncateTo: 0.01);
		nextPutAll: ')'! !

!NCLineCurveInfo methodsFor: 'accessing' stamp: 'nk 8/22/2003 12:06'!
segment
	^line segments at: segmentIndex! !

!NCLineCurveInfo methodsFor: 'accessing' stamp: 'nk 8/19/2003 10:27'!
segmentIndex
	^segmentIndex! !

!NCLineCurveInfo methodsFor: 'accessing' stamp: 'nk 8/22/2003 12:05'!
tangent
	^(line segments at: segmentIndex) tangentAt: minT.! !

!NCLineCurveInfo methodsFor: 'accessing' stamp: 'nk 8/19/2003 10:27'!
targetPoint
	^targetPoint! !

!NCLineCurveInfo methodsFor: 'geometry' stamp: 'nk 8/21/2003 10:02'!
vertexWithinProportion: proportion andDistance: dist
	"Answer which vertex is within proportion of the current segment (0..1), if any, and also within dist pixels.
	Answer 0 if none."
	| seg |
	seg := line segments at: segmentIndex ifAbsent: [ ^0 ].
	line isCurve ifTrue: [
		seg isBezier2Segment ifFalse: [ ^0 ].
		^(((minT - 0.5) abs <= proportion) and: [ minDist <= dist ])
			ifTrue: [ ((line segments select: [ :ea | ea isBezier2Segment ]) indexOf: seg) + 1 ]
			ifFalse: [ 0 ].
	] ifFalse: [ | atBegin atEnd |
		atBegin :=  minT between: 0.0 and: proportion / 2.0.
		atEnd :=  minT between: (1.0 - (proportion / 2.0)) and: 1.0.
		^(atEnd | atBegin and: [ minDist <= dist ]) 
			ifTrue: [ segmentIndex + (atEnd ifTrue: [ 1 ] ifFalse: [ 0 ]) ]
			ifFalse: [ 0 ]
	]! !

!Object methodsFor: '*connectors-debugging' stamp: 'nk 11/5/2003 14:34'!
haltOnlyOnce
"
Smalltalk at: #OneShotArmed ifAbsent: [true]
self rearmOneShot
"
	self doOnlyOnce: [self halt]! !

!Project class methodsFor: '*connectors-graph model' stamp: 'nk 1/26/2004 13:00'!
defaultDescriptiveName
	^'project'! !

!Project class methodsFor: '*connectors-projects' stamp: 'nk 3/10/2004 20:23'!
worldMorphClass
	^PasteUpMorph! !

!Project methodsFor: '*connectors-projects' stamp: 'nk 3/10/2004 20:22'!
worldMorphClass
	^self class worldMorphClass! !
ConnectorsInfo initialize!
NCLineMorph initialize!
NCTextMorph initialize!
NCAAConnectorMorph initialize!
ConnectorArrowNameType initialize!
